@@ -1,9 +1,17 @@
-{-# LANGUAGE NoImplicitPrelude #-}
 {-# LANGUAGE TemplateHaskell #-}
+{- |
+Module      : Text.Pandoc.Data
+Copyright   : Copyright (C) 2013-2020 John MacFarlane
+License     : GNU GPL, version 2 or above
 
+Maintainer  : John MacFarlane <jgm@berkeley@edu>
+Stability   : alpha
+Portability : portable
+
+Provide contents data files as Haskell values.
+-}
 module Text.Pandoc.Data (dataFiles) where
 
-import Prelude
 import qualified Data.ByteString as B
 import Data.FileEmbed
 import System.FilePath (splitDirectories)
