@@ -1,26 +1,8 @@
-{-
-Copyright © 2017-2018 Albert Krewinkel <tarleb+pandoc@moltkeplatz.de>
-
-This program is free software; you can redistribute it and/or modify
-it under the terms of the GNU General Public License as published by
-the Free Software Foundation; either version 2 of the License, or
-(at your option) any later version.
-
-This program is distributed in the hope that it will be useful,
-but WITHOUT ANY WARRANTY; without even the implied warranty of
-MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-GNU General Public License for more details.
-
-You should have received a copy of the GNU General Public License
-along with this program; if not, write to the Free Software
-Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--}
 {-# LANGUAGE FlexibleContexts #-}
-{-# LANGUAGE NoImplicitPrelude #-}
 {-# LANGUAGE ScopedTypeVariables #-}
 {- |
    Module      : Text.Pandoc.Lua.Packages
-   Copyright   : Copyright © 2017-2018 Albert Krewinkel
+   Copyright   : Copyright © 2017-2020 Albert Krewinkel
    License     : GNU GPL, version 2 or above
 
    Maintainer  : Albert Krewinkel <tarleb+pandoc@moltkeplatz.de>
@@ -33,24 +15,22 @@ module Text.Pandoc.Lua.Packages
   , installPandocPackageSearcher
   ) where
 
-import Prelude
 import Control.Monad (forM_)
 import Data.ByteString (ByteString)
-import Data.IORef (IORef)
 import Foreign.Lua (Lua, NumResults, liftIO)
-import Text.Pandoc.Class (CommonState, readDataFile, runIO, setUserDataDir)
-import Text.Pandoc.MediaBag (MediaBag)
+import Text.Pandoc.Class.PandocIO (runIO)
+import Text.Pandoc.Class.PandocMonad (readDataFile, setUserDataDir)
 
 import qualified Foreign.Lua as Lua
 import Text.Pandoc.Lua.Module.Pandoc as Pandoc
 import Text.Pandoc.Lua.Module.MediaBag as MediaBag
+import Text.Pandoc.Lua.Module.System as System
+import Text.Pandoc.Lua.Module.Types as Types
 import Text.Pandoc.Lua.Module.Utils as Utils
 
 -- | Parameters used to create lua packages/modules.
 data LuaPackageParams = LuaPackageParams
-  { luaPkgCommonState :: CommonState
-  , luaPkgDataDir :: Maybe FilePath
-  , luaPkgMediaBag :: IORef MediaBag
+  { luaPkgDataDir :: Maybe FilePath
   }
 
 -- | Insert pandoc's package loader as the first loader, making it the default.
@@ -68,15 +48,15 @@ installPandocPackageSearcher luaPkgParams = do
 
 -- | Load a pandoc module.
 pandocPackageSearcher :: LuaPackageParams -> String -> Lua NumResults
-pandocPackageSearcher luaPkgParams pkgName =
+pandocPackageSearcher pkgParams pkgName =
   case pkgName of
-    "pandoc"          -> let datadir = luaPkgDataDir luaPkgParams
+    "pandoc"          -> let datadir = luaPkgDataDir pkgParams
                          in pushWrappedHsFun (Pandoc.pushModule datadir)
-    "pandoc.mediabag" -> let st    = luaPkgCommonState luaPkgParams
-                             mbRef = luaPkgMediaBag luaPkgParams
-                         in pushWrappedHsFun (MediaBag.pushModule st mbRef)
-    "pandoc.utils"    -> let datadirMb = luaPkgDataDir luaPkgParams
-                         in pushWrappedHsFun (Utils.pushModule datadirMb)
+    "pandoc.mediabag" -> pushWrappedHsFun MediaBag.pushModule
+    "pandoc.system"   -> pushWrappedHsFun System.pushModule
+    "pandoc.types"    -> pushWrappedHsFun Types.pushModule
+    "pandoc.utils"    -> let datadir = luaPkgDataDir pkgParams
+                         in pushWrappedHsFun (Utils.pushModule datadir)
     _ -> searchPureLuaLoader
  where
   pushWrappedHsFun f = do
@@ -84,11 +64,11 @@ pandocPackageSearcher luaPkgParams pkgName =
     return 1
   searchPureLuaLoader = do
     let filename = pkgName ++ ".lua"
-    modScript <- liftIO (dataDirScript (luaPkgDataDir luaPkgParams) filename)
+    modScript <- liftIO (dataDirScript (luaPkgDataDir pkgParams) filename)
     case modScript of
       Just script -> pushWrappedHsFun (loadStringAsPackage pkgName script)
       Nothing -> do
-        Lua.push ("no file '" ++ filename ++ "' in pandoc's datadir")
+        Lua.push ("\n\tno file '" ++ filename ++ "' in pandoc's datadir")
         return 1
 
 loadStringAsPackage :: String -> ByteString -> Lua NumResults
