@@ -1,25 +1,8 @@
-{-# LANGUAGE NoImplicitPrelude #-}
-{-
-Copyright (C) 2015-2018 John MacFarlane <jgm@berkeley.edu>
-
-This program is free software; you can redistribute it and/or modify
-it under the terms of the GNU General Public License as published by
-the Free Software Foundation; either version 2 of the License, or
-(at your option) any later version.
-
-This program is distributed in the hope that it will be useful,
-but WITHOUT ANY WARRANTY; without even the implied warranty of
-MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-GNU General Public License for more details.
-
-You should have received a copy of the GNU General Public License
-along with this program; if not, write to the Free Software
-Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--}
-
+{-# LANGUAGE OverloadedStrings #-}
+{-# LANGUAGE ViewPatterns      #-}
 {- |
    Module      : Text.Pandoc.Readers.CommonMark
-   Copyright   : Copyright (C) 2015-2018 John MacFarlane
+   Copyright   : Copyright (C) 2015-2020 John MacFarlane
    License     : GNU GPL, version 2 or above
 
    Maintainer  : John MacFarlane <jgm@berkeley.edu>
@@ -33,17 +16,16 @@ CommonMark is a strongly specified variant of Markdown: http://commonmark.org.
 module Text.Pandoc.Readers.CommonMark (readCommonMark)
 where
 
-import Prelude
 import CMarkGFM
 import Control.Monad.State
-import Data.List (groupBy)
 import qualified Data.Set as Set
-import Data.Text (Text, unpack)
-import Text.Pandoc.Class (PandocMonad)
+import Data.Text (Text)
+import qualified Data.Text as T
+import Text.Pandoc.Class.PandocMonad (PandocMonad)
 import Text.Pandoc.Definition
 import Text.Pandoc.Emoji (emojiToInline)
 import Text.Pandoc.Options
-import Text.Pandoc.Shared (uniqueIdent)
+import Text.Pandoc.Shared (uniqueIdent, taskListItemFromAscii)
 import Text.Pandoc.Walk (walkM)
 
 -- | Parse a CommonMark formatted string into a 'Pandoc' structure.
@@ -58,24 +40,24 @@ readCommonMark opts s = return $
                [ extTable | isEnabled Ext_pipe_tables opts ] ++
                [ extAutolink | isEnabled Ext_autolink_bare_uris opts ]
 
-convertEmojis :: String -> [Inline]
-convertEmojis s@(':':xs) =
-   case break (==':') xs of
-        (ys,':':zs) ->
+convertEmojis :: Text -> [Inline]
+convertEmojis s@(T.uncons -> Just (':',xs)) =
+   case T.break (==':') xs of
+        (ys, T.uncons -> Just (':',zs)) ->
            case emojiToInline ys of
                 Just em -> em : convertEmojis zs
-                Nothing -> Str (':' : ys) : convertEmojis (':':zs)
+                Nothing -> Str (":" <> ys) : convertEmojis (":" <> zs)
         _ -> [Str s]
 convertEmojis s =
-  case break (==':') s of
+  case T.break (==':') s of
     ("","") -> []
     (_,"") -> [Str s]
-    (xs,ys) -> Str xs:convertEmojis ys
+    (xs,ys) -> Str xs : convertEmojis ys
 
 addHeaderIdentifiers :: ReaderOptions -> Pandoc -> Pandoc
 addHeaderIdentifiers opts doc = evalState (walkM (addHeaderId opts) doc) mempty
 
-addHeaderId :: ReaderOptions -> Block -> State (Set.Set String) Block
+addHeaderId :: ReaderOptions -> Block -> State (Set.Set Text) Block
 addHeaderId opts (Header lev (_,classes,kvs) ils) = do
   ids <- get
   let ident = uniqueIdent (readerExtensions opts) ils ids
@@ -100,23 +82,25 @@ addBlock _ (Node _ THEMATIC_BREAK _) =
 addBlock opts (Node _ BLOCK_QUOTE nodes) =
   (BlockQuote (addBlocks opts nodes) :)
 addBlock opts (Node _ (HTML_BLOCK t) _)
-  | isEnabled Ext_raw_html opts = (RawBlock (Format "html") (unpack t) :)
+  | isEnabled Ext_raw_html opts = (RawBlock (Format "html") t :)
   | otherwise                 = id
 -- Note:  the cmark parser will never generate CUSTOM_BLOCK,
 -- so we don't need to handle it:
 addBlock _ (Node _ (CUSTOM_BLOCK _onEnter _onExit) _nodes) =
   id
 addBlock _ (Node _ (CODE_BLOCK info t) _) =
-  (CodeBlock ("", take 1 (words (unpack info)), []) (unpack t) :)
+  (CodeBlock ("", take 1 (T.words info), []) t :)
 addBlock opts (Node _ (HEADING lev) nodes) =
   (Header lev ("",[],[]) (addInlines opts nodes) :)
 addBlock opts (Node _ (LIST listAttrs) nodes) =
-  (constructor (map (setTightness . addBlocks opts . children) nodes) :)
+  (constructor (map listItem nodes) :)
   where constructor = case listType listAttrs of
                        BULLET_LIST  -> BulletList
                        ORDERED_LIST -> OrderedList
                                          (start, DefaultStyle, delim)
         start = listStart listAttrs
+        listItem = taskListItemFromAscii exts . setTightness
+                     . addBlocks opts . children
         setTightness = if listTight listAttrs
                            then map paraToPlain
                            else id
@@ -125,6 +109,7 @@ addBlock opts (Node _ (LIST listAttrs) nodes) =
         delim = case listDelim listAttrs of
                      PERIOD_DELIM -> Period
                      PAREN_DELIM  -> OneParen
+        exts = readerExtensions opts
 addBlock opts (Node _ (TABLE alignments) nodes) =
   (Table [] aligns widths headers rows :)
   where aligns = map fromTableCellAlignment alignments
@@ -190,30 +175,29 @@ addInlines :: ReaderOptions -> [Node] -> [Inline]
 addInlines opts = foldr (addInline opts) []
 
 addInline :: ReaderOptions -> Node -> [Inline] -> [Inline]
-addInline opts (Node _ (TEXT t) _) = (foldr ((++) . toinl) [] clumps ++)
-  where raw = unpack t
-        clumps = groupBy samekind raw
+addInline opts (Node _ (TEXT t) _) = (concatMap toinl clumps ++)
+  where clumps = T.groupBy samekind t
         samekind ' ' ' ' = True
         samekind ' ' _   = False
         samekind _   ' ' = False
         samekind _  _    = True
-        toinl (' ':_) = [Space]
-        toinl xs      = if isEnabled Ext_emoji opts
-                        then convertEmojis xs
-                        else [Str xs]
+        toinl (T.uncons -> Just (' ', _)) = [Space]
+        toinl xs = if isEnabled Ext_emoji opts
+                   then convertEmojis xs
+                   else [Str xs]
 addInline _ (Node _ LINEBREAK _) = (LineBreak :)
 addInline opts (Node _ SOFTBREAK _)
   | isEnabled Ext_hard_line_breaks opts = (LineBreak :)
   | otherwise                           = (SoftBreak :)
 addInline opts (Node _ (HTML_INLINE t) _)
-  | isEnabled Ext_raw_html opts = (RawInline (Format "html") (unpack t) :)
+  | isEnabled Ext_raw_html opts = (RawInline (Format "html") t :)
   | otherwise                 = id
 -- Note:  the cmark parser will never generate CUSTOM_BLOCK,
 -- so we don't need to handle it:
 addInline _ (Node _ (CUSTOM_INLINE _onEnter _onExit) _nodes) =
   id
 addInline _ (Node _ (CODE t) _) =
-  (Code ("",[],[]) (unpack t) :)
+  (Code ("",[],[]) t :)
 addInline opts (Node _ EMPH nodes) =
   (Emph (addInlines opts nodes) :)
 addInline opts (Node _ STRONG nodes) =
@@ -221,7 +205,7 @@ addInline opts (Node _ STRONG nodes) =
 addInline opts (Node _ STRIKETHROUGH nodes) =
   (Strikeout (addInlines opts nodes) :)
 addInline opts (Node _ (LINK url title) nodes) =
-  (Link nullAttr (addInlines opts nodes) (unpack url, unpack title) :)
+  (Link nullAttr (addInlines opts nodes) (url, title) :)
 addInline opts (Node _ (IMAGE url title) nodes) =
-  (Image nullAttr (addInlines opts nodes) (unpack url, unpack title) :)
+  (Image nullAttr (addInlines opts nodes) (url, title) :)
 addInline _ _ = id
