@@ -1,28 +1,4 @@
-{-# LANGUAGE NoImplicitPrelude #-}
 {-# LANGUAGE OverloadedStrings #-}
-{-
-  Copyright (C) 2017 Sascha Wilde <wilde@sha-bang.de>
-
-  partly based on all the other readers, especially the work by
-  John MacFarlane <jgm@berkeley.edu> and
-  Alexander Sulfrian <alexander.sulfrian@fu-berlin.de>
-  all bugs are solely created by me.
-
-This program is free software; you can redistribute it and/or modify
-it under the terms of the GNU General Public License as published by
-the Free Software Foundation; either version 2 of the License, or
-(at your option) any later version.
-
-This program is distributed in the hope that it will be useful,
-but WITHOUT ANY WARRANTY; without even the implied warranty of
-MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-GNU General Public License for more details.
-
-You should have received a copy of the GNU General Public License
-along with this program; if not, write to the Free Software
-Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--}
-
 {- |
    Module      : Text.Pandoc.Readers.Creole
    Copyright   : Copyright (C) 2017 Sascha Wilde
@@ -37,13 +13,13 @@ Conversion of creole text to 'Pandoc' document.
 module Text.Pandoc.Readers.Creole ( readCreole
                                   ) where
 
-import Prelude
 import Control.Monad.Except (guard, liftM2, throwError)
 import qualified Data.Foldable as F
 import Data.Maybe (fromMaybe)
 import Data.Text (Text)
+import qualified Data.Text as T
 import qualified Text.Pandoc.Builder as B
-import Text.Pandoc.Class (PandocMonad (..))
+import Text.Pandoc.Class.PandocMonad (PandocMonad (..))
 import Text.Pandoc.Definition
 import Text.Pandoc.Options
 import Text.Pandoc.Parsing hiding (enclosed)
@@ -93,7 +69,6 @@ parseCreole = do
   eof
   return $ B.doc bs
 
-
 --
 -- block parsers
 --
@@ -115,9 +90,9 @@ nowiki = try $ fmap (B.codeBlock . mconcat) (nowikiStart
                                              >> manyTill content nowikiEnd)
   where
     content = brackets <|> line
-    brackets = try $ option "" ((:[]) <$> newline)
-               <+> (char ' ' >> (many (char ' ') <+> string "}}}") <* eol)
-    line = option "" ((:[]) <$> newline) <+> manyTill anyChar eol
+    brackets = try $ option "" (T.singleton <$> newline)
+               <+> (char ' ' >> (manyChar (char ' ') <+> textStr "}}}") <* eol)
+    line = option "" (T.singleton <$> newline) <+> manyTillChar anyChar eol
     eol = lookAhead $ try $ nowikiEnd <|> newline
     nowikiStart = optional newline >> string "{{{" >> skipMany spaceChar >> newline
     nowikiEnd = try $ linebreak >> string "}}}" >> skipMany spaceChar >> newline
@@ -129,7 +104,7 @@ header = try $ do
     fmap length (many1 (char '='))
   guard $ level <= 6
   skipSpaces
-  content <- B.str <$> manyTill (noneOf "\n") headerEnd
+  content <- B.str <$> manyTillChar (noneOf "\n") headerEnd
   return $ B.header level content
   where
     headerEnd = try $ skipSpaces >> many (char '=') >> skipSpaces >> newline
@@ -227,7 +202,7 @@ inline = choice [ whitespace
 
 escapedChar :: PandocMonad m => CRLParser m B.Inlines
 escapedChar =
-  fmap (B.str . (:[])) (try $ char '~' >> noneOf "\t\n ")
+  fmap (B.str . T.singleton) (try $ char '~' >> noneOf "\t\n ")
 
 escapedLink :: PandocMonad m => CRLParser m B.Inlines
 escapedLink = try $ do
@@ -240,8 +215,8 @@ image = try $ do
   (orig, src) <- wikiImg
   return $ B.image src "" (B.str orig)
   where
-    linkSrc = many $ noneOf "|}\n\r\t"
-    linkDsc = char '|' >> many (noneOf "}\n\r\t")
+    linkSrc = manyChar $ noneOf "|}\n\r\t"
+    linkDsc = char '|' >> manyChar (noneOf "}\n\r\t")
     wikiImg = try $ do
       string "{{"
       src <- linkSrc
@@ -254,11 +229,11 @@ link = try $ do
   (orig, src) <- uriLink <|> wikiLink
   return $ B.link src "" orig
   where
-    linkSrc = many $ noneOf "|]\n\r\t"
-    linkDsc :: PandocMonad m => String -> CRLParser m B.Inlines
+    linkSrc = manyChar $ noneOf "|]\n\r\t"
+    linkDsc :: PandocMonad m => Text -> CRLParser m B.Inlines
     linkDsc otxt = B.str
                    <$> try (option otxt
-                         (char '|' >> many (noneOf "]\n\r\t")))
+                         (char '|' >> manyChar (noneOf "]\n\r\t")))
     linkImg = try $ char '|' >> image
     wikiLink = try $ do
       string "[["
@@ -271,7 +246,7 @@ link = try $ do
       return (B.str orig, src)
 
 inlineNowiki :: PandocMonad m => CRLParser m B.Inlines
-inlineNowiki = B.code <$> (start >> manyTill (noneOf "\n\r") end)
+inlineNowiki = B.code <$> (start >> manyTillChar (noneOf "\n\r") end)
   where
     start = try $ string "{{{"
     end = try $ string "}}}" >> lookAhead (noneOf "}")
@@ -294,11 +269,11 @@ linebreak = newline >> notFollowedBy newline >> (lastNewline <|> innerNewline)
         innerNewline = return B.space
 
 symbol :: PandocMonad m => CRLParser m B.Inlines
-symbol = fmap (B.str . (:[])) (oneOf specialChars)
+symbol = fmap (B.str . T.singleton) (oneOf specialChars)
 
 str :: PandocMonad m => CRLParser m B.Inlines
 str = let strChar = noneOf ("\t\n " ++ specialChars) in
-        fmap B.str (many1 strChar)
+        fmap B.str (many1Char strChar)
 
 bold :: PandocMonad m => CRLParser m B.Inlines
 bold = B.strong . mconcat <$>
