@@ -1,25 +1,7 @@
-{-# LANGUAGE NoImplicitPrelude #-}
-{-
-Copyright (C) 2006-2018 John MacFarlane <jgm@berkeley.edu>
-
-This program is free software; you can redistribute it and/or modify
-it under the terms of the GNU General Public License as published by
-the Free Software Foundation; either version 2 of the License, or
-(at your option) any later version.
-
-This program is distributed in the hope that it will be useful,
-but WITHOUT ANY WARRANTY; without even the implied warranty of
-MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-GNU General Public License for more details.
-
-You should have received a copy of the GNU General Public License
-along with this program; if not, write to the Free Software
-Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--}
-
+{-# LANGUAGE OverloadedStrings #-}
 {- |
    Module      : Text.Pandoc.Readers.DocBook
-   Copyright   : Copyright (C) 2006-2018 John MacFarlane
+   Copyright   : Copyright (C) 2006-2020 John MacFarlane
    License     : GNU GPL, version 2 or above
 
    Maintainer  : John MacFarlane <jgm@berkeley.edu>
@@ -29,7 +11,6 @@ Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 Conversion of DocBook XML to 'Pandoc' document.
 -}
 module Text.Pandoc.Readers.DocBook ( readDocBook ) where
-import Prelude
 import Control.Monad.State.Strict
 import Data.Char (isSpace, toUpper)
 import Data.Default
@@ -42,8 +23,9 @@ import Data.Text (Text)
 import qualified Data.Text as T
 import Text.HTML.TagSoup.Entity (lookupEntity)
 import Text.Pandoc.Builder
-import Text.Pandoc.Class (PandocMonad)
+import Text.Pandoc.Class.PandocMonad (PandocMonad, report)
 import Text.Pandoc.Options
+import Text.Pandoc.Logging (LogMessage(..))
 import Text.Pandoc.Shared (crFilter, safeRead)
 import Text.TeXMath (readMathML, writeTeX)
 import Text.XML.Light
@@ -553,20 +535,22 @@ instance Default DBState where
 
 readDocBook :: PandocMonad m => ReaderOptions -> Text -> m Pandoc
 readDocBook _ inp = do
-  let tree = normalizeTree . parseXML . handleInstructions
-               $ T.unpack $ crFilter inp
+  let tree = normalizeTree . parseXML . handleInstructions $ crFilter inp
   (bs, st') <- flip runStateT (def{ dbContent = tree }) $ mapM parseBlock tree
   return $ Pandoc (dbMeta st') (toList . mconcat $ bs)
 
 -- We treat <?asciidoc-br?> specially (issue #1236), converting it
 -- to <br/>, since xml-light doesn't parse the instruction correctly.
 -- Other xml instructions are simply removed from the input stream.
-handleInstructions :: String -> String
-handleInstructions ('<':'?':'a':'s':'c':'i':'i':'d':'o':'c':'-':'b':'r':'?':'>':xs) = '<':'b':'r':'/':'>': handleInstructions xs
-handleInstructions xs = case break (=='<') xs of
+handleInstructions :: Text -> Text
+handleInstructions = T.pack . handleInstructions' . T.unpack
+
+handleInstructions' :: String -> String
+handleInstructions' ('<':'?':'a':'s':'c':'i':'i':'d':'o':'c':'-':'b':'r':'?':'>':xs) = '<':'b':'r':'/':'>': handleInstructions' xs
+handleInstructions' xs = case break (=='<') xs of
                              (ys, [])     -> ys
-                             ([], '<':zs) -> '<' : handleInstructions zs
-                             (ys, zs)     -> ys ++ handleInstructions zs
+                             ([], '<':zs) -> '<' : handleInstructions' zs
+                             (ys, zs)     -> ys ++ handleInstructions' zs
 
 getFigure :: PandocMonad m => Element -> DB m Blocks
 getFigure e = do
@@ -597,13 +581,13 @@ convertEntity :: String -> String
 convertEntity e = Data.Maybe.fromMaybe (map toUpper e) (lookupEntity e)
 
 -- convenience function to get an attribute value, defaulting to ""
-attrValue :: String -> Element -> String
+attrValue :: String -> Element -> Text
 attrValue attr elt =
-  fromMaybe "" (lookupAttrBy (\x -> qName x == attr) (elAttribs elt))
+  maybe "" T.pack (lookupAttrBy (\x -> qName x == attr) (elAttribs elt))
 
 -- convenience function
-named :: String -> Element -> Bool
-named s e = qName (elName e) == s
+named :: Text -> Element -> Bool
+named s e = qName (elName e) == T.unpack s
 
 --
 
@@ -628,7 +612,7 @@ addMetadataFromElement e = do
                    [z] -> getInlines z >>= addMeta fieldname
                    zs  -> mapM getInlines zs >>= addMeta fieldname
 
-addMeta :: PandocMonad m => ToMetaValue a => String -> a -> DB m ()
+addMeta :: PandocMonad m => ToMetaValue a => Text -> a -> DB m ()
 addMeta field val = modify (setMeta field val)
 
 instance HasMeta DBState where
@@ -636,25 +620,27 @@ instance HasMeta DBState where
   deleteMeta field s = s {dbMeta = deleteMeta field (dbMeta s)}
 
 isBlockElement :: Content -> Bool
-isBlockElement (Elem e) = qName (elName e) `elem` blocktags
-  where blocktags = ["toc","index","para","formalpara","simpara",
+isBlockElement (Elem e) = qName (elName e) `elem` blockTags
+isBlockElement _ = False
+
+blockTags :: [String]
+blockTags = ["toc","index","para","formalpara","simpara",
            "ackno","epigraph","blockquote","bibliography","bibliodiv",
            "biblioentry","glossee","glosseealso","glossary",
            "glossdiv","glosslist","chapter","appendix","preface",
            "bridgehead","sect1","sect2","sect3","sect4","sect5","section",
-           "refsect1","refsect2","refsect3","refsection",
-           "important","caution","note","tip","warning","qandadiv",
+           "refsect1","refsect2","refsect3","refsection", "qandadiv",
            "question","answer","abstract","itemizedlist","orderedlist",
            "variablelist","article","book","table","informaltable",
            "informalexample", "linegroup",
-           "screen","programlisting","example","calloutlist"]
-isBlockElement _ = False
+           "screen","programlisting","example","calloutlist"] ++ admonitionTags
+
+admonitionTags :: [String]
+admonitionTags = ["important","caution","note","tip","warning"]
 
 -- Trim leading and trailing newline characters
-trimNl :: String -> String
-trimNl = reverse . go . reverse . go
-  where go ('\n':xs) = xs
-        go xs        = xs
+trimNl :: Text -> Text
+trimNl = T.dropAround (== '\n')
 
 -- meld text into beginning of first paragraph of Blocks.
 -- assumes Blocks start with a Para; if not, does nothing.
@@ -681,7 +667,7 @@ getMediaobject e = do
                                    h = case atVal "depth" of
                                          "" -> []
                                          d  -> [("height", d)]
-                                   atr = (atVal "id", words $ atVal "role", w ++ h)
+                                   atr = (atVal "id", T.words $ atVal "role", w ++ h)
                                in  return (atVal "fileref", atr)
   let getCaption el = case filterChild (\x -> named "caption" x
                                             || named "textobject" x
@@ -704,12 +690,12 @@ parseBlock :: PandocMonad m => Content -> DB m Blocks
 parseBlock (Text (CData CDataRaw _ _)) = return mempty -- DOCTYPE
 parseBlock (Text (CData _ s _)) = if all isSpace s
                                      then return mempty
-                                     else return $ plain $ trimInlines $ text s
-parseBlock (CRef x) = return $ plain $ str $ map toUpper x
+                                     else return $ plain $ trimInlines $ text $ T.pack s
+parseBlock (CRef x) = return $ plain $ str $ T.toUpper $ T.pack x
 parseBlock (Elem e) =
   case qName (elName e) of
-        "toc"   -> return mempty -- skip TOC, since in pandoc it's autogenerated
-        "index" -> return mempty -- skip index, since page numbers meaningless
+        "toc"   -> skip -- skip TOC, since in pandoc it's autogenerated
+        "index" -> skip -- skip index, since page numbers meaningless
         "para"  -> parseMixed para (elContent e)
         "formalpara" -> do
            tit <- case filterChild (named "title") e of
@@ -721,9 +707,9 @@ parseBlock (Elem e) =
         "ackno"  -> parseMixed para (elContent e)
         "epigraph" -> parseBlockquote
         "blockquote" -> parseBlockquote
-        "attribution" -> return mempty
-        "titleabbrev" -> return mempty
-        "authorinitials" -> return mempty
+        "attribution" -> skip
+        "titleabbrev" -> skip
+        "authorinitials" -> skip
         "bibliography" -> sect 0
         "bibliodiv" -> sect 1
         "biblioentry" -> parseMixed para (elContent e)
@@ -739,7 +725,7 @@ parseBlock (Elem e) =
                   mapM parseGlossEntry (filterChildren (named "glossentry") e)
         "glosslist" -> definitionList <$>
                   mapM parseGlossEntry (filterChildren (named "glossentry") e)
-        "chapter" -> sect 0
+        "chapter" -> modify (\st -> st{ dbBook = True}) >> sect 0
         "appendix" -> sect 0
         "preface" -> sect 0
         "bridgehead" -> para . strong <$> getInlines e
@@ -753,19 +739,10 @@ parseBlock (Elem e) =
         "refsect2" -> sect 2
         "refsect3" -> sect 3
         "refsection" -> gets dbSectionLevel >>= sect . (+1)
-        "important" -> blockQuote . (para (strong $ str "Important") <>)
-                        <$> getBlocks e
-        "caution" -> blockQuote . (para (strong $ str "Caution") <>)
-                        <$> getBlocks e
-        "note" -> blockQuote . (para (strong $ str "Note") <>)
-                        <$> getBlocks e
-        "tip" -> blockQuote . (para (strong $ str "Tip") <>)
-                        <$> getBlocks e
-        "warning" -> blockQuote . (para (strong $ str "Warning") <>)
-                        <$> getBlocks e
-        "area" -> return mempty
-        "areaset" -> return mempty
-        "areaspec" -> return mempty
+        l | l `elem` admonitionTags -> parseAdmonition $ T.pack l
+        "area" -> skip
+        "areaset" -> skip
+        "areaspec" -> skip
         "qandadiv" -> gets dbSectionLevel >>= sect . (+1)
         "question" -> addToStart (strong (str "Q:") <> str " ") <$> getBlocks e
         "answer" -> addToStart (strong (str "A:") <> str " ") <$> getBlocks e
@@ -788,22 +765,22 @@ parseBlock (Elem e) =
         "variablelist" -> definitionList <$> deflistitems
         "figure" -> getFigure e
         "mediaobject" -> para <$> getMediaobject e
-        "caption" -> return mempty
+        "caption" -> skip
         "info" -> addMetadataFromElement e
         "articleinfo" -> addMetadataFromElement e
-        "sectioninfo" -> return mempty  -- keywords & other metadata
-        "refsectioninfo" -> return mempty  -- keywords & other metadata
-        "refsect1info" -> return mempty  -- keywords & other metadata
-        "refsect2info" -> return mempty  -- keywords & other metadata
-        "refsect3info" -> return mempty  -- keywords & other metadata
-        "sect1info" -> return mempty  -- keywords & other metadata
-        "sect2info" -> return mempty  -- keywords & other metadata
-        "sect3info" -> return mempty  -- keywords & other metadata
-        "sect4info" -> return mempty  -- keywords & other metadata
-        "sect5info" -> return mempty  -- keywords & other metadata
-        "chapterinfo" -> return mempty  -- keywords & other metadata
-        "glossaryinfo" -> return mempty  -- keywords & other metadata
-        "appendixinfo" -> return mempty  -- keywords & other metadata
+        "sectioninfo" -> skip -- keywords & other metadata
+        "refsectioninfo" -> skip -- keywords & other metadata
+        "refsect1info" -> skip -- keywords & other metadata
+        "refsect2info" -> skip -- keywords & other metadata
+        "refsect3info" -> skip -- keywords & other metadata
+        "sect1info" -> skip  -- keywords & other metadata
+        "sect2info" -> skip  -- keywords & other metadata
+        "sect3info" -> skip  -- keywords & other metadata
+        "sect4info" -> skip  -- keywords & other metadata
+        "sect5info" -> skip  -- keywords & other metadata
+        "chapterinfo" -> skip -- keywords & other metadata
+        "glossaryinfo" -> skip  -- keywords & other metadata
+        "appendixinfo" -> skip  -- keywords & other metadata
         "bookinfo" -> addMetadataFromElement e
         "article" -> modify (\st -> st{ dbBook = False }) >>
                            addMetadataFromElement e >> getBlocks e
@@ -820,8 +797,12 @@ parseBlock (Elem e) =
         "?xml"  -> return mempty
         "title" -> return mempty     -- handled in parent element
         "subtitle" -> return mempty  -- handled in parent element
-        _       -> getBlocks e
-   where parseMixed container conts = do
+        _       -> skip >> getBlocks e
+   where skip = do
+           lift $ report $ IgnoredElement $ T.pack $ qName (elName e)
+           return mempty
+
+         parseMixed container conts = do
            let (ils,rest) = break isBlockElement conts
            ils' <- (trimInlines . mconcat) <$> mapM parseInline ils
            let p = if ils' == mempty then mempty else container ils'
@@ -836,7 +817,7 @@ parseBlock (Elem e) =
                                 "" -> []
                                 x  -> [x]
            return $ codeBlockWith (attrValue "id" e, classes', [])
-                  $ trimNl $ strContentRecursive e
+                  $ trimNl $ T.pack $ strContentRecursive e
          parseBlockquote = do
             attrib <- case filterChild (named "attribution") e of
                              Nothing  -> return mempty
@@ -889,9 +870,9 @@ parseBlock (Elem e) =
                                                 _             -> AlignDefault
                       let toWidth c = case findAttr (unqual "colwidth") c of
                                                 Just w -> fromMaybe 0
-                                                   $ safeRead $ '0': filter (\x ->
+                                                   $ safeRead $ "0" <> T.filter (\x ->
                                                      (x >= '0' && x <= '9')
-                                                      || x == '.') w
+                                                      || x == '.') (T.pack w)
                                                 Nothing -> 0 :: Double
                       let numrows = case bodyrows of
                                          [] -> 0
@@ -926,6 +907,22 @@ parseBlock (Elem e) =
                      modify $ \st -> st{ dbSectionLevel = n - 1 }
                      return $ headerWith (ident,[],[]) n' headerText <> b
          lineItems = mapM getInlines $ filterChildren (named "line") e
+         -- | Admonitions are parsed into a div. Following other Docbook tools that output HTML,
+         -- we parse the optional title as a div with the @title@ class, and give the
+         -- block itself a class corresponding to the admonition name.
+         parseAdmonition label = do
+           -- <title> elements can be directly nested inside an admonition block, use
+           -- it if it's there. It is unclear whether we should include the label in
+           -- the title: docbook references are ambiguous on that, and some implementations of admonitions
+           -- (e.g. asciidoctor) just use an icon in all cases. To be conservative, we don't
+           -- include the label and leave it to styling.
+           title <- case filterChild (named "title") e of
+                        Just t  -> divWith ("", ["title"], []) . plain <$> getInlines t
+                        Nothing -> return mempty
+           -- this will ignore the title element if it is present
+           b <- getBlocks e
+           -- we also attach the label as a class, so it can be styled properly
+           return $ divWith (attrValue "id" e,[label],[]) (title <> b)
 
 getInlines :: PandocMonad m => Element -> DB m Inlines
 getInlines e' = (trimInlines . mconcat) <$>
@@ -940,9 +937,9 @@ elementToStr (Elem e') = Text $ CData CDataText (strContentRecursive e') Nothing
 elementToStr x = x
 
 parseInline :: PandocMonad m => Content -> DB m Inlines
-parseInline (Text (CData _ s _)) = return $ text s
+parseInline (Text (CData _ s _)) = return $ text $ T.pack s
 parseInline (CRef ref) =
-  return $ maybe (text $ map toUpper ref) text $ lookupEntity ref
+  return $ text $ maybe (T.toUpper $ T.pack ref) T.pack $ lookupEntity ref
 parseInline (Elem e) =
   case qName (elName e) of
         "equation" -> equation e displayMath
@@ -982,7 +979,7 @@ parseInline (Elem e) =
         "constant" -> codeWithLang
         "userinput" -> codeWithLang
         "varargs" -> return $ code "(...)"
-        "keycap" -> return (str $ strContent e)
+        "keycap" -> return (str $ T.pack $ strContent e)
         "keycombo" -> keycombo <$>
                          mapM parseInline (elContent e)
         "menuchoice" -> menuchoice <$>
@@ -994,20 +991,20 @@ parseInline (Elem e) =
             let title = case attrValue "endterm" e of
                             ""      -> maybe "???" xrefTitleByElem
                                          (findElementById linkend content)
-                            endterm -> maybe "???" strContent
+                            endterm -> maybe "???" (T.pack . strContent)
                                          (findElementById endterm content)
-            return $ link ('#' : linkend) "" (text title)
-        "email" -> return $ link ("mailto:" ++ strContent e) ""
-                          $ str $ strContent e
-        "uri" -> return $ link (strContent e) "" $ str $ strContent e
+            return $ link ("#" <> linkend) "" (text title)
+        "email" -> return $ link ("mailto:" <> T.pack (strContent e)) ""
+                          $ str $ T.pack $ strContent e
+        "uri" -> return $ link (T.pack $ strContent e) "" $ str $ T.pack $ strContent e
         "ulink" -> link (attrValue "url" e) "" <$> innerInlines
         "link" -> do
              ils <- innerInlines
              let href = case findAttr (QName "href" (Just "http://www.w3.org/1999/xlink") Nothing) e of
-                               Just h -> h
-                               _      -> '#' : attrValue "linkend" e
+                               Just h -> T.pack h
+                               _      -> "#" <> attrValue "linkend" e
              let ils' = if ils == mempty then str href else ils
-             let attr = (attrValue "id" e, words $ attrValue "role" e, [])
+             let attr = (attrValue "id" e, T.words $ attrValue "role" e, [])
              return $ linkWith attr href "" ils'
         "foreignphrase" -> emph <$> innerInlines
         "emphasis" -> case attrValue "role" e of
@@ -1018,19 +1015,23 @@ parseInline (Elem e) =
         "footnote" -> (note . mconcat) <$>
                          mapM parseBlock (elContent e)
         "title" -> return mempty
-        "affiliation" -> return mempty
+        "affiliation" -> skip
         -- Note: this isn't a real docbook tag; it's what we convert
         -- <?asciidor-br?> to in handleInstructions, above.  A kludge to
         -- work around xml-light's inability to parse an instruction.
         "br" -> return linebreak
-        _          -> innerInlines
-   where innerInlines = (trimInlines . mconcat) <$>
+        _          -> skip >> innerInlines
+   where skip = do
+           lift $ report $ IgnoredElement $ T.pack $ qName (elName e)
+           return mempty
+
+         innerInlines = (trimInlines . mconcat) <$>
                           mapM parseInline (elContent e)
          codeWithLang = do
            let classes' = case attrValue "language" e of
                                "" -> []
                                l  -> [l]
-           return $ codeWith (attrValue "id" e,classes',[]) $ strContentRecursive e
+           return $ codeWith (attrValue "id" e,classes',[]) $ T.pack $ strContentRecursive e
          simpleList = (mconcat . intersperse (str "," <> space)) <$> mapM getInlines
                          (filterChildren (named "member") e)
          segmentedList = do
@@ -1060,8 +1061,8 @@ parseInline (Elem e) =
          -- if there's no such attribute, employ some heuristics based on what
          -- docbook-xsl does.
          xrefTitleByElem el
-             | not (null xrefLabel) = xrefLabel
-             | otherwise            = case qName (elName el) of
+             | not (T.null xrefLabel) = xrefLabel
+             | otherwise              = case qName (elName el) of
                   "chapter"      -> descendantContent "title" el
                   "section"      -> descendantContent "title" el
                   "sect1"        -> descendantContent "title" el
@@ -1071,10 +1072,10 @@ parseInline (Elem e) =
                   "sect5"        -> descendantContent "title" el
                   "cmdsynopsis"  -> descendantContent "command" el
                   "funcsynopsis" -> descendantContent "function" el
-                  _              -> qName (elName el) ++ "_title"
+                  _              -> T.pack $ qName (elName el) ++ "_title"
           where
             xrefLabel = attrValue "xreflabel" el
-            descendantContent name = maybe "???" strContent
+            descendantContent name = maybe "???" (T.pack . strContent)
                                    . filterElementName (\n -> qName n == name)
 
 -- | Extract a math equation from an element
@@ -1086,34 +1087,32 @@ equation
   :: Monad m
   => Element
   -- ^ The element from which to extract a mathematical equation
-  -> (String -> Inlines)
+  -> (Text -> Inlines)
   -- ^ A constructor for some Inlines, taking the TeX code as input
   -> m Inlines
 equation e constructor =
-  return $ mconcat $ map constructor $ mathMLEquations ++ latexEquations
+  return $ mconcat $ map constructor $ mathMLEquations <> latexEquations
   where
-    mathMLEquations :: [String]
+    mathMLEquations :: [Text]
     mathMLEquations = map writeTeX $ rights $ readMath
       (\x -> qName (elName x) == "math" && qPrefix (elName x) == Just "mml")
-      (readMathML . showElement)
+      (readMathML . T.pack . showElement)
 
-    latexEquations :: [String]
+    latexEquations :: [Text]
     latexEquations = readMath (\x -> qName (elName x) == "mathphrase")
-                              (concat . fmap showVerbatimCData . elContent)
+                              (T.concat . fmap showVerbatimCData . elContent)
 
     readMath :: (Element -> Bool) -> (Element -> b) -> [b]
     readMath childPredicate fromElement =
-      ( map (fromElement . everywhere (mkT removePrefix))
+      map (fromElement . everywhere (mkT removePrefix))
       $ filterChildren childPredicate e
-      )
 
--- | Get the actual text stored in a verbatim CData block. 'showContent'
+-- | Get the actual text stored in a CData block. 'showContent'
 -- returns the text still surrounded by the [[CDATA]] tags.
---
--- Returns 'showContent' if this is not a verbatim CData
-showVerbatimCData :: Content -> String
-showVerbatimCData (Text (CData CDataVerbatim d _)) = d
-showVerbatimCData c = showContent c
+showVerbatimCData :: Content -> Text
+showVerbatimCData (Text (CData _ d _)) = T.pack d
+showVerbatimCData c = T.pack $ showContent c
+
 
 -- | Set the prefix of a name to 'Nothing'
 removePrefix :: QName -> QName
