@@ -1,26 +1,9 @@
 {-# LANGUAGE FlexibleContexts #-}
 {-# LANGUAGE TupleSections #-}
-{-
-Copyright (C) 2018 Alexander Krotov <ilabdsf@gmail.com>
-
-This program is free software; you can redistribute it and/or modify
-it under the terms of the GNU General Public License as published by
-the Free Software Foundation; either version 2 of the License, or
-(at your option) any later version.
-
-This program is distributed in the hope that it will be useful,
-but WITHOUT ANY WARRANTY; without even the implied warranty of
-MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-GNU General Public License for more details.
-
-You should have received a copy of the GNU General Public License
-along with this program; if not, write to the Free Software
-Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--}
-
+{-# LANGUAGE OverloadedStrings #-}
 {- |
    Module      : Text.Pandoc.Readers.FB2
-   Copyright   : Copyright (C) 2018 Alexander Krotov
+   Copyright   : Copyright (C) 2018-2020 Alexander Krotov
    License     : GNU GPL, version 2 or above
 
    Maintainer  : Alexander Krotov <ilabdsf@gmail.com>
@@ -40,21 +23,20 @@ TODO:
 -}
 
 module Text.Pandoc.Readers.FB2 ( readFB2 ) where
-import Prelude
 import Control.Monad.Except (throwError)
 import Control.Monad.State.Strict
 import Data.ByteString.Lazy.Char8 ( pack )
 import Data.ByteString.Base64.Lazy
-import Data.Char (isSpace, toUpper)
 import Data.Functor
-import Data.List (dropWhileEnd, intersperse)
-import Data.List.Split (splitOn)
+import Data.List (intersperse)
+import qualified Data.Map as M
 import Data.Text (Text)
+import qualified Data.Text as T
 import Data.Default
 import Data.Maybe
 import Text.HTML.TagSoup.Entity (lookupEntity)
 import Text.Pandoc.Builder
-import Text.Pandoc.Class (PandocMonad, insertMedia, report)
+import Text.Pandoc.Class.PandocMonad (PandocMonad, insertMedia, report)
 import Text.Pandoc.Error
 import Text.Pandoc.Logging
 import Text.Pandoc.Options
@@ -65,13 +47,15 @@ type FB2 m = StateT FB2State m
 
 data FB2State = FB2State{ fb2SectionLevel :: Int
                         , fb2Meta :: Meta
-                        , fb2Authors :: [String]
+                        , fb2Authors :: [Text]
+                        , fb2Notes :: M.Map Text Blocks
                         } deriving Show
 
 instance Default FB2State where
   def = FB2State{ fb2SectionLevel = 1
                 , fb2Meta = mempty
                 , fb2Authors = []
+                , fb2Notes = M.empty
                 }
 
 instance HasMeta FB2State where
@@ -79,28 +63,32 @@ instance HasMeta FB2State where
   deleteMeta field s = s {fb2Meta = deleteMeta field (fb2Meta s)}
 
 readFB2 :: PandocMonad m => ReaderOptions -> Text -> m Pandoc
-readFB2 _ inp  = do
-  (bs, st) <- runStateT (mapM parseBlock $ parseXML (crFilter inp)) def
-  let authors = if null $ fb2Authors st
-                then id
-                else setMeta "author" (map text $ reverse $ fb2Authors st)
-  pure $ Pandoc (authors $ fb2Meta st) (toList . mconcat $ bs)
+readFB2 _ inp =
+  case parseXMLDoc $ crFilter inp of
+    Nothing -> throwError $ PandocParseError "Not an XML document"
+    Just e ->  do
+      (bs, st) <- runStateT (parseRootElement e) def
+      let authors = if null $ fb2Authors st
+                    then id
+                    else setMeta "author" (map text $ reverse $ fb2Authors st)
+      pure $ Pandoc (authors $ fb2Meta st) $ toList bs
 
 -- * Utility functions
 
-trim :: String -> String
-trim = dropWhileEnd isSpace . dropWhile isSpace
+trim :: Text -> Text
+trim = T.strip
 
-removeHash :: String -> String
-removeHash ('#':xs) = xs
-removeHash xs = xs
+removeHash :: Text -> Text
+removeHash t = case T.uncons t of
+  Just ('#', xs) -> xs
+  _              -> t
 
-convertEntity :: String -> String
-convertEntity e = fromMaybe (map toUpper e) (lookupEntity e)
+convertEntity :: String -> Text
+convertEntity e = maybe (T.toUpper $ T.pack e) T.pack $ lookupEntity e
 
 parseInline :: PandocMonad m => Content -> FB2 m Inlines
 parseInline (Elem e) =
-  case qName $ elName e of
+  case T.pack $ qName $ elName e of
     "strong" -> strong <$> parseStyleType e
     "emphasis" -> emph <$> parseStyleType e
     "style" -> parseNamedStyle e
@@ -108,10 +96,12 @@ parseInline (Elem e) =
     "strikethrough" -> strikeout <$> parseStyleType e
     "sub" -> subscript <$> parseStyleType e
     "sup" -> superscript <$> parseStyleType e
-    "code" -> pure $ code $ strContent e
+    "code" -> pure $ code $ T.pack $ strContent e
     "image" -> parseInlineImageElement e
-    name -> throwError $ PandocParseError ("Couldn't parse FB2 file: unexpected element " ++ name ++ ".")
-parseInline (Text x) = pure $ text $ cdData x
+    name -> do
+      report $ IgnoredElement name
+      pure mempty
+parseInline (Text x) = pure $ text $ T.pack $ cdData x
 parseInline (CRef r) = pure $ str $ convertEntity r
 
 parseSubtitle :: PandocMonad m => Element -> FB2 m Blocks
@@ -119,40 +109,80 @@ parseSubtitle e = headerWith ("", ["unnumbered"], []) <$> gets fb2SectionLevel <
 
 -- * Root element parser
 
-parseBlock :: PandocMonad m => Content -> FB2 m Blocks
-parseBlock (Elem e) =
-  case qName $ elName e of
-    "?xml"  -> pure mempty
-    "FictionBook" -> mconcat <$> mapM parseFictionBookChild (elChildren e)
+parseRootElement :: PandocMonad m => Element -> FB2 m Blocks
+parseRootElement e =
+  case T.pack $ qName $ elName e of
+    "FictionBook" -> do
+      -- Parse notes before parsing the rest of the content.
+      case filterChild isNotesBody e of
+        Nothing -> pure ()
+        Just notesBody -> parseNotesBody notesBody
+      -- Parse metadata and content
+      mconcat <$> mapM parseFictionBookChild (elChildren e)
     name -> report (UnexpectedXmlElement name "root") $> mempty
-parseBlock _ = pure mempty
+
+-- | Parse notes
+parseNotesBody :: PandocMonad m => Element -> FB2 m ()
+parseNotesBody e = mempty <$ mapM parseNotesBodyChild (elChildren e)
+
+-- | Parse a child of @\<body name="notes">@ element.
+parseNotesBodyChild :: PandocMonad m => Element -> FB2 m ()
+parseNotesBodyChild e =
+  case qName $ elName e of
+    "section" -> parseNote e
+    _ -> pure ()
+
+isNotesBody :: Element -> Bool
+isNotesBody e =
+  qName (elName e) == "body" &&
+  findAttr (unqual "name") e == Just "notes"
+
+parseNote :: PandocMonad m => Element -> FB2 m ()
+parseNote e =
+  case findAttr (unqual "id") e of
+    Nothing -> pure ()
+    Just sectionId -> do
+      content <- mconcat <$> mapM parseSectionChild (dropTitle $ elChildren e)
+      oldNotes <- gets fb2Notes
+      modify $ \s -> s { fb2Notes = M.insert ("#" <> T.pack sectionId) content oldNotes }
+      pure ()
+  where
+    isTitle x = qName (elName x) == "title"
+    dropTitle (x:xs) = if isTitle x
+                         then xs -- Drop note section <title> if present
+                         else x:xs
+    dropTitle [] = []
 
 -- | Parse a child of @\<FictionBook>@ element.
 parseFictionBookChild :: PandocMonad m => Element -> FB2 m Blocks
 parseFictionBookChild e =
-  case qName $ elName e of
+  case T.pack $ qName $ elName e of
     "stylesheet" -> pure mempty -- stylesheet is ignored
     "description" -> mempty <$ mapM_ parseDescriptionChild (elChildren e)
-    "body" -> mconcat <$> mapM parseBodyChild (elChildren e)
+    "body" -> if isNotesBody e
+                then pure mempty
+                else mconcat <$> mapM parseBodyChild (elChildren e)
     "binary" -> mempty <$ parseBinaryElement e
     name -> report (UnexpectedXmlElement name "FictionBook") $> mempty
 
 -- | Parse a child of @\<description>@ element.
 parseDescriptionChild :: PandocMonad m => Element -> FB2 m ()
 parseDescriptionChild e =
-  case qName $ elName e of
+  case T.pack $ qName $ elName e of
     "title-info" -> mapM_ parseTitleInfoChild (elChildren e)
     "src-title-info" -> pure () -- ignore
     "document-info" -> pure ()
     "publish-info" -> pure ()
     "custom-info" -> pure ()
     "output" -> pure ()
-    name -> throwError $ PandocParseError ("Couldn't parse FB2 file: unexpected element " ++ name ++ "in description.")
+    name -> do
+      report $ IgnoredElement $ name <> " in description"
+      pure mempty
 
 -- | Parse a child of @\<body>@ element.
 parseBodyChild :: PandocMonad m => Element -> FB2 m Blocks
 parseBodyChild e =
-  case qName $ elName e of
+  case T.pack $ qName $ elName e of
     "image" -> parseImageElement e
     "title" -> header <$> gets fb2SectionLevel <*> parseTitleType (elContent e)
     "epigraph" -> parseEpigraph e
@@ -162,27 +192,30 @@ parseBodyChild e =
 -- | Parse a @\<binary>@ element.
 parseBinaryElement :: PandocMonad m => Element -> FB2 m ()
 parseBinaryElement e =
-  case (findAttr (QName "id" Nothing Nothing) e, findAttr (QName "content-type" Nothing Nothing) e) of
-    (Nothing, _) -> throwError $ PandocParseError "<binary> element must have an \"id\" attribute"
-    (Just _, Nothing) -> throwError $ PandocParseError "<binary> element must have a \"content-type\" attribute"
-    (Just filename, contentType) -> insertMedia filename contentType (decodeLenient (pack (strContent e)))
+  case (findAttr (unqual "id") e, findAttr (unqual "content-type") e) of
+    (Nothing, _) -> report $ IgnoredElement "binary without id attribute"
+    (Just _, Nothing) ->
+      report $ IgnoredElement "binary without content-type attribute"
+    (Just filename, contentType) -> insertMedia filename (T.pack <$> contentType) (decodeLenient (pack (strContent e)))
 
 -- * Type parsers
 
 -- | Parse @authorType@
-parseAuthor :: PandocMonad m => Element -> FB2 m String
-parseAuthor e = unwords <$> mapM parseAuthorChild (elChildren e)
+parseAuthor :: PandocMonad m => Element -> FB2 m Text
+parseAuthor e = T.unwords . catMaybes <$> mapM parseAuthorChild (elChildren e)
 
-parseAuthorChild :: PandocMonad m => Element -> FB2 m String
+parseAuthorChild :: PandocMonad m => Element -> FB2 m (Maybe Text)
 parseAuthorChild e =
-  case qName $ elName e of
-    "first-name" -> pure $ strContent e
-    "middle-name" -> pure $ strContent e
-    "last-name" -> pure $ strContent e
-    "nickname" -> pure $ strContent e
-    "home-page" -> pure $ strContent e
-    "email" -> pure $ strContent e
-    name -> throwError $ PandocParseError ("Couldn't parse FB2 file: unexpected element " ++ name ++ " in author.")
+  case T.pack $ qName $ elName e of
+    "first-name" -> pure $ Just $ T.pack $ strContent e
+    "middle-name" -> pure $ Just $ T.pack $ strContent e
+    "last-name" -> pure $ Just $ T.pack $ strContent e
+    "nickname" -> pure $ Just $ T.pack $ strContent e
+    "home-page" -> pure $ Just $ T.pack $ strContent e
+    "email" -> pure $ Just $ T.pack $ strContent e
+    name -> do
+      report $ IgnoredElement $ name <> " in author"
+      pure Nothing
 
 -- | Parse @titleType@
 parseTitle :: PandocMonad m => Element -> FB2 m Blocks
@@ -203,11 +236,13 @@ parseTitleContent _ = pure Nothing
 parseImageElement :: PandocMonad m => Element -> FB2 m Blocks
 parseImageElement e =
   case href of
-    Just src -> pure $ para $ imageWith (imgId, [], []) (removeHash src) title alt
-    Nothing -> throwError $ PandocParseError "Couldn't parse FB2 file: image without href."
-  where alt = maybe mempty str $ findAttr (QName "alt" Nothing Nothing) e
-        title = fromMaybe "" $ findAttr (QName "title" Nothing Nothing) e
-        imgId = fromMaybe "" $ findAttr (QName "id" Nothing Nothing) e
+    Just src -> pure $ para $ imageWith (imgId, [], []) (removeHash $ T.pack src) title alt
+    Nothing -> do
+      report $ IgnoredElement " image without href"
+      pure mempty
+  where alt = maybe mempty (str . T.pack) $ findAttr (unqual "alt") e
+        title = maybe "" T.pack $ findAttr (unqual "title") e
+        imgId = maybe "" T.pack $ findAttr (unqual "id") e
         href = findAttr (QName "href" (Just "http://www.w3.org/1999/xlink") Nothing) e
 
 -- | Parse @pType@
@@ -221,7 +256,7 @@ parseCite e = blockQuote . mconcat <$> mapM parseCiteChild (elChildren e)
 -- | Parse @citeType@ child
 parseCiteChild :: PandocMonad m => Element -> FB2 m Blocks
 parseCiteChild e =
-  case qName $ elName e of
+  case T.pack $ qName $ elName e of
     "p" -> para <$> parsePType e
     "poem" -> parsePoem e
     "empty-line" -> pure horizontalRule
@@ -236,13 +271,13 @@ parsePoem e = mconcat <$> mapM parsePoemChild (elChildren e)
 
 parsePoemChild :: PandocMonad m => Element -> FB2 m Blocks
 parsePoemChild e =
-  case qName $ elName e of
+  case T.pack $ qName $ elName e of
     "title" -> parseTitle e
     "subtitle" -> parseSubtitle e
     "epigraph" -> parseEpigraph e
     "stanza" -> parseStanza e
     "text-author" -> para <$> parsePType e
-    "date" -> pure $ para $ text $ strContent e
+    "date" -> pure $ para $ text $ T.pack $ strContent e
     name -> report (UnexpectedXmlElement name "poem") $> mempty
 
 parseStanza :: PandocMonad m => Element -> FB2 m Blocks
@@ -255,7 +290,7 @@ joinLineBlocks [] = []
 
 parseStanzaChild :: PandocMonad m => Element -> FB2 m Blocks
 parseStanzaChild e =
-  case qName $ elName e of
+  case T.pack $ qName $ elName e of
     "title" -> parseTitle e
     "subtitle" -> parseSubtitle e
     "v" -> lineBlock . (:[]) <$> parsePType e
@@ -265,11 +300,11 @@ parseStanzaChild e =
 parseEpigraph :: PandocMonad m => Element -> FB2 m Blocks
 parseEpigraph e =
   divWith (divId, ["epigraph"], []) . mconcat <$> mapM parseEpigraphChild (elChildren e)
-  where divId = fromMaybe "" $ findAttr (QName "id" Nothing Nothing) e
+  where divId = maybe "" T.pack $ findAttr (unqual "id") e
 
 parseEpigraphChild :: PandocMonad m => Element -> FB2 m Blocks
 parseEpigraphChild e =
-  case qName $ elName e of
+  case T.pack $ qName $ elName e of
     "p" -> para <$> parsePType e
     "poem" -> parsePoem e
     "cite" -> parseCite e
@@ -283,7 +318,7 @@ parseAnnotation e = mconcat <$> mapM parseAnnotationChild (elChildren e)
 
 parseAnnotationChild :: PandocMonad m => Element -> FB2 m Blocks
 parseAnnotationChild e =
-  case qName $ elName e of
+  case T.pack $ qName $ elName e of
     "p" -> para <$> parsePType e
     "poem" -> parsePoem e
     "cite" -> parseCite e
@@ -297,14 +332,14 @@ parseSection :: PandocMonad m => Element -> FB2 m Blocks
 parseSection e = do
   n <- gets fb2SectionLevel
   modify $ \st -> st{ fb2SectionLevel = n + 1 }
-  let sectionId = fromMaybe "" $ findAttr (QName "id" Nothing Nothing) e
+  let sectionId = maybe "" T.pack $ findAttr (unqual "id") e
   bs <- divWith (sectionId, ["section"], []) . mconcat <$> mapM parseSectionChild (elChildren e)
   modify $ \st -> st{ fb2SectionLevel = n }
   pure bs
 
 parseSectionChild :: PandocMonad m => Element -> FB2 m Blocks
 parseSectionChild e =
-  case qName $ elName e of
+  case T.pack $ qName $ elName e of
     "title" -> parseBodyChild e
     "epigraph" -> parseEpigraph e
     "image" -> parseImageElement e
@@ -326,14 +361,16 @@ parseStyleType e = mconcat <$> mapM parseInline (elContent e)
 parseNamedStyle :: PandocMonad m => Element -> FB2 m Inlines
 parseNamedStyle e = do
   content <- mconcat <$> mapM parseNamedStyleChild (elContent e)
-  let lang = maybeToList $ ("lang",) <$> findAttr (QName "lang" Nothing (Just "xml")) e
-  case findAttr (QName "name" Nothing Nothing) e of
-    Just name -> pure $ spanWith ("", [name], lang) content
-    Nothing -> throwError $ PandocParseError "Couldn't parse FB2 file: link without required name."
+  let lang = maybeToList $ ("lang",) . T.pack <$> findAttr (QName "lang" Nothing (Just "xml")) e
+  case findAttr (unqual "name") e of
+    Just name -> pure $ spanWith ("", [T.pack name], lang) content
+    Nothing -> do
+      report $ IgnoredElement "link without required name"
+      pure mempty
 
 parseNamedStyleChild :: PandocMonad m => Content -> FB2 m Inlines
 parseNamedStyleChild (Elem e) =
-  case qName (elName e) of
+  case T.pack $ qName (elName e) of
     "strong" -> strong <$> parseStyleType e
     "emphasis" -> emph <$> parseStyleType e
     "style" -> parseNamedStyle e
@@ -341,24 +378,35 @@ parseNamedStyleChild (Elem e) =
     "strikethrough" -> strikeout <$> parseStyleType e
     "sub" -> subscript <$> parseStyleType e
     "sup" -> superscript <$> parseStyleType e
-    "code" -> pure $ code $ strContent e
+    "code" -> pure $ code $ T.pack $ strContent e
     "image" -> parseInlineImageElement e
-    name -> throwError $ PandocParseError ("Couldn't parse FB2 file: unexpected element " ++ name ++ ".")
+    name -> do
+      report $ IgnoredElement $ name <> " in style"
+      pure mempty
 parseNamedStyleChild x = parseInline x
 
 -- | Parse @linkType@
 parseLinkType :: PandocMonad m => Element -> FB2 m Inlines
 parseLinkType e = do
   content <- mconcat <$> mapM parseStyleLinkType (elContent e)
-  case findAttr (QName "href" (Just "http://www.w3.org/1999/xlink") Nothing) e of
-    Just href -> pure $ link href "" content
-    Nothing -> throwError $ PandocParseError "Couldn't parse FB2 file: link without required href."
+  notes <- gets fb2Notes
+  case T.pack <$> findAttr (QName "href" (Just "http://www.w3.org/1999/xlink") Nothing) e of
+    Just href -> case findAttr (QName "type" Nothing Nothing) e of
+                   Just "note" -> case M.lookup href notes of
+                                    Nothing -> pure $ link href "" content
+                                    Just contents -> pure $ note contents
+                   _ -> pure $ link href "" content
+    Nothing -> do
+      report $ IgnoredElement "link without required href"
+      pure mempty
 
 -- | Parse @styleLinkType@
 parseStyleLinkType :: PandocMonad m => Content -> FB2 m Inlines
 parseStyleLinkType x@(Elem e) =
   case qName (elName e) of
-    "a" -> throwError $ PandocParseError "Couldn't parse FB2 file: links cannot be nested."
+    "a" -> do
+      report $ IgnoredElement "nested link"
+      pure mempty
     _ -> parseInline x
 parseStyleLinkType x = parseInline x
 
@@ -369,19 +417,21 @@ parseTable _ = pure mempty -- TODO: tables are not supported yet
 -- | Parse @title-infoType@
 parseTitleInfoChild :: PandocMonad m => Element -> FB2 m ()
 parseTitleInfoChild e =
-  case qName (elName e) of
+  case T.pack $ qName (elName e) of
     "genre" -> pure ()
     "author" -> parseAuthor e >>= \author -> modify (\st -> st {fb2Authors = author:fb2Authors st})
-    "book-title" -> modify (setMeta "title" (text $ strContent e))
+    "book-title" -> modify (setMeta "title" (text $ T.pack $ strContent e))
     "annotation" -> parseAnnotation e >>= modify . setMeta "abstract"
-    "keywords" -> modify (setMeta "keywords" (map (MetaString . trim) $ splitOn "," $ strContent e))
-    "date" -> modify (setMeta "date" (text $ strContent e))
+    "keywords" -> modify (setMeta "keywords" (map (MetaString . trim) $ T.splitOn ","
+                                                                      $ T.pack
+                                                                      $ strContent e))
+    "date" -> modify (setMeta "date" (text $ T.pack $ strContent e))
     "coverpage" -> parseCoverPage e
     "lang" -> pure ()
     "src-lang" -> pure ()
     "translator" -> pure ()
     "sequence" -> pure ()
-    name -> throwError $ PandocParseError ("Couldn't parse FB2 file: unexpected element " ++ name ++ " in title-info.")
+    name -> report $ IgnoredElement $ name <> " in title-info"
 
 parseCoverPage :: PandocMonad m => Element -> FB2 m ()
 parseCoverPage e =
@@ -389,7 +439,7 @@ parseCoverPage e =
     Just img -> case href of
                   Just src -> modify (setMeta "cover-image" (MetaString $ removeHash src))
                   Nothing -> pure ()
-                where href = findAttr (QName "href" (Just "http://www.w3.org/1999/xlink") Nothing) img
+                where href = T.pack <$> findAttr (QName "href" (Just "http://www.w3.org/1999/xlink") Nothing) img
     Nothing -> pure ()
 
 -- | Parse @inlineImageType@ element
@@ -399,6 +449,8 @@ parseInlineImageElement :: PandocMonad m
 parseInlineImageElement e =
   case href of
     Just src -> pure $ imageWith ("", [], []) (removeHash src) "" alt
-    Nothing -> throwError $ PandocParseError "Couldn't parse FB2 file: inline image without href."
-  where alt = maybe mempty str $ findAttr (QName "alt" Nothing Nothing) e
-        href = findAttr (QName "href" (Just "http://www.w3.org/1999/xlink") Nothing) e
+    Nothing -> do
+      report $ IgnoredElement "inline image without href"
+      pure mempty
+  where alt = maybe mempty (str . T.pack) $ findAttr (unqual "alt") e
+        href = T.pack <$> findAttr (QName "href" (Just "http://www.w3.org/1999/xlink") Nothing) e
