@@ -1,26 +1,8 @@
-{-# LANGUAGE NoImplicitPrelude #-}
 {-# LANGUAGE TupleSections #-}
-{-
-Copyright (C) 2017-2018 Hamish Mackenzie
-
-This program is free software; you can redistribute it and/or modify
-it under the terms of the GNU General Public License as published by
-the Free Software Foundation; either version 2 of the License, or
-(at your option) any later version.
-
-This program is distributed in the hope that it will be useful,
-but WITHOUT ANY WARRANTY; without even the implied warranty of
-MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-GNU General Public License for more details.
-
-You should have received a copy of the GNU General Public License
-along with this program; if not, write to the Free Software
-Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--}
-
+{-# LANGUAGE OverloadedStrings #-}
 {- |
    Module      : Text.Pandoc.Readers.JATS
-   Copyright   : Copyright (C) 2017-2018 Hamish Mackenzie
+   Copyright   : Copyright (C) 2017-2020 Hamish Mackenzie
    License     : GNU GPL, version 2 or above
 
    Maintainer  : John MacFarlane <jgm@berkeley.edu>
@@ -31,7 +13,6 @@ Conversion of JATS XML to 'Pandoc' document.
 -}
 
 module Text.Pandoc.Readers.JATS ( readJATS ) where
-import Prelude
 import Control.Monad.State.Strict
 import Data.Char (isDigit, isSpace, toUpper)
 import Data.Default
@@ -43,7 +24,7 @@ import Data.Text (Text)
 import qualified Data.Text as T
 import Text.HTML.TagSoup.Entity (lookupEntity)
 import Text.Pandoc.Builder
-import Text.Pandoc.Class (PandocMonad)
+import Text.Pandoc.Class.PandocMonad (PandocMonad)
 import Text.Pandoc.Options
 import Text.Pandoc.Shared (underlineSpan, crFilter, safeRead)
 import Text.TeXMath (readMathML, writeTeX)
@@ -57,7 +38,6 @@ data JATSState = JATSState{ jatsSectionLevel :: Int
                           , jatsQuoteType    :: QuoteType
                           , jatsMeta         :: Meta
                           , jatsBook         :: Bool
-                          , jatsFigureTitle  :: Inlines
                           , jatsContent      :: [Content]
                           } deriving Show
 
@@ -66,7 +46,6 @@ instance Default JATSState where
                  , jatsQuoteType = DoubleQuote
                  , jatsMeta = mempty
                  , jatsBook = False
-                 , jatsFigureTitle = mempty
                  , jatsContent = [] }
 
 
@@ -96,13 +75,13 @@ convertEntity :: String -> String
 convertEntity e = Data.Maybe.fromMaybe (map toUpper e) (lookupEntity e)
 
 -- convenience function to get an attribute value, defaulting to ""
-attrValue :: String -> Element -> String
+attrValue :: String -> Element -> Text
 attrValue attr =
   fromMaybe "" . maybeAttrValue attr
 
-maybeAttrValue :: String -> Element -> Maybe String
+maybeAttrValue :: String -> Element -> Maybe Text
 maybeAttrValue attr elt =
-  lookupAttrBy (\x -> qName x == attr) (elAttribs elt)
+  T.pack <$> lookupAttrBy (\x -> qName x == attr) (elAttribs elt)
 
 -- convenience function
 named :: String -> Element -> Bool
@@ -110,7 +89,7 @@ named s e = qName (elName e) == s
 
 --
 
-addMeta :: PandocMonad m => ToMetaValue a => String -> a -> JATS m ()
+addMeta :: PandocMonad m => ToMetaValue a => Text -> a -> JATS m ()
 addMeta field val = modify (setMeta field val)
 
 instance HasMeta JATSState where
@@ -146,28 +125,23 @@ isBlockElement (Elem e) = qName (elName e) `S.member` blocktags
 isBlockElement _ = False
 
 -- Trim leading and trailing newline characters
-trimNl :: String -> String
-trimNl = reverse . go . reverse . go
-  where go ('\n':xs) = xs
-        go xs        = xs
+trimNl :: Text -> Text
+trimNl = T.dropAround (== '\n')
 
 -- function that is used by both graphic (in parseBlock)
 -- and inline-graphic (in parseInline)
-getGraphic :: PandocMonad m => Element -> JATS m Inlines
-getGraphic e = do
+getGraphic :: PandocMonad m
+           => Maybe (Inlines, Text) -> Element -> JATS m Inlines
+getGraphic mbfigdata e = do
   let atVal a = attrValue a e
-      attr = (atVal "id", words $ atVal "role", [])
+      (ident, title, caption) =
+         case mbfigdata of
+           Just (capt, i) -> (i, "fig:" <> atVal "title", capt)
+           Nothing        -> (atVal "id", atVal "title",
+                              text (atVal "alt-text"))
+      attr = (ident, T.words $ atVal "role", [])
       imageUrl = atVal "href"
-      captionOrLabel = case filterChild (\x -> named "caption" x
-                                            || named "label" x) e of
-                        Nothing -> return mempty
-                        Just z  -> mconcat <$>
-                                         mapM parseInline (elContent z)
-  figTitle <- gets jatsFigureTitle
-  let (caption, title) = if isNull figTitle
-                            then (captionOrLabel, atVal "title")
-                            else (return figTitle, "fig:")
-  fmap (imageWith attr imageUrl title) caption
+  return $ imageWith attr imageUrl title caption
 
 getBlocks :: PandocMonad m => Element -> JATS m Blocks
 getBlocks e =  mconcat <$>
@@ -178,8 +152,8 @@ parseBlock :: PandocMonad m => Content -> JATS m Blocks
 parseBlock (Text (CData CDataRaw _ _)) = return mempty -- DOCTYPE
 parseBlock (Text (CData _ s _)) = if all isSpace s
                                      then return mempty
-                                     else return $ plain $ trimInlines $ text s
-parseBlock (CRef x) = return $ plain $ str $ map toUpper x
+                                     else return $ plain $ trimInlines $ text $ T.pack s
+parseBlock (CRef x) = return $ plain $ str $ T.toUpper $ T.pack x
 parseBlock (Elem e) =
   case qName (elName e) of
         "p" -> parseMixed para (elContent e)
@@ -190,21 +164,24 @@ parseBlock (Elem e) =
                     "bullet" -> bulletList <$> listitems
                     listType -> do
                       let start = fromMaybe 1 $
-                                  (strContent <$> (filterElement (named "list-item") e
+                                  (textContent <$> (filterElement (named "list-item") e
                                                >>= filterElement (named "label")))
                                    >>= safeRead
                       orderedListWith (start, parseListStyleType listType, DefaultDelim)
                         <$> listitems
         "def-list" -> definitionList <$> deflistitems
         "sec" -> gets jatsSectionLevel >>= sect . (+1)
-        "graphic" -> para <$> getGraphic e
+        "graphic" -> para <$> getGraphic Nothing e
         "journal-meta" -> parseMetadata e
         "article-meta" -> parseMetadata e
         "custom-meta" -> parseMetadata e
         "title" -> return mempty -- processed by header
         "table" -> parseTable
-        "fig" -> divWith (attrValue "id" e, ["fig"], []) <$> getBlocks e
-        "table-wrap" -> divWith (attrValue "id" e, ["table-wrap"], []) <$> getBlocks e
+        "fig" -> parseFigure
+        "fig-group" -> divWith (attrValue "id" e, ["fig-group"], [])
+                          <$> getBlocks e
+        "table-wrap" -> divWith (attrValue "id" e, ["table-wrap"], [])
+                          <$> getBlocks e
         "caption" -> divWith (attrValue "id" e, ["caption"], []) <$> sect 6
         "ref-list" -> parseRefList e
         "?xml"  -> return mempty
@@ -224,7 +201,7 @@ parseBlock (Elem e) =
                                 "" -> []
                                 x  -> [x]
            return $ codeBlockWith (attrValue "id" e, classes', [])
-                  $ trimNl $ strContentRecursive e
+                  $ trimNl $ textContentRecursive e
          parseBlockquote = do
             attrib <- case filterChild (named "attribution") e of
                              Nothing  -> return mempty
@@ -247,6 +224,20 @@ parseBlock (Elem e) =
                      terms' <- mapM getInlines terms
                      items' <- mapM getBlocks items
                      return (mconcat $ intersperse (str "; ") terms', items')
+         parseFigure =
+           -- if a simple caption and single graphic, we emit a standard
+           -- implicit figure.  otherwise, we emit a div with the contents
+           case filterChildren (named "graphic") e of
+                  [g] -> do
+                         caption <- case filterChild (named "caption") e of
+                                           Just t  -> mconcat .
+                                             intersperse linebreak <$>
+                                             mapM getInlines
+                                             (filterChildren (const True) t)
+                                           Nothing -> return mempty
+                         img <- getGraphic (Just (caption, attrValue "id" e)) g
+                         return $ para img
+                  _   -> divWith (attrValue "id" e, ["fig"], []) <$> getBlocks e
          parseTable = do
                       let isCaption x = named "title" x || named "caption" x
                       caption <- case filterChild isCaption e of
@@ -273,9 +264,9 @@ parseBlock (Elem e) =
                                                 Just "right"  -> AlignRight
                                                 Just "center" -> AlignCenter
                                                 _             -> AlignDefault
-                      let toWidth c = case findAttr (unqual "colwidth") c of
+                      let toWidth c = case findAttrText (unqual "colwidth") c of
                                                 Just w -> fromMaybe 0
-                                                   $ safeRead $ '0': filter (\x ->
+                                                   $ safeRead $ "0" <> T.filter (\x ->
                                                      isDigit x || x == '.') w
                                                 Nothing -> 0 :: Double
                       let numrows = foldl' max 0 $ map length bodyrows
@@ -369,7 +360,7 @@ parseRefList e = do
   return mempty
 
 parseRef :: PandocMonad m
-         => Element -> JATS m (Map.Map String MetaValue)
+         => Element -> JATS m (Map.Map Text MetaValue)
 parseRef e = do
   let refId = text $ attrValue "id" e
   let getInlineText n = maybe (return mempty) getInlines . filterChild (named n)
@@ -402,7 +393,7 @@ parseRef e = do
                family <- maybe (return mempty) getInlines
                          $ filterChild (named "surname") nm
                return $ toMetaValue $ Map.fromList [
-                   ("given", given)
+                   ("given" :: Text, given)
                  , ("family", family)
                  ]
          personGroups <- mapM (\pg ->
@@ -412,7 +403,7 @@ parseRef e = do
                                            toMetaValue names))
                          personGroups'
          return $ Map.fromList $
-           [ ("id", toMetaValue refId)
+           [ ("id" :: Text, toMetaValue refId)
            , ("type", toMetaValue refType)
            , ("title", toMetaValue refTitle)
            , ("container-title", toMetaValue refContainerTitle)
@@ -421,7 +412,7 @@ parseRef e = do
            , ("title", toMetaValue refTitle)
            , ("issued", toMetaValue
                         $ Map.fromList [
-                            ("year", refYear)
+                            ("year" :: Text, refYear)
                           ])
            , ("volume", toMetaValue refVolume)
            , ("page", toMetaValue refPages)
@@ -430,6 +421,15 @@ parseRef e = do
        Nothing -> return $ Map.insert "id" (toMetaValue refId) mempty
        -- TODO handle mixed-citation
 
+findAttrText :: QName -> Element -> Maybe Text
+findAttrText x = fmap T.pack . findAttr x
+
+textContent :: Element -> Text
+textContent = T.pack . strContent
+
+textContentRecursive :: Element -> Text
+textContentRecursive = T.pack . strContentRecursive
+
 strContentRecursive :: Element -> String
 strContentRecursive = strContent .
   (\e' -> e'{ elContent = map elementToStr $ elContent e' })
@@ -439,9 +439,9 @@ elementToStr (Elem e') = Text $ CData CDataText (strContentRecursive e') Nothing
 elementToStr x = x
 
 parseInline :: PandocMonad m => Content -> JATS m Inlines
-parseInline (Text (CData _ s _)) = return $ text s
+parseInline (Text (CData _ s _)) = return $ text $ T.pack s
 parseInline (CRef ref) =
-  return $ maybe (text $ map toUpper ref) text $ lookupEntity ref
+  return . text . maybe (T.toUpper $ T.pack ref) T.pack $ lookupEntity ref
 parseInline (Elem e) =
   case qName (elName e) of
         "italic" -> emph <$> innerInlines
@@ -456,7 +456,7 @@ parseInline (Elem e) =
         "code" -> codeWithLang
         "monospace" -> codeWithLang
 
-        "inline-graphic" -> getGraphic e
+        "inline-graphic" -> getGraphic Nothing e
         "disp-quote" -> do
             qt <- gets jatsQuoteType
             let qt' = if qt == SingleQuote then DoubleQuote else SingleQuote
@@ -470,23 +470,26 @@ parseInline (Elem e) =
         "xref" -> do
             ils <- innerInlines
             let rid = attrValue "rid" e
+            let rids = T.words rid
             let refType = ("ref-type",) <$> maybeAttrValue "ref-type" e
             let attr = (attrValue "id" e, [], maybeToList refType)
             return $ if refType == Just ("ref-type","bibr")
-                        then cite [Citation{
-                                         citationId = rid
-                                       , citationPrefix = []
-                                       , citationSuffix = []
-                                       , citationMode = NormalCitation
-                                       , citationNoteNum = 0
-                                       , citationHash = 0}] ils
-                        else linkWith attr ('#' : rid) "" ils
+                        then cite
+                             (map (\id' ->
+                                     Citation{ citationId = id'
+                                             , citationPrefix = []
+                                             , citationSuffix = []
+                                             , citationMode = NormalCitation
+                                             , citationNoteNum = 0
+                                             , citationHash = 0}) rids)
+                             ils
+                        else linkWith attr ("#" <> rid) "" ils
         "ext-link" -> do
              ils <- innerInlines
-             let title = fromMaybe "" $ findAttr (QName "title" (Just "http://www.w3.org/1999/xlink") Nothing) e
+             let title = fromMaybe "" $ findAttrText (QName "title" (Just "http://www.w3.org/1999/xlink") Nothing) e
              let href = case findAttr (QName "href" (Just "http://www.w3.org/1999/xlink") Nothing) e of
-                               Just h -> h
-                               _      -> '#' : attrValue "rid" e
+                               Just h -> T.pack h
+                               _      -> "#" <> attrValue "rid" e
              let ils' = if ils == mempty then str href else ils
              let attr = (attrValue "id" e, [], [])
              return $ linkWith attr href title ils'
@@ -494,23 +497,23 @@ parseInline (Elem e) =
         "disp-formula" -> formula displayMath
         "inline-formula" -> formula math
         "math" | qPrefix (elName e) == Just "mml" -> return . math $ mathML e
-        "tex-math" -> return . math $ strContent e
+        "tex-math" -> return . math $ textContent e
 
-        "email" -> return $ link ("mailto:" ++ strContent e) ""
-                          $ str $ strContent e
-        "uri" -> return $ link (strContent e) "" $ str $ strContent e
+        "email" -> return $ link ("mailto:" <> textContent e) ""
+                          $ str $ textContent e
+        "uri" -> return $ link (textContent e) "" $ str $ textContent e
         "fn" -> (note . mconcat) <$>
                          mapM parseBlock (elContent e)
         _          -> innerInlines
    where innerInlines = (trimInlines . mconcat) <$>
                           mapM parseInline (elContent e)
          mathML x =
-            case readMathML . showElement $ everywhere (mkT removePrefix) x of
+            case readMathML . T.pack . showElement $ everywhere (mkT removePrefix) x of
                 Left _ -> mempty
                 Right m -> writeTeX m
          formula constructor = do
             let whereToLook = fromMaybe e $ filterElement (named "alternatives") e
-                texMaths = map strContent $
+                texMaths = map textContent $
                             filterChildren (named  "tex-math") whereToLook
                 mathMLs = map mathML $
                             filterChildren isMathML whereToLook
@@ -523,4 +526,4 @@ parseInline (Elem e) =
            let classes' = case attrValue "language" e of
                                "" -> []
                                l  -> [l]
-           return $ codeWith (attrValue "id" e,classes',[]) $ strContentRecursive e
+           return $ codeWith (attrValue "id" e,classes',[]) $ textContentRecursive e
