@@ -1,29 +1,10 @@
-{-# LANGUAGE NoImplicitPrelude #-}
 {-# LANGUAGE FlexibleInstances     #-}
 {-# LANGUAGE MultiParamTypeClasses #-}
 {-# LANGUAGE OverloadedStrings     #-}
 {-# LANGUAGE ScopedTypeVariables   #-}
-{-
-Copyright (C) 2006-2018 John MacFarlane <jgm@berkeley.edu>
-
-This program is free software; you can redistribute it and/or modify
-it under the terms of the GNU General Public License as published by
-the Free Software Foundation; either version 2 of the License, or
-(at your option) any later version.
-
-This program is distributed in the hope that it will be useful,
-but WITHOUT ANY WARRANTY; without even the implied warranty of
-MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-GNU General Public License for more details.
-
-You should have received a copy of the GNU General Public License
-along with this program; if not, write to the Free Software
-Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--}
-
 {- |
    Module      : Text.Pandoc.Readers.LaTeX.Parsing
-   Copyright   : Copyright (C) 2006-2018 John MacFarlane
+   Copyright   : Copyright (C) 2006-2020 John MacFarlane
    License     : GNU GPL, version 2 or above
 
    Maintainer  : John MacFarlane <jgm@berkeley.edu>
@@ -87,7 +68,6 @@ module Text.Pandoc.Readers.LaTeX.Parsing
   , withRaw
   ) where
 
-import Prelude
 import Control.Applicative (many, (<|>))
 import Control.Monad
 import Control.Monad.Except (throwError)
@@ -100,7 +80,7 @@ import qualified Data.Set as Set
 import Data.Text (Text)
 import qualified Data.Text as T
 import Text.Pandoc.Builder
-import Text.Pandoc.Class (PandocMonad, report)
+import Text.Pandoc.Class.PandocMonad (PandocMonad, report)
 import Text.Pandoc.Error (PandocError (PandocMacroLoop))
 import Text.Pandoc.Logging
 import Text.Pandoc.Options
@@ -110,14 +90,13 @@ import Text.Pandoc.Readers.LaTeX.Types (ExpansionPoint (..), Macro (..),
                                         ArgSpec (..), Tok (..), TokType (..))
 import Text.Pandoc.Shared
 import Text.Parsec.Pos
-
--- import Debug.Trace (traceShowId)
+-- import Debug.Trace
 
 newtype DottedNum = DottedNum [Int]
   deriving (Show)
 
-renderDottedNum :: DottedNum -> String
-renderDottedNum (DottedNum xs) =
+renderDottedNum :: DottedNum -> T.Text
+renderDottedNum (DottedNum xs) = T.pack $
   intercalate "." (map show xs)
 
 incrementDottedNum :: Int -> DottedNum -> DottedNum
@@ -130,19 +109,20 @@ data LaTeXState = LaTeXState{ sOptions       :: ReaderOptions
                             , sMeta          :: Meta
                             , sQuoteContext  :: QuoteContext
                             , sMacros        :: M.Map Text Macro
-                            , sContainers    :: [String]
-                            , sHeaders       :: M.Map Inlines String
+                            , sContainers    :: [Text]
                             , sLogMessages   :: [LogMessage]
-                            , sIdentifiers   :: Set.Set String
+                            , sIdentifiers   :: Set.Set Text
                             , sVerbatimMode  :: Bool
-                            , sCaption       :: (Maybe Inlines, Maybe String)
+                            , sCaption       :: Maybe Inlines
                             , sInListItem    :: Bool
                             , sInTableCell   :: Bool
                             , sLastHeaderNum :: DottedNum
                             , sLastFigureNum :: DottedNum
-                            , sLabels        :: M.Map String [Inline]
+                            , sLastTableNum  :: DottedNum
+                            , sLastLabel     :: Maybe Text
+                            , sLabels        :: M.Map Text [Inline]
                             , sHasChapters   :: Bool
-                            , sToggles       :: M.Map String Bool
+                            , sToggles       :: M.Map Text Bool
                             , sExpanded      :: Bool
                             }
      deriving Show
@@ -153,15 +133,16 @@ defaultLaTeXState = LaTeXState{ sOptions       = def
                               , sQuoteContext  = NoQuote
                               , sMacros        = M.empty
                               , sContainers    = []
-                              , sHeaders       = M.empty
                               , sLogMessages   = []
                               , sIdentifiers   = Set.empty
                               , sVerbatimMode  = False
-                              , sCaption       = (Nothing, Nothing)
+                              , sCaption       = Nothing
                               , sInListItem    = False
                               , sInTableCell   = False
                               , sLastHeaderNum = DottedNum []
                               , sLastFigureNum = DottedNum []
+                              , sLastTableNum  = DottedNum []
+                              , sLastLabel     = Nothing
                               , sLabels        = M.empty
                               , sHasChapters   = False
                               , sToggles       = M.empty
@@ -192,10 +173,6 @@ instance HasIncludeFiles LaTeXState where
   addIncludeFile f s = s{ sContainers = f : sContainers s }
   dropLatestIncludeFile s = s { sContainers = drop 1 $ sContainers s }
 
-instance HasHeaderMap LaTeXState where
-  extractHeaderMap     = sHeaders
-  updateHeaderMap f st = st{ sHeaders = f $ sHeaders st }
-
 instance HasMacros LaTeXState where
   extractMacros  st  = sMacros st
   updateMacros f st  = st{ sMacros = f (sMacros st) }
@@ -226,10 +203,9 @@ withVerbatimMode parser = do
        return result
 
 rawLaTeXParser :: (PandocMonad m, HasMacros s, HasReaderOptions s)
-               => Bool -> LP m a -> LP m a -> ParserT String s m (a, String)
-rawLaTeXParser retokenize parser valParser = do
-  inp <- getInput
-  let toks = tokenize "source" $ T.pack inp
+               => [Tok] -> Bool -> LP m a -> LP m a
+               -> ParserT Text s m (a, Text)
+rawLaTeXParser toks retokenize parser valParser = do
   pstate <- getState
   let lstate = def{ sOptions = extractReaderOptions pstate }
   let lstate' = lstate { sMacros = extractMacros pstate }
@@ -250,18 +226,27 @@ rawLaTeXParser retokenize parser valParser = do
               Right ((val, raw), st) -> do
                 updateState (updateMacros (sMacros st <>))
                 _ <- takeP (T.length (untokenize toks'))
-                return (val, T.unpack (untokenize raw))
+                let result = untokenize raw
+                -- ensure we end with space if input did, see #4442
+                let result' =
+                      case reverse toks' of
+                        (Tok _ (CtrlSeq _) t : _)
+                         | " " `T.isSuffixOf` t
+                         , not (" " `T.isSuffixOf` result)
+                          -> result <> " "
+                        _ -> result
+                return (val, result')
 
 applyMacros :: (PandocMonad m, HasMacros s, HasReaderOptions s)
-            => String -> ParserT String s m String
+            => Text -> ParserT Text s m Text
 applyMacros s = (guardDisabled Ext_latex_macros >> return s) <|>
-   do let retokenize = toksToString <$> many (satisfyTok (const True))
+   do let retokenize = untokenize <$> many (satisfyTok (const True))
       pstate <- getState
       let lstate = def{ sOptions = extractReaderOptions pstate
                       , sMacros  = extractMacros pstate }
-      res <- runParserT retokenize lstate "math" (tokenize "math" (T.pack s))
+      res <- runParserT retokenize lstate "math" (tokenize "math" s)
       case res of
-           Left e   -> fail (show e)
+           Left e   -> Prelude.fail (show e)
            Right s' -> return s'
 
 tokenize :: SourceName -> Text -> [Tok]
@@ -324,7 +309,7 @@ totoks pos t =
                       : totoks (incSourceColumn pos 2) rest'
          | c == '#' ->
            let (t1, t2) = T.span (\d -> d >= '0' && d <= '9') rest
-           in  case safeRead (T.unpack t1) of
+           in  case safeRead t1 of
                     Just i ->
                        Tok pos (Arg i) ("#" <> t1)
                        : totoks (incSourceColumn pos (1 + T.length t1)) t2
@@ -368,10 +353,21 @@ isLowerHex :: Char -> Bool
 isLowerHex x = x >= '0' && x <= '9' || x >= 'a' && x <= 'f'
 
 untokenize :: [Tok] -> Text
-untokenize = mconcat . map untoken
+untokenize = foldr untokenAccum mempty
+
+untokenAccum :: Tok -> Text -> Text
+untokenAccum (Tok _ (CtrlSeq _) t) accum =
+  -- insert space to prevent breaking a control sequence; see #5836
+  case (T.unsnoc t, T.uncons accum) of
+    (Just (_,c), Just (d,_))
+      | isLetter c
+      , isLetter d
+      -> t <> " " <> accum
+    _ -> t <> accum
+untokenAccum (Tok _ _ t) accum = t <> accum
 
 untoken :: Tok -> Text
-untoken (Tok _ _ t) = t
+untoken t = untokenAccum t mempty
 
 toksToString :: [Tok] -> String
 toksToString = T.unpack . untokenize
@@ -397,7 +393,7 @@ doMacros = do
       updateState $ \st -> st{ sExpanded = True }
 
 doMacros' :: PandocMonad m => Int -> [Tok] -> LP m [Tok]
-doMacros' n inp = do
+doMacros' n inp =
   case inp of
      Tok spos (CtrlSeq "begin") _ : Tok _ Symbol "{" :
       Tok _ Word name : Tok _ Symbol "}" : ts
@@ -453,7 +449,7 @@ doMacros' n inp = do
 
     handleMacros n' spos name ts = do
       when (n' > 20)  -- detect macro expansion loops
-        $ throwError $ PandocMacroLoop (T.unpack name)
+        $ throwError $ PandocMacroLoop name
       macros <- sMacros <$> getState
       case M.lookup name macros of
            Nothing -> mzero
@@ -463,13 +459,13 @@ doMacros' n inp = do
                              Nothing -> getargs M.empty argspecs
                              Just o  -> do
                                 x <- option o bracketedToks
-                                getargs (M.singleton 1 x) argspecs
+                                getargs (M.singleton 1 x) $ drop 1 argspecs
                    rest <- getInput
                    return (args, rest)
              lstate <- getState
              res <- lift $ runParserT getargs' lstate "args" ts
              case res of
-               Left _ -> fail $ "Could not parse arguments for " ++
+               Left _ -> Prelude.fail $ "Could not parse arguments for " ++
                                 T.unpack name
                Right (args, rest) -> do
                  -- first boolean param is true if we're tokenizing
@@ -534,7 +530,9 @@ symbolIn cs = satisfyTok isInCs
         isInCs _ = False
 
 sp :: PandocMonad m => LP m ()
-sp = whitespace <|> endline
+sp = do
+  optional $ skipMany (whitespace <|> comment)
+  optional $ endline  *> skipMany (whitespace <|> comment)
 
 whitespace :: PandocMonad m => LP m ()
 whitespace = () <$ satisfyTok isSpaceTok
@@ -593,15 +591,15 @@ primEscape = do
                     Just (c, _)
                       | c >= '\64' && c <= '\127' -> return (chr (ord c - 64))
                       | otherwise                 -> return (chr (ord c + 64))
-                    Nothing -> fail "Empty content of Esc1"
-       Esc2 -> case safeRead ('0':'x':T.unpack (T.drop 2 t)) of
+                    Nothing -> Prelude.fail "Empty content of Esc1"
+       Esc2 -> case safeRead ("0x" <> T.drop 2 t) of
                     Just x  -> return (chr x)
-                    Nothing -> fail $ "Could not read: " ++ T.unpack t
-       _    -> fail "Expected an Esc1 or Esc2 token" -- should not happen
+                    Nothing -> Prelude.fail $ "Could not read: " ++ T.unpack t
+       _    -> Prelude.fail "Expected an Esc1 or Esc2 token" -- should not happen
 
 bgroup :: PandocMonad m => LP m Tok
 bgroup = try $ do
-  skipMany sp
+  optional sp
   symbol '{' <|> controlSeq "bgroup" <|> controlSeq "begingroup"
 
 egroup :: PandocMonad m => LP m Tok
@@ -618,19 +616,19 @@ braced' :: PandocMonad m => LP m Tok -> Int -> LP m [Tok]
 braced' getTok n =
   handleEgroup <|> handleBgroup <|> handleOther
   where handleEgroup = do
-          t <- egroup
+          t <- symbol '}'
           if n == 1
              then return []
              else (t:) <$> braced' getTok (n - 1)
         handleBgroup = do
-          t <- bgroup
+          t <- symbol '{'
           (t:) <$> braced' getTok (n + 1)
         handleOther = do
           t <- getTok
           (t:) <$> braced' getTok n
 
 braced :: PandocMonad m => LP m [Tok]
-braced = bgroup *> braced' anyTok 1
+braced = symbol '{' *> braced' anyTok 1
 
 -- URLs require special handling, because they can contain %
 -- characters.  So we retonenize comments as we go...
@@ -659,7 +657,8 @@ bracketed parser = try $ do
 bracketedToks :: PandocMonad m => LP m [Tok]
 bracketedToks = do
   symbol '['
-  mconcat <$> manyTill (braced <|> (:[]) <$> anyTok) (symbol ']')
+  concat <$> manyTill ((snd <$> withRaw (try braced)) <|> count 1 anyTok)
+                      (symbol ']')
 
 parenWrapped :: PandocMonad m => Monoid a => LP m a -> LP m a
 parenWrapped parser = try $ do
@@ -670,20 +669,19 @@ dimenarg :: PandocMonad m => LP m Text
 dimenarg = try $ do
   optional sp
   ch  <- option False $ True <$ symbol '='
+  minus <- option "" $ "-" <$ symbol '-'
   Tok _ _ s1 <- satisfyTok isWordTok
   s2 <- option "" $ try $ do
           symbol '.'
           Tok _ _ t <-  satisfyTok isWordTok
           return ("." <> t)
   let s = s1 <> s2
-  guard $ T.takeEnd 2 s `elem`
-           ["pt","pc","in","bp","cm","mm","dd","cc","sp"]
-  let num = T.dropEnd 2 s
+  let (num, rest) = T.span (\c -> isDigit c || c == '.') s
   guard $ T.length num > 0
-  guard $ T.all (\c -> isDigit c || c == '.') num
-  return $ T.pack ['=' | ch] <> s
+  guard $ rest `elem` ["", "pt","pc","in","bp","cm","mm","dd","cc","sp"]
+  return $ T.pack ['=' | ch] <> minus <> s
 
-ignore :: (Monoid a, PandocMonad m) => String -> ParserT s u m a
+ignore :: (Monoid a, PandocMonad m) => Text -> ParserT s u m a
 ignore raw = do
   pos <- getPosition
   report $ SkippedContent raw pos
@@ -693,6 +691,7 @@ withRaw :: PandocMonad m => LP m a -> LP m (a, [Tok])
 withRaw parser = do
   inp <- getInput
   result <- parser
-  nxt <- option (Tok (initialPos "source") Word "") (lookAhead anyTok)
-  let raw = takeWhile (/= nxt) inp
+  nxtpos <- option Nothing ((\(Tok pos' _ _) -> Just pos') <$> lookAhead anyTok)
+  let raw = takeWhile (\(Tok pos _ _) -> maybe True
+                  (\p -> sourceName p /= sourceName pos || pos < p) nxtpos) inp
   return (result, raw)
