@@ -1,27 +1,9 @@
-{-# LANGUAGE NoImplicitPrelude #-}
-{-# LANGUAGE RelaxedPolyRec #-}
+{-# LANGUAGE OverloadedStrings #-}
+{-# LANGUAGE RelaxedPolyRec    #-}
 -- RelaxedPolyRec needed for inlinesBetween on GHC < 7
-{-
-  Copyright (C) 2012-2018 John MacFarlane <jgm@berkeley.edu>
-
-This program is free software; you can redistribute it and/or modify
-it under the terms of the GNU General Public License as published by
-the Free Software Foundation; either version 2 of the License, or
-(at your option) any later version.
-
-This program is distributed in the hope that it will be useful,
-but WITHOUT ANY WARRANTY; without even the implied warranty of
-MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-GNU General Public License for more details.
-
-You should have received a copy of the GNU General Public License
-along with this program; if not, write to the Free Software
-Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--}
-
 {- |
    Module      : Text.Pandoc.Readers.MediaWiki
-   Copyright   : Copyright (C) 2012-2018 John MacFarlane
+   Copyright   : Copyright (C) 2012-2020 John MacFarlane
    License     : GNU GPL, version 2 or above
 
    Maintainer  : John MacFarlane <jgm@berkeley.edu>
@@ -37,28 +19,27 @@ _ parse templates?
 -}
 module Text.Pandoc.Readers.MediaWiki ( readMediaWiki ) where
 
-import Prelude
 import Control.Monad
 import Control.Monad.Except (throwError)
 import Data.Char (isDigit, isSpace)
 import qualified Data.Foldable as F
-import Data.List (intercalate, intersperse, isPrefixOf)
-import qualified Data.Map as M
+import Data.List (intersperse)
 import Data.Maybe (fromMaybe, maybeToList)
 import Data.Sequence (ViewL (..), viewl, (<|))
 import qualified Data.Set as Set
-import Data.Text (Text, unpack)
+import Data.Text (Text)
+import qualified Data.Text as T
 import Text.HTML.TagSoup
 import Text.Pandoc.Builder (Blocks, Inlines, trimInlines)
 import qualified Text.Pandoc.Builder as B
-import Text.Pandoc.Class (PandocMonad (..))
+import Text.Pandoc.Class.PandocMonad (PandocMonad (..))
 import Text.Pandoc.Definition
 import Text.Pandoc.Logging
 import Text.Pandoc.Options
 import Text.Pandoc.Parsing hiding (nested)
 import Text.Pandoc.Readers.HTML (htmlTag, isBlockTag, isCommentTag)
 import Text.Pandoc.Shared (crFilter, safeRead, stringify, stripTrailingNewlines,
-                           trim)
+                           trim, splitTextBy, tshow)
 import Text.Pandoc.Walk (walk)
 import Text.Pandoc.XML (fromEntities)
 
@@ -72,12 +53,11 @@ readMediaWiki opts s = do
                                             , mwMaxNestingLevel = 4
                                             , mwNextLinkNumber  = 1
                                             , mwCategoryLinks = []
-                                            , mwHeaderMap = M.empty
                                             , mwIdentifierList = Set.empty
                                             , mwLogMessages = []
                                             , mwInTT = False
                                             }
-            (unpack (crFilter s) ++ "\n")
+            (crFilter s <> "\n")
   case parsed of
     Right result -> return result
     Left e       -> throwError e
@@ -86,21 +66,16 @@ data MWState = MWState { mwOptions         :: ReaderOptions
                        , mwMaxNestingLevel :: Int
                        , mwNextLinkNumber  :: Int
                        , mwCategoryLinks   :: [Inlines]
-                       , mwHeaderMap       :: M.Map Inlines String
-                       , mwIdentifierList  :: Set.Set String
+                       , mwIdentifierList  :: Set.Set Text
                        , mwLogMessages     :: [LogMessage]
                        , mwInTT            :: Bool
                        }
 
-type MWParser m = ParserT [Char] MWState m
+type MWParser m = ParserT Text MWState m
 
 instance HasReaderOptions MWState where
   extractReaderOptions = mwOptions
 
-instance HasHeaderMap MWState where
-  extractHeaderMap     = mwHeaderMap
-  updateHeaderMap f st = st{ mwHeaderMap = f $ mwHeaderMap st }
-
 instance HasIdentifierList MWState where
   extractIdentifierList     = mwIdentifierList
   updateIdentifierList f st = st{ mwIdentifierList = f $ mwIdentifierList st }
@@ -130,58 +105,58 @@ specialChars = "'[]<=&*{}|\":\\"
 spaceChars :: [Char]
 spaceChars = " \n\t"
 
-sym :: PandocMonad m => String -> MWParser m ()
-sym s = () <$ try (string s)
+sym :: PandocMonad m => Text -> MWParser m ()
+sym s = () <$ try (string $ T.unpack s)
 
-newBlockTags :: [String]
+newBlockTags :: [Text]
 newBlockTags = ["haskell","syntaxhighlight","source","gallery","references"]
 
-isBlockTag' :: Tag String -> Bool
+isBlockTag' :: Tag Text -> Bool
 isBlockTag' tag@(TagOpen t _) = (isBlockTag tag || t `elem` newBlockTags) &&
   t `notElem` eitherBlockOrInline
 isBlockTag' tag@(TagClose t) = (isBlockTag tag || t `elem` newBlockTags) &&
   t `notElem` eitherBlockOrInline
 isBlockTag' tag = isBlockTag tag
 
-isInlineTag' :: Tag String -> Bool
+isInlineTag' :: Tag Text -> Bool
 isInlineTag' (TagComment _) = True
 isInlineTag' t              = not (isBlockTag' t)
 
-eitherBlockOrInline :: [String]
+eitherBlockOrInline :: [Text]
 eitherBlockOrInline = ["applet", "button", "del", "iframe", "ins",
                                "map", "area", "object"]
 
 htmlComment :: PandocMonad m => MWParser m ()
 htmlComment = () <$ htmlTag isCommentTag
 
-inlinesInTags :: PandocMonad m => String -> MWParser m Inlines
+inlinesInTags :: PandocMonad m => Text -> MWParser m Inlines
 inlinesInTags tag = try $ do
   (_,raw) <- htmlTag (~== TagOpen tag [])
-  if '/' `elem` raw   -- self-closing tag
+  if T.any (== '/') raw   -- self-closing tag
      then return mempty
      else trimInlines . mconcat <$>
             manyTill inline (htmlTag (~== TagClose tag))
 
-blocksInTags :: PandocMonad m => String -> MWParser m Blocks
+blocksInTags :: PandocMonad m => Text -> MWParser m Blocks
 blocksInTags tag = try $ do
   (_,raw) <- htmlTag (~== TagOpen tag [])
   let closer = if tag == "li"
-                  then htmlTag (~== TagClose "li")
+                  then htmlTag (~== TagClose ("li" :: Text))
                      <|> lookAhead (
-                              htmlTag (~== TagOpen "li" [])
-                          <|> htmlTag (~== TagClose "ol")
-                          <|> htmlTag (~== TagClose "ul"))
+                              htmlTag (~== TagOpen ("li" :: Text) [])
+                          <|> htmlTag (~== TagClose ("ol" :: Text))
+                          <|> htmlTag (~== TagClose ("ul" :: Text)))
                   else htmlTag (~== TagClose tag)
-  if '/' `elem` raw   -- self-closing tag
+  if T.any (== '/') raw   -- self-closing tag
      then return mempty
      else mconcat <$> manyTill block closer
 
-charsInTags :: PandocMonad m => String -> MWParser m [Char]
-charsInTags tag = try $ do
+textInTags :: PandocMonad m => Text -> MWParser m Text
+textInTags tag = try $ do
   (_,raw) <- htmlTag (~== TagOpen tag [])
-  if '/' `elem` raw   -- self-closing tag
+  if T.any (== '/') raw   -- self-closing tag
      then return ""
-     else manyTill anyChar (htmlTag (~== TagClose tag))
+     else T.pack <$> manyTill anyChar (htmlTag (~== TagClose tag))
 
 --
 -- main parser
@@ -217,7 +192,7 @@ block = do
      <|> blockTag
      <|> (B.rawBlock "mediawiki" <$> template)
      <|> para
-  trace (take 60 $ show $ B.toList res)
+  trace (T.take 60 $ tshow $ B.toList res)
   return res
 
 para :: PandocMonad m => MWParser m Blocks
@@ -233,6 +208,7 @@ table = do
   styles <- option [] $
                parseAttrs <* skipMany spaceChar <* optional (char '|')
   skipMany spaceChar
+  optional $ template >> skipMany spaceChar
   optional blanklines
   let tableWidth = case lookup "width" styles of
                          Just w  -> fromMaybe 1.0 $ parseWidth w
@@ -258,16 +234,16 @@ table = do
                           else (replicate cols mempty, hdr:rows')
   return $ B.table caption cellspecs headers rows
 
-parseAttrs :: PandocMonad m => MWParser m [(String,String)]
+parseAttrs :: PandocMonad m => MWParser m [(Text,Text)]
 parseAttrs = many1 parseAttr
 
-parseAttr :: PandocMonad m => MWParser m (String, String)
+parseAttr :: PandocMonad m => MWParser m (Text, Text)
 parseAttr = try $ do
   skipMany spaceChar
-  k <- many1 letter
+  k <- many1Char letter
   char '='
-  v <- (char '"' >> many1Till (satisfy (/='\n')) (char '"'))
-       <|> many1 (satisfy $ \c -> not (isSpace c) && c /= '|')
+  v <- (char '"' >> many1TillChar (satisfy (/='\n')) (char '"'))
+       <|> many1Char (satisfy $ \c -> not (isSpace c) && c /= '|')
   return (k,v)
 
 tableStart :: PandocMonad m => MWParser m ()
@@ -278,7 +254,7 @@ tableEnd = try $ guardColumnOne *> skipSpaces *> sym "|}"
 
 rowsep :: PandocMonad m => MWParser m ()
 rowsep = try $ guardColumnOne *> skipSpaces *> sym "|-" <*
-               many (char '-') <* optional parseAttr <* blanklines
+               many (char '-') <* optional parseAttrs <* blanklines
 
 cellsep :: PandocMonad m => MWParser m ()
 cellsep = try $ do
@@ -302,7 +278,7 @@ tableCaption = try $ do
   guardColumnOne
   skipSpaces
   sym "|+"
-  optional (try $ parseAttr *> skipSpaces *> char '|' *> blanklines)
+  optional (try $ parseAttrs *> skipSpaces *> char '|' *> blanklines)
   (trimInlines . mconcat) <$>
     many (notFollowedBy (cellsep <|> rowsep) *> inline)
 
@@ -317,8 +293,8 @@ tableCell = try $ do
                                  notFollowedBy (char '|')
   skipMany spaceChar
   pos' <- getPosition
-  ls <- concat <$> many (notFollowedBy (cellsep <|> rowsep <|> tableEnd) *>
-                     ((snd <$> withRaw table) <|> count 1 anyChar))
+  ls <- T.concat <$> many (notFollowedBy (cellsep <|> rowsep <|> tableEnd) *>
+                            ((snd <$> withRaw table) <|> countChar 1 anyChar))
   bs <- parseFromString (do setPosition pos'
                             mconcat <$> many block) ls
   let align = case lookup "align" attrs of
@@ -331,48 +307,49 @@ tableCell = try $ do
                     Nothing -> 0.0
   return ((align, width), bs)
 
-parseWidth :: String -> Maybe Double
+parseWidth :: Text -> Maybe Double
 parseWidth s =
-  case reverse s of
-       ('%':ds) | all isDigit ds -> safeRead ('0':'.':reverse ds)
-       _        -> Nothing
+  case T.unsnoc s of
+    Just (ds, '%') | T.all isDigit ds -> safeRead $ "0." <> ds
+    _ -> Nothing
 
-template :: PandocMonad m => MWParser m String
+template :: PandocMonad m => MWParser m Text
 template = try $ do
   string "{{"
   notFollowedBy (char '{')
   lookAhead $ letter <|> digit <|> char ':'
-  let chunk = template <|> variable <|> many1 (noneOf "{}") <|> count 1 anyChar
+  let chunk = template <|> variable <|> many1Char (noneOf "{}") <|> countChar 1 anyChar
   contents <- manyTill chunk (try $ string "}}")
-  return $ "{{" ++ concat contents ++ "}}"
+  return $ "{{" <> T.concat contents <> "}}"
 
 blockTag :: PandocMonad m => MWParser m Blocks
 blockTag = do
   (tag, _) <- lookAhead $ htmlTag isBlockTag'
   case tag of
       TagOpen "blockquote" _ -> B.blockQuote <$> blocksInTags "blockquote"
-      TagOpen "pre" _ -> B.codeBlock . trimCode <$> charsInTags "pre"
+      TagOpen "pre" _ -> B.codeBlock . trimCode <$> textInTags "pre"
       TagOpen "syntaxhighlight" attrs -> syntaxhighlight "syntaxhighlight" attrs
       TagOpen "source" attrs -> syntaxhighlight "source" attrs
       TagOpen "haskell" _ -> B.codeBlockWith ("",["haskell"],[]) . trimCode <$>
-                                charsInTags "haskell"
+                                textInTags "haskell"
       TagOpen "gallery" _ -> blocksInTags "gallery"
       TagOpen "p" _ -> mempty <$ htmlTag (~== tag)
       TagClose "p"  -> mempty <$ htmlTag (~== tag)
       _ -> B.rawBlock "html" . snd <$> htmlTag (~== tag)
 
-trimCode :: String -> String
-trimCode ('\n':xs) = stripTrailingNewlines xs
-trimCode xs        = stripTrailingNewlines xs
+trimCode :: Text -> Text
+trimCode t = case T.uncons t of
+  Just ('\n', xs) -> stripTrailingNewlines xs
+  _               -> stripTrailingNewlines t
 
-syntaxhighlight :: PandocMonad m => String -> [Attribute String] -> MWParser m Blocks
+syntaxhighlight :: PandocMonad m => Text -> [Attribute Text] -> MWParser m Blocks
 syntaxhighlight tag attrs = try $ do
   let mblang = lookup "lang" attrs
   let mbstart = lookup "start" attrs
   let mbline = lookup "line" attrs
   let classes = maybeToList mblang ++ maybe [] (const ["numberLines"]) mbline
   let kvs = maybe [] (\x -> [("startFrom",x)]) mbstart
-  contents <- charsInTags tag
+  contents <- textInTags tag
   return $ B.codeBlockWith ("",classes,kvs) $ trimCode contents
 
 hrule :: PandocMonad m => MWParser m Blocks
@@ -386,17 +363,17 @@ preformatted = try $ do
   guardColumnOne
   char ' '
   let endline' = B.linebreak <$ try (newline <* char ' ')
-  let whitespace' = B.str <$> many1 ('\160' <$ spaceChar)
+  let whitespace' = B.str <$> many1Char ('\160' <$ spaceChar)
   let spToNbsp ' ' = '\160'
       spToNbsp x   = x
   let nowiki' = mconcat . intersperse B.linebreak . map B.str .
-                lines . fromEntities . map spToNbsp <$> try
-                  (htmlTag (~== TagOpen "nowiki" []) *>
-                   manyTill anyChar (htmlTag (~== TagClose "nowiki")))
+                T.lines . fromEntities . T.map spToNbsp <$> try
+                  (htmlTag (~== TagOpen ("nowiki" :: Text) []) *>
+                   manyTillChar anyChar (htmlTag (~== TagClose ("nowiki" :: Text))))
   let inline' = whitespace' <|> endline' <|> nowiki'
                   <|> try (notFollowedBy newline *> inline)
   contents <- mconcat <$> many1 inline'
-  let spacesStr (Str xs) = all isSpace xs
+  let spacesStr (Str xs) = T.all isSpace xs
       spacesStr _        = False
   if F.all spacesStr contents
      then return mempty
@@ -409,7 +386,7 @@ encode = B.fromList . normalizeCode . B.toList . walk strToCode
         strToCode  x      = x
         normalizeCode []  = []
         normalizeCode (Code a1 x : Code a2 y : zs) | a1 == a2 =
-          normalizeCode $ Code a1 (x ++ y) : zs
+          normalizeCode $ Code a1 (x <> y) : zs
         normalizeCode (x:xs) = x : normalizeCode xs
 
 header :: PandocMonad m => MWParser m Blocks
@@ -418,23 +395,28 @@ header = try $ do
   lev <- length <$> many1 (char '=')
   guard $ lev <= 6
   contents <- trimInlines . mconcat <$> manyTill inline (count lev $ char '=')
-  attr <- registerHeader nullAttr contents
+  attr <- modifyIdentifier <$> registerHeader nullAttr contents
   return $ B.headerWith attr lev contents
 
+-- See #4731:
+modifyIdentifier :: Attr -> Attr
+modifyIdentifier (ident,cl,kv) = (ident',cl,kv)
+  where ident' = T.map (\c -> if c == '-' then '_' else c) ident
+
 bulletList :: PandocMonad m => MWParser m Blocks
 bulletList = B.bulletList <$>
    (   many1 (listItem '*')
-   <|> (htmlTag (~== TagOpen "ul" []) *> spaces *> many (listItem '*' <|> li) <*
-        optional (htmlTag (~== TagClose "ul"))) )
+   <|> (htmlTag (~== TagOpen ("ul" :: Text) []) *> spaces *> many (listItem '*' <|> li) <*
+        optional (htmlTag (~== TagClose ("ul" :: Text)))) )
 
 orderedList :: PandocMonad m => MWParser m Blocks
 orderedList =
        (B.orderedList <$> many1 (listItem '#'))
    <|> try
-       (do (tag,_) <- htmlTag (~== TagOpen "ol" [])
+       (do (tag,_) <- htmlTag (~== TagOpen ("ol" :: Text) [])
            spaces
            items <- many (listItem '#' <|> li)
-           optional (htmlTag (~== TagClose "ol"))
+           optional (htmlTag (~== TagClose ("ol" :: Text)))
            let start = fromMaybe 1 $ safeRead $ fromAttrib "start" tag
            return $ B.orderedListWith (start, DefaultStyle, DefaultDelim) items)
 
@@ -471,7 +453,7 @@ anyListStart :: PandocMonad m => MWParser m Char
 anyListStart = guardColumnOne >> oneOf "*#:;"
 
 li :: PandocMonad m => MWParser m Blocks
-li = lookAhead (htmlTag (~== TagOpen "li" [])) *>
+li = lookAhead (htmlTag (~== TagOpen ("li" :: Text) [])) *>
      (firstParaToPlain <$> blocksInTags "li") <* spaces
 
 listItem :: PandocMonad m => Char -> MWParser m Blocks
@@ -483,13 +465,13 @@ listItem c = try $ do
      else do
        skipMany spaceChar
        pos' <- getPosition
-       first <- concat <$> manyTill listChunk newline
+       first <- T.concat <$> manyTill listChunk newline
        rest <- many
                 (try $ string extras *> lookAhead listStartChar *>
-                       (concat <$> manyTill listChunk newline))
+                       (T.concat <$> manyTill listChunk newline))
        contents <- parseFromString (do setPosition pos'
                                        many1 $ listItem' c)
-                          (unlines (first : rest))
+                          (T.unlines (first : rest))
        case c of
            '*' -> return $ B.bulletList contents
            '#' -> return $ B.orderedList contents
@@ -503,20 +485,20 @@ listItem c = try $ do
 -- }}
 -- * next list item
 -- which seems to be valid mediawiki.
-listChunk :: PandocMonad m => MWParser m String
-listChunk = template <|> count 1 anyChar
+listChunk :: PandocMonad m => MWParser m Text
+listChunk = template <|> countChar 1 anyChar
 
 listItem' :: PandocMonad m => Char -> MWParser m Blocks
 listItem' c = try $ do
   listStart c
   skipMany spaceChar
   pos' <- getPosition
-  first <- concat <$> manyTill listChunk newline
+  first <- T.concat <$> manyTill listChunk newline
   rest <- many (try $ char c *> lookAhead listStartChar *>
-                   (concat <$> manyTill listChunk newline))
+                   (T.concat <$> manyTill listChunk newline))
   parseFromString (do setPosition pos'
                       firstParaToPlain . mconcat <$> many1 block)
-      $ unlines $ first : rest
+      $ T.unlines $ first : rest
 
 firstParaToPlain :: Blocks -> Blocks
 firstParaToPlain contents =
@@ -547,23 +529,23 @@ inline =  whitespace
       <|> special
 
 str :: PandocMonad m => MWParser m Inlines
-str = B.str <$> many1 (noneOf $ specialChars ++ spaceChars)
+str = B.str <$> many1Char (noneOf $ specialChars ++ spaceChars)
 
 math :: PandocMonad m => MWParser m Inlines
-math = (B.displayMath . trim <$> try (many1 (char ':') >> charsInTags "math"))
-   <|> (B.math . trim <$> charsInTags "math")
-   <|> (B.displayMath . trim <$> try (dmStart *> manyTill anyChar dmEnd))
-   <|> (B.math . trim <$> try (mStart *> manyTill (satisfy (/='\n')) mEnd))
+math = (B.displayMath . trim <$> try (many1 (char ':') >> textInTags "math"))
+   <|> (B.math . trim <$> textInTags "math")
+   <|> (B.displayMath . trim <$> try (dmStart *> manyTillChar anyChar dmEnd))
+   <|> (B.math . trim <$> try (mStart *> manyTillChar (satisfy (/='\n')) mEnd))
  where dmStart = string "\\["
        dmEnd   = try (string "\\]")
        mStart  = string "\\("
        mEnd    = try (string "\\)")
 
-variable :: PandocMonad m => MWParser m String
+variable :: PandocMonad m => MWParser m Text
 variable = try $ do
   string "{{{"
-  contents <- manyTill anyChar (try $ string "}}}")
-  return $ "{{{" ++ contents ++ "}}}"
+  contents <- manyTillChar anyChar (try $ string "}}}")
+  return $ "{{{" <> contents <> "}}}"
 
 inlineTag :: PandocMonad m => MWParser m Inlines
 inlineTag = do
@@ -572,11 +554,11 @@ inlineTag = do
        TagOpen "ref" _ -> B.note . B.plain <$> inlinesInTags "ref"
        TagOpen "nowiki" _ -> try $ do
           (_,raw) <- htmlTag (~== tag)
-          if '/' `elem` raw
+          if T.any (== '/') raw
              then return mempty
              else B.text . fromEntities <$>
-                       manyTill anyChar (htmlTag (~== TagClose "nowiki"))
-       TagOpen "br" _ -> B.linebreak <$ (htmlTag (~== TagOpen "br" []) -- will get /> too
+                       manyTillChar anyChar (htmlTag (~== TagClose ("nowiki" :: Text)))
+       TagOpen "br" _ -> B.linebreak <$ (htmlTag (~== TagOpen ("br" :: Text) []) -- will get /> too
                             *> optional blankline)
        TagOpen "strike" _ -> B.strikeout <$> inlinesInTags "strike"
        TagOpen "del" _ -> B.strikeout <$> inlinesInTags "del"
@@ -589,12 +571,12 @@ inlineTag = do
          result <- encode <$> inlinesInTags "tt"
          updateState $ \st -> st{ mwInTT = inTT }
          return result
-       TagOpen "hask" _ -> B.codeWith ("",["haskell"],[]) <$> charsInTags "hask"
+       TagOpen "hask" _ -> B.codeWith ("",["haskell"],[]) <$> textInTags "hask"
        _ -> B.rawInline "html" . snd <$> htmlTag (~== tag)
 
 special :: PandocMonad m => MWParser m Inlines
-special = B.str <$> count 1 (notFollowedBy' (htmlTag isBlockTag') *>
-                             oneOf specialChars)
+special = B.str <$> countChar 1 (notFollowedBy' (htmlTag isBlockTag') *>
+                                  oneOf specialChars)
 
 inlineHtml :: PandocMonad m => MWParser m Inlines
 inlineHtml = B.rawInline "html" . snd <$> htmlTag isInlineTag'
@@ -613,7 +595,7 @@ endline = () <$ try (newline <*
                      notFollowedBy anyListStart)
 
 imageIdentifiers :: PandocMonad m => [MWParser m ()]
-imageIdentifiers = [sym (identifier ++ ":") | identifier <- identifiers]
+imageIdentifiers = [sym (identifier <> ":") | identifier <- identifiers]
     where identifiers = ["File", "Image", "Archivo", "Datei", "Fichier",
                          "Bild"]
 
@@ -621,9 +603,9 @@ image :: PandocMonad m => MWParser m Inlines
 image = try $ do
   sym "[["
   choice imageIdentifiers
-  fname <- addUnderscores <$> many1 (noneOf "|]")
+  fname <- addUnderscores <$> many1Char (noneOf "|]")
   _ <- many imageOption
-  dims <- try (char '|' *> sepBy (many digit) (char 'x') <* string "px")
+  dims <- try (char '|' *> sepBy (manyChar digit) (char 'x') <* string "px")
           <|> return []
   _ <- many imageOption
   let kvs = case dims of
@@ -633,9 +615,9 @@ image = try $ do
   let attr = ("", [], kvs)
   caption <-   (B.str fname <$ sym "]]")
            <|> try (char '|' *> (mconcat <$> manyTill inline (sym "]]")))
-  return $ B.imageWith attr fname ("fig:" ++ stringify caption) caption
+  return $ B.imageWith attr fname ("fig:" <> stringify caption) caption
 
-imageOption :: PandocMonad m => MWParser m String
+imageOption :: PandocMonad m => MWParser m Text
 imageOption = try $ char '|' *> opt
   where
     opt = try (oneOfStrings [ "border", "thumbnail", "frameless"
@@ -643,30 +625,27 @@ imageOption = try $ char '|' *> opt
                             , "center", "none", "baseline", "sub"
                             , "super", "top", "text-top", "middle"
                             , "bottom", "text-bottom" ])
-      <|> try (string "frame")
+      <|> try (textStr "frame")
       <|> try (oneOfStrings ["link=","alt=","page=","class="] <* many (noneOf "|]"))
 
-collapseUnderscores :: String -> String
-collapseUnderscores []           = []
-collapseUnderscores ('_':'_':xs) = collapseUnderscores ('_':xs)
-collapseUnderscores (x:xs)       = x : collapseUnderscores xs
-
-addUnderscores :: String -> String
-addUnderscores = collapseUnderscores . intercalate "_" . words
+addUnderscores :: Text -> Text
+addUnderscores = T.intercalate "_" . splitTextBy sep
+  where
+    sep c = isSpace c || c == '_'
 
 internalLink :: PandocMonad m => MWParser m Inlines
 internalLink = try $ do
   sym "[["
-  pagename <- unwords . words <$> many (noneOf "|]")
+  pagename <- T.unwords . T.words <$> manyChar (noneOf "|]")
   label <- option (B.text pagename) $ char '|' *>
              (  (mconcat <$> many1 (notFollowedBy (char ']') *> inline))
              -- the "pipe trick"
              -- [[Help:Contents|] -> "Contents"
-             <|> return (B.text $ drop 1 $ dropWhile (/=':') pagename) )
+             <|> return (B.text $ T.drop 1 $ T.dropWhile (/=':') pagename) )
   sym "]]"
-  linktrail <- B.text <$> many letter
+  linktrail <- B.text <$> manyChar letter
   let link = B.link (addUnderscores pagename) "wikilink" (label <> linktrail)
-  if "Category:" `isPrefixOf` pagename
+  if "Category:" `T.isPrefixOf` pagename
      then do
        updateState $ \st -> st{ mwCategoryLinks = link : mwCategoryLinks st }
        return mempty
@@ -681,7 +660,7 @@ externalLink = try $ do
        <|> do char ']'
               num <- mwNextLinkNumber <$> getState
               updateState $ \st -> st{ mwNextLinkNumber = num + 1 }
-              return $ B.str $ show num
+              return $ B.str $ tshow num
   return $ B.link src "" lab
 
 url :: PandocMonad m => MWParser m Inlines
