@@ -1,25 +1,7 @@
-{-# LANGUAGE NoImplicitPrelude #-}
-{-
-Copyright (C) 2013-2018 John MacFarlane <jgm@berkeley.edu>
-
-This program is free software; you can redistribute it and/or modify
-it under the terms of the GNU General Public License as published by
-the Free Software Foundation; either version 2 of the License, or
-(at your option) any later version.
-
-This program is distributed in the hope that it will be useful,
-but WITHOUT ANY WARRANTY; without even the implied warranty of
-MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-GNU General Public License for more details.
-
-You should have received a copy of the GNU General Public License
-along with this program; if not, write to the Free Software
-Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--}
-
+{-# LANGUAGE OverloadedStrings #-}
 {- |
    Module      : Text.Pandoc.Readers.OPML
-   Copyright   : Copyright (C) 2013-2018 John MacFarlane
+   Copyright   : Copyright (C) 2013-2020 John MacFarlane
    License     : GNU GPL, version 2 or above
 
    Maintainer  : John MacFarlane <jgm@berkeley.edu>
@@ -30,16 +12,16 @@ Conversion of OPML to 'Pandoc' document.
 -}
 
 module Text.Pandoc.Readers.OPML ( readOPML ) where
-import Prelude
 import Control.Monad.State.Strict
 import Data.Char (toUpper)
 import Data.Default
 import Data.Generics
 import Data.Maybe (fromMaybe)
-import Data.Text (Text, pack, unpack)
+import Data.Text (Text)
+import qualified Data.Text as T
 import Text.HTML.TagSoup.Entity (lookupEntity)
 import Text.Pandoc.Builder
-import Text.Pandoc.Class (PandocMonad)
+import Text.Pandoc.Class.PandocMonad (PandocMonad)
 import Text.Pandoc.Options
 import Text.Pandoc.Readers.HTML (readHtml)
 import Text.Pandoc.Readers.Markdown (readMarkdown)
@@ -68,7 +50,7 @@ readOPML :: PandocMonad m => ReaderOptions -> Text -> m Pandoc
 readOPML opts inp  = do
   (bs, st') <- runStateT
                  (mapM parseBlock $ normalizeTree $
-                    parseXML (unpack (crFilter inp))) def{ opmlOptions = opts }
+                    parseXML (T.unpack (crFilter inp))) def{ opmlOptions = opts }
   return $
     setTitle (opmlDocTitle st') $
     setAuthors (opmlDocAuthors st') $
@@ -94,23 +76,26 @@ convertEntity :: String -> String
 convertEntity e = Data.Maybe.fromMaybe (map toUpper e) (lookupEntity e)
 
 -- convenience function to get an attribute value, defaulting to ""
-attrValue :: String -> Element -> String
+attrValue :: String -> Element -> Text
 attrValue attr elt =
-  fromMaybe "" (lookupAttrBy (\x -> qName x == attr) (elAttribs elt))
+  maybe "" T.pack (lookupAttrBy (\x -> qName x == attr) (elAttribs elt))
+
+textContent :: Element -> Text
+textContent = T.pack . strContent
 
 -- exceptT :: PandocMonad m => Either PandocError a -> OPML m a
 -- exceptT = either throwError return
 
-asHtml :: PandocMonad m => String -> OPML m Inlines
+asHtml :: PandocMonad m => Text -> OPML m Inlines
 asHtml s = do
   opts <- gets opmlOptions
-  Pandoc _ bs <- readHtml def{ readerExtensions = readerExtensions opts } (pack s)
+  Pandoc _ bs <- readHtml def{ readerExtensions = readerExtensions opts } s
   return $ blocksToInlines' bs
 
-asMarkdown :: PandocMonad m => String -> OPML m Blocks
+asMarkdown :: PandocMonad m => Text -> OPML m Blocks
 asMarkdown s = do
   opts <- gets opmlOptions
-  Pandoc _ bs <- readMarkdown def{ readerExtensions = readerExtensions opts }  (pack s)
+  Pandoc _ bs <- readMarkdown def{ readerExtensions = readerExtensions opts } s
   return $ fromList bs
 
 getBlocks :: PandocMonad m => Element -> OPML m Blocks
@@ -120,11 +105,11 @@ parseBlock :: PandocMonad m => Content -> OPML m Blocks
 parseBlock (Elem e) =
   case qName (elName e) of
         "ownerName"    -> mempty <$ modify (\st ->
-                              st{opmlDocAuthors = [text $ strContent e]})
+                              st{opmlDocAuthors = [text $ textContent e]})
         "dateModified" -> mempty <$ modify (\st ->
-                              st{opmlDocDate = text $ strContent e})
+                              st{opmlDocDate = text $ textContent e})
         "title"        -> mempty <$ modify (\st ->
-                              st{opmlDocTitle = text $ strContent e})
+                              st{opmlDocTitle = text $ textContent e})
         "outline" -> gets opmlSectionLevel >>= sect . (+1)
         "?xml"  -> return mempty
         _       -> getBlocks e
@@ -133,7 +118,7 @@ parseBlock (Elem e) =
                      modify $ \st -> st{ opmlSectionLevel = n }
                      bs <- getBlocks e
                      modify $ \st -> st{ opmlSectionLevel = n - 1 }
-                     let headerText' = case map toUpper (attrValue "type" e) of
+                     let headerText' = case T.toUpper (attrValue "type" e) of
                                              "LINK"  -> link
                                                (attrValue "url" e) "" headerText
                                              _ -> headerText
