@@ -1,28 +1,6 @@
-{-# LANGUAGE NoImplicitPrelude #-}
-{-# LANGUAGE Arrows          #-}
 {-# LANGUAGE TupleSections   #-}
 {-# LANGUAGE GADTs           #-}
 {-# LANGUAGE PatternGuards   #-}
-{-# LANGUAGE RecordWildCards #-}
-
-{-
-Copyright (C) 2015 Martin Linnemann <theCodingMarlin@googlemail.com>
-
-This program is free software; you can redistribute it and/or modify
-it under the terms of the GNU General Public License as published by
-the Free Software Foundation; either version 2 of the License, or
-(at your option) any later version.
-
-This program is distributed in the hope that it will be useful,
-but WITHOUT ANY WARRANTY; without even the implied warranty of
-MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-GNU General Public License for more details.
-
-You should have received a copy of the GNU General Public License
-along with this program; if not, write to the Free Software
-Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--}
-
 {- |
    Module      : Text.Pandoc.Readers.Odt.Generic.XMLConverter
    Copyright   : Copyright (C) 2015 Martin Linnemann
@@ -48,33 +26,40 @@ module Text.Pandoc.Readers.Odt.Generic.XMLConverter
 , modifyExtraState
 , producingExtraState
 , findChild'
+, filterChildrenName'
 , isSet'
 , isSetWithDefault
+, elName
 , searchAttr
 , lookupAttr
 , lookupAttr'
 , lookupDefaultingAttr
 , findAttr'
+, findAttrText'
 , findAttr
+, findAttrText
 , findAttrWithDefault
+, findAttrTextWithDefault
 , readAttr
 , readAttr'
 , readAttrWithDefault
 , getAttr
 , executeIn
+, executeInSub
 , withEveryL
 , tryAll
 , matchContent'
 , matchContent
 ) where
 
-import Prelude
 import           Control.Applicative  hiding ( liftA, liftA2 )
 import           Control.Monad               ( MonadPlus )
 import           Control.Arrow
 
+import           Data.Bool ( bool )
 import           Data.Either ( rights )
 import qualified Data.Map             as M
+import qualified Data.Text            as T
 import           Data.Default
 import           Data.Maybe
 
@@ -95,6 +80,7 @@ import           Text.Pandoc.Readers.Odt.Generic.Fallible
 type ElementName           = String
 type AttributeName         = String
 type AttributeValue        = String
+type TextAttributeValue    = T.Text
 
 --
 type NameSpacePrefix       = String
@@ -175,7 +161,7 @@ swapStack' state stack
 pushElement         :: XML.Element
                     -> XMLConverterState nsID extraState
                     -> XMLConverterState nsID extraState
-pushElement e state  = state { parentElements = e:(parentElements state) }
+pushElement e state  = state { parentElements = e:parentElements state }
 
 -- | Pop the top element from the call stack, unless it is the last one.
 popElement          :: XMLConverterState nsID extraState
@@ -328,34 +314,44 @@ readNSattributes         = fromState $ \state -> maybe (state, failEmpty     )
 
 -- | Given a namespace id and an element name, creates a 'XML.QName' for
 -- internal use
-elemName                 :: (NameSpaceID nsID)
+qualifyName              :: (NameSpaceID nsID)
                          => nsID -> ElementName
                          -> XMLConverter nsID extraState x XML.QName
-elemName nsID name       =         lookupNSiri nsID
+qualifyName nsID name    =         lookupNSiri nsID
                                &&& lookupNSprefix nsID
                            >>% XML.QName name
 
+-- | Checks if a given element matches both a specified namespace id
+-- and a predicate
+elemNameMatches          :: (NameSpaceID nsID)
+                         => nsID -> (ElementName -> Bool)
+                         -> XMLConverter nsID extraState XML.Element Bool
+elemNameMatches nsID f    = keepingTheValue (lookupNSiri nsID) >>% hasMatchingName
+  where hasMatchingName e iri = let name = XML.elName e
+                                in     f (XML.qName name)
+                                    && XML.qURI name == iri
+
 -- | Checks if a given element matches both a specified namespace id
 -- and a specified element name
 elemNameIs               :: (NameSpaceID nsID)
                          => nsID -> ElementName
                          -> XMLConverter nsID extraState XML.Element Bool
-elemNameIs nsID name     = keepingTheValue (lookupNSiri nsID) >>% hasThatName
-  where hasThatName e iri = let elName = XML.elName e
-                            in     XML.qName elName == name
-                                && XML.qURI  elName == iri
+elemNameIs nsID name     = elemNameMatches nsID (== name)
 
 --------------------------------------------------------------------------------
 -- General content
 --------------------------------------------------------------------------------
 
+elName :: XML.Element -> ElementName
+elName = XML.qName . XML.elName
+
 --
 elContent               :: XMLConverter nsID extraState x [XML.Content]
 elContent               =     getCurrentElement
                            >>^ XML.elContent
 
 --------------------------------------------------------------------------------
--- Chilren
+-- Children
 --------------------------------------------------------------------------------
 
 --
@@ -363,7 +359,7 @@ elContent               =     getCurrentElement
 findChildren             :: (NameSpaceID nsID)
                          => nsID -> ElementName
                          -> XMLConverter nsID extraState x [XML.Element]
-findChildren nsID name   =         elemName nsID name
+findChildren nsID name   =         qualifyName nsID name
                                &&& getCurrentElement
                            >>% XML.findChildren
 
@@ -372,7 +368,7 @@ findChild'              :: (NameSpaceID nsID)
                         => nsID
                         -> ElementName
                         -> XMLConverter nsID extraState x (Maybe XML.Element)
-findChild' nsID name    =         elemName nsID name
+findChild' nsID name    =         qualifyName nsID name
                               &&& getCurrentElement
                           >>% XML.findChild
 
@@ -383,6 +379,14 @@ findChild              :: (NameSpaceID nsID)
 findChild nsID name    =     findChild' nsID name
                          >>> maybeToChoice
 
+filterChildrenName'        :: (NameSpaceID nsID)
+                           => nsID
+                           -> (ElementName -> Bool)
+                           -> XMLConverter nsID extraState x [XML.Element]
+filterChildrenName' nsID f =     getCurrentElement
+                             >>> arr XML.elChildren
+                             >>> iterateS (keepingTheValue (elemNameMatches nsID f))
+                             >>> arr (map fst . filter snd)
 
 --------------------------------------------------------------------------------
 -- Attributes
@@ -460,10 +464,20 @@ lookupDefaultingAttr nsID attrName
 findAttr'               :: (NameSpaceID nsID)
                         => nsID -> AttributeName
                         -> XMLConverter nsID extraState x (Maybe AttributeValue)
-findAttr' nsID attrName =         elemName nsID attrName
+findAttr' nsID attrName =         qualifyName nsID attrName
                               &&& getCurrentElement
                           >>% XML.findAttr
 
+-- | Return value as a (Maybe Text)
+findAttrText'           :: (NameSpaceID nsID)
+                        => nsID -> AttributeName
+                        -> XMLConverter nsID extraState x (Maybe TextAttributeValue)
+findAttrText' nsID attrName
+                        =         qualifyName nsID attrName
+                              &&& getCurrentElement
+                          >>% XML.findAttr
+                          >>^ fmap T.pack
+
 -- | Return value as string or fail
 findAttr               :: (NameSpaceID nsID)
                        => nsID -> AttributeName
@@ -471,6 +485,15 @@ findAttr               :: (NameSpaceID nsID)
 findAttr nsID attrName =     findAttr' nsID attrName
                          >>> maybeToChoice
 
+-- | Return value as text or fail
+findAttrText           :: (NameSpaceID nsID)
+                       => nsID -> AttributeName
+                       -> FallibleXMLConverter nsID extraState x TextAttributeValue
+findAttrText nsID attrName
+                       = findAttr' nsID attrName
+                         >>^ fmap T.pack
+                         >>> maybeToChoice
+
 -- | Return value as string or return provided default value
 findAttrWithDefault    :: (NameSpaceID nsID)
                        => nsID -> AttributeName
@@ -480,6 +503,15 @@ findAttrWithDefault nsID attrName deflt
                        = findAttr' nsID attrName
                          >>^ fromMaybe deflt
 
+-- | Return value as string or return provided default value
+findAttrTextWithDefault :: (NameSpaceID nsID)
+                        => nsID -> AttributeName
+                        -> TextAttributeValue
+                        -> XMLConverter nsID extraState x TextAttributeValue
+findAttrTextWithDefault nsID attrName deflt
+                       = findAttr' nsID attrName
+                         >>^ maybe deflt T.pack
+
 -- | Read and return value or fail
 readAttr               :: (NameSpaceID nsID, Read attrValue)
                        => nsID -> AttributeName
@@ -556,17 +588,23 @@ executeThere a         =      second jumpThere
                           >>> jumpBack -- >>? jumpBack  would not ensure the jump.
                           >>^ collapseEither
 
--- | Do something in a sub-element, tnen come back
-executeIn              :: (NameSpaceID nsID)
-                       => nsID -> ElementName
-                       -> FallibleXMLConverter nsID extraState f s
-                       -> FallibleXMLConverter nsID extraState f s
-executeIn nsID name a  =     keepingTheValue
-                               (findChild nsID name)
-                         >>> ignoringState liftFailure
-                         >>? switchingTheStack a
-  where liftFailure (_, (Left  f)) = Left  f
-        liftFailure (x, (Right e)) = Right (x, e)
+
+-- | Do something in a specific element, then come back
+executeIn   :: XMLConverter nsID extraState XML.Element s
+            -> XMLConverter nsID extraState XML.Element s
+executeIn a = duplicate >>> switchingTheStack a
+
+-- | Do something in a sub-element, then come back
+executeInSub              :: (NameSpaceID nsID)
+                          => nsID -> ElementName
+                          -> FallibleXMLConverter nsID extraState f s
+                          -> FallibleXMLConverter nsID extraState f s
+executeInSub nsID name a  =     keepingTheValue
+                                  (findChild nsID name)
+                            >>> ignoringState liftFailure
+                            >>? switchingTheStack a
+  where liftFailure (_, Left  f) = Left  f
+        liftFailure (x, Right e) = Right (x, e)
 
 --------------------------------------------------------------------------------
 -- Iterating over children
@@ -662,7 +700,7 @@ prepareMatchersC      :: (NameSpaceID nsID)
                        => [(nsID, ElementName, FallibleXMLConverter nsID extraState x x)]
                        -> ContentMatchConverter nsID extraState x
 --prepareMatchersC      = foldSs . (map $ uncurry3  makeMatcherC)
-prepareMatchersC      = reverseComposition . (map $ uncurry3  makeMatcherC)
+prepareMatchersC      = reverseComposition . map (uncurry3  makeMatcherC)
 
 -- | Takes a list of element-data - converter groups and
 -- * Finds all content of the current element
