@@ -1,27 +1,7 @@
-{-# LANGUAGE NoImplicitPrelude #-}
 {-# LANGUAGE CPP             #-}
 {-# LANGUAGE Arrows          #-}
 {-# LANGUAGE RecordWildCards #-}
 {-# LANGUAGE TupleSections   #-}
-
-{-
-Copyright (C) 2015 Martin Linnemann <theCodingMarlin@googlemail.com>
-
-This program is free software; you can redistribute it and/or modify
-it under the terms of the GNU General Public License as published by
-the Free Software Foundation; either version 2 of the License, or
-(at your option) any later version.
-
-This program is distributed in the hope that it will be useful,
-but WITHOUT ANY WARRANTY; without even the implied warranty of
-MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-GNU General Public License for more details.
-
-You should have received a copy of the GNU General Public License
-along with this program; if not, write to the Free Software
-Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--}
-
 {- |
    Module      : Text.Pandoc.Readers.Odt.StyleReader
    Copyright   : Copyright (C) 2015 Martin Linnemann
@@ -58,11 +38,9 @@ module Text.Pandoc.Readers.Odt.StyleReader
 , readStylesAt
 ) where
 
-import Prelude
 import Control.Applicative hiding (liftA, liftA2, liftA3)
 import Control.Arrow
 
-import Data.Char (isDigit)
 import Data.Default
 import qualified Data.Foldable as F
 import Data.List (unfoldr)
@@ -72,6 +50,8 @@ import qualified Data.Set as S
 
 import qualified Text.XML.Light as XML
 
+import Text.Pandoc.Shared (safeRead)
+
 import Text.Pandoc.Readers.Odt.Arrows.Utils
 
 import Text.Pandoc.Readers.Odt.Generic.Fallible
@@ -131,14 +111,14 @@ type StyleReaderSafe a b  = XMLReaderSafe FontPitches a b
 
 -- | A reader for font pitches
 fontPitchReader :: XMLReader _s _x FontPitches
-fontPitchReader = executeIn NsOffice "font-face-decls" (
+fontPitchReader = executeInSub NsOffice "font-face-decls" (
                           withEveryL NsStyle "font-face" (liftAsSuccess (
                               findAttr' NsStyle "name"
                               &&&
                               lookupDefaultingAttr NsStyle "font-pitch"
                             ))
                     >>?^ ( M.fromList . foldl accumLegalPitches [] )
-                  ) `ifFailedDo` (returnV (Right M.empty))
+                  ) `ifFailedDo` returnV (Right M.empty)
   where accumLegalPitches ls (Nothing,_) = ls
         accumLegalPitches ls (Just n,p)  = (n,p):ls
 
@@ -441,7 +421,7 @@ readAllStyles = (      readFontPitches
 
 --
 readStyles :: StyleReader _x Styles
-readStyles = executeIn NsOffice "styles" $ liftAsSuccess
+readStyles = executeInSub NsOffice "styles" $ liftAsSuccess
   $ liftA3 Styles
     ( tryAll NsStyle "style"         readStyle        >>^ M.fromList )
     ( tryAll NsText  "list-style"    readListStyle    >>^ M.fromList )
@@ -449,7 +429,7 @@ readStyles = executeIn NsOffice "styles" $ liftAsSuccess
 
 --
 readAutomaticStyles :: StyleReader _x Styles
-readAutomaticStyles = executeIn NsOffice "automatic-styles" $ liftAsSuccess
+readAutomaticStyles = executeInSub NsOffice "automatic-styles" $ liftAsSuccess
   $ liftA3 Styles
     ( tryAll NsStyle "style"         readStyle        >>^ M.fromList )
     ( tryAll NsText  "list-style"    readListStyle    >>^ M.fromList )
@@ -480,7 +460,7 @@ readStyleProperties = liftA2 SProps
 --
 readTextProperties :: StyleReader _x TextProperties
 readTextProperties =
-  executeIn NsStyle "text-properties" $ liftAsSuccess
+  executeInSub NsStyle "text-properties" $ liftAsSuccess
     ( liftA6 PropT
        ( searchAttr   NsXSL_FO "font-style"  False isFontEmphasised )
        ( searchAttr   NsXSL_FO "font-weight" False isFontBold       )
@@ -519,7 +499,7 @@ readLineMode modeAttr styleAttr = proc x -> do
 --
 readParaProperties :: StyleReader _x ParaProperties
 readParaProperties =
-   executeIn NsStyle "paragraph-properties" $ liftAsSuccess
+   executeInSub NsStyle "paragraph-properties" $ liftAsSuccess
      ( liftA3 PropP
        ( liftA2 readNumbering
          ( isSet'           NsText   "number-lines"           )
@@ -566,21 +546,17 @@ readListLevelStyle :: ListLevelType -> StyleReader _x (Int, ListLevelStyle)
 readListLevelStyle levelType =      readAttr NsText "level"
                                >>?! keepingTheValue
                                     ( liftA5 toListLevelStyle
-                                      ( returnV  levelType              )
-                                      ( findAttr' NsStyle "num-prefix"  )
-                                      ( findAttr' NsStyle "num-suffix"  )
-                                      ( getAttr   NsStyle "num-format"  )
-                                      ( findAttr' NsText  "start-value" )
+                                      ( returnV       levelType             )
+                                      ( findAttr'     NsStyle "num-prefix"  )
+                                      ( findAttr'     NsStyle "num-suffix"  )
+                                      ( getAttr       NsStyle "num-format"  )
+                                      ( findAttrText' NsText  "start-value" )
                                     )
   where
   toListLevelStyle _ p s LinfNone b         = ListLevelStyle LltBullet p s LinfNone (startValue b)
   toListLevelStyle _ p s f@(LinfString _) b = ListLevelStyle LltBullet p s f (startValue b)
   toListLevelStyle t p s f b                = ListLevelStyle t      p s f (startValue b)
-  startValue (Just "") = 1
-  startValue (Just v)  = if all isDigit v
-                           then read v
-                           else 1
-  startValue Nothing   = 1
+  startValue mbx = fromMaybe 1 (mbx >>= safeRead)
 
 --
 chooseMostSpecificListLevelStyle :: S.Set ListLevelStyle -> Maybe ListLevelStyle
