@@ -1,26 +1,9 @@
-{-# LANGUAGE NoImplicitPrelude #-}
-{-
-Copyright (C) 2014-2018 Albert Krewinkel <tarleb+pandoc@moltkeplatz.de>
-
-This program is free software; you can redistribute it and/or modify
-it under the terms of the GNU General Public License as published by
-the Free Software Foundation; either version 2 of the License, or
-(at your option) any later version.
-
-This program is distributed in the hope that it will be useful,
-but WITHOUT ANY WARRANTY; without even the implied warranty of
-MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-GNU General Public License for more details.
-
-You should have received a copy of the GNU General Public License
-along with this program; if not, write to the Free Software
-Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--}
-{-# LANGUAGE FlexibleContexts #-}
-{-# LANGUAGE RecordWildCards  #-}
+{-# LANGUAGE FlexibleContexts  #-}
+{-# LANGUAGE RecordWildCards   #-}
+{-# LANGUAGE OverloadedStrings #-}
 {- |
    Module      : Text.Pandoc.Readers.Org.Blocks
-   Copyright   : Copyright (C) 2014-2018 Albert Krewinkel
+   Copyright   : Copyright (C) 2014-2020 Albert Krewinkel
    License     : GNU GPL, version 2 or above
 
    Maintainer  : Albert Krewinkel <tarleb+pandoc@moltkeplatz.de>
@@ -32,28 +15,30 @@ module Text.Pandoc.Readers.Org.Blocks
   , meta
   ) where
 
-import Prelude
 import Text.Pandoc.Readers.Org.BlockStarts
-import Text.Pandoc.Readers.Org.DocumentTree (documentTree, headlineToBlocks)
+import Text.Pandoc.Readers.Org.DocumentTree (documentTree,
+                                             unprunedHeadlineToBlocks)
 import Text.Pandoc.Readers.Org.Inlines
 import Text.Pandoc.Readers.Org.Meta (metaExport, metaKey, metaLine)
 import Text.Pandoc.Readers.Org.ParserState
 import Text.Pandoc.Readers.Org.Parsing
-import Text.Pandoc.Readers.Org.Shared (cleanLinkString, isImageFilename,
+import Text.Pandoc.Readers.Org.Shared (cleanLinkText, isImageFilename,
                                        originalLang, translateLang, exportsCode)
 
 import Text.Pandoc.Builder (Blocks, Inlines)
-import Text.Pandoc.Class (PandocMonad)
+import Text.Pandoc.Class.PandocMonad (PandocMonad)
 import Text.Pandoc.Definition
 import Text.Pandoc.Options
 import Text.Pandoc.Shared (compactify, compactifyDL, safeRead)
 
-import Control.Monad (foldM, guard, mzero, void)
-import Data.Char (isSpace, toLower, toUpper)
+import Control.Monad (foldM, guard, mplus, mzero, void)
+import Data.Char (isSpace)
 import Data.Default (Default)
-import Data.List (foldl', isPrefixOf)
+import Data.List (foldl')
 import Data.Maybe (fromMaybe, isJust, isNothing)
+import Data.Text (Text)
 
+import qualified Data.Text as T
 import qualified Text.Pandoc.Builder as B
 import qualified Text.Pandoc.Walk as Walk
 
@@ -64,11 +49,10 @@ import qualified Text.Pandoc.Walk as Walk
 -- | Get a list of blocks.
 blockList :: PandocMonad m => OrgParser m [Block]
 blockList = do
-  headlines      <- documentTree blocks inline
+  fHeadlineTree  <- documentTree blocks inline
   st             <- getState
-  headlineBlocks <- headlineToBlocks $ runF headlines st
-  -- ignore first headline, it's the document's title
-  return . drop 1 . B.toList $ headlineBlocks
+  let headlineTree = runF fHeadlineTree st
+  unprunedHeadlineToBlocks headlineTree st
 
 -- | Get the meta information saved in the state.
 meta :: Monad m => OrgParser m Meta
@@ -107,10 +91,10 @@ horizontalRule = return B.horizontalRule <$ try hline
 
 -- | Attributes that may be added to figures (like a name or caption).
 data BlockAttributes = BlockAttributes
-  { blockAttrName      :: Maybe String
-  , blockAttrLabel     :: Maybe String
+  { blockAttrName      :: Maybe Text
+  , blockAttrLabel     :: Maybe Text
   , blockAttrCaption   :: Maybe (F Inlines)
-  , blockAttrKeyValues :: [(String, String)]
+  , blockAttrKeyValues :: [(Text, Text)]
   }
 
 -- | Convert BlockAttributes into pandoc Attr
@@ -120,30 +104,32 @@ attrFromBlockAttributes BlockAttributes{..} =
     ident   = fromMaybe mempty $ lookup "id" blockAttrKeyValues
     classes = case lookup "class" blockAttrKeyValues of
                 Nothing     -> []
-                Just clsStr -> words clsStr
+                Just clsStr -> T.words clsStr
     kv      = filter ((`notElem` ["id", "class"]) . fst) blockAttrKeyValues
   in (ident, classes, kv)
 
-stringyMetaAttribute :: Monad m => OrgParser m (String, String)
+stringyMetaAttribute :: Monad m => OrgParser m (Text, Text)
 stringyMetaAttribute = try $ do
   metaLineStart
-  attrName <- map toUpper <$> many1Till nonspaceChar (char ':')
+  attrName <- T.toUpper <$> many1TillChar nonspaceChar (char ':')
   skipSpaces
   attrValue <- anyLine <|> ("" <$ newline)
   return (attrName, attrValue)
 
+-- | Parse a set of block attributes. Block attributes are given through
+-- lines like @#+CAPTION: block caption@ or @#+ATTR_HTML: :width 20@.
+-- Parsing will fail if any line contains an attribute different from
+-- those attributes known to work on blocks.
 blockAttributes :: PandocMonad m => OrgParser m BlockAttributes
 blockAttributes = try $ do
   kv <- many stringyMetaAttribute
-  guard $ all (attrCheck . fst) kv
+  guard $ all (isBlockAttr . fst) kv
   let caption = foldl' (appendValues "CAPTION") Nothing kv
   let kvAttrs = foldl' (appendValues "ATTR_HTML") Nothing kv
   let name    = lookup "NAME" kv
   let label   = lookup "LABEL" kv
-  caption' <- case caption of
-                   Nothing -> return Nothing
-                   Just s  -> Just <$> parseFromString inlines (s ++ "\n")
-  kvAttrs' <- parseFromString keyValues . (++ "\n") $ fromMaybe mempty kvAttrs
+  caption' <- traverse (parseFromString inlines . (<> "\n")) caption
+  kvAttrs' <- parseFromString keyValues . (<> "\n") $ fromMaybe mempty kvAttrs
   return BlockAttributes
            { blockAttrName = name
            , blockAttrLabel = label
@@ -151,27 +137,31 @@ blockAttributes = try $ do
            , blockAttrKeyValues = kvAttrs'
            }
  where
-   attrCheck :: String -> Bool
-   attrCheck x = x `elem` ["NAME", "LABEL", "CAPTION", "ATTR_HTML", "RESULTS"]
+   isBlockAttr :: Text -> Bool
+   isBlockAttr = flip elem
+                 [ "NAME", "LABEL", "CAPTION"
+                 , "ATTR_HTML", "ATTR_LATEX"
+                 , "RESULTS"
+                 ]
 
-   appendValues :: String -> Maybe String -> (String, String) -> Maybe String
+   appendValues :: Text -> Maybe Text -> (Text, Text) -> Maybe Text
    appendValues attrName accValue (key, value) =
      if key /= attrName
      then accValue
      else case accValue of
-            Just acc -> Just $ acc ++ ' ':value
+            Just acc -> Just $ acc <> " " <> value
             Nothing  -> Just value
 
 -- | Parse key-value pairs for HTML attributes
-keyValues :: Monad m => OrgParser m [(String, String)]
+keyValues :: Monad m => OrgParser m [(Text, Text)]
 keyValues = try $
   manyTill ((,) <$> key <*> value) newline
  where
-   key :: Monad m => OrgParser m String
-   key = try $ skipSpaces *> char ':' *> many1 nonspaceChar
+   key :: Monad m => OrgParser m Text
+   key = try $ skipSpaces *> char ':' *> many1Char nonspaceChar
 
-   value :: Monad m => OrgParser m String
-   value = skipSpaces *> manyTill anyChar endOfValue
+   value :: Monad m => OrgParser m Text
+   value = skipSpaces *> manyTillChar anyChar endOfValue
 
    endOfValue :: Monad m => OrgParser m ()
    endOfValue =
@@ -189,13 +179,13 @@ orgBlock = try $ do
   blockAttrs <- blockAttributes
   blkType <- blockHeaderStart
   ($ blkType) $
-    case map toLower blkType of
+    case T.toLower blkType of
       "export"  -> exportBlock
       "comment" -> rawBlockLines (const mempty)
       "html"    -> rawBlockLines (return . B.rawBlock (lowercase blkType))
       "latex"   -> rawBlockLines (return . B.rawBlock (lowercase blkType))
       "ascii"   -> rawBlockLines (return . B.rawBlock (lowercase blkType))
-      "example" -> rawBlockLines (return . exampleCode)
+      "example" -> exampleBlock blockAttrs
       "quote"   -> parseBlockLines (fmap B.blockQuote)
       "verse"   -> verseBlock
       "src"     -> codeBlock blockAttrs
@@ -203,61 +193,70 @@ orgBlock = try $ do
                    let (ident, classes, kv) = attrFromBlockAttributes blockAttrs
                    in fmap $ B.divWith (ident, classes ++ [blkType], kv)
  where
-   blockHeaderStart :: Monad m => OrgParser m String
+   blockHeaderStart :: Monad m => OrgParser m Text
    blockHeaderStart = try $ skipSpaces *> stringAnyCase "#+begin_" *> orgArgWord
 
-   lowercase :: String -> String
-   lowercase = map toLower
+   lowercase :: Text -> Text
+   lowercase = T.toLower
 
-rawBlockLines :: Monad m => (String   -> F Blocks) -> String -> OrgParser m (F Blocks)
+exampleBlock :: PandocMonad m => BlockAttributes -> Text -> OrgParser m (F Blocks)
+exampleBlock blockAttrs _label = do
+  skipSpaces
+  (classes, kv) <- switchesAsAttributes
+  newline
+  content <- rawBlockContent "example"
+  let id' = fromMaybe mempty $ blockAttrName blockAttrs
+  let codeBlck = B.codeBlockWith (id', "example":classes, kv) content
+  return . return $ codeBlck
+
+rawBlockLines :: Monad m => (Text   -> F Blocks) -> Text -> OrgParser m (F Blocks)
 rawBlockLines f blockType = ignHeaders *> (f <$> rawBlockContent blockType)
 
-parseBlockLines :: PandocMonad m => (F Blocks -> F Blocks) -> String -> OrgParser m (F Blocks)
+parseBlockLines :: PandocMonad m => (F Blocks -> F Blocks) -> Text -> OrgParser m (F Blocks)
 parseBlockLines f blockType = ignHeaders *> (f <$> parsedBlockContent)
  where
    parsedBlockContent :: PandocMonad m => OrgParser m (F Blocks)
    parsedBlockContent = try $ do
      raw <- rawBlockContent blockType
-     parseFromString blocks (raw ++ "\n")
+     parseFromString blocks (raw <> "\n")
 
 -- | Read the raw string content of a block
-rawBlockContent :: Monad m => String -> OrgParser m String
+rawBlockContent :: Monad m => Text -> OrgParser m Text
 rawBlockContent blockType = try $ do
   blkLines <- manyTill rawLine blockEnder
   tabLen <- getOption readerTabStop
-  return
-    . unlines
-    . stripIndent
-    . map (tabsToSpaces tabLen . commaEscaped)
-    $ blkLines
+  trimP <- orgStateTrimLeadBlkIndent <$> getState
+  let stripIndent strs = if trimP then map (T.drop (shortestIndent strs)) strs else strs
+  (T.unlines
+   . stripIndent
+   . map (tabsToSpaces tabLen . commaEscaped)
+   $ blkLines)
+   <$ updateState (\s -> s { orgStateTrimLeadBlkIndent = True })
  where
-   rawLine :: Monad m => OrgParser m String
+   rawLine :: Monad m => OrgParser m Text
    rawLine = try $ ("" <$ blankline) <|> anyLine
 
    blockEnder :: Monad m => OrgParser m ()
    blockEnder = try $ skipSpaces <* stringAnyCase ("#+end_" <> blockType)
 
-   stripIndent :: [String] -> [String]
-   stripIndent strs = map (drop (shortestIndent strs)) strs
+   shortestIndent :: [Text] -> Int
+   shortestIndent = foldr (min . T.length . T.takeWhile isSpace) maxBound
+                    . filter (not . T.null)
 
-   shortestIndent :: [String] -> Int
-   shortestIndent = foldr (min . length . takeWhile isSpace) maxBound
-                    . filter (not . null)
+   tabsToSpaces :: Int -> Text -> Text
+   tabsToSpaces tabStop t =
+     let (ind, suff) = T.span (\c -> c == ' ' || c == '\t') t
+         tabNum      = T.length $ T.filter (== '\n') ind
+         spaceNum    = T.length ind - tabNum
+     in  T.replicate (spaceNum + tabStop * tabNum) " " <> suff
 
-   tabsToSpaces :: Int -> String -> String
-   tabsToSpaces _      []         = []
-   tabsToSpaces tabLen cs'@(c:cs) =
-       case c of
-         ' '  -> ' ':tabsToSpaces tabLen cs
-         '\t' -> replicate tabLen ' ' ++ tabsToSpaces tabLen cs
-         _    -> cs'
-
-   commaEscaped :: String -> String
-   commaEscaped (',':cs@('*':_))     = cs
-   commaEscaped (',':cs@('#':'+':_)) = cs
-   commaEscaped (' ':cs)             = ' ':commaEscaped cs
-   commaEscaped ('\t':cs)            = '\t':commaEscaped cs
-   commaEscaped cs                   = cs
+   commaEscaped t =
+     let (ind, suff) = T.span (\c -> c == ' ' || c == '\t') t
+     in  case T.uncons suff of
+           Just (',', cs)
+             | "*"  <- T.take 1 cs -> ind <> cs
+             | "#+" <- T.take 2 cs -> ind <> cs
+           _                       -> t
 
 -- | Read but ignore all remaining block headers.
 ignHeaders :: Monad m => OrgParser m ()
@@ -265,34 +264,34 @@ ignHeaders = (() <$ newline) <|> (() <$ anyLine)
 
 -- | Read a block containing code intended for export in specific backends
 -- only.
-exportBlock :: Monad m => String -> OrgParser m (F Blocks)
+exportBlock :: Monad m => Text -> OrgParser m (F Blocks)
 exportBlock blockType = try $ do
   exportType <- skipSpaces *> orgArgWord <* ignHeaders
   contents   <- rawBlockContent blockType
-  returnF (B.rawBlock (map toLower exportType) contents)
+  returnF (B.rawBlock (T.toLower exportType) contents)
 
-verseBlock :: PandocMonad m => String -> OrgParser m (F Blocks)
+verseBlock :: PandocMonad m => Text -> OrgParser m (F Blocks)
 verseBlock blockType = try $ do
   ignHeaders
   content <- rawBlockContent blockType
   fmap B.lineBlock . sequence
-    <$> mapM parseVerseLine (lines content)
+    <$> mapM parseVerseLine (T.lines content)
  where
    -- replace initial spaces with nonbreaking spaces to preserve
    -- indentation, parse the rest as normal inline
-   parseVerseLine :: PandocMonad m => String -> OrgParser m (F Inlines)
+   parseVerseLine :: PandocMonad m => Text -> OrgParser m (F Inlines)
    parseVerseLine cs = do
-     let (initialSpaces, indentedLine) = span isSpace cs
-     let nbspIndent = if null initialSpaces
+     let (initialSpaces, indentedLine) = T.span isSpace cs
+     let nbspIndent = if T.null initialSpaces
                       then mempty
-                      else B.str $ map (const '\160') initialSpaces
-     line <- parseFromString inlines (indentedLine ++ "\n")
+                      else B.str $ T.map (const '\160') initialSpaces
+     line <- parseFromString inlines (indentedLine <> "\n")
      return (trimInlinesF $ pure nbspIndent <> line)
 
 -- | Read a code block and the associated results block if present.  Which of
 -- boths blocks is included in the output is determined using the "exports"
 -- argument in the block header.
-codeBlock :: PandocMonad m => BlockAttributes -> String -> OrgParser m (F Blocks)
+codeBlock :: PandocMonad m => BlockAttributes -> Text -> OrgParser m (F Blocks)
 codeBlock blockAttrs blockType = do
   skipSpaces
   (classes, kv)     <- codeHeaderArgs <|> (mempty <$ ignHeaders)
@@ -314,7 +313,7 @@ codeBlock blockAttrs blockType = do
    labelledBlock :: F Inlines -> F Blocks
    labelledBlock = fmap (B.plain . B.spanWith ("", ["label"], []))
 
-   exportsResults :: [(String, String)] -> Bool
+   exportsResults :: [(Text, Text)] -> Bool
    exportsResults = maybe False (`elem` ["results", "both"]) . lookup "exports"
 
 -- | Parse the result of an evaluated babel code block.
@@ -329,7 +328,7 @@ babelResultsBlock = try $ do
   resultsMarker = try . void $ stringAnyCase "#+RESULTS:" *> blankline
 
 -- | Parse code block arguments
-codeHeaderArgs :: Monad m => OrgParser m ([String], [(String, String)])
+codeHeaderArgs :: Monad m => OrgParser m ([Text], [(Text, Text)])
 codeHeaderArgs = try $ do
   language   <- skipSpaces *> orgArgWord
   (switchClasses, switchKv) <- switchesAsAttributes
@@ -338,14 +337,14 @@ codeHeaderArgs = try $ do
          , originalLang language <> switchKv <> parameters
          )
 
-switchesAsAttributes :: Monad m => OrgParser m ([String], [(String, String)])
+switchesAsAttributes :: Monad m => OrgParser m ([Text], [(Text, Text)])
 switchesAsAttributes = try $ do
   switches <- skipSpaces *> try (switch `sepBy` many1 spaceChar)
   return $ foldr addToAttr ([], []) switches
  where
-  addToAttr :: (Char, Maybe String, SwitchPolarity)
-            -> ([String], [(String, String)])
-            -> ([String], [(String, String)])
+  addToAttr :: (Char, Maybe Text, SwitchPolarity)
+            -> ([Text], [(Text, Text)])
+            -> ([Text], [(Text, Text)])
   addToAttr ('n', lineNum, pol) (cls, kv) =
     let kv' = case lineNum of
                 Just num -> ("startFrom", num):kv
@@ -365,18 +364,25 @@ switchPolarity :: Monad m => OrgParser m SwitchPolarity
 switchPolarity = (SwitchMinus <$ char '-') <|> (SwitchPlus <$ char '+')
 
 -- | Parses a source block switch option.
-switch :: Monad m => OrgParser m (Char, Maybe String, SwitchPolarity)
-switch = try $ lineNumberSwitch <|> labelSwitch <|> simpleSwitch
+switch :: Monad m => OrgParser m (Char, Maybe Text, SwitchPolarity)
+switch = try $ lineNumberSwitch <|> labelSwitch
+               <|> whitespaceSwitch <|> simpleSwitch
  where
    simpleSwitch = (\pol c -> (c, Nothing, pol)) <$> switchPolarity <*> letter
    labelSwitch = genericSwitch 'l' $
-     char '"' *> many1Till nonspaceChar (char '"')
+     char '"' *> many1TillChar nonspaceChar (char '"')
+
+whitespaceSwitch :: Monad m => OrgParser m (Char, Maybe Text, SwitchPolarity)
+whitespaceSwitch = do
+  string "-i"
+  updateState $ \s -> s { orgStateTrimLeadBlkIndent = False }
+  return ('i', Nothing, SwitchMinus)
 
 -- | Generic source block switch-option parser.
 genericSwitch :: Monad m
               => Char
-              -> OrgParser m String
-              -> OrgParser m (Char, Maybe String, SwitchPolarity)
+              -> OrgParser m Text
+              -> OrgParser m (Char, Maybe Text, SwitchPolarity)
 genericSwitch c p = try $ do
   polarity <- switchPolarity <* char c <* skipSpaces
   arg <- optionMaybe p
@@ -384,17 +390,17 @@ genericSwitch c p = try $ do
 
 -- | Reads a line number switch option. The line number switch can be used with
 -- example and source blocks.
-lineNumberSwitch :: Monad m => OrgParser m (Char, Maybe String, SwitchPolarity)
-lineNumberSwitch = genericSwitch 'n' (many digit)
+lineNumberSwitch :: Monad m => OrgParser m (Char, Maybe Text, SwitchPolarity)
+lineNumberSwitch = genericSwitch 'n' (manyChar digit)
 
-blockOption :: Monad m => OrgParser m (String, String)
+blockOption :: Monad m => OrgParser m (Text, Text)
 blockOption = try $ do
   argKey <- orgArgKey
   paramValue <- option "yes" orgParamValue
   return (argKey, paramValue)
 
-orgParamValue :: Monad m => OrgParser m String
-orgParamValue = try $
+orgParamValue :: Monad m => OrgParser m Text
+orgParamValue = try $ fmap T.pack $
   skipSpaces
     *> notFollowedBy orgArgKey
     *> noneOf "\n\r" `many1Till` endOfValue
@@ -413,7 +419,7 @@ orgParamValue = try $
 -- export setting.
 genericDrawer :: PandocMonad m => OrgParser m (F Blocks)
 genericDrawer = try $ do
-  name    <- map toUpper <$> drawerStart
+  name    <- T.toUpper <$> drawerStart
   content <- manyTill drawerLine (try drawerEnd)
   state   <- getState
   -- Include drawer if it is explicitly included in or not explicitly excluded
@@ -425,16 +431,16 @@ genericDrawer = try $ do
     Right names | name `notElem` names -> return mempty
     _           -> drawerDiv name <$> parseLines content
  where
-  parseLines :: PandocMonad m => [String] -> OrgParser m (F Blocks)
-  parseLines = parseFromString blocks . (++ "\n") . unlines
+  parseLines :: PandocMonad m => [Text] -> OrgParser m (F Blocks)
+  parseLines = parseFromString blocks . (<> "\n") . T.unlines
 
-  drawerDiv :: String -> F Blocks -> F Blocks
+  drawerDiv :: Text -> F Blocks -> F Blocks
   drawerDiv drawerName = fmap $ B.divWith (mempty, [drawerName, "drawer"], mempty)
 
-drawerLine :: Monad m => OrgParser m String
+drawerLine :: Monad m => OrgParser m Text
 drawerLine = anyLine
 
-drawerEnd :: Monad m => OrgParser m String
+drawerEnd :: Monad m => OrgParser m Text
 drawerEnd = try $
   skipSpaces *> stringAnyCase ":END:" <* skipSpaces <* newline
 
@@ -449,17 +455,17 @@ figure :: PandocMonad m => OrgParser m (F Blocks)
 figure = try $ do
   figAttrs <- blockAttributes
   src <- skipSpaces *> selfTarget <* skipSpaces <* endOfParagraph
-  case cleanLinkString src of
+  case cleanLinkText src of
     Nothing     -> mzero
     Just imgSrc -> do
       guard (isImageFilename imgSrc)
       let isFigure = isJust $ blockAttrCaption figAttrs
       return $ imageBlock isFigure figAttrs imgSrc
  where
-   selfTarget :: PandocMonad m => OrgParser m String
+   selfTarget :: PandocMonad m => OrgParser m Text
    selfTarget = try $ char '[' *> linkTarget <* char ']'
 
-   imageBlock :: Bool -> BlockAttributes -> String -> F Blocks
+   imageBlock :: Bool -> BlockAttributes -> Text -> F Blocks
    imageBlock isFigure figAttrs imgSrc =
      let
        figName    = fromMaybe mempty $ blockAttrName figAttrs
@@ -471,11 +477,11 @@ figure = try $ do
      in
        B.para . B.imageWith attr imgSrc figTitle <$> figCaption
 
-   withFigPrefix :: String -> String
+   withFigPrefix :: Text -> Text
    withFigPrefix cs =
-     if "fig:" `isPrefixOf` cs
+     if "fig:" `T.isPrefixOf` cs
      then cs
-     else "fig:" ++ cs
+     else "fig:" <> cs
 
 -- | Succeeds if looking at the end of the current paragraph
 endOfParagraph :: Monad m => OrgParser m ()
@@ -488,12 +494,12 @@ endOfParagraph = try $ skipSpaces *> newline *> endOfBlock
 
 -- | Example code marked up by a leading colon.
 example :: Monad m => OrgParser m (F Blocks)
-example = try $ returnF . exampleCode =<< unlines <$> many1 exampleLine
+example = try $ returnF . exampleCode =<< T.unlines <$> many1 exampleLine
  where
-   exampleLine :: Monad m => OrgParser m String
+   exampleLine :: Monad m => OrgParser m Text
    exampleLine = try $ exampleLineStart *> anyLine
 
-exampleCode :: String -> Blocks
+exampleCode :: Text -> Blocks
 exampleCode = B.codeBlockWith ("", ["example"], [])
 
 
@@ -509,7 +515,7 @@ include :: PandocMonad m => OrgParser m (F Blocks)
 include = try $ do
   metaLineStart <* stringAnyCase "include:" <* skipSpaces
   filename <- includeTarget
-  includeArgs <- many (try $ skipSpaces *> many1 alphaNum)
+  includeArgs <- many (try $ skipSpaces *> many1Char alphaNum)
   params <- keyValues
   blocksParser <- case includeArgs of
       ("example" : _) -> return $ pure . B.codeBlock <$> parseRaw
@@ -528,10 +534,10 @@ include = try $ do
     char '"'
     manyTill (noneOf "\n\r\t") (char '"')
 
-  parseRaw :: PandocMonad m => OrgParser m String
-  parseRaw = many anyChar
+  parseRaw :: PandocMonad m => OrgParser m Text
+  parseRaw = manyChar anyChar
 
-  blockFilter :: [(String, String)] -> [Block] -> [Block]
+  blockFilter :: [(Text, Text)] -> [Block] -> [Block]
   blockFilter params blks =
     let minlvl = lookup "minlevel" params
     in case (minlvl >>= safeRead :: Maybe Int) of
@@ -547,10 +553,10 @@ include = try $ do
 
   shiftHeader :: Int -> Block -> Block
   shiftHeader shift blk =
-    if shift <= 0
-    then blk
-    else case blk of
-      (Header lvl attr content) -> Header (lvl - shift) attr content
+    case blk of
+      (Header lvl attr content)
+       | lvl - shift > 0  -> Header (lvl - shift) attr content
+       | otherwise        -> Para content
       _ -> blk
 
 rawExportLine :: PandocMonad m => OrgParser m Blocks
@@ -603,16 +609,22 @@ orgTable = try $ do
   guard =<< not . isFirstInListItem <$> getState
   blockAttrs <- blockAttributes
   lookAhead tableStart
-  do
-    rows <- tableRows
-    let caption = fromMaybe (return mempty) $ blockAttrCaption blockAttrs
-    return $ (<$> caption) . orgToPandocTable . normalizeTable =<< rowsToTable rows
+  rows <- tableRows
+
+  let caption = fromMaybe mempty (blockAttrCaption blockAttrs)
+  let orgTbl = normalizeTable <$> rowsToTable rows
+  -- wrap table in div if a name or label is given
+  let identMb = blockAttrName blockAttrs `mplus` blockAttrLabel blockAttrs
+  let wrap = case identMb of
+        Just ident -> B.divWith (ident, mempty, mempty)
+        Nothing    -> id
+  return . fmap wrap $ (orgToPandocTable <$> orgTbl <*> caption)
 
 orgToPandocTable :: OrgTable
                  -> Inlines
                  -> Blocks
 orgToPandocTable (OrgTable colProps heads lns) caption =
-  let totalWidth = if any isJust (map columnRelWidth colProps)
+  let totalWidth = if any (isJust . columnRelWidth) colProps
                    then Just . sum $ map (fromMaybe 1 . columnRelWidth) colProps
                    else Nothing
   in B.table caption (map (convertColProp totalWidth) colProps) heads lns
@@ -653,7 +665,7 @@ columnPropertyCell = emptyCell <|> propCell <?> "alignment info"
                  <$> (skipSpaces
                       *> char '<'
                       *> optionMaybe tableAlignFromChar)
-                 <*> (optionMaybe (many1 digit >>= safeRead)
+                 <*> (optionMaybe (many1Char digit >>= safeRead)
                       <* char '>'
                       <* emptyCell)
 
@@ -732,10 +744,10 @@ latexFragment = try $ do
                               , "\\end{", e, "}\n"
                               ]
 
-latexEnd :: Monad m => String -> OrgParser m ()
+latexEnd :: Monad m => Text -> OrgParser m ()
 latexEnd envName = try $
   () <$ skipSpaces
-     <* string ("\\end{" ++ envName ++ "}")
+     <* textStr ("\\end{" <> envName <> "}")
      <* blankline
 
 
@@ -789,7 +801,7 @@ bulletList = try $ do
   fmap (B.bulletList . compactify) . sequence
     <$> many1 (listItem (bulletListStart `indented` indent))
 
-indented :: Monad m => OrgParser m Int -> Int -> OrgParser m Int
+indented :: OrgParser m Int -> Int -> OrgParser m Int
 indented indentedMarker minIndent = try $ do
   n <- indentedMarker
   guard (minIndent <= n)
@@ -806,12 +818,12 @@ definitionListItem :: PandocMonad m
                    -> OrgParser m (F (Inlines, [Blocks]))
 definitionListItem parseIndentedMarker = try $ do
   markerLength <- parseIndentedMarker
-  term <- manyTill (noneOf "\n\r") (try definitionMarker)
+  term <- manyTillChar (noneOf "\n\r") (try definitionMarker)
   line1 <- anyLineNewline
   blank <- option "" ("\n" <$ blankline)
-  cont <- concat <$> many (listContinuation markerLength)
+  cont <- T.concat <$> many (listContinuation markerLength)
   term' <- parseFromString inlines term
-  contents' <- parseFromString blocks $ line1 ++ blank ++ cont
+  contents' <- parseFromString blocks $ line1 <> blank <> cont
   return $ (,) <$> term' <*> fmap (:[]) contents'
  where
    definitionMarker =
@@ -825,16 +837,27 @@ listItem parseIndentedMarker = try . withContext ListItemState $ do
   markerLength <- try parseIndentedMarker
   firstLine <- anyLineNewline
   blank <- option "" ("\n" <$ blankline)
-  rest <- concat <$> many (listContinuation markerLength)
-  parseFromString blocks $ firstLine ++ blank ++ rest
+  rest <- T.concat <$> many (listContinuation markerLength)
+  parseFromString blocks $ firstLine <> blank <> rest
 
 -- continuation of a list item - indented and separated by blankline or endline.
 -- Note: nested lists are parsed as continuations.
-listContinuation :: Monad m => Int
-                 -> OrgParser m String
+listContinuation :: PandocMonad m => Int -> OrgParser m Text
 listContinuation markerLength = try $ do
   notFollowedBy' blankline
-  mappend <$> (concat <$> many1 listLine)
-          <*> many blankline
+  mappend <$> (T.concat <$> many1 (listContinuation' markerLength))
+          <*> manyChar blankline
  where
-   listLine = try $ indentWith markerLength *> anyLineNewline
+   listContinuation' indentation =
+      blockLines indentation <|> listLine indentation
+   listLine indentation = try $ indentWith indentation *> anyLineNewline
+  -- The block attributes and start must be appropriately indented,
+  -- but the contents, and end do not.
+   blockLines indentation =
+      try $ lookAhead (indentWith indentation
+                       >> blockAttributes
+                       >>= (\blockAttrs ->
+                              case attrFromBlockAttributes blockAttrs of
+                                ("", [], []) -> countChar 1 anyChar
+                                _ -> indentWith indentation))
+            >> (snd <$> withRaw orgBlock)
