@@ -1,25 +1,9 @@
-{-
-Copyright (C) 2014-2018 Albert Krewinkel <tarleb+pandoc@moltkeplatz.de>
-
-This program is free software; you can redistribute it and/or modify
-it under the terms of the GNU General Public License as published by
-the Free Software Foundation; either version 2 of the License, or
-(at your option) any later version.
-
-This program is distributed in the hope that it will be useful,
-but WITHOUT ANY WARRANTY; without even the implied warranty of
-MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-GNU General Public License for more details.
-
-You should have received a copy of the GNU General Public License
-along with this program; if not, write to the Free Software
-Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--}
-{-# LANGUAGE FlexibleContexts #-}
-{-# LANGUAGE NoImplicitPrelude #-}
+{-# LANGUAGE FlexibleContexts  #-}
+{-# LANGUAGE OverloadedStrings #-}
+{-# LANGUAGE TupleSections     #-}
 {- |
    Module      : Text.Pandoc.Readers.Org.DocumentTree
-   Copyright   : Copyright (C) 2014-2018 Albert Krewinkel
+   Copyright   : Copyright (C) 2014-2020 Albert Krewinkel
    License     : GNU GPL, version 2 or above
 
    Maintainer  : Albert Krewinkel <tarleb+pandoc@moltkeplatz.de>
@@ -28,22 +12,23 @@ Parsers for org-mode headlines and document subtrees
 -}
 module Text.Pandoc.Readers.Org.DocumentTree
   ( documentTree
-  , headlineToBlocks
+  , unprunedHeadlineToBlocks
   ) where
 
-import Prelude
-import Control.Arrow ((***))
-import Control.Monad (guard, void)
-import Data.Char (toLower, toUpper)
+import Control.Arrow ((***), first)
+import Control.Monad (guard)
 import Data.List (intersperse)
+import Data.Maybe (mapMaybe)
+import Data.Text (Text)
 import Text.Pandoc.Builder (Blocks, Inlines)
-import Text.Pandoc.Class (PandocMonad)
+import Text.Pandoc.Class.PandocMonad (PandocMonad)
 import Text.Pandoc.Definition
 import Text.Pandoc.Readers.Org.BlockStarts
 import Text.Pandoc.Readers.Org.ParserState
 import Text.Pandoc.Readers.Org.Parsing
 
 import qualified Data.Set as Set
+import qualified Data.Text as T
 import qualified Text.Pandoc.Builder as B
 
 --
@@ -75,28 +60,28 @@ documentTree blocks inline = do
       }
 
 -- | Create a tag containing the given string.
-toTag :: String -> Tag
+toTag :: Text -> Tag
 toTag = Tag
 
 -- | The key (also called name or type) of a property.
-newtype PropertyKey = PropertyKey { fromKey :: String }
+newtype PropertyKey = PropertyKey { fromKey :: Text }
   deriving (Show, Eq, Ord)
 
 -- | Create a property key containing the given string.  Org mode keys are
 -- case insensitive and are hence converted to lower case.
-toPropertyKey :: String -> PropertyKey
-toPropertyKey = PropertyKey . map toLower
+toPropertyKey :: Text -> PropertyKey
+toPropertyKey = PropertyKey . T.toLower
 
 -- | The value assigned to a property.
-newtype PropertyValue = PropertyValue { fromValue :: String }
+newtype PropertyValue = PropertyValue { fromValue :: Text }
 
 -- | Create a property value containing the given string.
-toPropertyValue :: String -> PropertyValue
+toPropertyValue :: Text -> PropertyValue
 toPropertyValue = PropertyValue
 
 -- | Check whether the property value is non-nil (i.e. truish).
 isNonNil :: PropertyValue -> Bool
-isNonNil p = map toLower (fromValue p) `notElem` ["()", "{}", "nil"]
+isNonNil p = T.toLower (fromValue p) `notElem` ["()", "{}", "nil"]
 
 -- | Key/value pairs from a PROPERTIES drawer
 type Properties = [(PropertyKey, PropertyValue)]
@@ -124,15 +109,13 @@ headline blocks inline lvl = try $ do
   level <- headerStart
   guard (lvl <= level)
   todoKw <- optionMaybe todoKeyword
-  title <- trimInlinesF . mconcat <$> manyTill inline endOfTitle
-  tags  <- option [] headerTags
-  newline
+  (title, tags) <- manyThen inline endOfTitle
   planning   <- option emptyPlanning planningInfo
   properties <- option mempty propertiesDrawer
   contents   <- blocks
   children   <- many (headline blocks inline (level + 1))
   return $ do
-    title'    <- title
+    title'    <- trimInlinesF (mconcat title)
     contents' <- contents
     children' <- sequence children
     return Headline
@@ -146,13 +129,45 @@ headline blocks inline lvl = try $ do
       , headlineChildren = children'
       }
  where
-   endOfTitle :: Monad m => OrgParser m ()
-   endOfTitle = void . lookAhead $ optional headerTags *> newline
+   endOfTitle :: Monad m => OrgParser m [Tag]
+   endOfTitle = try $ do
+     skipSpaces
+     tags <- option [] (headerTags <* skipSpaces)
+     newline
+     return tags
 
    headerTags :: Monad m => OrgParser m [Tag]
-   headerTags = try $
-     let tag = orgTagWord <* char ':'
-     in map toTag <$> (skipSpaces *> char ':' *> many1 tag <* skipSpaces)
+   headerTags = try $ do
+     char ':'
+     endBy1 (toTag <$> orgTagWord) (char ':')
+
+   manyThen :: Monad m
+            => OrgParser m a
+            -> OrgParser m b
+            -> OrgParser m ([a], b)
+   manyThen p end = (([],) <$> try end) <|> do
+     x <- p
+     first (x:) <$> manyThen p end
+
+   -- titleFollowedByTags :: Monad m => OrgParser m (Inlines, [Tag])
+   -- titleFollowedByTags = do
+
+
+unprunedHeadlineToBlocks :: Monad m => Headline -> OrgParserState -> OrgParser m [Block]
+unprunedHeadlineToBlocks hdln st =
+  let usingSelectedTags = docContainsSelectTags hdln st
+      rootNode = if not usingSelectedTags
+                   then hdln
+                   else includeRootAndSelected hdln st
+      rootNode' = removeExplicitlyExcludedNodes rootNode st
+  in if not usingSelectedTags ||
+        any (`Set.member` orgStateSelectTags st) (headlineTags rootNode')
+        then do headlineBlocks <- headlineToBlocks rootNode'
+                -- ignore first headline, it's the document's title
+                return . drop 1 . B.toList $ headlineBlocks
+        else do headlineBlocks <- mconcat <$> mapM headlineToBlocks
+                                                   (headlineChildren rootNode')
+                return . B.toList $ headlineBlocks
 
 -- | Convert an Org mode headline (i.e. a document tree) into pandoc's Blocks
 headlineToBlocks :: Monad m => Headline -> OrgParser m Blocks
@@ -161,17 +176,46 @@ headlineToBlocks hdln = do
   let tags = headlineTags hdln
   let text = headlineText hdln
   let level = headlineLevel hdln
-  shouldNotExport <- hasDoNotExportTag tags
   case () of
-    _ | shouldNotExport -> return mempty
     _ | any isArchiveTag  tags -> archivedHeadlineToBlocks hdln
     _ | isCommentTitle text    -> return mempty
     _ | maxLevel <= level      -> headlineToHeaderWithList hdln
     _ | otherwise              -> headlineToHeaderWithContents hdln
 
-hasDoNotExportTag :: Monad m => [Tag] -> OrgParser m Bool
-hasDoNotExportTag tags = containsExcludedTag . orgStateExcludedTags <$> getState
-  where containsExcludedTag s = any (`Set.member` s) tags
+removeExplicitlyExcludedNodes :: Headline -> OrgParserState -> Headline
+removeExplicitlyExcludedNodes hdln st =
+  hdln { headlineChildren =
+           [removeExplicitlyExcludedNodes childHdln st |
+              childHdln <- headlineChildren hdln,
+              not $ headlineContainsExcludeTags childHdln st] }
+
+includeRootAndSelected :: Headline -> OrgParserState -> Headline
+includeRootAndSelected hdln st =
+  hdln { headlineChildren = mapMaybe (`includeAncestorsAndSelected` st)
+                                     (headlineChildren hdln)}
+
+docContainsSelectTags :: Headline -> OrgParserState -> Bool
+docContainsSelectTags hdln st =
+  headlineContainsSelectTags hdln st ||
+  any (`docContainsSelectTags` st) (headlineChildren hdln)
+
+includeAncestorsAndSelected :: Headline -> OrgParserState -> Maybe Headline
+includeAncestorsAndSelected hdln st =
+  if headlineContainsSelectTags hdln st
+    then Just hdln
+    else let children = mapMaybe (`includeAncestorsAndSelected` st)
+                                 (headlineChildren hdln)
+         in case children of
+              [] -> Nothing
+              _ -> Just $ hdln { headlineChildren = children }
+
+headlineContainsSelectTags :: Headline -> OrgParserState -> Bool
+headlineContainsSelectTags hdln st =
+  any (`Set.member` orgStateSelectTags st) (headlineTags hdln)
+
+headlineContainsExcludeTags :: Headline -> OrgParserState -> Bool
+headlineContainsExcludeTags hdln st =
+  any (`Set.member` orgStateExcludeTags st) (headlineTags hdln)
 
 isArchiveTag :: Tag -> Bool
 isArchiveTag = (== toTag "ARCHIVE")
@@ -206,8 +250,8 @@ headlineToHeaderWithList hdln = do
                    else flattenHeader header
   return . mconcat $
     [ headerText
-    , headlineContents hdln
     , planningBlock
+    , headlineContents hdln
     , listBlock
     ]
  where
@@ -244,32 +288,34 @@ headlineToHeader hdln = do
 todoKeyword :: Monad m => OrgParser m TodoMarker
 todoKeyword = try $ do
   taskStates <- activeTodoMarkers <$> getState
-  let kwParser tdm = try (tdm <$ string (todoMarkerName tdm) <* spaceChar)
+  let kwParser tdm = try (tdm <$ textStr (todoMarkerName tdm)
+                              <* spaceChar
+                              <* updateLastPreCharPos)
   choice (map kwParser taskStates)
 
 todoKeywordToInlines :: TodoMarker -> Inlines
 todoKeywordToInlines tdm =
   let todoText  = todoMarkerName tdm
-      todoState = map toLower . show $ todoMarkerState tdm
+      todoState = T.toLower . T.pack . show $ todoMarkerState tdm
       classes = [todoState, todoText]
   in B.spanWith (mempty, classes, mempty) (B.str todoText)
 
 propertiesToAttr :: Properties -> Attr
 propertiesToAttr properties =
   let
-    toStringPair = fromKey *** fromValue
+    toTextPair = fromKey *** fromValue
     customIdKey = toPropertyKey "custom_id"
     classKey    = toPropertyKey "class"
     unnumberedKey = toPropertyKey "unnumbered"
     specialProperties = [customIdKey, classKey, unnumberedKey]
     id'  = maybe mempty fromValue . lookup customIdKey $ properties
     cls  = maybe mempty fromValue . lookup classKey    $ properties
-    kvs' = map toStringPair . filter ((`notElem` specialProperties) . fst)
+    kvs' = map toTextPair . filter ((`notElem` specialProperties) . fst)
            $ properties
     isUnnumbered =
       maybe False isNonNil . lookup unnumberedKey $ properties
   in
-    (id', words cls ++ ["unnumbered" | isUnnumbered], kvs')
+    (id', T.words cls ++ ["unnumbered" | isUnnumbered], kvs')
 
 tagsToInlines :: [Tag] -> Inlines
 tagsToInlines [] = mempty
@@ -305,15 +351,15 @@ planningToBlock planning = do
                   <> B.emph (B.str time)
 
 -- | An Org timestamp, including repetition marks. TODO: improve
-type Timestamp = String
+type Timestamp = Text
 
 timestamp :: Monad m => OrgParser m Timestamp
 timestamp = try $ do
   openChar <- oneOf "<["
   let isActive = openChar == '<'
   let closeChar = if isActive then '>' else ']'
-  content <- many1Till anyChar (char closeChar)
-  return (openChar : content ++ [closeChar])
+  content <- many1TillChar anyChar (char closeChar)
+  return $ T.cons openChar $ content `T.snoc` closeChar
 
 -- | Planning information for a subtree/headline.
 data PlanningInfo = PlanningInfo
@@ -343,7 +389,7 @@ planningInfo = try $ do
 propertiesDrawer :: Monad m => OrgParser m Properties
 propertiesDrawer = try $ do
   drawerType <- drawerStart
-  guard $ map toUpper drawerType == "PROPERTIES"
+  guard $ T.toUpper drawerType == "PROPERTIES"
   manyTill property (try endOfDrawer)
  where
    property :: Monad m => OrgParser m (PropertyKey, PropertyValue)
@@ -351,12 +397,12 @@ propertiesDrawer = try $ do
 
    key :: Monad m => OrgParser m PropertyKey
    key = fmap toPropertyKey . try $
-         skipSpaces *> char ':' *> many1Till nonspaceChar (char ':')
+         skipSpaces *> char ':' *> many1TillChar nonspaceChar (char ':')
 
    value :: Monad m => OrgParser m PropertyValue
    value = fmap toPropertyValue . try $
-           skipSpaces *> manyTill anyChar (try $ skipSpaces *> newline)
+           skipSpaces *> manyTillChar anyChar (try $ skipSpaces *> newline)
 
-   endOfDrawer :: Monad m => OrgParser m String
+   endOfDrawer :: Monad m => OrgParser m Text
    endOfDrawer = try $
      skipSpaces *> stringAnyCase ":END:" <* skipSpaces <* newline
