@@ -1,27 +1,9 @@
-{-# LANGUAGE NoImplicitPrelude #-}
-{-# LANGUAGE FlexibleContexts #-}
-{-# LANGUAGE TupleSections    #-}
-{-
-Copyright (C) 2014-2018 Albert Krewinkel <tarleb+pandoc@moltkeplatz.de>
-
-This program is free software; you can redistribute it and/or modify
-it under the terms of the GNU General Public License as published by
-the Free Software Foundation; either version 2 of the License, or
-(at your option) any later version.
-
-This program is distributed in the hope that it will be useful,
-but WITHOUT ANY WARRANTY; without even the implied warranty of
-MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-GNU General Public License for more details.
-
-You should have received a copy of the GNU General Public License
-along with this program; if not, write to the Free Software
-Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--}
-
+{-# LANGUAGE FlexibleContexts  #-}
+{-# LANGUAGE TupleSections     #-}
+{-# LANGUAGE OverloadedStrings #-}
 {- |
    Module      : Text.Pandoc.Readers.Org.Meta
-   Copyright   : Copyright (C) 2014-2018 Albert Krewinkel
+   Copyright   : Copyright (C) 2014-2020 Albert Krewinkel
    License     : GNU GPL, version 2 or above
 
    Maintainer  : Albert Krewinkel <tarleb+pandoc@moltkeplatz.de>
@@ -34,7 +16,6 @@ module Text.Pandoc.Readers.Org.Meta
   , metaLine
   ) where
 
-import Prelude
 import Text.Pandoc.Readers.Org.BlockStarts
 import Text.Pandoc.Readers.Org.ExportSettings (exportSettings)
 import Text.Pandoc.Readers.Org.Inlines
@@ -43,16 +24,17 @@ import Text.Pandoc.Readers.Org.Parsing
 
 import Text.Pandoc.Builder (Blocks, Inlines)
 import qualified Text.Pandoc.Builder as B
-import Text.Pandoc.Class (PandocMonad)
+import Text.Pandoc.Class.PandocMonad (PandocMonad)
 import Text.Pandoc.Definition
 import Text.Pandoc.Shared (safeRead)
 
 import Control.Monad (mzero, void, when)
-import Data.Char (toLower)
 import Data.List (intersperse)
 import Data.Maybe (fromMaybe)
 import qualified Data.Map as M
 import qualified Data.Set as Set
+import Data.Text (Text)
+import qualified Data.Text as T
 import Network.HTTP (urlEncode)
 
 -- | Returns the current meta, respecting export options.
@@ -65,7 +47,7 @@ metaExport = do
          . (if exportWithEmail   settings then id else removeMeta "email")
         <$> orgStateMeta st
 
-removeMeta :: String -> Meta -> Meta
+removeMeta :: Text -> Meta -> Meta
 removeMeta key meta' =
   let metaMap = unMeta meta'
   in Meta $ M.delete key metaMap
@@ -78,18 +60,18 @@ metaLine = mempty <$ metaLineStart <* (optionLine <|> declarationLine)
 
 declarationLine :: PandocMonad m => OrgParser m ()
 declarationLine = try $ do
-  key   <- map toLower <$> metaKey
+  key   <- T.toLower <$> metaKey
   (key', value) <- metaValue key
   let addMetaValue st =
         st { orgStateMeta = B.setMeta key' <$> value <*> orgStateMeta st }
   when (key' /= "results") $ updateState addMetaValue
 
-metaKey :: Monad m => OrgParser m String
-metaKey = map toLower <$> many1 (noneOf ": \n\r")
-                      <*  char ':'
-                      <*  skipSpaces
+metaKey :: Monad m => OrgParser m Text
+metaKey = T.toLower <$> many1Char (noneOf ": \n\r")
+                    <*  char ':'
+                    <*  skipSpaces
 
-metaValue :: PandocMonad m => String -> OrgParser m (String, F MetaValue)
+metaValue :: PandocMonad m => Text -> OrgParser m (Text, F MetaValue)
 metaValue key =
   let inclKey = "header-includes"
   in case key of
@@ -106,7 +88,7 @@ metaValue key =
     -- Org-mode expects class options to contain the surrounding brackets,
     -- pandoc does not.
     "latex_class_options" -> ("classoption",) <$>
-                             metaModifiedString (filter (`notElem` "[]"))
+                             metaModifiedString (T.filter (`notElem` ("[]" :: String)))
     "html_head"       -> (inclKey,) <$>
                          accumulatingList inclKey (metaExportSnippet "html")
     _                 -> (key,) <$> metaString
@@ -116,25 +98,25 @@ metaInlines = fmap (MetaInlines . B.toList) <$> inlinesTillNewline
 
 metaInlinesCommaSeparated :: PandocMonad m => OrgParser m (F MetaValue)
 metaInlinesCommaSeparated = do
-  itemStrs <- many1 (noneOf ",\n") `sepBy1` char ','
+  itemStrs <- many1Char (noneOf ",\n") `sepBy1` char ','
   newline
-  items <- mapM (parseFromString inlinesTillNewline . (++ "\n")) itemStrs
+  items <- mapM (parseFromString inlinesTillNewline . (<> "\n")) itemStrs
   let toMetaInlines = MetaInlines . B.toList
   return $ MetaList . map toMetaInlines <$> sequence items
 
 metaString :: Monad m => OrgParser m (F MetaValue)
 metaString = metaModifiedString id
 
-metaModifiedString :: Monad m => (String -> String) -> OrgParser m (F MetaValue)
+metaModifiedString :: Monad m => (Text -> Text) -> OrgParser m (F MetaValue)
 metaModifiedString f = return . MetaString . f <$> anyLine
 
 -- | Read an format specific meta definition
-metaExportSnippet :: Monad m => String -> OrgParser m (F MetaValue)
+metaExportSnippet :: Monad m => Text -> OrgParser m (F MetaValue)
 metaExportSnippet format =
   return . MetaInlines . B.toList . B.rawInline format <$> anyLine
 
 -- | Accumulate the result of the @parser@ in a list under @key@.
-accumulatingList :: Monad m => String
+accumulatingList :: Monad m => Text
                  -> OrgParser m (F MetaValue)
                  -> OrgParser m (F MetaValue)
 accumulatingList key p = do
@@ -149,7 +131,7 @@ accumulatingList key p = do
 --
 -- export options
 --
-optionLine :: Monad m => OrgParser m ()
+optionLine :: PandocMonad m => OrgParser m ()
 optionLine = try $ do
   key <- metaKey
   case key of
@@ -159,50 +141,58 @@ optionLine = try $ do
     "seq_todo" -> todoSequence >>= updateState . registerTodoSequence
     "typ_todo" -> todoSequence >>= updateState . registerTodoSequence
     "macro"    -> macroDefinition >>= updateState . registerMacro
-    "exclude_tags" -> excludedTagList >>= updateState . setExcludedTags
+    "exclude_tags" -> tagList >>= updateState . setExcludedTags
+    "select_tags" -> tagList >>= updateState . setSelectedTags
     "pandoc-emphasis-pre" -> emphChars >>= updateState . setEmphasisPreChar
     "pandoc-emphasis-post" -> emphChars >>= updateState . setEmphasisPostChar
     _          -> mzero
 
-addLinkFormat :: Monad m => String
-              -> (String -> String)
+addLinkFormat :: Monad m => Text
+              -> (Text -> Text)
               -> OrgParser m ()
 addLinkFormat key formatter = updateState $ \s ->
   let fs = orgStateLinkFormatters s
   in s{ orgStateLinkFormatters = M.insert key formatter fs }
 
-parseLinkFormat :: Monad m => OrgParser m (String, String -> String)
+parseLinkFormat :: Monad m => OrgParser m (Text, Text -> Text)
 parseLinkFormat = try $ do
-  linkType <- (:) <$> letter <*> many (alphaNum <|> oneOf "-_") <* skipSpaces
+  linkType <- T.cons <$> letter <*> manyChar (alphaNum <|> oneOf "-_") <* skipSpaces
   linkSubst <- parseFormat
   return (linkType, linkSubst)
 
 -- | An ad-hoc, single-argument-only implementation of a printf-style format
 -- parser.
-parseFormat :: Monad m => OrgParser m (String -> String)
+parseFormat :: Monad m => OrgParser m (Text -> Text)
 parseFormat = try $ replacePlain <|> replaceUrl <|> justAppend
  where
    -- inefficient, but who cares
-   replacePlain = try $ (\x -> concat . flip intersperse x)
+   replacePlain = try $ (\x -> T.concat . flip intersperse x)
                      <$> sequence [tillSpecifier 's', rest]
-   replaceUrl   = try $ (\x -> concat . flip intersperse x . urlEncode)
+   replaceUrl   = try $ (\x -> T.concat . flip intersperse x . T.pack . urlEncode . T.unpack)
                      <$> sequence [tillSpecifier 'h', rest]
-   justAppend   = try $ (++) <$> rest
+   justAppend   = try $ (<>) <$> rest
 
-   rest            = manyTill anyChar         (eof <|> () <$ oneOf "\n\r")
-   tillSpecifier c = manyTill (noneOf "\n\r") (try $ string ('%':c:""))
+   rest            = manyTillChar anyChar         (eof <|> () <$ oneOf "\n\r")
+   tillSpecifier c = manyTillChar (noneOf "\n\r") (try $ string ('%':c:""))
 
-excludedTagList :: Monad m => OrgParser m [Tag]
-excludedTagList = do
+tagList :: Monad m => OrgParser m [Tag]
+tagList = do
   skipSpaces
   map Tag <$> many (orgTagWord <* skipSpaces) <* newline
 
 setExcludedTags :: [Tag] -> OrgParserState -> OrgParserState
-setExcludedTags tagList st =
-  let finalSet = if orgStateExcludedTagsChanged st
-                   then foldr Set.insert (orgStateExcludedTags st) tagList
-                   else Set.fromList tagList
-  in st { orgStateExcludedTags = finalSet, orgStateExcludedTagsChanged = True }
+setExcludedTags tags st =
+  let finalSet = if orgStateExcludeTagsChanged st
+                   then foldr Set.insert (orgStateExcludeTags st) tags
+                   else Set.fromList tags
+  in st { orgStateExcludeTags = finalSet, orgStateExcludeTagsChanged = True }
+
+setSelectedTags :: [Tag] -> OrgParserState -> OrgParserState
+setSelectedTags tags st =
+  let finalSet = if orgStateSelectTagsChanged st
+                   then foldr Set.insert (orgStateSelectTags st) tags
+                   else Set.fromList tags
+  in st { orgStateSelectTags = finalSet, orgStateSelectTagsChanged = True }
 
 setEmphasisPreChar :: Maybe [Char] -> OrgParserState -> OrgParserState
 setEmphasisPreChar csMb st =
@@ -241,41 +231,41 @@ todoSequence = try $ do
                     (x:xs) -> return $ keywordsToSequence (reverse xs) [x]
 
  where
-   todoKeywords :: Monad m => OrgParser m [String]
+   todoKeywords :: Monad m => OrgParser m [Text]
    todoKeywords = try $
-     let keyword = many1 nonspaceChar <* skipSpaces
+     let keyword = many1Char nonspaceChar <* skipSpaces
          endOfKeywords = todoDoneSep <|> void newline
      in manyTill keyword (lookAhead endOfKeywords)
 
    todoDoneSep :: Monad m => OrgParser m ()
    todoDoneSep = void . try $ skipSpaces *> char '|' <* skipSpaces1
 
-   keywordsToSequence :: [String] -> [String] -> TodoSequence
+   keywordsToSequence :: [Text] -> [Text] -> TodoSequence
    keywordsToSequence todo done =
      let todoMarkers = map (TodoMarker Todo) todo
          doneMarkers = map (TodoMarker Done) done
      in todoMarkers ++ doneMarkers
 
-macroDefinition :: Monad m => OrgParser m (String, [String] -> String)
+macroDefinition :: Monad m => OrgParser m (Text, [Text] -> Text)
 macroDefinition = try $ do
-  macroName <- many1 nonspaceChar <* skipSpaces
+  macroName <- many1Char nonspaceChar <* skipSpaces
   firstPart <- expansionPart
   (elemOrder, parts) <- unzip <$> many ((,) <$> placeholder <*> expansionPart)
   let expander = mconcat . alternate (firstPart:parts) . reorder elemOrder
   return (macroName, expander)
  where
   placeholder :: Monad m => OrgParser m Int
-  placeholder = try . fmap read $ char '$' *> many1 digit
+  placeholder = try . fmap (fromMaybe 1 . safeRead) $ char '$' *> many1Char digit
 
-  expansionPart :: Monad m => OrgParser m String
-  expansionPart = try $ many (notFollowedBy placeholder *> noneOf "\n\r")
+  expansionPart :: Monad m => OrgParser m Text
+  expansionPart = try $ manyChar (notFollowedBy placeholder *> noneOf "\n\r")
 
   alternate :: [a] -> [a] -> [a]
   alternate []     ys     = ys
   alternate xs     []     = xs
   alternate (x:xs) (y:ys) = x : y : alternate xs ys
 
-  reorder :: [Int] -> [String] -> [String]
+  reorder :: [Int] -> [Text] -> [Text]
   reorder perm xs =
     let element n = take 1 $ drop (n - 1) xs
     in concatMap element perm
