@@ -1,25 +1,6 @@
-{-# LANGUAGE NoImplicitPrelude #-}
-{-
-Copyright (C) 2014-2018 Albert Krewinkel <tarleb+pandoc@moltkeplatz.de>
-
-This program is free software; you can redistribute it and/or modify
-it under the terms of the GNU General Public License as published by
-the Free Software Foundation; either version 2 of the License, or
-(at your option) any later version.
-
-This program is distributed in the hope that it will be useful,
-but WITHOUT ANY WARRANTY; without even the implied warranty of
-MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-GNU General Public License for more details.
-
-You should have received a copy of the GNU General Public License
-along with this program; if not, write to the Free Software
-Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--}
-
 {- |
    Module      : Text.Pandoc.Readers.Org.Parsing
-   Copyright   : Copyright (C) 2014-2018 Albert Krewinkel
+   Copyright   : Copyright (C) 2014-2020 Albert Krewinkel
    License     : GNU GPL, version 2 or above
 
    Maintainer  : Albert Krewinkel <tarleb+pandoc@moltkeplatz.de>
@@ -50,7 +31,13 @@ module Text.Pandoc.Readers.Org.Parsing
   , orgTagWordChar
   -- * Re-exports from Text.Pandoc.Parser
   , ParserContext (..)
+  , textStr
+  , countChar
+  , manyChar
+  , many1Char
+  , manyTillChar
   , many1Till
+  , many1TillChar
   , notFollowedBy'
   , spaceChar
   , nonspaceChar
@@ -106,6 +93,7 @@ module Text.Pandoc.Readers.Org.Parsing
   , sepBy
   , sepBy1
   , sepEndBy1
+  , endBy1
   , option
   , optional
   , optionMaybe
@@ -115,7 +103,7 @@ module Text.Pandoc.Readers.Org.Parsing
   , getPosition
   ) where
 
-import Prelude
+import Data.Text (Text)
 import Text.Pandoc.Readers.Org.ParserState
 
 import Text.Pandoc.Parsing hiding (F, anyLine, blanklines, newline,
@@ -126,14 +114,14 @@ import Control.Monad (guard)
 import Control.Monad.Reader (ReaderT)
 
 -- | The parser used to read org files.
-type OrgParser m = ParserT [Char] OrgParserState (ReaderT OrgParserLocal m)
+type OrgParser m = ParserT Text OrgParserState (ReaderT OrgParserLocal m)
 
 --
 -- Adaptions and specializations of parsing utilities
 --
 
 -- | Parse any line of text
-anyLine :: Monad m => OrgParser m String
+anyLine :: Monad m => OrgParser m Text
 anyLine =
   P.anyLine
     <* updateLastPreCharPos
@@ -141,7 +129,7 @@ anyLine =
 
 -- | Like @'Text.Pandoc.Parsing'@, but resets the position of the last character
 -- allowed before emphasised text.
-parseFromString :: Monad m => OrgParser m a -> String -> OrgParser m a
+parseFromString :: Monad m => OrgParser m a -> Text -> OrgParser m a
 parseFromString parser str' = do
   updateState $ \s -> s{ orgStateLastPreCharPos = Nothing }
   result <- P.parseFromString parser str'
@@ -160,7 +148,7 @@ newline =
        <* updateLastForbiddenCharPos
 
 -- | Like @Text.Parsec.Char.blanklines@, but causes additional state changes.
-blanklines :: Monad m => OrgParser m [Char]
+blanklines :: Monad m => OrgParser m Text
 blanklines =
   P.blanklines
        <* updateLastPreCharPos
@@ -210,21 +198,21 @@ updateLastPreCharPos = getPosition >>= \p ->
 --
 
 -- | Read the key of a plist style key-value list.
-orgArgKey :: Monad m => OrgParser m String
+orgArgKey :: Monad m => OrgParser m Text
 orgArgKey = try $
   skipSpaces *> char ':'
-             *> many1 orgArgWordChar
+             *> many1Char orgArgWordChar
 
 -- | Read the value of a plist style key-value list.
-orgArgWord :: Monad m => OrgParser m String
-orgArgWord = many1 orgArgWordChar
+orgArgWord :: Monad m => OrgParser m Text
+orgArgWord = many1Char orgArgWordChar
 
 -- | Chars treated as part of a word in plists.
 orgArgWordChar :: Monad m => OrgParser m Char
 orgArgWordChar = alphaNum <|> oneOf "-_"
 
-orgTagWord :: Monad m => OrgParser m String
-orgTagWord = many1 orgTagWordChar
+orgTagWord :: Monad m => OrgParser m Text
+orgTagWord = many1Char orgTagWordChar
 
 orgTagWordChar :: Monad m => OrgParser m Char
 orgTagWordChar = alphaNum <|> oneOf "@%#_"
