@@ -1,26 +1,8 @@
-{-# LANGUAGE NoImplicitPrelude #-}
 {-# LANGUAGE OverloadedStrings #-}
-{-
-Copyright (C) 2011-2018 John MacFarlane <jgm@berkeley.edu>
-
-This program is free software; you can redistribute it and/or modify
-it under the terms of the GNU General Public License as published by
-the Free Software Foundation; either version 2 of the License, or
-(at your option) any later version.
-
-This program is distributed in the hope that it will be useful,
-but WITHOUT ANY WARRANTY; without even the implied warranty of
-MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-GNU General Public License for more details.
-
-You should have received a copy of the GNU General Public License
-along with this program; if not, write to the Free Software
-Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--}
-
+{-# LANGUAGE TupleSections     #-}
 {- |
    Module      : Text.Pandoc.SelfContained
-   Copyright   : Copyright (C) 2011-2018 John MacFarlane
+   Copyright   : Copyright (C) 2011-2020 John MacFarlane
    License     : GNU GPL, version 2 or above
 
    Maintainer  : John MacFarlane <jgm@berkeley.edu>
@@ -32,7 +14,6 @@ offline, by incorporating linked images, CSS, and scripts into
 the HTML using data URIs.
 -}
 module Text.Pandoc.SelfContained ( makeDataURI, makeSelfContained ) where
-import Prelude
 import Codec.Compression.GZip as Gzip
 import Control.Applicative ((<|>))
 import Control.Monad.Except (throwError)
@@ -41,35 +22,35 @@ import Data.ByteString (ByteString)
 import Data.ByteString.Base64
 import qualified Data.ByteString.Char8 as B
 import qualified Data.ByteString.Lazy as L
-import Data.Char (isAlphaNum, isAscii, toLower)
-import Data.List (isPrefixOf)
+import qualified Data.Text as T
+import Data.Char (isAlphaNum, isAscii)
 import Network.URI (escapeURIString)
 import System.FilePath (takeDirectory, takeExtension, (</>))
 import Text.HTML.TagSoup
-import Text.Pandoc.Class (PandocMonad (..), fetchItem, getInputFiles, report,
-                          setInputFiles)
+import Text.Pandoc.Class.PandocMonad (PandocMonad (..), fetchItem,
+                                      getInputFiles, report, setInputFiles)
 import Text.Pandoc.Error
 import Text.Pandoc.Logging
 import Text.Pandoc.MIME (MimeType)
 import Text.Pandoc.Shared (isURI, renderTags', trim)
-import Text.Pandoc.UTF8 (toString)
+import Text.Pandoc.UTF8 (toString, toText, fromText)
 import Text.Parsec (ParsecT, runParserT)
 import qualified Text.Parsec as P
 
 isOk :: Char -> Bool
 isOk c = isAscii c && isAlphaNum c
 
-makeDataURI :: (MimeType, ByteString) -> String
+makeDataURI :: (MimeType, ByteString) -> T.Text
 makeDataURI (mime, raw) =
   if textual
-     then "data:" ++ mime' ++ "," ++ escapeURIString isOk (toString raw)
-     else "data:" ++ mime' ++ ";base64," ++ toString (encode raw)
-  where textual = "text/" `Data.List.isPrefixOf` mime
-        mime' = if textual && ';' `notElem` mime
-                   then mime ++ ";charset=utf-8"
+     then "data:" <> mime' <> "," <> T.pack (escapeURIString isOk (toString raw))
+     else "data:" <> mime' <> ";base64," <> toText (encode raw)
+  where textual = "text/" `T.isPrefixOf` mime
+        mime' = if textual && T.any (== ';') mime
+                   then mime <> ";charset=utf-8"
                    else mime  -- mime type already has charset
 
-convertTags :: PandocMonad m => [Tag String] -> m [Tag String]
+convertTags :: PandocMonad m => [Tag T.Text] -> m [Tag T.Text]
 convertTags [] = return []
 convertTags (t@TagOpen{}:ts)
   | fromAttrib "data-external" t == "1" = (t:) <$> convertTags ts
@@ -88,8 +69,8 @@ convertTags (t@(TagOpen tagname as):ts)
               else return (x,y)
 convertTags (t@(TagOpen "script" as):TagClose "script":ts) =
   case fromAttrib "src" t of
-       []  -> (t:) <$> convertTags ts
-       src    -> do
+       ""  -> (t:) <$> convertTags ts
+       src -> do
            let typeAttr = fromAttrib "type" t
            res <- getData typeAttr src
            rest <- convertTags ts
@@ -98,13 +79,13 @@ convertTags (t@(TagOpen "script" as):TagClose "script":ts) =
                      (("src",dataUri) : [(x,y) | (x,y) <- as, x /= "src"]) :
                      TagClose "script" : rest
                 Right (mime, bs)
-                  | ("text/javascript" `isPrefixOf` mime ||
-                     "application/javascript" `isPrefixOf` mime ||
-                     "application/x-javascript" `isPrefixOf` mime) &&
+                  | ("text/javascript" `T.isPrefixOf` mime ||
+                     "application/javascript" `T.isPrefixOf` mime ||
+                     "application/x-javascript" `T.isPrefixOf` mime) &&
                      not ("</script" `B.isInfixOf` bs) ->
                      return $
-                       TagOpen "script" [("type", typeAttr)|not (null typeAttr)]
-                       : TagText (toString bs)
+                       TagOpen "script" [("type", typeAttr)|not (T.null typeAttr)]
+                       : TagText (toText bs)
                        : TagClose "script"
                        : rest
                   | otherwise ->
@@ -114,7 +95,7 @@ convertTags (t@(TagOpen "script" as):TagClose "script":ts) =
                         TagClose "script" : rest
 convertTags (t@(TagOpen "link" as):ts) =
   case fromAttrib "href" t of
-       []  -> (t:) <$> convertTags ts
+       ""  -> (t:) <$> convertTags ts
        src -> do
            res <- getData (fromAttrib "type" t) src
            case res of
@@ -124,14 +105,14 @@ convertTags (t@(TagOpen "link" as):ts) =
                      (("href",dataUri) : [(x,y) | (x,y) <- as, x /= "href"]) :
                      rest
                 Right (mime, bs)
-                  | "text/css" `isPrefixOf` mime
-                    && null (fromAttrib "media" t)
+                  | "text/css" `T.isPrefixOf` mime
+                    && T.null (fromAttrib "media" t)
                     && not ("</" `B.isInfixOf` bs) -> do
                       rest <- convertTags $
                                  dropWhile (==TagClose "link") ts
                       return $
-                       TagOpen "style" [("type", mime)]
-                       : TagText (toString bs)
+                       TagOpen "style" [("type", "text/css")] -- see #5725
+                       : TagText (toText bs)
                        : TagClose "style"
                        : rest
                   | otherwise -> do
@@ -147,7 +128,7 @@ cssURLs d orig = do
   res <- runParserT (parseCSSUrls d) () "css" orig
   case res of
        Left e    -> do
-         report $ CouldNotParseCSS (show e)
+         report $ CouldNotParseCSS $ T.pack $ show e
          return orig
        Right bs  -> return bs
 
@@ -193,50 +174,52 @@ pCSSUrl d = P.try $ do
        Left b -> return b
        Right (mt,b) -> do
          let enc = makeDataURI (mt, b)
-         return (B.pack $ "url(" ++ enc ++ ")")
+         return $ fromText $ "url(" <> enc <> ")"
 
 pQuoted :: PandocMonad m
-        => ParsecT ByteString () m (String, ByteString)
+        => ParsecT ByteString () m (T.Text, ByteString)
 pQuoted = P.try $ do
   quote <- P.oneOf "\"'"
-  url <- P.manyTill P.anyChar (P.char quote)
-  let fallback = B.pack ([quote] ++ trim url ++ [quote])
+  url <- T.pack <$> P.manyTill P.anyChar (P.char quote)
+  let fallback = fromText $ T.singleton quote <> trim url <> T.singleton quote
   return (url, fallback)
 
 pUrl :: PandocMonad m
-     => ParsecT ByteString () m (String, ByteString)
+     => ParsecT ByteString () m (T.Text, ByteString)
 pUrl = P.try $ do
   P.string "url("
   P.spaces
   quote <- P.option Nothing (Just <$> P.oneOf "\"'")
-  url <- P.manyTill P.anyChar (maybe (P.lookAhead (P.char ')')) P.char quote)
+  url <- T.pack <$> P.manyTill P.anyChar (maybe (P.lookAhead (P.char ')')) P.char quote)
   P.spaces
   P.char ')'
-  let fallback = B.pack ("url(" ++ maybe "" (:[]) quote ++ trim url ++
-                            maybe "" (:[]) quote ++ ")")
+  let fallback = fromText ("url(" <> maybe "" T.singleton quote <> trim url <>
+                            maybe "" T.singleton quote <> ")")
   return (url, fallback)
 
 handleCSSUrl :: PandocMonad m
-             => FilePath -> (String, ByteString)
+             => FilePath -> (T.Text, ByteString)
              -> ParsecT ByteString () m
                   (Either ByteString (MimeType, ByteString))
 handleCSSUrl d (url, fallback) =
-  case escapeURIString (/='|') (trim url) of
+  case escapeURIString (/='|') (T.unpack $ trim url) of
       '#':_ -> return $ Left fallback
       'd':'a':'t':'a':':':_ -> return $ Left fallback
-      u ->  do let url' = if isURI u then u else d </> u
+      u ->  do let url' = if isURI (T.pack u) then T.pack u else T.pack (d </> u)
                res <- lift $ getData "" url'
                case res of
-                    Left uri -> return $ Left (B.pack $ "url(" ++ uri ++ ")")
-                    Right (mt, raw) -> do
+                    Left uri -> return $ Left (fromText $ "url(" <> uri <> ")")
+                    Right (mt', raw) -> do
                       -- note that the downloaded CSS may
                       -- itself contain url(...).
-                      b <- if "text/css" `isPrefixOf` mt
-                              then cssURLs d raw
-                              else return raw
+                      (mt, b) <- if "text/css" `T.isPrefixOf` mt'
+                                    -- see #5725: in HTML5, content type
+                                    -- isn't allowed on style type attribute
+                                    then ("text/css",) <$> cssURLs d raw
+                                    else return (mt', raw)
                       return $ Right (mt, b)
 
-getDataURI :: PandocMonad m => MimeType -> String -> m String
+getDataURI :: PandocMonad m => MimeType -> T.Text -> m T.Text
 getDataURI mimetype src = do
   res <- getData mimetype src
   case res of
@@ -244,35 +227,36 @@ getDataURI mimetype src = do
        Right x  -> return $ makeDataURI x
 
 getData :: PandocMonad m
-        => MimeType -> String
-        -> m (Either String (MimeType, ByteString))
-getData _ src@('d':'a':'t':'a':':':_) = return $ Left src-- already data: uri
-getData mimetype src = do
-  let ext = map toLower $ takeExtension src
-  (raw, respMime) <- fetchItem src
-  let raw' = if ext == ".gz"
-                then B.concat $ L.toChunks $ Gzip.decompress $ L.fromChunks [raw]
-                else raw
-  mime <- case (mimetype, respMime) of
-                  ("",Nothing) -> throwError $ PandocSomeError
-                         $ "Could not determine mime type for `" ++ src ++ "'"
-                  (x, Nothing) -> return x
-                  (_, Just x ) -> return x
-  result <- if "text/css" `isPrefixOf` mime
-               then do
-                 oldInputs <- getInputFiles
-                 setInputFiles [src]
-                 res <- cssURLs (takeDirectory src) raw'
-                 setInputFiles oldInputs
-                 return res
+        => MimeType -> T.Text
+        -> m (Either T.Text (MimeType, ByteString))
+getData mimetype src
+  | "data:" `T.isPrefixOf` src = return $ Left src -- already data: uri
+  | otherwise = do
+      let ext = T.toLower $ T.pack $ takeExtension $ T.unpack src
+      (raw, respMime) <- fetchItem src
+      let raw' = if ext `elem` [".gz", ".svgz"]
+                 then B.concat $ L.toChunks $ Gzip.decompress $ L.fromChunks [raw]
+                 else raw
+      mime <- case (mimetype, respMime) of
+                ("",Nothing) -> throwError $ PandocSomeError
+                  $ "Could not determine mime type for `" <> src <> "'"
+                (x, Nothing) -> return x
+                (_, Just x ) -> return x
+      result <- if "text/css" `T.isPrefixOf` mime
+                then do
+                  oldInputs <- getInputFiles
+                  setInputFiles [T.unpack src]
+                  res <- cssURLs (takeDirectory $ T.unpack src) raw'
+                  setInputFiles oldInputs
+                  return res
                else return raw'
-  return $ Right (mime, result)
+      return $ Right (mime, result)
 
 
 
 -- | Convert HTML into self-contained HTML, incorporating images,
 -- scripts, and CSS using data: URIs.
-makeSelfContained :: PandocMonad m => String -> m String
+makeSelfContained :: PandocMonad m => T.Text -> m T.Text
 makeSelfContained inp = do
   let tags = parseTags inp
   out' <- convertTags tags
