@@ -1,26 +1,7 @@
-{-# LANGUAGE NoImplicitPrelude #-}
 {-# LANGUAGE OverloadedStrings #-}
-{-
-Copyright (C) 2006-2018 John MacFarlane <jgm@berkeley.edu>
-
-This program is free software; you can redistribute it and/or modify
-it under the terms of the GNU General Public License as published by
-the Free Software Foundation; either version 2 of the License, or
-(at your option) any later version.
-
-This program is distributed in the hope that it will be useful,
-but WITHOUT ANY WARRANTY; without even the implied warranty of
-MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-GNU General Public License for more details.
-
-You should have received a copy of the GNU General Public License
-along with this program; if not, write to the Free Software
-Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--}
-
 {- |
    Module      : Text.Pandoc.Writers.AsciiDoc
-   Copyright   : Copyright (C) 2006-2018 John MacFarlane
+   Copyright   : Copyright (C) 2006-2020 John MacFarlane
    License     : GNU GPL, version 2 or above
 
    Maintainer  : John MacFarlane <jgm@berkeley.edu>
@@ -37,42 +18,55 @@ that it has omitted the construct.
 
 AsciiDoc:  <http://www.methods.co.nz/asciidoc/>
 -}
-module Text.Pandoc.Writers.AsciiDoc (writeAsciiDoc) where
-import Prelude
+module Text.Pandoc.Writers.AsciiDoc (writeAsciiDoc, writeAsciiDoctor) where
 import Control.Monad.State.Strict
-import Data.Char (isPunctuation, isSpace, toLower)
-import Data.List (intercalate, intersperse, stripPrefix)
-import Data.Maybe (fromMaybe, isJust, listToMaybe)
+import Data.Char (isPunctuation, isSpace)
+import Data.List (intercalate, intersperse)
+import Data.Maybe (fromMaybe, isJust)
 import qualified Data.Set as Set
+import qualified Data.Text as T
 import Data.Text (Text)
-import Text.Pandoc.Class (PandocMonad, report)
+import Text.Pandoc.Class.PandocMonad (PandocMonad, report)
 import Text.Pandoc.Definition
 import Text.Pandoc.ImageSize
 import Text.Pandoc.Logging
 import Text.Pandoc.Options
 import Text.Pandoc.Parsing hiding (blankline, space)
-import Text.Pandoc.Pretty
+import Text.DocLayout
 import Text.Pandoc.Shared
-import Text.Pandoc.Templates (renderTemplate')
+import Text.Pandoc.Templates (renderTemplate)
 import Text.Pandoc.Writers.Shared
 
-data WriterState = WriterState { defListMarker    :: String
-                               , orderedListLevel :: Int
-                               , bulletListLevel  :: Int
-                               , intraword        :: Bool
-                               , autoIds          :: Set.Set String
+data WriterState = WriterState { defListMarker       :: Text
+                               , orderedListLevel    :: Int
+                               , bulletListLevel     :: Int
+                               , intraword           :: Bool
+                               , autoIds             :: Set.Set Text
+                               , asciidoctorVariant  :: Bool
+                               , inList              :: Bool
+                               , hasMath             :: Bool
                                }
 
+defaultWriterState :: WriterState
+defaultWriterState = WriterState { defListMarker      = "::"
+                                 , orderedListLevel   = 0
+                                 , bulletListLevel    = 0
+                                 , intraword          = False
+                                 , autoIds            = Set.empty
+                                 , asciidoctorVariant = False
+                                 , inList             = False
+                                 , hasMath            = False
+                                 }
+
 -- | Convert Pandoc to AsciiDoc.
 writeAsciiDoc :: PandocMonad m => WriterOptions -> Pandoc -> m Text
 writeAsciiDoc opts document =
-  evalStateT (pandocToAsciiDoc opts document) WriterState{
-      defListMarker = "::"
-    , orderedListLevel = 1
-    , bulletListLevel = 1
-    , intraword = False
-    , autoIds = Set.empty
-    }
+  evalStateT (pandocToAsciiDoc opts document) defaultWriterState
+
+-- | Convert Pandoc to AsciiDoctor compatible AsciiDoc.
+writeAsciiDoctor :: PandocMonad m => WriterOptions -> Pandoc -> m Text
+writeAsciiDoctor opts document =
+  evalStateT (pandocToAsciiDoc opts document) defaultWriterState{ asciidoctorVariant = True }
 
 type ADW = StateT WriterState
 
@@ -84,38 +78,30 @@ pandocToAsciiDoc opts (Pandoc meta blocks) = do
   let colwidth = if writerWrapText opts == WrapAuto
                     then Just $ writerColumns opts
                     else Nothing
-  let render' :: Doc -> Text
-      render' = render colwidth
-  metadata <- metaToJSON opts
-              (fmap render' . blockListToAsciiDoc opts)
-              (fmap render' . inlineListToAsciiDoc opts)
+  metadata <- metaToContext opts
+              (blockListToAsciiDoc opts)
+              (fmap chomp . inlineListToAsciiDoc opts)
               meta
-  body <- vcat <$> mapM (elementToAsciiDoc 1 opts) (hierarchicalize blocks)
-  let main = render colwidth body
+  main <- blockListToAsciiDoc opts $ makeSections False (Just 1) blocks
+  st <- get
   let context  = defField "body" main
                $ defField "toc"
                   (writerTableOfContents opts &&
                    isJust (writerTemplate opts))
-               $defField "titleblock" titleblock metadata
-  case writerTemplate opts of
-       Nothing  -> return main
-       Just tpl -> renderTemplate' tpl context
-
-elementToAsciiDoc :: PandocMonad m
-                  => Int -> WriterOptions -> Element -> ADW m Doc
-elementToAsciiDoc _ opts (Blk b) = blockToAsciiDoc opts b
-elementToAsciiDoc nestlevel opts (Sec _lvl _num attr label children) = do
-  hdr <- blockToAsciiDoc opts (Header nestlevel attr label)
-  rest <- vcat <$> mapM (elementToAsciiDoc (nestlevel + 1) opts) children
-  return $ hdr $$ rest
+               $ defField "math" (hasMath st)
+               $ defField "titleblock" titleblock metadata
+  return $ render colwidth $
+    case writerTemplate opts of
+       Nothing  -> main
+       Just tpl -> renderTemplate tpl context
 
 -- | Escape special characters for AsciiDoc.
-escapeString :: String -> String
+escapeString :: Text -> Text
 escapeString = escapeStringUsing escs
   where escs = backslashEscapes "{"
 
 -- | Ordered list start parser for use in Para below.
-olMarker :: Parser [Char] ParserState Char
+olMarker :: Parser Text ParserState Char
 olMarker = do (start, style', delim) <- anyOrderedListMarker
               if delim == Period &&
                           (style' == UpperAlpha || (style' == UpperRoman &&
@@ -125,27 +111,36 @@ olMarker = do (start, style', delim) <- anyOrderedListMarker
 
 -- | True if string begins with an ordered list marker
 -- or would be interpreted as an AsciiDoc option command
-needsEscaping :: String -> Bool
+needsEscaping :: Text -> Bool
 needsEscaping s = beginsWithOrderedListMarker s || isBracketed s
   where
     beginsWithOrderedListMarker str =
-      case runParser olMarker defaultParserState "para start" (take 10 str) of
+      case runParser olMarker defaultParserState "para start" (T.take 10 str) of
              Left  _ -> False
              Right _ -> True
-    isBracketed ('[':cs) = listToMaybe (reverse cs) == Just ']'
-    isBracketed _ = False
+    isBracketed t
+      | Just ('[', t') <- T.uncons t
+      , Just (_, ']')  <- T.unsnoc t'
+      = True
+      | otherwise = False
 
 -- | Convert Pandoc block element to asciidoc.
 blockToAsciiDoc :: PandocMonad m
                 => WriterOptions -- ^ Options
                 -> Block         -- ^ Block element
-                -> ADW m Doc
+                -> ADW m (Doc Text)
 blockToAsciiDoc _ Null = return empty
+blockToAsciiDoc opts (Div (id',"section":_,_)
+                       (Header level (_,cls,kvs) ils : xs)) = do
+  hdr <- blockToAsciiDoc opts (Header level (id',cls,kvs) ils)
+  rest <- blockListToAsciiDoc opts xs
+  return $ hdr $$ rest
 blockToAsciiDoc opts (Plain inlines) = do
   contents <- inlineListToAsciiDoc opts inlines
   return $ contents <> blankline
-blockToAsciiDoc opts (Para [Image attr alt (src,'f':'i':'g':':':tit)]) =
-  blockToAsciiDoc opts (Para [Image attr alt (src,tit)])
+blockToAsciiDoc opts (Para [Image attr alt (src,tgt)])
+  | Just tit <- T.stripPrefix "fig:" tgt
+  = blockToAsciiDoc opts (Para [Image attr alt (src,tit)])
 blockToAsciiDoc opts (Para inlines) = do
   contents <- inlineListToAsciiDoc opts inlines
   -- escape if para starts with ordered list marker
@@ -161,7 +156,7 @@ blockToAsciiDoc opts (LineBlock lns) = do
   contents <- joinWithLinefeeds <$> mapM docify lns
   return $ "[verse]" $$ text "--" $$ contents $$ text "--" $$ blankline
 blockToAsciiDoc _ b@(RawBlock f s)
-  | f == "asciidoc" = return $ text s
+  | f == "asciidoc" = return $ literal s
   | otherwise         = do
       report $ BlockNotRendered b
       return empty
@@ -172,20 +167,20 @@ blockToAsciiDoc opts (Header level (ident,_,_) inlines) = do
   ids <- gets autoIds
   let autoId = uniqueIdent (writerExtensions opts) inlines ids
   modify $ \st -> st{ autoIds = Set.insert autoId ids }
-  let identifier = if null ident ||
+  let identifier = if T.null ident ||
                       (isEnabled Ext_auto_identifiers opts && ident == autoId)
                       then empty
-                      else "[[" <> text ident <> "]]"
+                      else "[[" <> literal ident <> "]]"
   return $ identifier $$
            nowrap (text (replicate (level + 1) '=') <> space <> contents) <>
            blankline
 
 blockToAsciiDoc _ (CodeBlock (_,classes,_) str) = return $ flush (
   if null classes
-     then "...." $$ text str $$ "...."
-     else attrs $$ "----" $$ text str $$ "----")
+     then "...." $$ literal str $$ "...."
+     else attrs $$ "----" $$ literal str $$ "----")
   <> blankline
-    where attrs = "[" <> text (intercalate "," ("source" : classes)) <> "]"
+    where attrs = "[" <> literal (T.intercalate "," ("source" : classes)) <> "]"
 blockToAsciiDoc opts (BlockQuote blocks) = do
   contents <- blockListToAsciiDoc opts blocks
   let isBlock (BlockQuote _) = True
@@ -194,8 +189,7 @@ blockToAsciiDoc opts (BlockQuote blocks) = do
   let contents' = if any isBlock blocks
                      then "--" $$ contents $$ "--"
                      else contents
-  let cols = offset contents'
-  let bar = text $ replicate cols '_'
+  let bar = text "____"
   return $ bar $$ chomp contents' $$ bar <> blankline
 blockToAsciiDoc opts (Table caption aligns widths headers rows) =  do
   caption' <- inlineListToAsciiDoc opts caption
@@ -252,70 +246,114 @@ blockToAsciiDoc opts (Table caption aligns widths headers rows) =  do
                     else 100000
   let maxwidth = maximum $ map offset (head':rows')
   let body = if maxwidth > colwidth then vsep rows' else vcat rows'
-  let border = text $ "|" ++ replicate (max 5 (min maxwidth colwidth) - 1) '='
+  let border = text "|==="
   return $
     caption'' $$ tablespec $$ border $$ head'' $$ body $$ border $$ blankline
 blockToAsciiDoc opts (BulletList items) = do
+  inlist <- gets inList
+  modify $ \st -> st{ inList = True }
   contents <- mapM (bulletListItemToAsciiDoc opts) items
-  return $ cat contents <> blankline
+  modify $ \st -> st{ inList = inlist }
+  return $ mconcat contents <> blankline
 blockToAsciiDoc opts (OrderedList (start, sty, _delim) items) = do
   let listStyle = case sty of
                        DefaultStyle -> []
                        Decimal      -> ["arabic"]
                        Example      -> []
-                       _            -> [map toLower (show sty)]
-  let listStart = if start == 1 then [] else ["start=" ++ show start]
-  let listoptions = case intercalate ", " (listStyle ++ listStart) of
-                          [] -> empty
-                          x  -> brackets (text x)
+                       _            -> [T.toLower (tshow sty)]
+  let listStart = ["start=" <> tshow start | start /= 1]
+  let listoptions = case T.intercalate ", " (listStyle ++ listStart) of
+                          "" -> empty
+                          x  -> brackets (literal x)
+  inlist <- gets inList
+  modify $ \st -> st{ inList = True }
   contents <- mapM (orderedListItemToAsciiDoc opts) items
-  return $ listoptions $$ cat contents <> blankline
+  modify $ \st -> st{ inList = inlist }
+  return $ listoptions $$ mconcat contents <> blankline
 blockToAsciiDoc opts (DefinitionList items) = do
+  inlist <- gets inList
+  modify $ \st -> st{ inList = True }
   contents <- mapM (definitionListItemToAsciiDoc opts) items
-  return $ cat contents <> blankline
-blockToAsciiDoc opts (Div (ident,_,_) bs) = do
-  let identifier = if null ident then empty else "[[" <> text ident <> "]]"
-  contents <- blockListToAsciiDoc opts bs
+  modify $ \st -> st{ inList = inlist }
+  return $ mconcat contents <> blankline
+blockToAsciiDoc opts (Div (ident,classes,_) bs) = do
+  let identifier = if T.null ident then empty else "[[" <> literal ident <> "]]"
+  let admonitions = ["attention","caution","danger","error","hint",
+                     "important","note","tip","warning"]
+  contents <-
+       case classes of
+         (l:_) | l `elem` admonitions -> do
+             let (titleBs, bodyBs) =
+                     case bs of
+                       (Div (_,["title"],_) ts : rest) -> (ts, rest)
+                       _ -> ([], bs)
+             admonitionTitle <- if null titleBs
+                                   then return mempty
+                                   else ("." <>) <$>
+                                         blockListToAsciiDoc opts titleBs
+             admonitionBody <- blockListToAsciiDoc opts bodyBs
+             return $ "[" <> literal (T.toUpper l) <> "]" $$
+                      chomp admonitionTitle $$
+                      "====" $$
+                      chomp admonitionBody $$
+                      "===="
+         _ -> blockListToAsciiDoc opts bs
   return $ identifier $$ contents
 
 -- | Convert bullet list item (list of blocks) to asciidoc.
 bulletListItemToAsciiDoc :: PandocMonad m
-                         => WriterOptions -> [Block] -> ADW m Doc
+                         => WriterOptions -> [Block] -> ADW m (Doc Text)
 bulletListItemToAsciiDoc opts blocks = do
   lev <- gets bulletListLevel
   modify $ \s -> s{ bulletListLevel = lev + 1 }
   contents <- foldM (addBlock opts) empty blocks
   modify $ \s -> s{ bulletListLevel = lev }
-  let marker = text (replicate lev '*')
-  return $ marker <> text " " <> contents <> cr
+  let marker = text (replicate (lev + 1) '*')
+  return $ marker <> text " " <> listBegin blocks <>
+    contents <> cr
 
-addBlock :: PandocMonad m => WriterOptions -> Doc -> Block -> ADW m Doc
-addBlock opts d b | isEmpty d    = chomp `fmap` blockToAsciiDoc opts b
-addBlock opts d b@(BulletList _) = do x <- blockToAsciiDoc opts b
-                                      return $ d <> cr <> chomp x
-addBlock opts d b@(OrderedList _ _) = do x <- blockToAsciiDoc opts b
-                                         return $ d <> cr <> chomp x
-addBlock opts d b = do x <- blockToAsciiDoc opts b
-                       return $ d <> cr <> text "+" <> cr <> chomp x
+addBlock :: PandocMonad m
+         => WriterOptions -> Doc Text -> Block -> ADW m (Doc Text)
+addBlock opts d b = do
+  x <- chomp <$> blockToAsciiDoc opts b
+  return $
+    case b of
+        BulletList{} -> d <> cr <> x
+        OrderedList{} -> d <> cr <> x
+        Para (Math DisplayMath _:_) -> d <> cr <> x
+        Plain (Math DisplayMath _:_) -> d <> cr <> x
+        Para{} | isEmpty d -> x
+        Plain{} | isEmpty d -> x
+        _ -> d <> cr <> text "+" <> cr <> x
+
+listBegin :: [Block] -> Doc Text
+listBegin blocks =
+        case blocks of
+          Para (Math DisplayMath _:_) : _  -> "{blank}"
+          Plain (Math DisplayMath _:_) : _ -> "{blank}"
+          Para _ : _                       -> empty
+          Plain _ : _                      -> empty
+          _ : _                            -> "{blank}"
+          []                               -> "{blank}"
 
 -- | Convert ordered list item (a list of blocks) to asciidoc.
 orderedListItemToAsciiDoc :: PandocMonad m
                           => WriterOptions -- ^ options
                           -> [Block]       -- ^ list item (list of blocks)
-                          -> ADW m Doc
+                          -> ADW m (Doc Text)
 orderedListItemToAsciiDoc opts blocks = do
   lev <- gets orderedListLevel
   modify $ \s -> s{ orderedListLevel = lev + 1 }
   contents <- foldM (addBlock opts) empty blocks
   modify $ \s -> s{ orderedListLevel = lev }
-  let marker = text (replicate lev '.')
-  return $ marker <> text " " <> contents <> cr
+  let marker = text (replicate (lev + 1) '.')
+  return $ marker <> text " " <> listBegin blocks <> contents <> cr
 
 -- | Convert definition list item (label, list of blocks) to asciidoc.
 definitionListItemToAsciiDoc :: PandocMonad m
                              => WriterOptions
                              -> ([Inline],[[Block]])
-                             -> ADW m Doc
+                             -> ADW m (Doc Text)
 definitionListItemToAsciiDoc opts (label, defs) = do
   labelText <- inlineListToAsciiDoc opts label
   marker <- gets defListMarker
@@ -323,25 +361,26 @@ definitionListItemToAsciiDoc opts (label, defs) = do
      then modify (\st -> st{ defListMarker = ";;"})
      else modify (\st -> st{ defListMarker = "::"})
   let divider = cr <> text "+" <> cr
-  let defsToAsciiDoc :: PandocMonad m => [Block] -> ADW m Doc
+  let defsToAsciiDoc :: PandocMonad m => [Block] -> ADW m (Doc Text)
       defsToAsciiDoc ds = (vcat . intersperse divider . map chomp)
            `fmap` mapM (blockToAsciiDoc opts) ds
   defs' <- mapM defsToAsciiDoc defs
   modify (\st -> st{ defListMarker = marker })
   let contents = nest 2 $ vcat $ intersperse divider $ map chomp defs'
-  return $ labelText <> text marker <> cr <> contents <> cr
+  return $ labelText <> literal marker <> cr <> contents <> cr
 
 -- | Convert list of Pandoc block elements to asciidoc.
 blockListToAsciiDoc :: PandocMonad m
                     => WriterOptions -- ^ Options
                     -> [Block]       -- ^ List of block elements
-                    -> ADW m Doc
-blockListToAsciiDoc opts blocks = cat `fmap` mapM (blockToAsciiDoc opts) blocks
+                    -> ADW m (Doc Text)
+blockListToAsciiDoc opts blocks =
+  mconcat `fmap` mapM (blockToAsciiDoc opts) blocks
 
 data SpacyLocation = End | Start
 
 -- | Convert list of Pandoc inline elements to asciidoc.
-inlineListToAsciiDoc :: PandocMonad m => WriterOptions -> [Inline] -> ADW m Doc
+inlineListToAsciiDoc :: PandocMonad m => WriterOptions -> [Inline] -> ADW m (Doc Text)
 inlineListToAsciiDoc opts lst = do
   oldIntraword <- gets intraword
   setIntraword False
@@ -371,10 +410,11 @@ inlineListToAsciiDoc opts lst = do
        isSpacy _ SoftBreak = True
        -- Note that \W characters count as spacy in AsciiDoc
        -- for purposes of determining interword:
-       isSpacy End (Str xs) = case reverse xs of
-                                   c:_ -> isPunctuation c || isSpace c
-                                   _   -> False
-       isSpacy Start (Str (c:_)) = isPunctuation c || isSpace c
+       isSpacy End (Str xs) = case T.unsnoc xs of
+                                   Just (_, c) -> isPunctuation c || isSpace c
+                                   _           -> False
+       isSpacy Start (Str xs)
+         | Just (c, _) <- T.uncons xs = isPunctuation c || isSpace c
        isSpacy _ _ = False
 
 setIntraword :: PandocMonad m => Bool -> ADW m ()
@@ -384,7 +424,9 @@ withIntraword :: PandocMonad m => ADW m a -> ADW m a
 withIntraword p = setIntraword True *> p <* setIntraword False
 
 -- | Convert Pandoc inline element to asciidoc.
-inlineToAsciiDoc :: PandocMonad m => WriterOptions -> Inline -> ADW m Doc
+inlineToAsciiDoc :: PandocMonad m => WriterOptions -> Inline -> ADW m (Doc Text)
+inlineToAsciiDoc opts (Emph [Strong xs]) =
+  inlineToAsciiDoc opts (Strong [Emph xs])  -- see #5565
 inlineToAsciiDoc opts (Emph lst) = do
   contents <- inlineListToAsciiDoc opts lst
   isIntraword <- gets intraword
@@ -405,23 +447,49 @@ inlineToAsciiDoc opts (Subscript lst) = do
   contents <- inlineListToAsciiDoc opts lst
   return $ "~" <> contents <> "~"
 inlineToAsciiDoc opts (SmallCaps lst) = inlineListToAsciiDoc opts lst
-inlineToAsciiDoc opts (Quoted SingleQuote lst) =
-  inlineListToAsciiDoc opts (Str "`" : lst ++ [Str "'"])
-inlineToAsciiDoc opts (Quoted DoubleQuote lst) =
-  inlineListToAsciiDoc opts (Str "``" : lst ++ [Str "''"])
-inlineToAsciiDoc _ (Code _ str) = return $
-  text "`" <> text (escapeStringUsing (backslashEscapes "`") str) <> "`"
-inlineToAsciiDoc _ (Str str) = return $ text $ escapeString str
-inlineToAsciiDoc _ (Math InlineMath str) =
-  return $ "latexmath:[$" <> text str <> "$]"
-inlineToAsciiDoc _ (Math DisplayMath str) =
-  return $ "latexmath:[\\[" <> text str <> "\\]]"
+inlineToAsciiDoc opts (Quoted qt lst) = do
+  isAsciidoctor <- gets asciidoctorVariant
+  inlineListToAsciiDoc opts $
+    case qt of
+      SingleQuote
+        | isAsciidoctor -> [Str "'`"] ++ lst ++ [Str "`'"]
+        | otherwise     -> [Str "`"] ++ lst ++ [Str "'"]
+      DoubleQuote
+        | isAsciidoctor -> [Str "\"`"] ++ lst ++ [Str "`\""]
+        | otherwise     -> [Str "``"] ++ lst ++ [Str "''"]
+inlineToAsciiDoc _ (Code _ str) = do
+  isAsciidoctor <- gets asciidoctorVariant
+  let contents = literal (escapeStringUsing (backslashEscapes "`") str)
+  return $
+    if isAsciidoctor
+       then text "`+" <> contents <> "+`"
+       else text "`"  <> contents <> "`"
+inlineToAsciiDoc _ (Str str) = return $ literal $ escapeString str
+inlineToAsciiDoc _ (Math InlineMath str) = do
+  isAsciidoctor <- gets asciidoctorVariant
+  modify $ \st -> st{ hasMath = True }
+  let content = if isAsciidoctor
+                then literal str
+                else "$" <> literal str <> "$"
+  return $ "latexmath:[" <> content <> "]"
+inlineToAsciiDoc _ (Math DisplayMath str) = do
+  isAsciidoctor <- gets asciidoctorVariant
+  modify $ \st -> st{ hasMath = True }
+  let content = if isAsciidoctor
+                then literal str
+                else "\\[" <> literal str <> "\\]"
+  inlist <- gets inList
+  let sepline = if inlist
+                   then text "+"
+                   else blankline
+  return $
+      (cr <> sepline) $$ "[latexmath]" $$ "++++" $$
+      content $$ "++++" <> cr
 inlineToAsciiDoc _ il@(RawInline f s)
-  | f == "asciidoc" = return $ text s
+  | f == "asciidoc" = return $ literal s
   | otherwise         = do
       report $ InlineNotRendered il
       return empty
-  | otherwise       = return empty
 inlineToAsciiDoc _ LineBreak = return $ " +" <> cr
 inlineToAsciiDoc _ Space = return space
 inlineToAsciiDoc opts SoftBreak =
@@ -435,38 +503,38 @@ inlineToAsciiDoc opts (Link _ txt (src, _tit)) = do
 -- abs:  http://google.cod[Google]
 -- or my@email.com[email john]
   linktext <- inlineListToAsciiDoc opts txt
-  let isRelative = ':' `notElem` src
+  let isRelative = T.all (/= ':') src
   let prefix = if isRelative
                   then text "link:"
                   else empty
-  let srcSuffix = fromMaybe src (stripPrefix "mailto:" src)
+  let srcSuffix = fromMaybe src (T.stripPrefix "mailto:" src)
   let useAuto = case txt of
                       [Str s] | escapeURI s == srcSuffix -> True
                       _       -> False
   return $ if useAuto
-              then text srcSuffix
-              else prefix <> text src <> "[" <> linktext <> "]"
+              then literal srcSuffix
+              else prefix <> literal src <> "[" <> linktext <> "]"
 inlineToAsciiDoc opts (Image attr alternate (src, tit)) = do
 -- image:images/logo.png[Company logo, title="blah"]
   let txt = if null alternate || (alternate == [Str ""])
                then [Str "image"]
                else alternate
   linktext <- inlineListToAsciiDoc opts txt
-  let linktitle = if null tit
+  let linktitle = if T.null tit
                      then empty
-                     else ",title=\"" <> text tit <> "\""
+                     else ",title=\"" <> literal tit <> "\""
       showDim dir = case dimension dir attr of
                       Just (Percent a) ->
                         ["scaledwidth=" <> text (show (Percent a))]
                       Just dim         ->
-                        [text (show dir) <> "=" <> text (showInPixel opts dim)]
+                        [text (show dir) <> "=" <> literal (showInPixel opts dim)]
                       Nothing          ->
                         []
       dimList = showDim Width ++ showDim Height
       dims = if null dimList
                 then empty
-                else "," <> cat (intersperse "," dimList)
-  return $ "image:" <> text src <> "[" <> linktext <> linktitle <> dims <> "]"
+                else "," <> mconcat (intersperse "," dimList)
+  return $ "image:" <> literal src <> "[" <> linktext <> linktitle <> dims <> "]"
 inlineToAsciiDoc opts (Note [Para inlines]) =
   inlineToAsciiDoc opts (Note [Plain inlines])
 inlineToAsciiDoc opts (Note [Plain inlines]) = do
@@ -476,9 +544,11 @@ inlineToAsciiDoc opts (Note [Plain inlines]) = do
 inlineToAsciiDoc _ (Note _) = return "[multiblock footnote omitted]"
 inlineToAsciiDoc opts (Span (ident,classes,_) ils) = do
   contents <- inlineListToAsciiDoc opts ils
-  if null ident && null classes
+  isIntraword <- gets intraword
+  let marker = if isIntraword then "##" else "#"
+  if T.null ident && null classes
      then return contents
      else do
-       let modifier = brackets $ text $ unwords $
-            [ '#':ident | not (null ident)] ++ map ('.':) classes
-       return $ modifier <> "#" <> contents <> "#"
+       let modifier = brackets $ literal $ T.unwords $
+            [ "#" <> ident | not (T.null ident)] ++ map ("." <>) classes
+       return $ modifier <> marker <> contents <> marker
