@@ -1,26 +1,8 @@
-{-# LANGUAGE NoImplicitPrelude #-}
 {-# LANGUAGE OverloadedStrings #-}
-{-
-Copyright (C) 2015-2018 John MacFarlane <jgm@berkeley.edu>
-
-This program is free software; you can redistribute it and/or modify
-it under the terms of the GNU General Public License as published by
-the Free Software Foundation; either version 2 of the License, or
-(at your option) any later version.
-
-This program is distributed in the hope that it will be useful,
-but WITHOUT ANY WARRANTY; without even the implied warranty of
-MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-GNU General Public License for more details.
-
-You should have received a copy of the GNU General Public License
-along with this program; if not, write to the Free Software
-Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--}
-
+{-# LANGUAGE ViewPatterns      #-}
 {- |
    Module      : Text.Pandoc.Writers.CommonMark
-   Copyright   : Copyright (C) 2015-2018 John MacFarlane
+   Copyright   : Copyright (C) 2015-2020 John MacFarlane
    License     : GNU GPL, version 2 or above
 
    Maintainer  : John MacFarlane <jgm@berkeley.edu>
@@ -33,42 +15,50 @@ CommonMark:  <http://commonmark.org>
 -}
 module Text.Pandoc.Writers.CommonMark (writeCommonMark) where
 
-import Prelude
 import CMarkGFM
 import Control.Monad.State.Strict (State, get, modify, runState)
 import Data.Char (isAscii)
 import Data.Foldable (foldrM)
 import Data.List (transpose)
-import Data.Monoid (Any (..))
 import Data.Text (Text)
 import qualified Data.Text as T
 import Network.HTTP (urlEncode)
-import Text.Pandoc.Class (PandocMonad)
+import Text.Pandoc.Class.PandocMonad (PandocMonad)
 import Text.Pandoc.Definition
 import Text.Pandoc.Options
-import Text.Pandoc.Shared (isTightList, linesToPara, substitute, capitalize)
-import Text.Pandoc.Templates (renderTemplate')
-import Text.Pandoc.Walk (query, walk, walkM)
+import Text.Pandoc.Shared (capitalize, isTightList,
+    linesToPara, onlySimpleTableCells, taskListItemToAscii, tshow)
+import Text.Pandoc.Templates (renderTemplate)
+import Text.Pandoc.Walk (walk, walkM)
 import Text.Pandoc.Writers.HTML (writeHtml5String, tagWithAttributes)
 import Text.Pandoc.Writers.Shared
 import Text.Pandoc.XML (toHtml5Entities)
+import Text.DocLayout (literal, render)
 
 -- | Convert Pandoc to CommonMark.
 writeCommonMark :: PandocMonad m => WriterOptions -> Pandoc -> m Text
 writeCommonMark opts (Pandoc meta blocks) = do
+  toc <- if writerTableOfContents opts
+            then blocksToCommonMark opts [ toTableOfContents opts blocks ]
+            else return mempty
+
   let (blocks', notes) = runState (walkM processNotes blocks) []
-      notes' = if null notes
-               then []
-               else [OrderedList (1, Decimal, Period) $ reverse notes]
+      notes' = [OrderedList (1, Decimal, Period) $ reverse notes | not (null notes)]
   main <-  blocksToCommonMark opts (blocks' ++ notes')
-  metadata <- metaToJSON opts
-              (blocksToCommonMark opts)
-              (inlinesToCommonMark opts)
+  metadata <- metaToContext opts
+              (fmap (literal . T.stripEnd) . blocksToCommonMark opts)
+              (fmap (literal . T.stripEnd) . inlinesToCommonMark opts)
               meta
-  let context = defField "body" main metadata
-  case writerTemplate opts of
-       Nothing  -> return main
-       Just tpl -> renderTemplate' tpl context
+  let context =
+          -- for backwards compatibility we populate toc
+          -- with the contents of the toc, rather than a boolean:
+          defField "toc" toc
+        $ defField "table-of-contents" toc
+        $ defField "body" main metadata
+  return $
+    case writerTemplate opts of
+       Nothing  -> main
+       Just tpl -> render Nothing $ renderTemplate tpl context
 
 softBreakToSpace :: Inline -> Inline
 softBreakToSpace SoftBreak = Space
@@ -78,7 +68,7 @@ processNotes :: Inline -> State [[Block]] Inline
 processNotes (Note bs) = do
   modify (bs :)
   notes <- get
-  return $ Str $ "[" ++ show (length notes) ++ "]"
+  return $ Str $ "[" <> tshow (length notes) <> "]"
 processNotes x = return x
 
 node :: NodeType -> [Node] -> Node
@@ -114,25 +104,29 @@ blockToNodes opts (Para xs) ns =
   return (node PARAGRAPH (inlinesToNodes opts xs) : ns)
 blockToNodes opts (LineBlock lns) ns = blockToNodes opts (linesToPara lns) ns
 blockToNodes _ (CodeBlock (_,classes,_) xs) ns = return
-  (node (CODE_BLOCK (T.pack (unwords classes)) (T.pack xs)) [] : ns)
-blockToNodes opts (RawBlock fmt xs) ns
-  | fmt == Format "html" && isEnabled Ext_raw_html opts
-              = return (node (HTML_BLOCK (T.pack xs)) [] : ns)
-  | (fmt == Format "latex" || fmt == Format "tex") && isEnabled Ext_raw_tex opts
-              = return (node (CUSTOM_BLOCK (T.pack xs) T.empty) [] : ns)
+  (node (CODE_BLOCK (T.unwords classes) xs) [] : ns)
+blockToNodes opts (RawBlock (Format f) xs) ns
+  | f == "html" && isEnabled Ext_raw_html opts
+              = return (node (HTML_BLOCK xs) [] : ns)
+  | (f == "latex" || f == "tex") && isEnabled Ext_raw_tex opts
+              = return (node (CUSTOM_BLOCK xs T.empty) [] : ns)
+  | f == "markdown"
+              = return (node (CUSTOM_BLOCK xs T.empty) [] : ns)
   | otherwise = return ns
 blockToNodes opts (BlockQuote bs) ns = do
   nodes <- blocksToNodes opts bs
   return (node BLOCK_QUOTE nodes : ns)
 blockToNodes opts (BulletList items) ns = do
-  nodes <- mapM (blocksToNodes opts) items
+  let exts = writerExtensions opts
+  nodes <- mapM (blocksToNodes opts . taskListItemToAscii exts) items
   return (node (LIST ListAttributes{
                    listType = BULLET_LIST,
                    listDelim = PERIOD_DELIM,
                    listTight = isTightList items,
                    listStart = 1 }) (map (node ITEM) nodes) : ns)
 blockToNodes opts (OrderedList (start, _sty, delim) items) ns = do
-  nodes <- mapM (blocksToNodes opts) items
+  let exts = writerExtensions opts
+  nodes <- mapM (blocksToNodes opts . taskListItemToAscii exts) items
   return (node (LIST ListAttributes{
                    listType = ORDERED_LIST,
                    listDelim = case delim of
@@ -160,17 +154,8 @@ blockToNodes opts (DefinitionList items) ns =
           Plain (term ++ [LineBreak] ++ xs) : ys ++ concat zs
         dlToBullet (term, xs) =
           Para term : concat xs
-blockToNodes opts t@(Table capt aligns _widths headers rows) ns = do
-  let allcells = concat (headers:rows)
-  let isLineBreak LineBreak = Any True
-      isLineBreak _         = Any False
-  let isPlainOrPara [Para _]  = True
-      isPlainOrPara [Plain _] = True
-      isPlainOrPara []        = True
-      isPlainOrPara _         = False
-  let isSimple = all isPlainOrPara allcells &&
-                 not ( getAny (query isLineBreak allcells) )
-  if isEnabled Ext_pipe_tables opts && isSimple
+blockToNodes opts t@(Table capt aligns _widths headers rows) ns =
+  if isEnabled Ext_pipe_tables opts && onlySimpleTableCells (headers:rows)
      then do
        -- We construct a table manually as a CUSTOM_BLOCK, for
        -- two reasons:  (1) cmark-gfm currently doesn't support
@@ -179,9 +164,9 @@ blockToNodes opts t@(Table capt aligns _widths headers rows) ns = do
        let capt' = node PARAGRAPH (inlinesToNodes opts capt)
        -- backslash | in code and raw:
        let fixPipe (Code attr xs) =
-             Code attr (substitute "|" "\\|" xs)
+             Code attr (T.replace "|" "\\|" xs)
            fixPipe (RawInline format xs) =
-             RawInline format (substitute "|" "\\|" xs)
+             RawInline format (T.replace "|" "\\|" xs)
            fixPipe x = x
        let toCell [Plain ils] = T.strip
                                 $ nodeToCommonmark [] Nothing
@@ -247,18 +232,16 @@ inlineToNodes opts (Str s) = stringToNodes opts s'
 inlineToNodes _ Space   = (node (TEXT (T.pack " ")) [] :)
 inlineToNodes _ LineBreak = (node LINEBREAK [] :)
 inlineToNodes opts SoftBreak
-  | isEnabled Ext_hard_line_breaks opts = (node LINEBREAK [] :)
+  | isEnabled Ext_hard_line_breaks opts = (node (TEXT " ") [] :)
   | writerWrapText opts == WrapNone     = (node (TEXT " ") [] :)
   | otherwise                           = (node SOFTBREAK [] :)
 inlineToNodes opts (Emph xs) = (node EMPH (inlinesToNodes opts xs) :)
 inlineToNodes opts (Strong xs) = (node STRONG (inlinesToNodes opts xs) :)
-inlineToNodes opts (Strikeout xs) =
-  if isEnabled Ext_strikeout opts
-     then (node (CUSTOM_INLINE "~~" "~~") (inlinesToNodes opts xs) :)
-     else if isEnabled Ext_raw_html opts
-            then ((node (HTML_INLINE (T.pack "<s>")) [] : inlinesToNodes opts xs ++
-                  [node (HTML_INLINE (T.pack "</s>")) []]) ++ )
-            else (inlinesToNodes opts xs ++)
+inlineToNodes opts (Strikeout xs)
+  | isEnabled Ext_strikeout opts = (node (CUSTOM_INLINE "~~" "~~") (inlinesToNodes opts xs) :)
+  | isEnabled Ext_raw_html opts = ((node (HTML_INLINE (T.pack "<s>")) [] : inlinesToNodes opts xs ++
+        [node (HTML_INLINE (T.pack "</s>")) []]) ++ )
+  | otherwise = (inlinesToNodes opts xs ++)
 inlineToNodes opts (Superscript xs) =
   if isEnabled Ext_raw_html opts
     then ((node (HTML_INLINE (T.pack "<sup>")) [] : inlinesToNodes opts xs ++
@@ -286,17 +269,19 @@ inlineToNodes opts (SmallCaps xs) =
            [node (HTML_INLINE (T.pack "</span>")) []]) ++ )
     else (inlinesToNodes opts (capitalize xs) ++)
 inlineToNodes opts (Link _ ils (url,tit)) =
-  (node (LINK (T.pack url) (T.pack tit)) (inlinesToNodes opts ils) :)
+  (node (LINK url tit) (inlinesToNodes opts ils) :)
 -- title beginning with fig: indicates implicit figure
-inlineToNodes opts (Image alt ils (url,'f':'i':'g':':':tit)) =
+inlineToNodes opts (Image alt ils (url,T.stripPrefix "fig:" -> Just tit)) =
   inlineToNodes opts (Image alt ils (url,tit))
 inlineToNodes opts (Image _ ils (url,tit)) =
-  (node (IMAGE (T.pack url) (T.pack tit)) (inlinesToNodes opts ils) :)
-inlineToNodes opts (RawInline fmt xs)
-  | fmt == Format "html" && isEnabled Ext_raw_html opts
-              = (node (HTML_INLINE (T.pack xs)) [] :)
-  | (fmt == Format "latex" || fmt == Format "tex") && isEnabled Ext_raw_tex opts
-              = (node (CUSTOM_INLINE (T.pack xs) T.empty) [] :)
+  (node (IMAGE url tit) (inlinesToNodes opts ils) :)
+inlineToNodes opts (RawInline (Format f) xs)
+  | f == "html" && isEnabled Ext_raw_html opts
+              = (node (HTML_INLINE xs) [] :)
+  | (f == "latex" || f == "tex") && isEnabled Ext_raw_tex opts
+              = (node (CUSTOM_INLINE xs T.empty) [] :)
+  | f == "markdown"
+              = (node (CUSTOM_INLINE xs T.empty) [] :)
   | otherwise = id
 inlineToNodes opts (Quoted qt ils) =
   ((node (HTML_INLINE start) [] :
@@ -312,12 +297,12 @@ inlineToNodes opts (Quoted qt ils) =
                             | writerPreferAscii opts ->
                                      ("&ldquo;", "&rdquo;")
                             | otherwise -> ("“", "”")
-inlineToNodes _ (Code _ str) = (node (CODE (T.pack str)) [] :)
+inlineToNodes _ (Code _ str) = (node (CODE str) [] :)
 inlineToNodes opts (Math mt str) =
   case writerHTMLMathMethod opts of
        WebTeX url ->
            let core = inlineToNodes opts
-                        (Image nullAttr [Str str] (url ++ urlEncode str, str))
+                        (Image nullAttr [Str str] (url <> T.pack (urlEncode $ T.unpack str), str))
                sep = if mt == DisplayMath
                         then (node LINEBREAK [] :)
                         else id
@@ -325,14 +310,14 @@ inlineToNodes opts (Math mt str) =
        _  ->
            case mt of
             InlineMath  ->
-              (node (HTML_INLINE (T.pack ("\\(" ++ str ++ "\\)"))) [] :)
+              (node (HTML_INLINE ("\\(" <> str <> "\\)")) [] :)
             DisplayMath ->
-              (node (HTML_INLINE (T.pack ("\\[" ++ str ++ "\\]"))) [] :)
-inlineToNodes opts (Span ("",["emoji"],kvs) [Str s]) = do
+              (node (HTML_INLINE ("\\[" <> str <> "\\]")) [] :)
+inlineToNodes opts (Span ("",["emoji"],kvs) [Str s]) =
   case lookup "data-emoji" kvs of
        Just emojiname | isEnabled Ext_emoji opts ->
-            (node (TEXT (":" <> T.pack emojiname <> ":")) [] :)
-       _ -> (node (TEXT (T.pack s)) [] :)
+            (node (TEXT (":" <> emojiname <> ":")) [] :)
+       _ -> (node (TEXT s) [] :)
 inlineToNodes opts (Span attr ils) =
   let nodes = inlinesToNodes opts ils
       op = tagWithAttributes opts True False "span" attr
@@ -344,17 +329,17 @@ inlineToNodes opts (Cite _ ils) = (inlinesToNodes opts ils ++)
 inlineToNodes _ (Note _) = id -- should not occur
 -- we remove Note elements in preprocessing
 
-stringToNodes :: WriterOptions -> String -> [Node] -> [Node]
+stringToNodes :: WriterOptions -> Text -> [Node] -> [Node]
 stringToNodes opts s
-  | not (writerPreferAscii opts) = (node (TEXT (T.pack s)) [] :)
+  | not (writerPreferAscii opts) = (node (TEXT s) [] :)
   | otherwise = step s
   where
     step input =
-      let (ascii, rest) = span isAscii input
-          this = node (TEXT (T.pack ascii)) []
-          nodes = case rest of
-            [] -> id
-            (nonAscii : rest') ->
+      let (ascii, rest) = T.span isAscii input
+          this = node (TEXT ascii) []
+          nodes = case T.uncons rest of
+            Nothing -> id
+            Just (nonAscii, rest') ->
               let escaped = toHtml5Entities (T.singleton nonAscii)
               in (node (HTML_INLINE escaped) [] :) . step rest'
       in (this :) . nodes
@@ -362,7 +347,7 @@ stringToNodes opts s
 toSubscriptInline :: Inline -> Maybe Inline
 toSubscriptInline Space = Just Space
 toSubscriptInline (Span attr ils) = Span attr <$> traverse toSubscriptInline ils
-toSubscriptInline (Str s) = Str <$> traverse toSubscript s
+toSubscriptInline (Str s) = Str . T.pack <$> traverse toSubscript (T.unpack s)
 toSubscriptInline LineBreak = Just LineBreak
 toSubscriptInline SoftBreak = Just SoftBreak
 toSubscriptInline _ = Nothing
@@ -370,7 +355,7 @@ toSubscriptInline _ = Nothing
 toSuperscriptInline :: Inline -> Maybe Inline
 toSuperscriptInline Space = Just Space
 toSuperscriptInline (Span attr ils) = Span attr <$> traverse toSuperscriptInline ils
-toSuperscriptInline (Str s) = Str <$> traverse toSuperscript s
+toSuperscriptInline (Str s) = Str . T.pack <$> traverse toSuperscript (T.unpack s)
 toSuperscriptInline LineBreak = Just LineBreak
 toSuperscriptInline SoftBreak = Just SoftBreak
 toSuperscriptInline _ = Nothing
