@@ -1,27 +1,9 @@
-{-# LANGUAGE NoImplicitPrelude #-}
 {-# LANGUAGE OverloadedStrings   #-}
 {-# LANGUAGE ScopedTypeVariables #-}
-{-
-Copyright (C) 2007-2018 John MacFarlane <jgm@berkeley.edu>
-
-This program is free software; you can redistribute it and/or modify
-it under the terms of the GNU General Public License as published by
-the Free Software Foundation; either version 2 of the License, or
-(at your option) any later version.
-
-This program is distributed in the hope that it will be useful,
-but WITHOUT ANY WARRANTY; without even the implied warranty of
-MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-GNU General Public License for more details.
-
-You should have received a copy of the GNU General Public License
-along with this program; if not, write to the Free Software
-Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--}
-
+{-# LANGUAGE ViewPatterns        #-}
 {- |
    Module      : Text.Pandoc.Writers.ConTeXt
-   Copyright   : Copyright (C) 2007-2018 John MacFarlane
+   Copyright   : Copyright (C) 2007-2020 John MacFarlane
    License     : GNU GPL, version 2 or above
 
    Maintainer  : John MacFarlane <jgm@berkeley.edu>
@@ -31,22 +13,22 @@ Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 Conversion of 'Pandoc' format into ConTeXt.
 -}
 module Text.Pandoc.Writers.ConTeXt ( writeConTeXt ) where
-import Prelude
 import Control.Monad.State.Strict
 import Data.Char (ord, isDigit)
-import Data.List (intercalate, intersperse)
+import Data.List (intersperse)
 import Data.Maybe (mapMaybe)
 import Data.Text (Text)
+import qualified Data.Text as T
 import Network.URI (unEscapeString)
 import Text.Pandoc.BCP47
-import Text.Pandoc.Class (PandocMonad, report, toLang)
+import Text.Pandoc.Class.PandocMonad (PandocMonad, report, toLang)
 import Text.Pandoc.Definition
 import Text.Pandoc.ImageSize
 import Text.Pandoc.Logging
 import Text.Pandoc.Options
-import Text.Pandoc.Pretty
+import Text.DocLayout
 import Text.Pandoc.Shared
-import Text.Pandoc.Templates (renderTemplate')
+import Text.Pandoc.Templates (renderTemplate)
 import Text.Pandoc.Walk (query)
 import Text.Pandoc.Writers.Shared
 import Text.Printf (printf)
@@ -55,6 +37,8 @@ data WriterState =
   WriterState { stNextRef          :: Int  -- number of next URL reference
               , stOrderedListLevel :: Int  -- level of ordered list
               , stOptions          :: WriterOptions -- writer options
+              , stHasCslRefs       :: Bool -- has CSL citations
+              , stCslHangingIndent :: Bool -- CSL hanging indent
               }
 
 data Tabl = Xtb | Ntb deriving (Show, Eq)
@@ -68,6 +52,8 @@ writeConTeXt options document =
   let defaultWriterState = WriterState { stNextRef = 1
                                        , stOrderedListLevel = 0
                                        , stOptions = options
+                                       , stHasCslRefs = False
+                                       , stCslHangingIndent = False
                                        }
   in evalStateT (pandocToConTeXt options document) defaultWriterState
 
@@ -78,24 +64,24 @@ pandocToConTeXt options (Pandoc meta blocks) = do
   let colwidth = if writerWrapText options == WrapAuto
                     then Just $ writerColumns options
                     else Nothing
-  let render' :: Doc -> Text
-      render' = render colwidth
-  metadata <- metaToJSON options
-              (fmap render' . blockListToConTeXt)
-              (fmap render' . inlineListToConTeXt)
+  metadata <- metaToContext options
+              blockListToConTeXt
+              (fmap chomp . inlineListToConTeXt)
               meta
-  body <- mapM (elementToConTeXt options) $ hierarchicalize blocks
-  let main = (render' . vcat) body
-  let layoutFromMargins = intercalate [','] $ mapMaybe (\(x,y) ->
-                                ((x ++ "=") ++) <$> getField y metadata)
+  main <- blockListToConTeXt $ makeSections False Nothing blocks
+  let layoutFromMargins = mconcat $ intersperse ("," :: Doc Text) $
+                          mapMaybe (\(x,y) ->
+                                ((x <> "=") <>) <$> getField y metadata)
                               [("leftmargin","margin-left")
                               ,("rightmargin","margin-right")
                               ,("top","margin-top")
                               ,("bottom","margin-bottom")
                               ]
   mblang <- fromBCP47 (getLang options meta)
+  st <- get
   let context =   defField "toc" (writerTableOfContents options)
-                $ defField "placelist" (intercalate ("," :: String) $
+                $ defField "placelist"
+                   (mconcat . intersperse ("," :: Doc Text) $
                      take (writerTOCDepth options +
                            case writerTopLevelDivision options of
                              TopLevelPart    -> 0
@@ -106,25 +92,35 @@ pandocToConTeXt options (Pandoc meta blocks) = do
                 $ defField "body" main
                 $ defField "layout" layoutFromMargins
                 $ defField "number-sections" (writerNumberSections options)
-                $ maybe id (defField "context-lang") mblang
-                $ (case getField "papersize" metadata of
+                $ defField "csl-refs" (stHasCslRefs st)
+                $ defField "csl-hanging-indent" (stCslHangingIndent st)
+                $ maybe id (\l ->
+                     defField "context-lang" (literal l :: Doc Text)) mblang
+                $ (case T.unpack . render Nothing <$>
+                      getField "papersize" metadata of
                         Just (('a':d:ds) :: String)
                           | all isDigit (d:ds) -> resetField "papersize"
-                                                     (('A':d:ds) :: String)
+                                                   (T.pack ('A':d:ds))
+                        _                     -> id)
+                $ (case T.toLower $ lookupMetaString "pdfa" meta of
+                        "true" -> resetField "pdfa" (T.pack "1b:2005")
                         _                     -> id) metadata
-  let context' = defField "context-dir" (toContextDir
+  let context' = defField "context-dir" (maybe mempty toContextDir
                                          $ getField "dir" context) context
-  case writerTemplate options of
-       Nothing  -> return main
-       Just tpl -> renderTemplate' tpl context'
+  return $ render colwidth $
+    case writerTemplate options of
+       Nothing  -> main
+       Just tpl -> renderTemplate tpl context'
 
-toContextDir :: Maybe String -> String
-toContextDir (Just "rtl") = "r2l"
-toContextDir (Just "ltr") = "l2r"
-toContextDir _            = ""
+-- change rtl to r2l, ltr to l2r
+toContextDir :: Doc Text -> Doc Text
+toContextDir = fmap (\t -> case t of
+                              "ltr" -> "l2r"
+                              "rtl" -> "r2l"
+                              _     -> t)
 
 -- | escape things as needed for ConTeXt
-escapeCharForConTeXt :: WriterOptions -> Char -> String
+escapeCharForConTeXt :: WriterOptions -> Char -> Text
 escapeCharForConTeXt opts ch =
  let ligatures = isEnabled Ext_smart opts in
  case ch of
@@ -143,41 +139,40 @@ escapeCharForConTeXt opts ch =
     '\x2013' | ligatures -> "--"
     '\x2019' | ligatures -> "'"
     '\x2026' -> "\\ldots{}"
-    x        -> [x]
+    x        -> T.singleton x
 
 -- | Escape string for ConTeXt
-stringToConTeXt :: WriterOptions -> String -> String
-stringToConTeXt opts = concatMap (escapeCharForConTeXt opts)
+stringToConTeXt :: WriterOptions -> Text -> Text
+stringToConTeXt opts = T.concatMap (escapeCharForConTeXt opts)
 
 -- | Sanitize labels
-toLabel :: String -> String
-toLabel z = concatMap go z
+toLabel :: Text -> Text
+toLabel z = T.concatMap go z
  where go x
-         | x `elem` ("\\#[]\",{}%()|=" :: String) = "ux" ++ printf "%x" (ord x)
-         | otherwise = [x]
-
--- | Convert Elements to ConTeXt
-elementToConTeXt :: PandocMonad m => WriterOptions -> Element -> WM m Doc
-elementToConTeXt _ (Blk block) = blockToConTeXt block
-elementToConTeXt opts (Sec level _ attr title' elements) = do
-  header' <- sectionHeader attr level title'
-  footer' <- sectionFooter attr level
-  innerContents <- mapM (elementToConTeXt opts) elements
-  return $ header' $$ vcat innerContents $$ footer'
+         | x `elem` ("\\#[]\",{}%()|=" :: String) = "ux" <> T.pack (printf "%x" (ord x))
+         | otherwise = T.singleton x
 
 -- | Convert Pandoc block element to ConTeXt.
-blockToConTeXt :: PandocMonad m => Block -> WM m Doc
+blockToConTeXt :: PandocMonad m => Block -> WM m (Doc Text)
 blockToConTeXt Null = return empty
+blockToConTeXt (Div attr@(_,"section":_,_)
+                 (Header level _ title' : xs)) = do
+  header' <- sectionHeader attr level title'
+  footer' <- sectionFooter attr level
+  innerContents <- blockListToConTeXt xs
+  return $ header' $$ innerContents $$ footer'
 blockToConTeXt (Plain lst) = inlineListToConTeXt lst
 -- title beginning with fig: indicates that the image is a figure
-blockToConTeXt (Para [Image attr txt (src,'f':'i':'g':':':_)]) = do
-  capt <- inlineListToConTeXt txt
-  img  <- inlineToConTeXt (Image attr txt (src, ""))
-  let (ident, _, _) = attr
-      label = if null ident
-                 then empty
-                 else "[]" <> brackets (text $ toLabel ident)
-  return $ blankline $$ "\\placefigure" <> label <> braces capt <> img <> blankline
+blockToConTeXt (Para [Image attr txt (src,tgt)])
+  | Just _ <- T.stripPrefix "fig:" tgt
+  = do
+      capt <- inlineListToConTeXt txt
+      img  <- inlineToConTeXt (Image attr txt (src, ""))
+      let (ident, _, _) = attr
+          label = if T.null ident
+                  then empty
+                  else "[]" <> brackets (literal $ toLabel ident)
+      return $ blankline $$ "\\placefigure" <> label <> braces capt <> img <> blankline
 blockToConTeXt (Para lst) = do
   contents <- inlineListToConTeXt lst
   return $ contents <> blankline
@@ -188,17 +183,22 @@ blockToConTeXt (BlockQuote lst) = do
   contents <- blockListToConTeXt lst
   return $ "\\startblockquote" $$ nest 0 contents $$ "\\stopblockquote" <> blankline
 blockToConTeXt (CodeBlock _ str) =
-  return $ flush ("\\starttyping" <> cr <> text str <> cr <> "\\stoptyping") $$ blankline
+  return $ flush ("\\starttyping" <> cr <> literal str <> cr <> "\\stoptyping") $$ blankline
   -- blankline because \stoptyping can't have anything after it, inc. '}'
 blockToConTeXt b@(RawBlock f str)
-  | f == Format "context" || f == Format "tex" = return $ text str <> blankline
+  | f == Format "context" || f == Format "tex" = return $ literal str <> blankline
   | otherwise = empty <$ report (BlockNotRendered b)
+blockToConTeXt (Div ("refs",classes,_) bs) = do
+  modify $ \st -> st{ stHasCslRefs = True
+                    , stCslHangingIndent = "hanging-indent" `elem` classes }
+  inner <- blockListToConTeXt bs
+  return $ "\\startcslreferences" $$ inner $$ "\\stopcslreferences"
 blockToConTeXt (Div (ident,_,kvs) bs) = do
   let align dir txt = "\\startalignment[" <> dir <> "]" $$ txt $$ "\\stopalignment"
   mblang <- fromBCP47 (lookup "lang" kvs)
-  let wrapRef txt = if null ident
+  let wrapRef txt = if T.null ident
                        then txt
-                       else ("\\reference" <> brackets (text $ toLabel ident) <>
+                       else ("\\reference" <> brackets (literal $ toLabel ident) <>
                               braces empty <> "%") $$ txt
       wrapDir = case lookup "dir" kvs of
                   Just "rtl" -> align "righttoleft"
@@ -206,7 +206,7 @@ blockToConTeXt (Div (ident,_,kvs) bs) = do
                   _          -> id
       wrapLang txt = case mblang of
                        Just lng -> "\\start\\language["
-                                     <> text lng <> "]" $$ txt $$ "\\stop"
+                                     <> literal lng <> "]" $$ txt $$ "\\stop"
                        Nothing  -> txt
       wrapBlank txt = blankline <> txt <> blankline
   (wrapBlank . wrapLang . wrapDir . wrapRef) <$> blockListToConTeXt bs
@@ -215,29 +215,29 @@ blockToConTeXt (BulletList lst) = do
   return $ ("\\startitemize" <> if isTightList lst
                                    then brackets "packed"
                                    else empty) $$
-    vcat contents $$ text "\\stopitemize" <> blankline
+    vcat contents $$ literal "\\stopitemize" <> blankline
 blockToConTeXt (OrderedList (start, style', delim) lst) = do
     st <- get
     let level = stOrderedListLevel st
     put st {stOrderedListLevel = level + 1}
     contents <- mapM listItemToConTeXt lst
     put st {stOrderedListLevel = level}
-    let start' = if start == 1 then "" else "start=" ++ show start
+    let start' = if start == 1 then "" else "start=" <> tshow start
     let delim' = case delim of
                         DefaultDelim -> ""
                         Period       -> "stopper=."
                         OneParen     -> "stopper=)"
                         TwoParens    -> "left=(,stopper=)"
-    let width = maximum $ map length $ take (length contents)
+    let width = maximum $ map T.length $ take (length contents)
                           (orderedListMarkers (start, style', delim))
     let width' = (toEnum width + 1) / 2
     let width'' = if width' > (1.5 :: Double)
-                     then "width=" ++ show width' ++ "em"
+                     then "width=" <> tshow width' <> "em"
                      else ""
-    let specs2Items = filter (not . null) [start', delim', width'']
+    let specs2Items = filter (not . T.null) [start', delim', width'']
     let specs2 = if null specs2Items
                     then ""
-                    else "[" ++ intercalate "," specs2Items ++ "]"
+                    else "[" <> T.intercalate "," specs2Items <> "]"
     let style'' = '[': (case style' of
                           DefaultStyle -> orderedListStyles !! level
                           Decimal      -> 'n'
@@ -247,8 +247,8 @@ blockToConTeXt (OrderedList (start, style', delim) lst) = do
                           LowerAlpha   -> 'a'
                           UpperAlpha   -> 'A') :
                        if isTightList lst then ",packed]" else "]"
-    let specs = style'' ++ specs2
-    return $ "\\startitemize" <> text specs $$ vcat contents $$
+    let specs = T.pack style'' <> specs2
+    return $ "\\startitemize" <> literal specs $$ vcat contents $$
              "\\stopitemize" <> blankline
 blockToConTeXt (DefinitionList lst) =
   liftM vcat $ mapM defListItemToConTeXt lst
@@ -272,7 +272,8 @@ blockToConTeXt (Table caption aligns widths heads rows) = do
         else "title=" <> braces captionText
       ) $$ body $$ "\\stopplacetable" <> blankline
 
-tableToConTeXt :: PandocMonad m => Tabl -> Doc -> [Doc] -> WM m Doc
+tableToConTeXt :: PandocMonad m
+               => Tabl -> Doc Text -> [Doc Text] -> WM m (Doc Text)
 tableToConTeXt Xtb heads rows =
   return $ "\\startxtable" $$
     (if isEmpty heads
@@ -294,7 +295,7 @@ tableToConTeXt Ntb heads rows =
            "\\startTABLEfoot" $$ last rows $$ "\\stopTABLEfoot") $$
     "\\stopTABLE"
 
-tableRowToConTeXt :: PandocMonad m => Tabl -> [Alignment] -> [Double] -> [[Block]] -> WM m Doc
+tableRowToConTeXt :: PandocMonad m => Tabl -> [Alignment] -> [Double] -> [[Block]] -> WM m (Doc Text)
 tableRowToConTeXt Xtb aligns widths cols = do
   cells <- mapM (tableColToConTeXt Xtb) $ zip3 aligns widths cols
   return $ "\\startxrow" $$ vcat cells $$ "\\stopxrow"
@@ -302,36 +303,37 @@ tableRowToConTeXt Ntb aligns widths cols = do
   cells <- mapM (tableColToConTeXt Ntb) $ zip3 aligns widths cols
   return $ vcat cells $$ "\\NC\\NR"
 
-tableColToConTeXt :: PandocMonad m => Tabl -> (Alignment, Double, [Block]) -> WM m Doc
+tableColToConTeXt :: PandocMonad m => Tabl -> (Alignment, Double, [Block]) -> WM m (Doc Text)
 tableColToConTeXt tabl (align, width, blocks) = do
   cellContents <- blockListToConTeXt blocks
   let colwidth = if width == 0
         then empty
         else "width=" <> braces (text (printf "%.2f\\textwidth" width))
   let halign = alignToConTeXt align
-  let options = (if keys == empty
+  let options = (if isEmpty keys
                  then empty
                  else brackets keys) <> space
-        where keys = hcat $ intersperse "," $ filter (empty /=) [halign, colwidth]
+        where keys = hcat $ intersperse "," $ filter (not . isEmpty) [halign, colwidth]
   tableCellToConTeXt tabl options cellContents
 
-tableCellToConTeXt :: PandocMonad m => Tabl -> Doc -> Doc -> WM m Doc
+tableCellToConTeXt :: PandocMonad m
+                   => Tabl -> Doc Text -> Doc Text -> WM m (Doc Text)
 tableCellToConTeXt Xtb options cellContents =
   return $ "\\startxcell" <> options <> cellContents <> " \\stopxcell"
 tableCellToConTeXt Ntb options cellContents =
   return $ "\\NC" <> options <> cellContents
 
-alignToConTeXt :: Alignment -> Doc
+alignToConTeXt :: Alignment -> Doc Text
 alignToConTeXt align = case align of
                          AlignLeft    -> "align=right"
                          AlignRight   -> "align=left"
                          AlignCenter  -> "align=middle"
                          AlignDefault -> empty
 
-listItemToConTeXt :: PandocMonad m => [Block] -> WM m Doc
+listItemToConTeXt :: PandocMonad m => [Block] -> WM m (Doc Text)
 listItemToConTeXt list = (("\\item" $$) . nest 2) <$> blockListToConTeXt list
 
-defListItemToConTeXt :: PandocMonad m => ([Inline], [[Block]]) -> WM m Doc
+defListItemToConTeXt :: PandocMonad m => ([Inline], [[Block]]) -> WM m (Doc Text)
 defListItemToConTeXt (term, defs) = do
   term' <- inlineListToConTeXt term
   def'  <- liftM vsep $ mapM blockListToConTeXt defs
@@ -339,13 +341,13 @@ defListItemToConTeXt (term, defs) = do
            "\\stopdescription" <> blankline
 
 -- | Convert list of block elements to ConTeXt.
-blockListToConTeXt :: PandocMonad m => [Block] -> WM m Doc
+blockListToConTeXt :: PandocMonad m => [Block] -> WM m (Doc Text)
 blockListToConTeXt lst = liftM vcat $ mapM blockToConTeXt lst
 
 -- | Convert list of inline elements to ConTeXt.
 inlineListToConTeXt :: PandocMonad m
                     => [Inline]  -- ^ Inlines to convert
-                    -> WM m Doc
+                    -> WM m (Doc Text)
 inlineListToConTeXt lst = liftM hcat $ mapM inlineToConTeXt $ addStruts lst
   -- We add a \strut after a line break that precedes a space,
   -- or the space gets swallowed
@@ -354,14 +356,14 @@ inlineListToConTeXt lst = liftM hcat $ mapM inlineToConTeXt $ addStruts lst
              addStruts xs
         addStruts (x:xs) = x : addStruts xs
         addStruts [] = []
-        isSpacey Space            = True
-        isSpacey (Str ('\160':_)) = True
-        isSpacey _                = False
+        isSpacey Space                               = True
+        isSpacey (Str (T.uncons -> Just ('\160',_))) = True
+        isSpacey _                                   = False
 
 -- | Convert inline element to ConTeXt
 inlineToConTeXt :: PandocMonad m
                 => Inline    -- ^ Inline to convert
-                -> WM m Doc
+                -> WM m (Doc Text)
 inlineToConTeXt (Emph lst) = do
   contents <- inlineListToConTeXt lst
   return $ braces $ "\\em " <> contents
@@ -380,11 +382,11 @@ inlineToConTeXt (Subscript lst) = do
 inlineToConTeXt (SmallCaps lst) = do
   contents <- inlineListToConTeXt lst
   return $ braces $ "\\sc " <> contents
-inlineToConTeXt (Code _ str) | not ('{' `elem` str || '}' `elem` str) =
-  return $ "\\type" <> braces (text str)
+inlineToConTeXt (Code _ str) | not ('{' `elemText` str || '}' `elemText` str) =
+  return $ "\\type" <> braces (literal str)
 inlineToConTeXt (Code _ str) = do
   opts <- gets stOptions
-  return $ "\\mono" <> braces (text $ stringToConTeXt opts str)
+  return $ "\\mono" <> braces (literal $ stringToConTeXt opts str)
 inlineToConTeXt (Quoted SingleQuote lst) = do
   contents <- inlineListToConTeXt lst
   return $ "\\quote" <> braces contents
@@ -394,15 +396,15 @@ inlineToConTeXt (Quoted DoubleQuote lst) = do
 inlineToConTeXt (Cite _ lst) = inlineListToConTeXt lst
 inlineToConTeXt (Str str) = do
   opts <- gets stOptions
-  return $ text $ stringToConTeXt opts str
+  return $ literal $ stringToConTeXt opts str
 inlineToConTeXt (Math InlineMath str) =
-  return $ char '$' <> text str <> char '$'
+  return $ char '$' <> literal str <> char '$'
 inlineToConTeXt (Math DisplayMath str) =
-  return $ text "\\startformula "  <> text str <> text " \\stopformula" <> space
+  return $ literal "\\startformula "  <> literal str <> literal " \\stopformula" <> space
 inlineToConTeXt il@(RawInline f str)
-  | f == Format "tex" || f == Format "context" = return $ text str
+  | f == Format "tex" || f == Format "context" = return $ literal str
   | otherwise = empty <$ report (InlineNotRendered il)
-inlineToConTeXt LineBreak = return $ text "\\crlf" <> cr
+inlineToConTeXt LineBreak = return $ literal "\\crlf" <> cr
 inlineToConTeXt SoftBreak = do
   wrapText <- gets (writerWrapText . stOptions)
   return $ case wrapText of
@@ -411,66 +413,66 @@ inlineToConTeXt SoftBreak = do
                WrapPreserve -> cr
 inlineToConTeXt Space = return space
 -- Handle HTML-like internal document references to sections
-inlineToConTeXt (Link _ txt ('#' : ref, _)) = do
+inlineToConTeXt (Link _ txt (T.uncons -> Just ('#', ref), _)) = do
   opts <- gets stOptions
   contents <-  inlineListToConTeXt txt
   let ref' = toLabel $ stringToConTeXt opts ref
-  return $ text "\\goto"
+  return $ literal "\\goto"
            <> braces contents
-           <> brackets (text ref')
+           <> brackets (literal ref')
 
 inlineToConTeXt (Link _ txt (src, _)) = do
-  let isAutolink = txt == [Str (unEscapeString src)]
+  let isAutolink = txt == [Str (T.pack $ unEscapeString $ T.unpack src)]
   st <- get
   let next = stNextRef st
   put $ st {stNextRef = next + 1}
-  let ref = "url" ++ show next
+  let ref = "url" <> tshow next
   contents <-  inlineListToConTeXt txt
   return $ "\\useURL"
-           <> brackets (text ref)
-           <> brackets (text $ escapeStringUsing [('#',"\\#"),('%',"\\%")] src)
+           <> brackets (literal ref)
+           <> brackets (literal $ escapeStringUsing [('#',"\\#"),('%',"\\%")] src)
            <> (if isAutolink
                   then empty
                   else brackets empty <> brackets contents)
            <> "\\from"
-           <> brackets (text ref)
+           <> brackets (literal ref)
 inlineToConTeXt (Image attr@(_,cls,_) _ (src, _)) = do
   opts <- gets stOptions
-  let showDim dir = let d = text (show dir) <> "="
+  let showDim dir = let d = literal (tshow dir) <> "="
                     in case dimension dir attr of
                          Just (Pixel a)   ->
-                           [d <> text (showInInch opts (Pixel a)) <> "in"]
+                           [d <> literal (showInInch opts (Pixel a)) <> "in"]
                          Just (Percent a) ->
-                           [d <> text (showFl (a / 100)) <> "\\textwidth"]
+                           [d <> literal (showFl (a / 100)) <> "\\textwidth"]
                          Just dim         ->
-                           [d <> text (show dim)]
+                           [d <> literal (tshow dim)]
                          Nothing          ->
                            []
       dimList = showDim Width ++ showDim Height
       dims = if null dimList
                 then empty
-                else brackets $ cat (intersperse "," dimList)
+                else brackets $ mconcat (intersperse "," dimList)
       clas = if null cls
                 then empty
-                else brackets $ text $ toLabel $ head cls
+                else brackets $ literal $ toLabel $ head cls
       -- Use / for path separators on Windows; see #4918
-      fixPathSeparators = map $ \c -> case c of
-                                           '\\' -> '/'
-                                           _    -> c
+      fixPathSeparators = T.map $ \c -> case c of
+                                          '\\' -> '/'
+                                          _    -> c
       src' = fixPathSeparators $
              if isURI src
                 then src
-                else unEscapeString src
-  return $ braces $ "\\externalfigure" <> brackets (text src') <> dims <> clas
+                else T.pack $ unEscapeString $ T.unpack src
+  return $ braces $ "\\externalfigure" <> brackets (literal src') <> dims <> clas
 inlineToConTeXt (Note contents) = do
   contents' <- blockListToConTeXt contents
   let codeBlock x@(CodeBlock _ _) = [x]
       codeBlock _                 = []
   let codeBlocks = query codeBlock contents
   return $ if null codeBlocks
-              then text "\\footnote{" <> nest 2 contents' <> char '}'
-              else text "\\startbuffer " <> nest 2 contents' <>
-                   text "\\stopbuffer\\footnote{\\getbuffer}"
+              then literal "\\footnote{" <> nest 2 (chomp contents') <> char '}'
+              else literal "\\startbuffer " <> nest 2 (chomp contents') <>
+                   literal "\\stopbuffer\\footnote{\\getbuffer}"
 inlineToConTeXt (Span (_,_,kvs) ils) = do
   mblang <- fromBCP47 (lookup "lang" kvs)
   let wrapDir txt = case lookup "dir" kvs of
@@ -478,8 +480,8 @@ inlineToConTeXt (Span (_,_,kvs) ils) = do
                       Just "ltr" -> braces $ "\\lefttoright " <> txt
                       _          -> txt
       wrapLang txt = case mblang of
-                       Just lng -> "\\start\\language[" <> text lng
-                                      <> "]" <> txt <> "\\stop "
+                       Just lng -> braces ("\\language" <>
+                                           brackets (literal lng) <> txt)
                        Nothing -> txt
   (wrapLang . wrapDir) <$> inlineListToConTeXt ils
 
@@ -488,28 +490,28 @@ sectionHeader :: PandocMonad m
               => Attr
               -> Int
               -> [Inline]
-              -> WM m Doc
+              -> WM m (Doc Text)
 sectionHeader (ident,classes,kvs) hdrLevel lst = do
   opts <- gets stOptions
   contents <- inlineListToConTeXt lst
   levelText <- sectionLevelToText opts (ident,classes,kvs) hdrLevel
-  let ident' = if null ident
+  let ident' = if T.null ident
                then empty
-               else "reference=" <> braces (text (toLabel ident))
-  let contents' = if contents == empty
+               else "reference=" <> braces (literal (toLabel ident))
+  let contents' = if isEmpty contents
                   then empty
                   else "title=" <> braces contents
-  let options = if keys == empty || levelText == empty
+  let options = if isEmpty keys || isEmpty levelText
                 then empty
                 else brackets keys
-        where keys = hcat $ intersperse "," $ filter (empty /=) [contents', ident']
+        where keys = hcat $ intersperse "," $ filter (not . isEmpty) [contents', ident']
   let starter = if writerSectionDivs opts
                 then "\\start"
                 else "\\"
   return $ starter <> levelText <> options <> blankline
 
 -- | Craft the section footer
-sectionFooter :: PandocMonad m => Attr -> Int -> WM m Doc
+sectionFooter :: PandocMonad m => Attr -> Int -> WM m (Doc Text)
 sectionFooter attr hdrLevel = do
   opts <- gets stOptions
   levelText <- sectionLevelToText opts attr hdrLevel
@@ -518,7 +520,7 @@ sectionFooter attr hdrLevel = do
            else empty
 
 -- | Generate a textual representation of the section level
-sectionLevelToText :: PandocMonad m => WriterOptions -> Attr -> Int -> WM m Doc
+sectionLevelToText :: PandocMonad m => WriterOptions -> Attr -> Int -> WM m (Doc Text)
 sectionLevelToText opts (_,classes,_) hdrLevel = do
   let level' = case writerTopLevelDivision opts of
                  TopLevelPart    -> hdrLevel - 2
@@ -526,23 +528,23 @@ sectionLevelToText opts (_,classes,_) hdrLevel = do
                  TopLevelSection -> hdrLevel
                  TopLevelDefault -> hdrLevel
   let (section, chapter) = if "unnumbered" `elem` classes
-                              then (text "subject", text "title")
-                              else (text "section", text "chapter")
+                              then (literal "subject", literal "title")
+                              else (literal "section", literal "chapter")
   return $ case level' of
-             -1         -> text "part"
+             -1         -> literal "part"
              0          -> chapter
              n | n >= 1 -> text (concat (replicate (n - 1) "sub"))
                            <> section
              _          -> empty -- cannot happen
 
-fromBCP47 :: PandocMonad m => Maybe String -> WM m (Maybe String)
+fromBCP47 :: PandocMonad m => Maybe Text -> WM m (Maybe Text)
 fromBCP47 mbs = fromBCP47' <$> toLang mbs
 
 -- Takes a list of the constituents of a BCP 47 language code
 -- and irons out ConTeXt's exceptions
 -- https://tools.ietf.org/html/bcp47#section-2.1
 -- http://wiki.contextgarden.net/Language_Codes
-fromBCP47' :: Maybe Lang -> Maybe String
+fromBCP47' :: Maybe Lang -> Maybe Text
 fromBCP47' (Just (Lang "ar" _ "SY" _)     ) = Just "ar-sy"
 fromBCP47' (Just (Lang "ar" _ "IQ" _)     ) = Just "ar-iq"
 fromBCP47' (Just (Lang "ar" _ "JO" _)     ) = Just "ar-jo"
