@@ -1,26 +1,8 @@
-{-# LANGUAGE NoImplicitPrelude #-}
 {-# LANGUAGE OverloadedStrings #-}
-{-
-Copyright (C) 2007-2018 John MacFarlane <jgm@berkeley.edu>
-
-This program is free software; you can redistribute it and/or modify
-it under the terms of the GNU General Public License as published by
-the Free Software Foundation; either version 2 of the License, or
-(at your option) any later version.
-
-This program is distributed in the hope that it will be useful,
-but WITHOUT ANY WARRANTY; without even the implied warranty of
-MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-GNU General Public License for more details.
-
-You should have received a copy of the GNU General Public License
-along with this program; if not, write to the Free Software
-Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--}
-
+{-# LANGUAGE ViewPatterns      #-}
 {- |
    Module      : Text.Pandoc.Writers.Man
-   Copyright   : Copyright (C) 2007-2018 John MacFarlane
+   Copyright   : Copyright (C) 2007-2020 John MacFarlane
    License     : GNU GPL, version 2 or above
 
    Maintainer  : John MacFarlane <jgm@berkeley.edu>
@@ -30,21 +12,21 @@ Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 Conversion of 'Pandoc' documents to roff man page format.
 
 -}
-module Text.Pandoc.Writers.Man ( writeMan) where
-import Prelude
+module Text.Pandoc.Writers.Man ( writeMan ) where
 import Control.Monad.State.Strict
-import Data.List (intersperse, stripPrefix)
+import Data.List (intersperse)
 import Data.Maybe (fromMaybe)
 import Data.Text (Text)
 import qualified Data.Text as T
 import Text.Pandoc.Builder (deleteMeta)
-import Text.Pandoc.Class (PandocMonad, report)
+import Text.Pandoc.Class.PandocMonad (PandocMonad, report)
 import Text.Pandoc.Definition
 import Text.Pandoc.Logging
 import Text.Pandoc.Options
-import Text.Pandoc.Pretty
+import Text.DocLayout
 import Text.Pandoc.Shared
-import Text.Pandoc.Templates
+import Text.Pandoc.Walk (walk)
+import Text.Pandoc.Templates (renderTemplate)
 import Text.Pandoc.Writers.Math
 import Text.Pandoc.Writers.Shared
 import Text.Pandoc.Writers.Roff
@@ -61,10 +43,8 @@ pandocToMan opts (Pandoc meta blocks) = do
   let colwidth = if writerWrapText opts == WrapAuto
                     then Just $ writerColumns opts
                     else Nothing
-  let render' :: Doc -> Text
-      render' = render colwidth
   titleText <- inlineListToMan opts $ docTitle meta
-  let title' = render' titleText
+  let title' = render Nothing titleText
   let setFieldsFromTitle =
        case T.break (== ' ') title' of
            (cmdName, rest) -> case T.break (=='(') cmdName of
@@ -79,37 +59,37 @@ pandocToMan opts (Pandoc meta blocks) = do
                                                (T.strip $ mconcat hds)
                                           [] -> id
                                    _  -> defField "title" title'
-  metadata <- metaToJSON opts
-              (fmap render' . blockListToMan opts)
-              (fmap render' . inlineListToMan opts)
+  metadata <- metaToContext opts
+              (blockListToMan opts)
+              (fmap chomp . inlineListToMan opts)
               $ deleteMeta "title" meta
   body <- blockListToMan opts blocks
   notes <- gets stNotes
   notes' <- notesToMan opts (reverse notes)
-  let main = render' $ body $$ notes' $$ text ""
+  let main = body $$ notes' $$ text ""
   hasTables <- gets stHasTables
   let context = defField "body" main
               $ setFieldsFromTitle
               $ defField "has-tables" hasTables
               $ defField "hyphenate" True
               $ defField "pandoc-version" pandocVersion metadata
-  case writerTemplate opts of
-       Nothing  -> return main
-       Just tpl -> renderTemplate' tpl context
+  return $ render colwidth $
+    case writerTemplate opts of
+       Nothing  -> main
+       Just tpl -> renderTemplate tpl context
 
-escString :: WriterOptions -> String -> String
+escString :: WriterOptions -> Text -> Text
 escString _ = escapeString AsciiOnly -- for better portability
 
 -- | Return man representation of notes.
-notesToMan :: PandocMonad m => WriterOptions -> [[Block]] -> StateT WriterState m Doc
+notesToMan :: PandocMonad m => WriterOptions -> [[Block]] -> StateT WriterState m (Doc Text)
 notesToMan opts notes =
   if null notes
      then return empty
-     else zipWithM (noteToMan opts) [1..] notes >>=
-          return . (text ".SH NOTES" $$) . vcat
+     else (text ".SH NOTES" $$) . vcat <$> zipWithM (noteToMan opts) [1..] notes
 
 -- | Return man representation of a note.
-noteToMan :: PandocMonad m => WriterOptions -> Int -> [Block] -> StateT WriterState m Doc
+noteToMan :: PandocMonad m => WriterOptions -> Int -> [Block] -> StateT WriterState m (Doc Text)
 noteToMan opts num note = do
   contents <- blockListToMan opts note
   let marker = cr <> text ".SS " <> brackets (text (show num))
@@ -123,7 +103,7 @@ noteToMan opts num note = do
 blockToMan :: PandocMonad m
            => WriterOptions -- ^ Options
            -> Block         -- ^ Block element
-           -> StateT WriterState m Doc
+           -> StateT WriterState m (Doc Text)
 blockToMan _ Null = return empty
 blockToMan opts (Div _ bs) = blockListToMan opts bs
 blockToMan opts (Plain inlines) =
@@ -135,30 +115,30 @@ blockToMan opts (Para inlines) = do
 blockToMan opts (LineBlock lns) =
   blockToMan opts $ linesToPara lns
 blockToMan _ b@(RawBlock f str)
-  | f == Format "man" = return $ text str
+  | f == Format "man" = return $ literal str
   | otherwise         = do
       report $ BlockNotRendered b
       return empty
-blockToMan _ HorizontalRule = return $ text ".PP" $$ text "   *   *   *   *   *"
+blockToMan _ HorizontalRule = return $ literal ".PP" $$ literal "   *   *   *   *   *"
 blockToMan opts (Header level _ inlines) = do
   contents <- inlineListToMan opts inlines
   let heading = case level of
                   1 -> ".SH "
                   _ -> ".SS "
-  return $ nowrap $ text heading <> contents
+  return $ nowrap $ literal heading <> contents
 blockToMan opts (CodeBlock _ str) = return $
-  text ".IP" $$
-  text ".nf" $$
-  text "\\f[C]" $$
-  ((case str of
-    '.':_ -> text "\\&"
-    _     -> mempty) <>
-   text (escString opts str)) $$
-  text "\\f[R]" $$
-  text ".fi"
+  literal ".IP" $$
+  literal ".nf" $$
+  literal "\\f[C]" $$
+  ((case T.uncons str of
+    Just ('.',_) -> literal "\\&"
+    _            -> mempty) <>
+   literal (escString opts str)) $$
+  literal "\\f[R]" $$
+  literal ".fi"
 blockToMan opts (BlockQuote blocks) = do
   contents <- blockListToMan opts blocks
-  return $ text ".RS" $$ contents $$ text ".RE"
+  return $ literal ".RS" $$ contents $$ literal ".RE"
 blockToMan opts (Table caption alignments widths headers rows) =
   let aligncode AlignLeft    = "l"
       aligncode AlignRight   = "r"
@@ -169,24 +149,24 @@ blockToMan opts (Table caption alignments widths headers rows) =
   modify $ \st -> st{ stHasTables = True }
   let iwidths = if all (== 0) widths
                    then repeat ""
-                   else map (printf "w(%0.1fn)" . (70 *)) widths
+                   else map (T.pack . printf "w(%0.1fn)" . (70 *)) widths
   -- 78n default width - 8n indent = 70n
-  let coldescriptions = text $ unwords
-                        (zipWith (\align width -> aligncode align ++ width)
-                        alignments iwidths) ++ "."
+  let coldescriptions = literal $ T.unwords
+                        (zipWith (\align width -> aligncode align <> width)
+                        alignments iwidths) <> "."
   colheadings <- mapM (blockListToMan opts) headers
-  let makeRow cols = text "T{" $$
-                     vcat (intersperse (text "T}@T{") cols) $$
-                     text "T}"
+  let makeRow cols = literal "T{" $$
+                     vcat (intersperse (literal "T}@T{") cols) $$
+                     literal "T}"
   let colheadings' = if all null headers
                         then empty
                         else makeRow colheadings $$ char '_'
   body <- mapM (\row -> do
                          cols <- mapM (blockListToMan opts) row
                          return $ makeRow cols) rows
-  return $ text ".PP" $$ caption' $$
-           text ".TS" $$ text "tab(@);" $$ coldescriptions $$
-           colheadings' $$ vcat body $$ text ".TE"
+  return $ literal ".PP" $$ caption' $$
+           literal ".TS" $$ literal "tab(@);" $$ coldescriptions $$
+           colheadings' $$ vcat body $$ literal ".TE"
 
 blockToMan opts (BulletList items) = do
   contents <- mapM (bulletListItemToMan opts) items
@@ -194,7 +174,7 @@ blockToMan opts (BulletList items) = do
 blockToMan opts (OrderedList attribs items) = do
   let markers = take (length items) $ orderedListMarkers attribs
   let indent = 1 +
-                     maximum (map length markers)
+                     maximum (map T.length markers)
   contents <- mapM (\(num, item) -> orderedListItemToMan opts num indent item) $
               zip markers items
   return (vcat contents)
@@ -203,30 +183,30 @@ blockToMan opts (DefinitionList items) = do
   return (vcat contents)
 
 -- | Convert bullet list item (list of blocks) to man.
-bulletListItemToMan :: PandocMonad m => WriterOptions -> [Block] -> StateT WriterState m Doc
+bulletListItemToMan :: PandocMonad m => WriterOptions -> [Block] -> StateT WriterState m (Doc Text)
 bulletListItemToMan _ [] = return empty
 bulletListItemToMan opts (Para first:rest) =
   bulletListItemToMan opts (Plain first:rest)
 bulletListItemToMan opts (Plain first:rest) = do
   first' <- blockToMan opts (Plain first)
   rest' <- blockListToMan opts rest
-  let first'' = text ".IP \\[bu] 2" $$ first'
+  let first'' = literal ".IP \\[bu] 2" $$ first'
   let rest''  = if null rest
                    then empty
-                   else text ".RS 2" $$ rest' $$ text ".RE"
+                   else literal ".RS 2" $$ rest' $$ literal ".RE"
   return (first'' $$ rest'')
 bulletListItemToMan opts (first:rest) = do
   first' <- blockToMan opts first
   rest' <- blockListToMan opts rest
-  return $ text "\\[bu] .RS 2" $$ first' $$ rest' $$ text ".RE"
+  return $ literal "\\[bu] .RS 2" $$ first' $$ rest' $$ literal ".RE"
 
 -- | Convert ordered list item (a list of blocks) to man.
 orderedListItemToMan :: PandocMonad m
                      => WriterOptions -- ^ options
-                     -> String   -- ^ order marker for list item
+                     -> Text   -- ^ order marker for list item
                      -> Int      -- ^ number of spaces to indent
                      -> [Block]  -- ^ list item (list of blocks)
-                     -> StateT WriterState m Doc
+                     -> StateT WriterState m (Doc Text)
 orderedListItemToMan _ _ _ [] = return empty
 orderedListItemToMan opts num indent (Para first:rest) =
   orderedListItemToMan opts num indent (Plain first:rest)
@@ -234,19 +214,21 @@ orderedListItemToMan opts num indent (first:rest) = do
   first' <- blockToMan opts first
   rest' <- blockListToMan opts rest
   let num' = printf ("%" ++ show (indent - 1) ++ "s") num
-  let first'' = text (".IP \"" ++ num' ++ "\" " ++ show indent) $$ first'
+  let first'' = literal (".IP \"" <> T.pack num' <> "\" " <> tshow indent) $$ first'
   let rest''  = if null rest
                    then empty
-                   else text ".RS 4" $$ rest' $$ text ".RE"
+                   else literal ".RS 4" $$ rest' $$ literal ".RE"
   return $ first'' $$ rest''
 
 -- | Convert definition list item (label, list of blocks) to man.
 definitionListItemToMan :: PandocMonad m
                         => WriterOptions
                         -> ([Inline],[[Block]])
-                        -> StateT WriterState m Doc
+                        -> StateT WriterState m (Doc Text)
 definitionListItemToMan opts (label, defs) = do
-  labelText <- inlineListToMan opts label
+  -- in most man pages, option and other code in option lists is boldface,
+  -- but not other things, so we try to reproduce this style:
+  labelText <- inlineListToMan opts $ makeCodeBold label
   contents <- if null defs
                  then return empty
                  else liftM vcat $ forM defs $ \blocks ->
@@ -261,24 +243,29 @@ definitionListItemToMan opts (label, defs) = do
                             return $ first' $$
                                      if null xs
                                         then empty
-                                        else text ".RS" $$ rest' $$ text ".RE"
+                                        else literal ".RS" $$ rest' $$ literal ".RE"
                           [] -> return empty
-  return $ text ".TP" $$ nowrap (text ".B " <> labelText) $$ contents
+  return $ literal ".TP" $$ nowrap labelText $$ contents
+
+makeCodeBold :: [Inline] -> [Inline]
+makeCodeBold = walk go
+  where go x@Code{} = Strong [x]
+        go x        = x
 
 -- | Convert list of Pandoc block elements to man.
 blockListToMan :: PandocMonad m
                => WriterOptions -- ^ Options
                -> [Block]       -- ^ List of block elements
-               -> StateT WriterState m Doc
+               -> StateT WriterState m (Doc Text)
 blockListToMan opts blocks =
   vcat <$> mapM (blockToMan opts) blocks
 
 -- | Convert list of Pandoc inline elements to man.
-inlineListToMan :: PandocMonad m => WriterOptions -> [Inline] -> StateT WriterState m Doc
+inlineListToMan :: PandocMonad m => WriterOptions -> [Inline] -> StateT WriterState m (Doc Text)
 inlineListToMan opts lst = hcat <$> mapM (inlineToMan opts) lst
 
 -- | Convert Pandoc inline element to man.
-inlineToMan :: PandocMonad m => WriterOptions -> Inline -> StateT WriterState m Doc
+inlineToMan :: PandocMonad m => WriterOptions -> Inline -> StateT WriterState m (Doc Text)
 inlineToMan opts (Span _ ils) = inlineListToMan opts ils
 inlineToMan opts (Emph lst) =
   withFontFeature 'I' (inlineListToMan opts lst)
@@ -286,7 +273,7 @@ inlineToMan opts (Strong lst) =
   withFontFeature 'B' (inlineListToMan opts lst)
 inlineToMan opts (Strikeout lst) = do
   contents <- inlineListToMan opts lst
-  return $ text "[STRIKEOUT:" <> contents <> char ']'
+  return $ literal "[STRIKEOUT:" <> contents <> char ']'
 inlineToMan opts (Superscript lst) = do
   contents <- inlineListToMan opts lst
   return $ char '^' <> contents <> char '^'
@@ -299,48 +286,48 @@ inlineToMan opts (Quoted SingleQuote lst) = do
   return $ char '`' <> contents <> char '\''
 inlineToMan opts (Quoted DoubleQuote lst) = do
   contents <- inlineListToMan opts lst
-  return $ text "\\[lq]" <> contents <> text "\\[rq]"
+  return $ literal "\\[lq]" <> contents <> literal "\\[rq]"
 inlineToMan opts (Cite _ lst) =
   inlineListToMan opts lst
 inlineToMan opts (Code _ str) =
-  withFontFeature 'C' (return (text $ escString opts str))
-inlineToMan opts (Str str@('.':_)) =
-  return $ afterBreak "\\&" <> text (escString opts str)
-inlineToMan opts (Str str) = return $ text $ escString opts str
+  withFontFeature 'C' (return (literal $ escString opts str))
+inlineToMan opts (Str str@(T.uncons -> Just ('.',_))) =
+  return $ afterBreak "\\&" <> literal (escString opts str)
+inlineToMan opts (Str str) = return $ literal $ escString opts str
 inlineToMan opts (Math InlineMath str) =
   lift (texMathToInlines InlineMath str) >>= inlineListToMan opts
 inlineToMan opts (Math DisplayMath str) = do
   contents <- lift (texMathToInlines DisplayMath str) >>= inlineListToMan opts
-  return $ cr <> text ".RS" $$ contents $$ text ".RE"
+  return $ cr <> literal ".RS" $$ contents $$ literal ".RE"
 inlineToMan _ il@(RawInline f str)
-  | f == Format "man" = return $ text str
+  | f == Format "man" = return $ literal str
   | otherwise         = do
       report $ InlineNotRendered il
       return empty
 inlineToMan _ LineBreak = return $
-  cr <> text ".PD 0" $$ text ".P" $$ text ".PD" <> cr
+  cr <> literal ".PD 0" $$ literal ".P" $$ literal ".PD" <> cr
 inlineToMan _ SoftBreak = return space
 inlineToMan _ Space = return space
-inlineToMan opts (Link _ txt ('#':_, _)) =
-  inlineListToMan opts txt -- skip internal links
-inlineToMan opts (Link _ txt (src, _)) = do
+inlineToMan opts (Link _ txt (src, _))
+  | not (isURI src) = inlineListToMan opts txt -- skip relative links
+  | otherwise       = do
   linktext <- inlineListToMan opts txt
-  let srcSuffix = fromMaybe src (stripPrefix "mailto:" src)
+  let srcSuffix = fromMaybe src (T.stripPrefix "mailto:" src)
   return $ case txt of
            [Str s]
              | escapeURI s == srcSuffix ->
-                                 char '<' <> text srcSuffix <> char '>'
-           _                  -> linktext <> text " (" <> text src <> char ')'
+                                 char '<' <> literal srcSuffix <> char '>'
+           _                  -> linktext <> literal " (" <> literal src <> char ')'
 inlineToMan opts (Image attr alternate (source, tit)) = do
   let txt = if null alternate || (alternate == [Str ""]) ||
                (alternate == [Str source]) -- to prevent autolinks
                then [Str "image"]
                else alternate
   linkPart <- inlineToMan opts (Link attr txt (source, tit))
-  return $ char '[' <> text "IMAGE: " <> linkPart <> char ']'
+  return $ char '[' <> literal "IMAGE: " <> linkPart <> char ']'
 inlineToMan _ (Note contents) = do
   -- add to notes in state
   modify $ \st -> st{ stNotes = contents : stNotes st }
   notes <- gets stNotes
-  let ref = show (length notes)
-  return $ char '[' <> text ref <> char ']'
+  let ref = tshow (length notes)
+  return $ char '[' <> literal ref <> char ']'
