@@ -1,26 +1,8 @@
-{-# LANGUAGE NoImplicitPrelude #-}
 {-# LANGUAGE OverloadedStrings #-}
-{-
-Copyright (C) 2017-2018 Alexander Krotov <ilabdsf@gmail.com>
-
-This program is free software; you can redistribute it and/or modify
-it under the terms of the GNU General Public License as published by
-the Free Software Foundation; either version 2 of the License, or
-(at your option) any later version.
-
-This program is distributed in the hope that it will be useful,
-but WITHOUT ANY WARRANTY; without even the implied warranty of
-MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-GNU General Public License for more details.
-
-You should have received a copy of the GNU General Public License
-along with this program; if not, write to the Free Software
-Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--}
-
+{-# LANGUAGE ViewPatterns      #-}
 {- |
    Module      : Text.Pandoc.Writers.Muse
-   Copyright   : Copyright (C) 2017-2018 Alexander Krotov
+   Copyright   : Copyright (C) 2017-2020 Alexander Krotov
    License     : GNU GPL, version 2 or above
 
    Maintainer  : Alexander Krotov <ilabdsf@gmail.com>
@@ -43,26 +25,26 @@ However, @\<literal style="html">@ tag is used for HTML raw blocks
 even though it is supported only in Emacs Muse.
 -}
 module Text.Pandoc.Writers.Muse (writeMuse) where
-import Prelude
+import Control.Monad.Except (throwError)
 import Control.Monad.Reader
 import Control.Monad.State.Strict
 import Data.Char (isAlphaNum, isAsciiLower, isAsciiUpper, isDigit, isSpace)
 import Data.Default
-import Data.List (intersperse, isInfixOf, transpose)
-import Data.Monoid (Any (..))
+import Data.List (intersperse, transpose)
 import qualified Data.Set as Set
+import qualified Data.Text as T
 import Data.Text (Text)
 import System.FilePath (takeExtension)
-import Text.Pandoc.Class (PandocMonad)
+import Text.Pandoc.Class.PandocMonad (PandocMonad)
 import Text.Pandoc.Definition
+import Text.Pandoc.Error
 import Text.Pandoc.ImageSize
 import Text.Pandoc.Options
-import Text.Pandoc.Pretty
+import Text.DocLayout
 import Text.Pandoc.Shared
-import Text.Pandoc.Templates (renderTemplate')
+import Text.Pandoc.Templates (renderTemplate)
 import Text.Pandoc.Writers.Math
 import Text.Pandoc.Writers.Shared
-import Text.Pandoc.Walk
 
 type Notes = [[Block]]
 
@@ -83,7 +65,7 @@ data WriterEnv =
 data WriterState =
   WriterState { stNotes   :: Notes
               , stNoteNum :: Int
-              , stIds     :: Set.Set String
+              , stIds     :: Set.Set Text
               , stUseTags :: Bool -- ^ Use tags for emphasis, for example because previous character is a letter
               }
 
@@ -124,26 +106,25 @@ pandocToMuse (Pandoc meta blocks) = do
   let colwidth = if writerWrapText opts == WrapAuto
                     then Just $ writerColumns opts
                     else Nothing
-  let render' :: Doc -> Text
-      render' = render Nothing
-  metadata <- metaToJSON opts
-               (fmap render' . blockListToMuse)
-               (fmap render' . inlineListToMuse)
+  metadata <- metaToContext opts
+               blockListToMuse
+               (fmap chomp . inlineListToMuse)
                meta
   body <- blockListToMuse blocks
   notes <- currentNotesToMuse
-  let main = render colwidth $ body $+$ notes
+  let main = body $+$ notes
   let context = defField "body" main metadata
-  case writerTemplate opts of
-       Nothing  -> return main
-       Just tpl -> renderTemplate' tpl context
+  return $ render colwidth $
+    case writerTemplate opts of
+       Nothing  -> main
+       Just tpl -> renderTemplate tpl context
 
 -- | Helper function for flatBlockListToMuse
 -- | Render all blocks and insert blank lines between the first two
 catWithBlankLines :: PandocMonad m
                   => [Block]       -- ^ List of block elements
                   -> Int           -- ^ Number of blank lines
-                  -> Muse m Doc
+                  -> Muse m (Doc Text)
 catWithBlankLines (b : bs) n = do
   b' <- blockToMuseWithNotes b
   bs' <- flatBlockListToMuse bs
@@ -154,7 +135,7 @@ catWithBlankLines _ _ = error "Expected at least one block"
 -- | without setting envTopLevel.
 flatBlockListToMuse :: PandocMonad m
                 => [Block]       -- ^ List of block elements
-                -> Muse m Doc
+                -> Muse m (Doc Text)
 flatBlockListToMuse bs@(BulletList _ : BulletList _ : _) = catWithBlankLines bs 2
 flatBlockListToMuse bs@(OrderedList (_, style1, _) _ : OrderedList (_, style2, _) _ : _) =
   catWithBlankLines bs (if style1' == style2' then 2 else 0)
@@ -171,7 +152,7 @@ simpleTable :: PandocMonad m
             => [Inline]
             -> [TableCell]
             -> [[TableCell]]
-            -> Muse m Doc
+            -> Muse m (Doc Text)
 simpleTable caption headers rows = do
   topLevel <- asks envTopLevel
   caption' <- inlineListToMuse caption
@@ -179,7 +160,7 @@ simpleTable caption headers rows = do
   rows' <- mapM (mapM blockListToMuse) rows
   let widthsInChars = maximum . map offset <$> transpose (headers' : rows')
   let hpipeBlocks sep blocks = hcat $ intersperse sep' blocks
-        where sep' = lblock (length sep) $ text sep
+        where sep' = lblock (T.length sep) $ literal sep
   let makeRow sep = hpipeBlocks sep . zipWith lblock widthsInChars
   let head' = makeRow " || " headers'
   rows'' <- mapM (\row -> makeRow rowSeparator <$> mapM blockListToMuse row) rows
@@ -194,7 +175,7 @@ simpleTable caption headers rows = do
 -- | Convert list of Pandoc block elements to Muse.
 blockListToMuse :: PandocMonad m
                 => [Block]       -- ^ List of block elements
-                -> Muse m Doc
+                -> Muse m (Doc Text)
 blockListToMuse =
   local (\env -> env { envTopLevel = not (envInsideBlock env)
                      , envInsideBlock = True
@@ -203,19 +184,19 @@ blockListToMuse =
 -- | Convert Pandoc block element to Muse.
 blockToMuse :: PandocMonad m
             => Block         -- ^ Block element
-            -> Muse m Doc
+            -> Muse m (Doc Text)
 blockToMuse (Plain inlines) = inlineListToMuse' inlines
 blockToMuse (Para inlines) = do
   contents <- inlineListToMuse' inlines
   return $ contents <> blankline
 blockToMuse (LineBlock lns) = do
   lns' <- local (\env -> env { envOneLine = True }) $ mapM inlineListToMuse lns
-  return $ nowrap $ vcat (map (text "> " <>) lns') <> blankline
+  return $ nowrap $ vcat (map (literal "> " <>) lns') <> blankline
 blockToMuse (CodeBlock (_,_,_) str) =
-  return $ "<example>" $$ text str $$ "</example>" $$ blankline
+  return $ "<example>" $$ literal str $$ "</example>" $$ blankline
 blockToMuse (RawBlock (Format format) str) =
-  return $ blankline $$ "<literal style=\"" <> text format <> "\">" $$
-           text str $$ "</literal>" $$ blankline
+  return $ blankline $$ "<literal style=\"" <> literal format <> "\">" $$
+           literal str $$ "</literal>" $$ blankline
 blockToMuse (BlockQuote blocks) = do
   contents <- flatBlockListToMuse blocks
   return $ blankline
@@ -230,10 +211,10 @@ blockToMuse (OrderedList (start, style, _) items) = do
   topLevel <- asks envTopLevel
   return $ (if topLevel then nest 1 else id) (vcat contents) $$ blankline
   where orderedListItemToMuse :: PandocMonad m
-                              => String   -- ^ marker for list item
+                              => Text     -- ^ marker for list item
                               -> [Block]  -- ^ list item (list of blocks)
-                              -> Muse m Doc
-        orderedListItemToMuse marker item = hang (length marker + 1) (text marker <> space)
+                              -> Muse m (Doc Text)
+        orderedListItemToMuse marker item = hang (T.length marker + 1) (literal marker <> space)
           <$> blockListToMuse item
 blockToMuse (BulletList items) = do
   contents <- mapM bulletListItemToMuse items
@@ -241,7 +222,7 @@ blockToMuse (BulletList items) = do
   return $ (if topLevel then nest 1 else id) (vcat contents) $$ blankline
   where bulletListItemToMuse :: PandocMonad m
                              => [Block]
-                             -> Muse m Doc
+                             -> Muse m (Doc Text)
         bulletListItemToMuse item = do
           modify $ \st -> st { stUseTags = False }
           hang 2 "- " <$> blockListToMuse item
@@ -251,16 +232,17 @@ blockToMuse (DefinitionList items) = do
   return $ (if topLevel then nest 1 else id) (vcat contents) $$ blankline
   where definitionListItemToMuse :: PandocMonad m
                                  => ([Inline], [[Block]])
-                                 -> Muse m Doc
+                                 -> Muse m (Doc Text)
         definitionListItemToMuse (label, defs) = do
           modify $ \st -> st { stUseTags = False }
           label' <- local (\env -> env { envOneLine = True, envAfterSpace = True }) $ inlineListToMuse' label
-          let ind = offset' label' -- using Text.Pandoc.Pretty.offset results in round trip failures
+          let ind = offset' label' -- using Text.DocLayout.offset results in round trip failures
           hang ind (nowrap label') . vcat <$> mapM descriptionToMuse defs
-          where offset' d = maximum (0: map length (lines $ render Nothing d))
+          where offset' d = maximum (0: map T.length
+                                         (T.lines $ render Nothing d))
         descriptionToMuse :: PandocMonad m
                           => [Block]
-                          -> Muse m Doc
+                          -> Muse m (Doc Text)
         descriptionToMuse desc = hang 4 " :: " <$> blockListToMuse desc
 blockToMuse (Header level (ident,_,_) inlines) = do
   opts <- asks envOptions
@@ -270,10 +252,10 @@ blockToMuse (Header level (ident,_,_) inlines) = do
   let autoId = uniqueIdent (writerExtensions opts) inlines ids
   modify $ \st -> st{ stIds = Set.insert autoId ids }
 
-  let attr' = if null ident || (isEnabled Ext_auto_identifiers opts && ident == autoId)
+  let attr' = if T.null ident || (isEnabled Ext_auto_identifiers opts && ident == autoId)
                  then empty
-                 else "#" <> text ident <> cr
-  let header' = if topLevel then text (replicate level '*') <> space else mempty
+                 else "#" <> literal ident <> cr
+  let header' = if topLevel then literal (T.replicate level "*") <> space else mempty
   return $ blankline <> attr' $$ nowrap (header' <> contents) <> blankline
 -- https://www.gnu.org/software/emacs-muse/manual/muse.html#Horizontal-Rules-and-Anchors
 blockToMuse HorizontalRule = return $ blankline $$ "----" $$ blankline
@@ -287,21 +269,13 @@ blockToMuse (Table caption aligns widths headers rows) =
     blocksToDoc opts blocks =
       local (\env -> env { envOptions = opts }) $ blockListToMuse blocks
     numcols = maximum (length aligns : length widths : map length (headers:rows))
-    hasSimpleCells = all isSimpleCell (concat (headers:rows))
-    isLineBreak LineBreak = Any True
-    isLineBreak _         = Any False
-    hasLineBreak = getAny . query isLineBreak
-    isSimple = hasSimpleCells && all (== 0) widths
-    isSimpleCell [Plain ils] = not (hasLineBreak ils)
-    isSimpleCell [Para ils ] = not (hasLineBreak ils)
-    isSimpleCell []          = True
-    isSimpleCell _           = False
+    isSimple = onlySimpleTableCells (headers:rows) && all (== 0) widths
 blockToMuse (Div _ bs) = flatBlockListToMuse bs
 blockToMuse Null = return empty
 
 -- | Return Muse representation of notes collected so far.
 currentNotesToMuse :: PandocMonad m
-                   => Muse m Doc
+                   => Muse m (Doc Text)
 currentNotesToMuse = do
   notes <- reverse <$> gets stNotes
   modify $ \st -> st { stNotes = mempty }
@@ -310,7 +284,7 @@ currentNotesToMuse = do
 -- | Return Muse representation of notes.
 notesToMuse :: PandocMonad m
             => Notes
-            -> Muse m Doc
+            -> Muse m (Doc Text)
 notesToMuse notes = do
   n <- gets stNoteNum
   modify $ \st -> st { stNoteNum = stNoteNum st + length notes }
@@ -320,21 +294,21 @@ notesToMuse notes = do
 noteToMuse :: PandocMonad m
            => Int
            -> [Block]
-           -> Muse m Doc
+           -> Muse m (Doc Text)
 noteToMuse num note = do
-  res <- hang (length marker) (text marker) <$>
+  res <- hang (T.length marker) (literal marker) <$>
     local (\env -> env { envInsideBlock = True
                        , envInlineStart = True
                        , envAfterSpace = True
                        }) (blockListToMuse note)
   return $ res <> blankline
   where
-    marker = "[" ++ show num ++ "] "
+    marker = "[" <> tshow num <> "] "
 
 -- | Return Muse representation of block and accumulated notes.
 blockToMuseWithNotes :: PandocMonad m
                      => Block
-                     -> Muse m Doc
+                     -> Muse m (Doc Text)
 blockToMuseWithNotes blk = do
   topLevel <- asks envTopLevel
   opts <- asks envOptions
@@ -355,33 +329,29 @@ blockToMuseWithNotes blk = do
     else return b
 
 -- | Escape special characters for Muse.
-escapeString :: String -> String
-escapeString s =
-  "<verbatim>" ++
-  substitute "</verbatim>" "<</verbatim><verbatim>/verbatim>" s ++
+escapeText :: Text -> Text
+escapeText s =
+  "<verbatim>" <>
+  T.replace "</verbatim>" "<</verbatim><verbatim>/verbatim>" s <>
   "</verbatim>"
 
 -- | Replace newlines with spaces
-replaceNewlines :: String -> String
-replaceNewlines ('\n':xs) = ' ':replaceNewlines xs
-replaceNewlines (x:xs)    = x:replaceNewlines xs
-replaceNewlines []        = []
+replaceNewlines :: Text -> Text
+replaceNewlines = T.map $ \c ->
+  if c == '\n' then ' ' else c
 
-startsWithMarker :: (Char -> Bool) -> String -> Bool
-startsWithMarker f (' ':xs) = startsWithMarker f xs
-startsWithMarker f (x:xs) =
-  f x && (startsWithMarker f xs || startsWithDot xs)
+startsWithMarker :: (Char -> Bool) -> Text -> Bool
+startsWithMarker f t = case T.uncons $ T.dropWhile f' t of
+  Just ('.', xs) -> T.null xs || isSpace (T.head xs)
+  _              -> False
   where
-    startsWithDot ['.']     = True
-    startsWithDot ('.':c:_) = isSpace c
-    startsWithDot _         = False
-startsWithMarker _ [] = False
+    f' c = c == ' ' || f c
 
--- | Escape special characters for Muse if needed.
-containsFootnotes :: String -> Bool
-containsFootnotes = p
-  where p ('[':xs) = q xs || p xs
-        p (_:xs)   = p xs
+containsNotes :: Char -> Char -> Text -> Bool
+containsNotes left right = p . T.unpack -- This ought to be a parser
+  where p (left':xs)
+          | left' == left = q xs || p xs
+          | otherwise = p xs
         p ""       = False
         q (x:xs)
           | x `elem` ("123456789"::String) = r xs || p xs
@@ -389,31 +359,35 @@ containsFootnotes = p
         q [] = False
         r ('0':xs) = r xs || p xs
         r xs       = s xs || q xs || p xs
-        s (']':_) = True
-        s (_:xs)  = p xs
+        s (right':xs)
+          | right' == right = True
+          | otherwise = p xs
         s []      = False
 
 -- | Return True if string should be escaped with <verbatim> tags
-shouldEscapeString :: PandocMonad m
-                   => String
+shouldEscapeText :: PandocMonad m
+                   => Text
                    -> Muse m Bool
-shouldEscapeString s = do
+shouldEscapeText s = do
   insideLink <- asks envInsideLinkDescription
-  return $ null s ||
-           any (`elem` ("#*<=|" :: String)) s ||
-           "::" `isInfixOf` s ||
-           "~~" `isInfixOf` s ||
-           "[[" `isInfixOf` s ||
-           ("]" `isInfixOf` s && insideLink) ||
-           containsFootnotes s
+  return $ T.null s ||
+           T.any (`elem` ("#*<=|" :: String)) s ||
+           "::" `T.isInfixOf` s ||
+           "~~" `T.isInfixOf` s ||
+           "[[" `T.isInfixOf` s ||
+           ">>>" `T.isInfixOf` s ||
+           ("]" `T.isInfixOf` s && insideLink) ||
+           containsNotes '[' ']' s ||
+           containsNotes '{' '}' s
 
-conditionalEscapeString :: PandocMonad m
-                        => String
-                        -> Muse m String
-conditionalEscapeString s = do
-  shouldEscape <- shouldEscapeString s
+-- | Escape special characters for Muse if needed.
+conditionalEscapeText :: PandocMonad m
+                        => Text
+                        -> Muse m Text
+conditionalEscapeText s = do
+  shouldEscape <- shouldEscapeText s
   return $ if shouldEscape
-             then escapeString s
+             then escapeText s
              else s
 
 -- Expand Math and Cite before normalizing inline list
@@ -437,7 +411,7 @@ removeKeyValues :: Inline -> Inline
 removeKeyValues (Code (i, cls, _) xs) = Code (i, cls, []) xs
 -- Do not remove attributes from Link
 -- Do not remove attributes, such as "width", from Image
-removeKeyValues (Span (i, cls, _) xs) = Span (i, cls, []) xs
+-- Do not remove attributes, such as "dir", from Span
 removeKeyValues x                     = x
 
 normalizeInlineList :: [Inline] -> [Inline]
@@ -446,23 +420,23 @@ normalizeInlineList (Str "" : xs)
 normalizeInlineList (x : Str "" : xs)
   = normalizeInlineList (x:xs)
 normalizeInlineList (Str x1 : Str x2 : xs)
-  = normalizeInlineList $ Str (x1 ++ x2) : xs
+  = normalizeInlineList $ Str (x1 <> x2) : xs
 normalizeInlineList (Emph x1 : Emph x2 : ils)
-  = normalizeInlineList $ Emph (x1 ++ x2) : ils
+  = normalizeInlineList $ Emph (x1 <> x2) : ils
 normalizeInlineList (Strong x1 : Strong x2 : ils)
-  = normalizeInlineList $ Strong (x1 ++ x2) : ils
+  = normalizeInlineList $ Strong (x1 <> x2) : ils
 normalizeInlineList (Strikeout x1 : Strikeout x2 : ils)
-  = normalizeInlineList $ Strikeout (x1 ++ x2) : ils
+  = normalizeInlineList $ Strikeout (x1 <> x2) : ils
 normalizeInlineList (Superscript x1 : Superscript x2 : ils)
-  = normalizeInlineList $ Superscript (x1 ++ x2) : ils
+  = normalizeInlineList $ Superscript (x1 <> x2) : ils
 normalizeInlineList (Subscript x1 : Subscript x2 : ils)
-  = normalizeInlineList $ Subscript (x1 ++ x2) : ils
+  = normalizeInlineList $ Subscript (x1 <> x2) : ils
 normalizeInlineList (SmallCaps x1 : SmallCaps x2 : ils)
-  = normalizeInlineList $ SmallCaps (x1 ++ x2) : ils
+  = normalizeInlineList $ SmallCaps (x1 <> x2) : ils
 normalizeInlineList (Code _ x1 : Code _ x2 : ils)
-  = normalizeInlineList $ Code nullAttr (x1 ++ x2) : ils
+  = normalizeInlineList $ Code nullAttr (x1 <> x2) : ils
 normalizeInlineList (RawInline f1 x1 : RawInline f2 x2 : ils) | f1 == f2
-  = normalizeInlineList $ RawInline f1 (x1 ++ x2) : ils
+  = normalizeInlineList $ RawInline f1 (x1 <> x2) : ils
 -- Do not join Span's during normalization
 normalizeInlineList (x:xs) = x : normalizeInlineList xs
 normalizeInlineList [] = []
@@ -482,33 +456,41 @@ startsWithSpace _             = False
 endsWithSpace :: [Inline] -> Bool
 endsWithSpace [Space]     = True
 endsWithSpace [SoftBreak] = True
-endsWithSpace [Str s]     = stringStartsWithSpace $ reverse s
+endsWithSpace [Str s]     = stringEndsWithSpace s
 endsWithSpace (_:xs)      = endsWithSpace xs
 endsWithSpace []          = False
 
-urlEscapeBrackets :: String -> String
-urlEscapeBrackets (']':xs) = '%':'5':'D':urlEscapeBrackets xs
-urlEscapeBrackets (x:xs)   = x:urlEscapeBrackets xs
-urlEscapeBrackets []       = []
+urlEscapeBrackets :: Text -> Text
+urlEscapeBrackets = T.concatMap $ \c -> case c of
+  ']' -> "%5D"
+  _   -> T.singleton c
 
-isHorizontalRule :: String -> Bool
-isHorizontalRule s = length s >= 4 && all (== '-') s
+isHorizontalRule :: Text -> Bool
+isHorizontalRule s = T.length s >= 4 && T.all (== '-') s
 
-stringStartsWithSpace :: String -> Bool
-stringStartsWithSpace (x:_) = isSpace x
-stringStartsWithSpace ""    = False
+stringStartsWithSpace :: Text -> Bool
+stringStartsWithSpace = maybe False (isSpace . fst) . T.uncons
+
+stringEndsWithSpace :: Text -> Bool
+stringEndsWithSpace = maybe False (isSpace . snd) . T.unsnoc
 
 fixOrEscape :: Bool -> Inline -> Bool
-fixOrEscape sp (Str "-") = sp
-fixOrEscape sp (Str s@('-':x:_)) = (sp && isSpace x) || isHorizontalRule s
-fixOrEscape sp (Str (";")) = not sp
-fixOrEscape sp (Str (';':x:_)) = not sp && isSpace x
-fixOrEscape _ (Str (">")) = True
-fixOrEscape _ (Str ('>':x:_)) = isSpace x
-fixOrEscape sp (Str s) = (sp && (startsWithMarker isDigit s ||
-                                startsWithMarker isAsciiLower s ||
-                                startsWithMarker isAsciiUpper s))
-                         || stringStartsWithSpace s
+fixOrEscape b (Str s) = fixOrEscapeStr b s
+  where
+    fixOrEscapeStr sp t = case T.uncons t of
+      Just ('-', xs)
+        | T.null xs -> sp
+        | otherwise -> (sp && isSpace (T.head xs)) || isHorizontalRule t
+      Just (';', xs)
+        | T.null xs -> not sp
+        | otherwise -> not sp && isSpace (T.head xs)
+      Just ('>', xs)
+        | T.null xs -> True
+        | otherwise -> isSpace (T.head xs)
+      _             -> (sp && (startsWithMarker isDigit s ||
+                               startsWithMarker isAsciiLower s ||
+                               startsWithMarker isAsciiUpper s))
+                       || stringStartsWithSpace s
 fixOrEscape _ Space = True
 fixOrEscape _ SoftBreak = True
 fixOrEscape _ _ = False
@@ -517,17 +499,15 @@ inlineListStartsWithAlnum :: PandocMonad m
                           => [Inline]
                           -> Muse m Bool
 inlineListStartsWithAlnum (Str s:_) = do
-  esc <- shouldEscapeString s
-  return $ esc || isAlphaNum (head s)
+  esc <- shouldEscapeText s
+  return $ esc || isAlphaNum (T.head s)
 inlineListStartsWithAlnum _ = return False
 
 -- | Convert list of Pandoc inline elements to Muse
 renderInlineList :: PandocMonad m
                  => [Inline]
-                 -> Muse m Doc
-renderInlineList [] = do
-  start <- asks envInlineStart
-  pure $ if start then "<verbatim></verbatim>" else ""
+                 -> Muse m (Doc Text)
+renderInlineList [] = pure ""
 renderInlineList (x:xs) = do
   start <- asks envInlineStart
   afterSpace <- asks envAfterSpace
@@ -550,20 +530,23 @@ renderInlineList (x:xs) = do
                              , envNearAsterisks = False
                              }) $ renderInlineList xs
   if start && fixOrEscape afterSpace x
-    then pure (text "<verbatim></verbatim>" <> r <> lst')
+    then pure (literal "<verbatim></verbatim>" <> r <> lst')
     else pure (r <> lst')
 
 -- | Normalize and convert list of Pandoc inline elements to Muse.
 inlineListToMuse :: PandocMonad m
                  => [Inline]
-                 -> Muse m Doc
+                 -> Muse m (Doc Text)
 inlineListToMuse lst = do
   lst' <- normalizeInlineList . fixNotes <$> preprocessInlineList (map (removeKeyValues . replaceSmallCaps) lst)
   insideAsterisks <- asks envInsideAsterisks
+  start <- asks envInlineStart
   modify $ \st -> st { stUseTags = False } -- Previous character is likely a '>' or some other markup
-  local (\env -> env { envNearAsterisks = insideAsterisks }) $ renderInlineList lst'
+  if start && null lst'
+    then pure "<verbatim></verbatim>"
+    else local (\env -> env { envNearAsterisks = insideAsterisks }) $ renderInlineList lst'
 
-inlineListToMuse' :: PandocMonad m => [Inline] -> Muse m Doc
+inlineListToMuse' :: PandocMonad m => [Inline] -> Muse m (Doc Text)
 inlineListToMuse' lst = do
   topLevel <- asks envTopLevel
   afterSpace <- asks envAfterSpace
@@ -571,23 +554,23 @@ inlineListToMuse' lst = do
                      , envAfterSpace = afterSpace || not topLevel
                      }) $ inlineListToMuse lst
 
-emphasis :: PandocMonad m => String -> String -> [Inline] -> Muse m Doc
+emphasis :: PandocMonad m => Text -> Text -> [Inline] -> Muse m (Doc Text)
 emphasis b e lst = do
   contents <- local (\env -> env { envInsideAsterisks = inAsterisks }) $ inlineListToMuse lst
   modify $ \st -> st { stUseTags = useTags }
-  return $ text b <> contents <> text e
-  where inAsterisks = last b == '*' || head e == '*'
-        useTags = last e /= '>'
+  return $ literal b <> contents <> literal e
+  where inAsterisks = T.last b == '*' || T.head e == '*'
+        useTags = T.last e /= '>'
 
 -- | Convert Pandoc inline element to Muse.
 inlineToMuse :: PandocMonad m
              => Inline
-             -> Muse m Doc
+             -> Muse m (Doc Text)
 inlineToMuse (Str str) = do
-  escapedStr <- conditionalEscapeString $ replaceNewlines str
-  let useTags = isAlphaNum $ last escapedStr -- escapedStr is never empty because empty strings are escaped
+  escapedStr <- conditionalEscapeText $ replaceNewlines str
+  let useTags = isAlphaNum $ T.last escapedStr -- escapedStr is never empty because empty strings are escaped
   modify $ \st -> st { stUseTags = useTags }
-  return $ text escapedStr
+  return $ literal escapedStr
 inlineToMuse (Emph [Strong lst]) = do
   useTags <- gets stUseTags
   let lst' = normalizeInlineList lst
@@ -629,7 +612,8 @@ inlineToMuse (Subscript lst) = do
   modify $ \st -> st { stUseTags = False }
   return $ "<sub>" <> contents <> "</sub>"
 inlineToMuse SmallCaps {} =
-  fail "SmallCaps should be expanded before normalization"
+  throwError $ PandocShouldNeverHappenError
+    "SmallCaps should be expanded before normalization"
 inlineToMuse (Quoted SingleQuote lst) = do
   contents <- inlineListToMuse lst
   modify $ \st -> st { stUseTags = False }
@@ -639,18 +623,21 @@ inlineToMuse (Quoted DoubleQuote lst) = do
   modify $ \st -> st { stUseTags = False }
   return $ "“" <> contents <> "”"
 inlineToMuse Cite {} =
-  fail "Citations should be expanded before normalization"
+  throwError $ PandocShouldNeverHappenError
+               "Citations should be expanded before normalization"
 inlineToMuse (Code _ str) = do
   useTags <- gets stUseTags
   modify $ \st -> st { stUseTags = False }
-  return $ if useTags || null str || '=' `elem` str || isSpace (head str) || isSpace (last str)
-             then "<code>" <> text (substitute "</code>" "<</code><code>/code>" str) <> "</code>"
-             else "=" <> text str <> "="
+  return $ if useTags || T.null str || T.any (== '=') str
+              || isSpace (T.head str) || isSpace (T.last str)
+             then "<code>" <> literal (T.replace "</code>" "<</code><code>/code>" str) <> "</code>"
+             else "=" <> literal str <> "="
 inlineToMuse Math{} =
-  fail "Math should be expanded before normalization"
+  throwError $ PandocShouldNeverHappenError
+    "Math should be expanded before normalization"
 inlineToMuse (RawInline (Format f) str) = do
   modify $ \st -> st { stUseTags = False }
-  return $ "<literal style=\"" <> text f <> "\">" <> text str <> "</literal>"
+  return $ "<literal style=\"" <> literal f <> "\">" <> literal str <> "</literal>"
 inlineToMuse LineBreak = do
   oneline <- asks envOneLine
   modify $ \st -> st { stUseTags = False }
@@ -667,27 +654,27 @@ inlineToMuse (Link _ txt (src, _)) =
   case txt of
         [Str x] | escapeURI x == src -> do
              modify $ \st -> st { stUseTags = False }
-             return $ "[[" <> text (escapeLink x) <> "]]"
+             return $ "[[" <> literal (escapeLink x) <> "]]"
         _ -> do contents <- local (\env -> env { envInsideLinkDescription = True }) $ inlineListToMuse txt
                 modify $ \st -> st { stUseTags = False }
-                return $ "[[" <> text (escapeLink src) <> "][" <> contents <> "]]"
-  where escapeLink lnk = if isImageUrl lnk then "URL:" ++ urlEscapeBrackets lnk else urlEscapeBrackets lnk
+                return $ "[[" <> literal (escapeLink src) <> "][" <> contents <> "]]"
+  where escapeLink lnk = if isImageUrl lnk then "URL:" <> urlEscapeBrackets lnk else urlEscapeBrackets lnk
         -- Taken from muse-image-regexp defined in Emacs Muse file lisp/muse-regexps.el
         imageExtensions = [".eps", ".gif", ".jpg", ".jpeg", ".pbm", ".png", ".tiff", ".xbm", ".xpm"]
-        isImageUrl = (`elem` imageExtensions) . takeExtension
-inlineToMuse (Image attr alt (source,'f':'i':'g':':':title)) =
+        isImageUrl = (`elem` imageExtensions) . takeExtension . T.unpack
+inlineToMuse (Image attr alt (source,T.stripPrefix "fig:" -> Just title)) =
   inlineToMuse (Image attr alt (source,title))
 inlineToMuse (Image attr@(_, classes, _) inlines (source, title)) = do
   opts <- asks envOptions
   alt <- local (\env -> env { envInsideLinkDescription = True }) $ inlineListToMuse inlines
-  title' <- if null title
+  title' <- if T.null title
             then if null inlines
                  then return ""
                  else return $ "[" <> alt <> "]"
-            else do s <- local (\env -> env { envInsideLinkDescription = True }) $ conditionalEscapeString title
-                    return $ "[" <> text s <> "]"
+            else do s <- local (\env -> env { envInsideLinkDescription = True }) $ conditionalEscapeText title
+                    return $ "[" <> literal s <> "]"
   let width = case dimension Width attr of
-                Just (Percent x) | isEnabled Ext_amuse opts -> " " ++ show (round x :: Integer)
+                Just (Percent x) | isEnabled Ext_amuse opts -> " " <> tshow (round x :: Integer)
                 _ -> ""
   let leftalign = if "align-left" `elem` classes
                   then " l"
@@ -696,7 +683,7 @@ inlineToMuse (Image attr@(_, classes, _) inlines (source, title)) = do
                    then " r"
                    else ""
   modify $ \st -> st { stUseTags = False }
-  return $ "[[" <> text (urlEscapeBrackets source ++ width ++ leftalign ++ rightalign) <> "]" <> title' <> "]"
+  return $ "[[" <> literal (urlEscapeBrackets source <> width <> leftalign <> rightalign) <> "]" <> title' <> "]"
 inlineToMuse (Note contents) = do
   -- add to notes in state
   notes <- gets stNotes
@@ -704,16 +691,19 @@ inlineToMuse (Note contents) = do
                      , stUseTags = False
                      }
   n <- gets stNoteNum
-  let ref = show $ n + length notes
-  return $ "[" <> text ref <> "]"
-inlineToMuse (Span (anchor,names,_) inlines) = do
+  let ref = tshow $ n + length notes
+  return $ "[" <> literal ref <> "]"
+inlineToMuse (Span (anchor,names,kvs) inlines) = do
   contents <- inlineListToMuse inlines
-  let anchorDoc = if null anchor
+  let (contents', hasDir) = case lookup "dir" kvs of
+                              Just "rtl" -> ("<<<" <> contents <> ">>>", True)
+                              Just "ltr" -> (">>>" <> contents <> "<<<", True)
+                              _ -> (contents, False)
+  let anchorDoc = if T.null anchor
                      then mempty
-                     else text ('#':anchor) <> space
+                     else literal ("#" <> anchor) <> space
   modify $ \st -> st { stUseTags = False }
-  return $ anchorDoc <> (if null inlines && not (null anchor)
+  return $ anchorDoc <> (if null inlines && not (T.null anchor)
                          then mempty
-                         else (if null names
-                               then "<class>"
-                               else "<class name=\"" <> text (head names) <> "\">") <> contents <> "</class>")
+                         else (if null names then (if hasDir then contents' else "<class>" <> contents' <> "</class>")
+                               else "<class name=\"" <> literal (head names) <> "\">" <> contents' <> "</class>"))
