@@ -1,26 +1,8 @@
-{-# LANGUAGE NoImplicitPrelude #-}
 {-# LANGUAGE ScopedTypeVariables #-}
-{-
-Copyright (C) 2008-2018 John MacFarlane <jgm@berkeley.edu>
-
-This program is free software; you can redistribute it and/or modify
-it under the terms of the GNU General Public License as published by
-the Free Software Foundation; either version 2 of the License, or
-(at your option) any later version.
-
-This program is distributed in the hope that it will be useful,
-but WITHOUT ANY WARRANTY; without even the implied warranty of
-MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-GNU General Public License for more details.
-
-You should have received a copy of the GNU General Public License
-along with this program; if not, write to the Free Software
-Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--}
-
+{-# LANGUAGE OverloadedStrings   #-}
 {- |
    Module      : Text.Pandoc.Writers.ODT
-   Copyright   : Copyright (C) 2008-2018 John MacFarlane
+   Copyright   : Copyright (C) 2008-2020 John MacFarlane
    License     : GNU GPL, version 2 or above
 
    Maintainer  : John MacFarlane <jgm@berkeley.edu>
@@ -30,28 +12,29 @@ Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 Conversion of 'Pandoc' documents to ODT.
 -}
 module Text.Pandoc.Writers.ODT ( writeODT ) where
-import Prelude
 import Codec.Archive.Zip
 import Control.Monad.Except (catchError)
 import Control.Monad.State.Strict
 import qualified Data.ByteString.Lazy as B
 import Data.Generics (everywhere', mkT)
-import Data.List (isPrefixOf, intercalate)
-import Data.Maybe (fromMaybe)
+import Data.List (isPrefixOf)
 import qualified Data.Map as Map
+import qualified Data.Text as T
 import qualified Data.Text.Lazy as TL
+import Data.Time
 import System.FilePath (takeDirectory, takeExtension, (<.>))
 import Text.Pandoc.BCP47 (Lang (..), getLang, renderLang)
-import Text.Pandoc.Class (PandocMonad, report, toLang)
-import qualified Text.Pandoc.Class as P
+import Text.Pandoc.Class.PandocMonad (PandocMonad, report, toLang)
+import qualified Text.Pandoc.Class.PandocMonad as P
 import Text.Pandoc.Definition
 import Text.Pandoc.ImageSize
 import Text.Pandoc.Logging
 import Text.Pandoc.MIME (extensionFromMimeType, getMimeType)
 import Text.Pandoc.Options (WrapOption (..), WriterOptions (..))
-import Text.Pandoc.Pretty
-import Text.Pandoc.Shared (stringify, normalizeDate)
-import Text.Pandoc.Writers.Shared (lookupMetaString, fixDisplayMath)
+import Text.DocLayout
+import Text.Pandoc.Shared (stringify, pandocVersion, tshow)
+import Text.Pandoc.Writers.Shared (lookupMetaString, lookupMetaBlocks,
+                                   fixDisplayMath)
 import Text.Pandoc.UTF8 (fromStringLazy, fromTextLazy, toStringLazy)
 import Text.Pandoc.Walk
 import Text.Pandoc.Writers.OpenDocument (writeOpenDocument)
@@ -83,6 +66,7 @@ pandocToODT :: PandocMonad m
 pandocToODT opts doc@(Pandoc meta _) = do
   let title = docTitle meta
   let authors = docAuthors meta
+  utctime <- P.getCurrentTime
   lang <- toLang (getLang opts meta)
   refArchive <-
        case writerReferenceDoc opts of
@@ -104,7 +88,7 @@ pandocToODT opts doc@(Pandoc meta _) = do
                         Nothing  -> empty
                         Just m   -> selfClosingTag "manifest:file-entry"
                                      [("manifest:media-type", m)
-                                     ,("manifest:full-path", fp)
+                                     ,("manifest:full-path", T.pack fp)
                                      ,("manifest:version", "1.2")
                                      ]
   let files = [ ent | ent <- filesInArchive archive,
@@ -115,7 +99,7 @@ pandocToODT opts doc@(Pandoc meta _) = do
         $ fromStringLazy $ render Nothing
         $ text "<?xml version=\"1.0\" encoding=\"utf-8\"?>"
         $$
-         (inTags True "manifest:manifest"
+         inTags True "manifest:manifest"
             [("xmlns:manifest","urn:oasis:names:tc:opendocument:xmlns:manifest:1.0")
             ,("manifest:version","1.2")] ( selfClosingTag "manifest:file-entry"
                  [("manifest:media-type","application/vnd.oasis.opendocument.text")
@@ -123,46 +107,59 @@ pandocToODT opts doc@(Pandoc meta _) = do
                 $$ vcat ( map toFileEntry files )
                 $$ vcat ( map toFileEntry formulas )
               )
-         )
   let archive' = addEntryToArchive manifestEntry archive
+  -- create meta.xml
   let userDefinedMetaFields = [k | k <- Map.keys (unMeta meta)
-                              , k `notElem` ["title", "lang", "author", "date"]]
-  let escapedText = text . escapeStringForXML
+                              , k `notElem` ["title", "lang", "author"
+                                           , "description", "subject", "keywords"]]
+  let escapedText = text . T.unpack . escapeStringForXML
+  let keywords = case lookupMeta "keywords" meta of
+                      Just (MetaList xs) -> map stringify xs
+                      _                  -> []
   let userDefinedMeta =
         map (\k -> inTags False "meta:user-defined"
-              [ ("meta_name", escapeStringForXML k)
-              ,("meta-value-type", "string")
+              [ ("meta:name", escapeStringForXML k)
+              ,("meta:value-type", "string")
               ] (escapedText $ lookupMetaString k meta)) userDefinedMetaFields
   let metaTag metafield = inTagsSimple metafield . escapedText
   let metaEntry = toEntry "meta.xml" epochtime
        $ fromStringLazy $ render Nothing
        $ text "<?xml version=\"1.0\" encoding=\"utf-8\"?>"
        $$
-        (inTags True "office:document-meta"
+        inTags True "office:document-meta"
            [("xmlns:office","urn:oasis:names:tc:opendocument:xmlns:office:1.0")
            ,("xmlns:xlink","http://www.w3.org/1999/xlink")
            ,("xmlns:dc","http://purl.org/dc/elements/1.1/")
            ,("xmlns:meta","urn:oasis:names:tc:opendocument:xmlns:meta:1.0")
            ,("xmlns:ooo","http://openoffice.org/2004/office")
            ,("xmlns:grddl","http://www.w3.org/2003/g/data-view#")
-           ,("office:version","1.2")] ( inTags True "office:meta" [] $
-                 ( metaTag "dc:title" (stringify title)
+           ,("office:version","1.2")] ( inTags True "office:meta" []
+                 ( metaTag "meta:generator" ("Pandoc/" <> pandocVersion)
+                   $$
+                   metaTag "dc:title" (stringify title)
+                   $$
+                   metaTag "dc:description"
+                          (T.intercalate "\n" (map stringify $
+                                         lookupMetaBlocks "description" meta))
+                   $$
+                   metaTag "dc:subject" (lookupMetaString "subject" meta)
+                   $$
+                   metaTag "meta:keyword" (T.intercalate ", " keywords)
                    $$
                    case lang of
                         Just l  -> metaTag "dc:language" (renderLang l)
                         Nothing -> empty
                    $$
-                   metaTag "dc:creator"
-                     (intercalate "; " (map stringify authors))
-                   $$
-                   maybe mempty
-                     (metaTag "dc:date")
-                       (normalizeDate (lookupMetaString "date" meta))
+                   (\d a -> metaTag "meta:initial-creator" a
+                         $$ metaTag "dc:creator" a
+                         $$ metaTag "meta:creation-date" d
+                         $$ metaTag "dc:date" d
+                   ) (T.pack $ formatTime defaultTimeLocale "%FT%XZ" utctime)
+                     (T.intercalate "; " (map stringify authors))
                    $$
                    vcat userDefinedMeta
                  )
              )
-        )
   -- make sure mimetype is first
   let mimetypeEntry = toEntry "mimetype" epochtime
                       $ fromStringLazy "application/vnd.oasis.opendocument.text"
@@ -190,9 +187,9 @@ updateStyleWithLang (Just lang) arch = do
 addLang :: Lang -> Element -> Element
 addLang lang = everywhere' (mkT updateLangAttr)
     where updateLangAttr (Attr n@(QName "language" _ (Just "fo")) _)
-                           = Attr n (langLanguage lang)
+                           = Attr n (T.unpack $ langLanguage lang)
           updateLangAttr (Attr n@(QName "country" _ (Just "fo")) _)
-                           = Attr n (langRegion lang)
+                           = Attr n (T.unpack $ langRegion lang)
           updateLangAttr x = x
 
 -- | transform both Image and Math elements
@@ -206,12 +203,12 @@ transformPicMath opts (Image attr@(id', cls, _) lab (src,t)) = catchError
                          return (100, 100)
        let dims =
              case (getDim Width, getDim Height) of
-               (Just w, Just h)              -> [("width", show w), ("height", show h)]
-               (Just w@(Percent _), Nothing) -> [("rel-width", show w),("rel-height", "scale"),("width", show ptX ++ "pt"),("height", show ptY ++ "pt")]
-               (Nothing, Just h@(Percent _)) -> [("rel-width", "scale"),("rel-height", show h),("width", show ptX ++ "pt"),("height", show ptY ++ "pt")]
-               (Just w@(Inch i), Nothing)    -> [("width", show w), ("height", show (i / ratio) ++ "in")]
-               (Nothing, Just h@(Inch i))    -> [("width", show (i * ratio) ++ "in"), ("height", show h)]
-               _                             -> [("width", show ptX ++ "pt"), ("height", show ptY ++ "pt")]
+               (Just w, Just h)              -> [("width", tshow w), ("height", tshow h)]
+               (Just w@(Percent _), Nothing) -> [("rel-width", tshow w),("rel-height", "scale"),("width", tshow ptX <> "pt"),("height", tshow ptY <> "pt")]
+               (Nothing, Just h@(Percent _)) -> [("rel-width", "scale"),("rel-height", tshow h),("width", tshow ptX <> "pt"),("height", tshow ptY <> "pt")]
+               (Just w@(Inch i), Nothing)    -> [("width", tshow w), ("height", tshow (i / ratio) <> "in")]
+               (Nothing, Just h@(Inch i))    -> [("width", tshow (i * ratio) <> "in"), ("height", tshow h)]
+               _                             -> [("width", tshow ptX <> "pt"), ("height", tshow ptY <> "pt")]
              where
                ratio = ptX / ptY
                getDim dir = case dimension dir attr of
@@ -220,16 +217,16 @@ transformPicMath opts (Image attr@(id', cls, _) lab (src,t)) = catchError
                               Nothing          -> Nothing
        let  newattr = (id', cls, dims)
        entries <- gets stEntries
-       let extension = fromMaybe (takeExtension $ takeWhile (/='?') src)
+       let extension = maybe (takeExtension $ takeWhile (/='?') $ T.unpack src) T.unpack
                            (mbMimeType >>= extensionFromMimeType)
        let newsrc = "Pictures/" ++ show (length entries) <.> extension
        let toLazy = B.fromChunks . (:[])
        epochtime <- floor `fmap` lift P.getPOSIXTime
        let entry = toEntry newsrc epochtime $ toLazy img
        modify $ \st -> st{ stEntries = entry : entries }
-       return $ Image newattr lab (newsrc, t))
+       return $ Image newattr lab (T.pack newsrc, t))
    (\e -> do
-       report $ CouldNotFetchResource src (show e)
+       report $ CouldNotFetchResource src $ T.pack (show e)
        return $ Emph lab)
 
 transformPicMath _ (Math t math) = do
@@ -240,7 +237,7 @@ transformPicMath _ (Math t math) = do
        Right r -> do
          let conf = useShortEmptyTags (const False) defaultConfigPP
          let mathml = ppcTopElement conf r
-         epochtime <- floor `fmap` (lift P.getPOSIXTime)
+         epochtime <- floor `fmap` lift P.getPOSIXTime
          let dirname = "Formula-" ++ show (length entries) ++ "/"
          let fname = dirname ++ "content.xml"
          let entry = toEntry fname epochtime (fromStringLazy mathml)
@@ -257,7 +254,7 @@ transformPicMath _ (Math t math) = do
                                            ,("text:anchor-type","paragraph")]
                                       else [("draw:style-name","fr1")
                                            ,("text:anchor-type","as-char")]) $
-             selfClosingTag "draw:object" [("xlink:href", dirname)
+             selfClosingTag "draw:object" [("xlink:href", T.pack dirname)
                                         , ("xlink:type", "simple")
                                         , ("xlink:show", "embed")
                                         , ("xlink:actuate", "onLoad")]
@@ -268,12 +265,12 @@ documentSettings :: Bool -> B.ByteString
 documentSettings isTextMode = fromStringLazy $ render Nothing
     $ text "<?xml version=\"1.0\" encoding=\"utf-8\"?>"
     $$
-    (inTags True "office:document-settings"
+    inTags True "office:document-settings"
       [("xmlns:office","urn:oasis:names:tc:opendocument:xmlns:office:1.0")
       ,("xmlns:xlink","http://www.w3.org/1999/xlink")
       ,("xmlns:config","urn:oasis:names:tc:opendocument:xmlns:config:1.0")
       ,("xmlns:ooo","http://openoffice.org/2004/office")
-      ,("office:version","1.2")] $
+      ,("office:version","1.2")] (
        inTagsSimple "office:settings" $
          inTags False "config:config-item-set"
            [("config:name", "ooo:configuration-settings")] $
