@@ -1,29 +1,10 @@
-{-# LANGUAGE NoImplicitPrelude #-}
 {-# LANGUAGE FlexibleContexts  #-}
 {-# LANGUAGE OverloadedStrings #-}
 {-# LANGUAGE PatternGuards     #-}
-{-
-Copyright (C) 2008-2018 Andrea Rossato <andrea.rossato@ing.unitn.it>
-                        and John MacFarlane.
-
-This program is free software; you can redistribute it and/or modify
-it under the terms of the GNU General Public License as published by
-the Free Software Foundation; either version 2 of the License, or
-(at your option) any later version.
-
-This program is distributed in the hope that it will be useful,
-but WITHOUT ANY WARRANTY; without even the implied warranty of
-MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-GNU General Public License for more details.
-
-You should have received a copy of the GNU General Public License
-along with this program; if not, write to the Free Software
-Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--}
-
+{-# LANGUAGE ViewPatterns      #-}
 {- |
    Module      : Text.Pandoc.Writers.OpenDocument
-   Copyright   : Copyright (C) 2008-2018 Andrea Rossato and John MacFarlane
+   Copyright   : Copyright (C) 2008-2020 Andrea Rossato and John MacFarlane
    License     : GNU GPL, version 2 or above
 
    Maintainer  : Andrea Rossato <andrea.rossato@ing.unitn.it>
@@ -33,25 +14,25 @@ Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 Conversion of 'Pandoc' documents to OpenDocument XML.
 -}
 module Text.Pandoc.Writers.OpenDocument ( writeOpenDocument ) where
-import Prelude
 import Control.Arrow ((***), (>>>))
 import Control.Monad.State.Strict hiding (when)
 import Data.Char (chr)
-import Data.List (sortBy)
+import Data.List (sortOn, sortBy, foldl')
 import qualified Data.Map as Map
 import Data.Maybe (fromMaybe)
 import Data.Ord (comparing)
 import qualified Data.Set as Set
 import Data.Text (Text)
+import qualified Data.Text as T
 import Text.Pandoc.BCP47 (Lang (..), parseBCP47)
-import Text.Pandoc.Class (PandocMonad, report, translateTerm,
-                          setTranslations, toLang)
+import Text.Pandoc.Class.PandocMonad (PandocMonad, report, translateTerm,
+                                      setTranslations, toLang)
 import Text.Pandoc.Definition
 import Text.Pandoc.Logging
 import Text.Pandoc.Options
-import Text.Pandoc.Pretty
-import Text.Pandoc.Shared (linesToPara)
-import Text.Pandoc.Templates (renderTemplate')
+import Text.DocLayout
+import Text.Pandoc.Shared (linesToPara, tshow)
+import Text.Pandoc.Templates (renderTemplate)
 import qualified Text.Pandoc.Translations as Term (Term(Figure, Table))
 import Text.Pandoc.Writers.Math
 import Text.Pandoc.Writers.Shared
@@ -70,11 +51,12 @@ plainToPara x         = x
 type OD m = StateT WriterState m
 
 data WriterState =
-    WriterState { stNotes          :: [Doc]
-                , stTableStyles    :: [Doc]
-                , stParaStyles     :: [Doc]
-                , stListStyles     :: [(Int, [Doc])]
-                , stTextStyles     :: Map.Map (Set.Set TextStyle) (String, Doc)
+    WriterState { stNotes          :: [Doc Text]
+                , stTableStyles    :: [Doc Text]
+                , stParaStyles     :: [Doc Text]
+                , stListStyles     :: [(Int, [Doc Text])]
+                , stTextStyles     :: Map.Map (Set.Set TextStyle)
+                                        (Text, Doc Text)
                 , stTextStyleAttr  :: Set.Set TextStyle
                 , stIndentPara     :: Int
                 , stInDefinition   :: Bool
@@ -102,19 +84,20 @@ defaultWriterState =
                 , stImageCaptionId = 1
                 }
 
-when :: Bool -> Doc -> Doc
+when :: Bool -> Doc Text -> Doc Text
 when p a = if p then a else empty
 
-addTableStyle :: PandocMonad m => Doc -> OD m ()
+addTableStyle :: PandocMonad m => Doc Text -> OD m ()
 addTableStyle i = modify $ \s -> s { stTableStyles = i : stTableStyles s }
 
-addNote :: PandocMonad m => Doc -> OD m ()
+addNote :: PandocMonad m => Doc Text -> OD m ()
 addNote i = modify $ \s -> s { stNotes = i : stNotes s }
 
-addParaStyle :: PandocMonad m => Doc -> OD m ()
+addParaStyle :: PandocMonad m => Doc Text -> OD m ()
 addParaStyle i = modify $ \s -> s { stParaStyles = i : stParaStyles s }
 
-addTextStyle :: PandocMonad m => Set.Set TextStyle -> (String, Doc) -> OD m ()
+addTextStyle :: PandocMonad m
+             => Set.Set TextStyle -> (Text, Doc Text) -> OD m ()
 addTextStyle attrs i = modify $ \s ->
   s { stTextStyles = Map.insert attrs i (stTextStyles s) }
 
@@ -138,7 +121,7 @@ setInDefinitionList b = modify $  \s -> s { stInDefinition = b }
 setFirstPara :: PandocMonad m => OD m ()
 setFirstPara =  modify $  \s -> s { stFirstPara = True }
 
-inParagraphTags :: PandocMonad m => Doc -> OD m Doc
+inParagraphTags :: PandocMonad m => Doc Text -> OD m (Doc Text)
 inParagraphTags d = do
   b <- gets stFirstPara
   a <- if b
@@ -147,10 +130,10 @@ inParagraphTags d = do
        else    return   [("text:style-name", "Text_20_body")]
   return $ inTags False "text:p" a d
 
-inParagraphTagsWithStyle :: String -> Doc -> Doc
+inParagraphTagsWithStyle :: Text -> Doc Text -> Doc Text
 inParagraphTagsWithStyle sty = inTags False "text:p" [("text:style-name", sty)]
 
-inSpanTags :: String -> Doc -> Doc
+inSpanTags :: Text -> Doc Text -> Doc Text
 inSpanTags s = inTags False "text:span" [("text:style-name",s)]
 
 withTextStyle :: PandocMonad m => TextStyle -> OD m a -> OD m a
@@ -161,7 +144,7 @@ withTextStyle s f = do
   modify $ \st -> st{ stTextStyleAttr = oldTextStyleAttr }
   return res
 
-inTextStyle :: PandocMonad m => Doc -> OD m Doc
+inTextStyle :: PandocMonad m => Doc Text -> OD m (Doc Text)
 inTextStyle d = do
   at <- gets stTextStyleAttr
   if Set.null at
@@ -172,20 +155,21 @@ inTextStyle d = do
             Just (styleName, _) -> return $
               inTags False "text:span" [("text:style-name",styleName)] d
             Nothing -> do
-              let styleName = "T" ++ show (Map.size styles + 1)
+              let styleName = "T" <> tshow (Map.size styles + 1)
               addTextStyle at (styleName,
                      inTags False "style:style"
                        [("style:name", styleName)
                        ,("style:family", "text")]
                        $ selfClosingTag "style:text-properties"
-                          (concatMap textStyleAttr (Set.toList at)))
+                          (sortOn fst . Map.toList
+                                $ foldl' textStyleAttr mempty (Set.toList at)))
               return $ inTags False
                   "text:span" [("text:style-name",styleName)] d
 
-formulaStyles :: [Doc]
+formulaStyles :: [Doc Text]
 formulaStyles = [formulaStyle InlineMath, formulaStyle DisplayMath]
 
-formulaStyle :: MathType -> Doc
+formulaStyle :: MathType -> Doc Text
 formulaStyle mt = inTags False "style:style"
   [("style:name", if mt == InlineMath then "fr1" else "fr2")
   ,("style:family", "graphic")
@@ -200,32 +184,33 @@ formulaStyle mt = inTags False "style:style"
                                                   ,("style:horizontal-rel", "paragraph-content")
                                                   ,("style:wrap",           "none")]
 
-inHeaderTags :: PandocMonad m => Int -> String -> Doc -> OD m Doc
+inHeaderTags :: PandocMonad m => Int -> Text -> Doc Text -> OD m (Doc Text)
 inHeaderTags i ident d =
-  return $ inTags False "text:h" [ ("text:style-name", "Heading_20_" ++ show i)
-                                 , ("text:outline-level", show i)]
-         $ if null ident
+  return $ inTags False "text:h" [ ("text:style-name", "Heading_20_" <> tshow i)
+                                 , ("text:outline-level", tshow i)]
+         $ if T.null ident
               then d
               else selfClosingTag "text:bookmark-start" [ ("text:name", ident) ]
                    <> d <>
                    selfClosingTag "text:bookmark-end" [ ("text:name", ident) ]
 
-inQuotes :: QuoteType -> Doc -> Doc
+inQuotes :: QuoteType -> Doc Text -> Doc Text
 inQuotes SingleQuote s = char '\8216' <> s <> char '\8217'
 inQuotes DoubleQuote s = char '\8220' <> s <> char '\8221'
 
-handleSpaces :: String -> Doc
-handleSpaces s
-    | ( ' ':_) <- s = genTag s
-    | ('\t':x) <- s = selfClosingTag "text:tab" [] <> rm x
-    | otherwise     = rm s
-    where
-        genTag       = span (==' ') >>> tag . length *** rm >>> uncurry (<>)
-        tag n        = when (n /= 0) $ selfClosingTag "text:s" [("text:c", show n)]
-        rm ( ' ':xs) = char ' ' <> genTag xs
-        rm ('\t':xs) = selfClosingTag "text:tab" [] <> genTag xs
-        rm (   x:xs) = char  x  <> rm xs
-        rm        [] = empty
+handleSpaces :: Text -> Doc Text
+handleSpaces s = case T.uncons s of
+  Just (' ', _) -> genTag s
+  Just ('\t',x) -> selfClosingTag "text:tab" [] <> rm x
+  _             -> rm s
+  where
+    genTag = T.span (==' ') >>> tag . T.length *** rm >>> uncurry (<>)
+    tag n  = when (n /= 0) $ selfClosingTag "text:s" [("text:c", tshow n)]
+    rm t   = case T.uncons t of
+      Just ( ' ',xs) -> char ' ' <> genTag xs
+      Just ('\t',xs) -> selfClosingTag "text:tab" [] <> genTag xs
+      Just (   x,xs) -> char x <> rm xs
+      Nothing        -> empty
 
 -- | Convert Pandoc document to string in OpenDocument format.
 writeOpenDocument :: PandocMonad m => WriterOptions -> Pandoc -> m Text
@@ -238,54 +223,54 @@ writeOpenDocument opts (Pandoc meta blocks) = do
   let colwidth = if writerWrapText opts == WrapAuto
                     then Just $ writerColumns opts
                     else Nothing
-  let render' :: Doc -> Text
-      render' = render colwidth
   ((body, metadata),s) <- flip runStateT
         defaultWriterState $ do
-           m <- metaToJSON opts
-                  (fmap render' . blocksToOpenDocument opts)
-                  (fmap render' . inlinesToOpenDocument opts)
+           m <- metaToContext opts
+                  (blocksToOpenDocument opts)
+                  (fmap chomp . inlinesToOpenDocument opts)
                   meta
-           b <- render' `fmap` blocksToOpenDocument opts blocks
+           b <- blocksToOpenDocument opts blocks
            return (b, m)
   let styles   = stTableStyles s ++ stParaStyles s ++ formulaStyles ++
                      map snd (sortBy (flip (comparing fst)) (
                         Map.elems (stTextStyles s)))
       listStyle (n,l) = inTags True "text:list-style"
-                          [("style:name", "L" ++ show n)] (vcat l)
+                          [("style:name", "L" <> tshow n)] (vcat l)
   let listStyles  = map listStyle (stListStyles s)
   let automaticStyles = vcat $ reverse $ styles ++ listStyles
   let context = defField "body" body
-              $ defField "toc" (writerTableOfContents opts)
-              $defField "automatic-styles" (render' automaticStyles) metadata
-  case writerTemplate opts of
-       Nothing  -> return body
-       Just tpl -> renderTemplate' tpl context
+              . defField "toc" (writerTableOfContents opts)
+              . defField "automatic-styles" automaticStyles
+              $ metadata
+  return $ render colwidth $
+    case writerTemplate opts of
+       Nothing  -> body
+       Just tpl -> renderTemplate tpl context
 
 withParagraphStyle :: PandocMonad m
-                   => WriterOptions -> String -> [Block] -> OD m Doc
+                   => WriterOptions -> Text -> [Block] -> OD m (Doc Text)
 withParagraphStyle  o s (b:bs)
     | Para l <- b = go =<< inParagraphTagsWithStyle s <$> inlinesToOpenDocument o l
     | otherwise   = go =<< blockToOpenDocument o b
     where go i = (<>) i <$>  withParagraphStyle o s bs
 withParagraphStyle _ _ [] = return empty
 
-inPreformattedTags :: PandocMonad m => String -> OD m Doc
+inPreformattedTags :: PandocMonad m => Text -> OD m (Doc Text)
 inPreformattedTags s = do
   n <- paraStyle [("style:parent-style-name","Preformatted_20_Text")]
-  return . inParagraphTagsWithStyle ("P" ++ show n) . handleSpaces $ s
+  return . inParagraphTagsWithStyle ("P" <> tshow n) . handleSpaces $ s
 
 orderedListToOpenDocument :: PandocMonad m
-                          => WriterOptions -> Int -> [[Block]] -> OD m Doc
+                          => WriterOptions -> Int -> [[Block]] -> OD m (Doc Text)
 orderedListToOpenDocument o pn bs =
     vcat . map (inTagsIndented "text:list-item") <$>
     mapM (orderedItemToOpenDocument o pn . map plainToPara) bs
 
 orderedItemToOpenDocument :: PandocMonad m
-                          => WriterOptions -> Int -> [Block] -> OD m Doc
+                          => WriterOptions -> Int -> [Block] -> OD m (Doc Text)
 orderedItemToOpenDocument  o n bs = vcat <$> mapM go bs
  where go (OrderedList a l) = newLevel a l
-       go (Para          l) = inParagraphTagsWithStyle ("P" ++ show n) <$>
+       go (Para          l) = inParagraphTagsWithStyle ("P" <> tshow n) <$>
                                 inlinesToOpenDocument o l
        go b                 = blockToOpenDocument o b
        newLevel a l = do
@@ -311,21 +296,21 @@ newOrderedListStyle b a = do
   return (ln,pn)
 
 bulletListToOpenDocument :: PandocMonad m
-                         => WriterOptions -> [[Block]] -> OD m Doc
+                         => WriterOptions -> [[Block]] -> OD m (Doc Text)
 bulletListToOpenDocument o b = do
   ln <- (+) 1 . length <$> gets stListStyles
   (pn,ns) <- if isTightList b then inTightList (bulletListStyle ln) else bulletListStyle ln
   modify $ \s -> s { stListStyles = ns : stListStyles s }
-  is <- listItemsToOpenDocument ("P" ++ show pn) o b
-  return $ inTags True "text:list" [("text:style-name", "L" ++ show ln)] is
+  is <- listItemsToOpenDocument ("P" <> tshow pn) o b
+  return $ inTags True "text:list" [("text:style-name", "L" <> tshow ln)] is
 
 listItemsToOpenDocument :: PandocMonad m
-                        => String -> WriterOptions -> [[Block]] -> OD m Doc
+                        => Text -> WriterOptions -> [[Block]] -> OD m (Doc Text)
 listItemsToOpenDocument s o is =
     vcat . map (inTagsIndented "text:list-item") <$> mapM (withParagraphStyle o s . map plainToPara) is
 
 deflistItemToOpenDocument :: PandocMonad m
-                          => WriterOptions -> ([Inline],[[Block]]) -> OD m Doc
+                          => WriterOptions -> ([Inline],[[Block]]) -> OD m (Doc Text)
 deflistItemToOpenDocument o (t,d) = do
   let ts = if isTightList d
            then "Definition_20_Term_20_Tight"       else "Definition_20_Term"
@@ -336,28 +321,28 @@ deflistItemToOpenDocument o (t,d) = do
   return $ t' $$ d'
 
 inBlockQuote :: PandocMonad m
-             => WriterOptions -> Int -> [Block] -> OD m Doc
+             => WriterOptions -> Int -> [Block] -> OD m (Doc Text)
 inBlockQuote  o i (b:bs)
     | BlockQuote l <- b = do increaseIndent
                              ni <- paraStyle
                                    [("style:parent-style-name","Quotations")]
                              go =<< inBlockQuote o ni (map plainToPara l)
-    | Para       l <- b = go =<< inParagraphTagsWithStyle ("P" ++ show  i) <$> inlinesToOpenDocument o l
+    | Para       l <- b = go =<< inParagraphTagsWithStyle ("P" <> tshow i) <$> inlinesToOpenDocument o l
     | otherwise         = go =<< blockToOpenDocument o b
     where go  block  = ($$) block <$> inBlockQuote o i bs
 inBlockQuote     _ _ [] =  resetIndent >> return empty
 
 -- | Convert a list of Pandoc blocks to OpenDocument.
-blocksToOpenDocument :: PandocMonad m => WriterOptions -> [Block] -> OD m Doc
+blocksToOpenDocument :: PandocMonad m => WriterOptions -> [Block] -> OD m (Doc Text)
 blocksToOpenDocument o b = vcat <$> mapM (blockToOpenDocument o) b
 
 -- | Convert a Pandoc block element to OpenDocument.
-blockToOpenDocument :: PandocMonad m => WriterOptions -> Block -> OD m Doc
+blockToOpenDocument :: PandocMonad m => WriterOptions -> Block -> OD m (Doc Text)
 blockToOpenDocument o bs
     | Plain          b <- bs = if null b
                                   then return empty
                                   else inParagraphTags =<< inlinesToOpenDocument o b
-    | Para [Image attr c (s,'f':'i':'g':':':t)] <- bs
+    | Para [Image attr c (s,T.stripPrefix "fig:" -> Just t)] <- bs
                              = figure attr c s t
     | Para           b <- bs = if null b &&
                                     not (isEnabled Ext_empty_paragraphs o)
@@ -378,7 +363,7 @@ blockToOpenDocument o bs
     | HorizontalRule   <- bs = setFirstPara >> return (selfClosingTag "text:p"
                                 [ ("text:style-name", "Horizontal_20_Line") ])
     | RawBlock f     s <- bs = if f == Format "opendocument"
-                                  then return $ text s
+                                  then return $ text $ T.unpack s
                                   else do
                                     report $ BlockNotRendered bs
                                     return empty
@@ -389,21 +374,21 @@ blockToOpenDocument o bs
                            r <- vcat  <$> mapM (deflistItemToOpenDocument o) b
                            setInDefinitionList False
                            return r
-      preformatted  s = (flush . vcat) <$> mapM (inPreformattedTags . escapeStringForXML) (lines s)
+      preformatted  s = (flush . vcat) <$> mapM (inPreformattedTags . escapeStringForXML) (T.lines s)
       mkBlockQuote  b = do increaseIndent
                            i <- paraStyle
                                  [("style:parent-style-name","Quotations")]
                            inBlockQuote o i (map plainToPara b)
       orderedList a b = do (ln,pn) <- newOrderedListStyle (isTightList b) a
-                           inTags True "text:list" [ ("text:style-name", "L" ++ show ln)]
+                           inTags True "text:list" [ ("text:style-name", "L" <> tshow ln)]
                                       <$> orderedListToOpenDocument o pn b
       table c a w h r = do
         tn <- length <$> gets stTableStyles
         pn <- length <$> gets stParaStyles
         let  genIds      = map chr [65..]
-             name        = "Table" ++ show (tn + 1)
+             name        = "Table" <> tshow (tn + 1)
              columnIds   = zip genIds w
-             mkColumn  n = selfClosingTag "table:table-column" [("table:style-name", name ++ "." ++ [fst n])]
+             mkColumn  n = selfClosingTag "table:table-column" [("table:style-name", name <> "." <> T.singleton (fst n))]
              columns     = map mkColumn columnIds
              paraHStyles = paraTableStyles "Heading"  pn a
              paraStyles  = paraTableStyles "Contents" (pn + length (newPara paraHStyles)) a
@@ -412,64 +397,75 @@ blockToOpenDocument o bs
         mapM_ addParaStyle . newPara $ paraHStyles ++ paraStyles
         captionDoc <- if null c
                       then return empty
-                      else inlinesToOpenDocument o c >>= numberedTableCaption
+                      else inlinesToOpenDocument o c >>=
+                             if isEnabled Ext_native_numbering o
+                                then numberedTableCaption
+                                else unNumberedCaption "TableCaption"
         th <- if all null h
                  then return empty
                  else colHeadsToOpenDocument o (map fst paraHStyles) h
         tr <- mapM (tableRowToOpenDocument o (map fst paraStyles)) r
-        return $ inTags True "table:table" [ ("table:name"      , name)
-                                           , ("table:style-name", name)
-                                           ] (vcat columns $$ th $$ vcat tr) $$ captionDoc
+        let tableDoc = inTags True "table:table" [
+                            ("table:name"      , name)
+                          , ("table:style-name", name)
+                          ] (vcat columns $$ th $$ vcat tr)
+        return $ captionDoc $$ tableDoc
       figure attr caption source title | null caption =
         withParagraphStyle o "Figure" [Para [Image attr caption (source,title)]]
                                   | otherwise    = do
         imageDoc <- withParagraphStyle o "FigureWithCaption" [Para [Image attr caption (source,title)]]
-        captionDoc <- inlinesToOpenDocument o caption >>= numberedFigureCaption
+        captionDoc <- inlinesToOpenDocument o caption >>=
+                         if isEnabled Ext_native_numbering o
+                            then numberedFigureCaption
+                            else unNumberedCaption "FigureCaption"
         return $ imageDoc $$ captionDoc
 
 
-numberedTableCaption :: PandocMonad m => Doc -> OD m Doc
+numberedTableCaption :: PandocMonad m => Doc Text -> OD m (Doc Text)
 numberedTableCaption caption = do
     id' <- gets stTableCaptionId
     modify (\st -> st{ stTableCaptionId = id' + 1 })
     capterm <- translateTerm Term.Table
-    return $ numberedCaption "Table" capterm "Table" id' caption
+    return $ numberedCaption "TableCaption" capterm "Table" id' caption
 
-numberedFigureCaption :: PandocMonad m => Doc -> OD m Doc
+numberedFigureCaption :: PandocMonad m => Doc Text -> OD m (Doc Text)
 numberedFigureCaption caption = do
     id' <- gets stImageCaptionId
     modify (\st -> st{ stImageCaptionId = id' + 1 })
     capterm <- translateTerm Term.Figure
     return $ numberedCaption "FigureCaption" capterm "Illustration" id' caption
 
-numberedCaption :: String -> String -> String -> Int -> Doc -> Doc
+numberedCaption :: Text -> Text -> Text -> Int -> Doc Text -> Doc Text
 numberedCaption style term name num caption =
-    let t = text term
+    let t = text $ T.unpack term
         r = num - 1
-        s = inTags False "text:sequence" [ ("text:ref-name", "ref" ++ name ++ show r),
+        s = inTags False "text:sequence" [ ("text:ref-name", "ref" <> name <> tshow r),
                                            ("text:name", name),
-                                           ("text:formula", "ooow:" ++ name ++ "+1"),
+                                           ("text:formula", "ooow:" <> name <> "+1"),
                                            ("style:num-format", "1") ] $ text $ show num
         c = text ": "
     in inParagraphTagsWithStyle style $ hcat [ t, text " ", s, c, caption ]
 
+unNumberedCaption :: Monad m => Text -> Doc Text -> OD m (Doc Text)
+unNumberedCaption style caption = return $ inParagraphTagsWithStyle style caption
+
 colHeadsToOpenDocument :: PandocMonad m
-                       => WriterOptions -> [String] -> [[Block]]
-                       -> OD m Doc
+                       => WriterOptions -> [Text] -> [[Block]]
+                       -> OD m (Doc Text)
 colHeadsToOpenDocument o ns hs =
     inTagsIndented "table:table-header-rows" . inTagsIndented "table:table-row" . vcat <$>
     mapM (tableItemToOpenDocument o "TableHeaderRowCell") (zip ns hs)
 
 tableRowToOpenDocument :: PandocMonad m
-                       => WriterOptions -> [String] -> [[Block]]
-                       -> OD m Doc
+                       => WriterOptions -> [Text] -> [[Block]]
+                       -> OD m (Doc Text)
 tableRowToOpenDocument o ns cs =
     inTagsIndented "table:table-row" . vcat <$>
     mapM (tableItemToOpenDocument o "TableRowCell") (zip ns cs)
 
 tableItemToOpenDocument :: PandocMonad m
-                        => WriterOptions -> String -> (String,[Block])
-                        -> OD m Doc
+                        => WriterOptions -> Text -> (Text,[Block])
+                        -> OD m (Doc Text)
 tableItemToOpenDocument o s (n,i) =
   let a = [ ("table:style-name" , s )
           , ("office:value-type", "string"     )
@@ -478,10 +474,10 @@ tableItemToOpenDocument o s (n,i) =
       withParagraphStyle o n (map plainToPara i)
 
 -- | Convert a list of inline elements to OpenDocument.
-inlinesToOpenDocument :: PandocMonad m => WriterOptions -> [Inline] -> OD m Doc
+inlinesToOpenDocument :: PandocMonad m => WriterOptions -> [Inline] -> OD m (Doc Text)
 inlinesToOpenDocument o l = hcat <$> toChunks o l
 
-toChunks :: PandocMonad m => WriterOptions -> [Inline] -> OD m [Doc]
+toChunks :: PandocMonad m => WriterOptions -> [Inline] -> OD m [Doc Text]
 toChunks _ [] = return []
 toChunks o (x : xs)
   | isChunkable x = do
@@ -502,7 +498,7 @@ isChunkable SoftBreak = True
 isChunkable _         = False
 
 -- | Convert an inline element to OpenDocument.
-inlineToOpenDocument :: PandocMonad m => WriterOptions -> Inline -> OD m Doc
+inlineToOpenDocument :: PandocMonad m => WriterOptions -> Inline -> OD m (Doc Text)
 inlineToOpenDocument o ils
   = case ils of
     Space         -> return space
@@ -525,7 +521,7 @@ inlineToOpenDocument o ils
                          inlinesToOpenDocument o
     Cite      _ l -> inlinesToOpenDocument o l
     RawInline f s -> if f == Format "opendocument"
-                       then return $ text s
+                       then return $ text $ T.unpack s
                        else do
                          report $ InlineNotRendered ils
                          return empty
@@ -549,7 +545,7 @@ inlineToOpenDocument o ils
                    getDims (("rel-height", w):xs) = ("style:rel-height", w) : getDims xs
                    getDims (_:xs) =                             getDims xs
                return $ inTags False "draw:frame"
-                        (("draw:name", "img" ++ show id') : getDims kvs) $
+                        (("draw:name", "img" <> tshow id') : getDims kvs) $
                      selfClosingTag "draw:image" [ ("xlink:href"   , s       )
                                                  , ("xlink:type"   , "simple")
                                                  , ("xlink:show"   , "embed" )
@@ -557,7 +553,7 @@ inlineToOpenDocument o ils
       mkNote     l = do
         n <- length <$> gets stNotes
         let footNote t = inTags False "text:note"
-                         [ ("text:id"        , "ftn" ++ show n)
+                         [ ("text:id"        , "ftn" <> tshow n)
                          , ("text:note-class", "footnote"     )] $
                          inTagsSimple "text:note-citation" (text . show $ n + 1) <>
                          inTagsSimple "text:note-body" t
@@ -565,20 +561,20 @@ inlineToOpenDocument o ils
         addNote nn
         return nn
 
-bulletListStyle :: PandocMonad m => Int -> OD m (Int,(Int,[Doc]))
+bulletListStyle :: PandocMonad m => Int -> OD m (Int,(Int,[Doc Text]))
 bulletListStyle l = do
   let doStyles  i = inTags True "text:list-level-style-bullet"
-                    [ ("text:level"      , show (i + 1)       )
-                    , ("text:style-name" , "Bullet_20_Symbols")
-                    , ("style:num-suffix", "."                )
-                    , ("text:bullet-char", [bulletList !! i]  )
+                    [ ("text:level"      , tshow (i + 1))
+                    , ("text:style-name" , "Bullet_20_Symbols"  )
+                    , ("style:num-suffix", "."                  )
+                    , ("text:bullet-char", T.singleton (bulletList !! i))
                     ] (listLevelStyle (1 + i))
       bulletList  = map chr $ cycle [8226,9702,9642]
       listElStyle = map doStyles [0..9]
   pn <- paraListStyle l
   return (pn, (l, listElStyle))
 
-orderedListLevelStyle :: ListAttributes -> (Int, [Doc]) -> (Int,[Doc])
+orderedListLevelStyle :: ListAttributes -> (Int, [Doc Text]) -> (Int,[Doc Text])
 orderedListLevelStyle (s,n, d) (l,ls) =
     let suffix    = case d of
                       OneParen  -> [("style:num-suffix", ")")]
@@ -592,16 +588,16 @@ orderedListLevelStyle (s,n, d) (l,ls) =
                       LowerRoman -> "i"
                       _          -> "1"
         listStyle = inTags True "text:list-level-style-number"
-                    ([ ("text:level"      , show $ 1 + length ls  )
+                    ([ ("text:level"      , tshow $ 1 + length ls )
                      , ("text:style-name" , "Numbering_20_Symbols")
                      , ("style:num-format", format                )
-                     , ("text:start-value", show s                )
+                     , ("text:start-value", tshow s               )
                      ] ++ suffix) (listLevelStyle (1 + length ls))
     in  (l, ls ++ [listStyle])
 
-listLevelStyle :: Int -> Doc
+listLevelStyle :: Int -> Doc Text
 listLevelStyle i =
-    let indent = show (0.25 + (0.25 * fromIntegral i :: Double)) in
+    let indent = tshow (0.25 + (0.25 * fromIntegral i :: Double)) in
     inTags True "style:list-level-properties"
                        [ ("text:list-level-position-and-space-mode",
                           "label-alignment")
@@ -609,27 +605,27 @@ listLevelStyle i =
                        ] $
        selfClosingTag "style:list-level-label-alignment"
                       [ ("text:label-followed-by", "listtab")
-                      , ("text:list-tab-stop-position", indent ++ "in")
+                      , ("text:list-tab-stop-position", indent <> "in")
                       , ("fo:text-indent", "-0.25in")
-                      , ("fo:margin-left", indent ++ "in")
+                      , ("fo:margin-left", indent <> "in")
                       ]
 
-tableStyle :: Int -> [(Char,Double)] -> Doc
+tableStyle :: Int -> [(Char,Double)] -> Doc Text
 tableStyle num wcs =
-    let tableId        = "Table" ++ show (num + 1)
+    let tableId        = "Table" <> tshow (num + 1)
         table          = inTags True "style:style"
                          [("style:name", tableId)
                          ,("style:family", "table")] $
                          selfClosingTag "style:table-properties"
                          [("table:align"    , "center")]
         colStyle (c,0) = selfClosingTag "style:style"
-                         [ ("style:name"  , tableId ++ "." ++ [c])
+                         [ ("style:name"  , tableId <> "." <> T.singleton c)
                          , ("style:family", "table-column"       )]
         colStyle (c,w) = inTags True "style:style"
-                         [ ("style:name"  , tableId ++ "." ++ [c])
+                         [ ("style:name"  , tableId <> "." <> T.singleton c)
                          , ("style:family", "table-column"       )] $
                          selfClosingTag "style:table-column-properties"
-                         [("style:rel-column-width", printf "%d*" (floor $ w * 65535 :: Integer))]
+                         [("style:rel-column-width", T.pack $ printf "%d*" (floor $ w * 65535 :: Integer))]
         headerRowCellStyle = inTags True "style:style"
                          [ ("style:name"  , "TableHeaderRowCell")
                          , ("style:family", "table-cell"    )] $
@@ -646,15 +642,15 @@ tableStyle num wcs =
         columnStyles   = map colStyle wcs
     in cellStyles $$ table $$ vcat columnStyles
 
-paraStyle :: PandocMonad m => [(String,String)] -> OD m Int
+paraStyle :: PandocMonad m => [(Text,Text)] -> OD m Int
 paraStyle attrs = do
   pn <- (+)   1 . length       <$> gets stParaStyles
   i  <- (*) (0.5 :: Double) . fromIntegral <$> gets stIndentPara
   b  <- gets stInDefinition
   t  <- gets stTight
-  let styleAttr = [ ("style:name"             , "P" ++ show pn)
+  let styleAttr = [ ("style:name"             , "P" <> tshow pn)
                   , ("style:family"           , "paragraph"   )]
-      indentVal = flip (++) "in" . show $ if b then max 0.5 i else i
+      indentVal = flip (<>) "in" . tshow $ if b then max 0.5 i else i
       tight     = if t then [ ("fo:margin-top"          , "0in"    )
                             , ("fo:margin-bottom"       , "0in"    )]
                        else []
@@ -664,30 +660,30 @@ paraStyle attrs = do
                            , ("fo:text-indent"         , "0in"    )
                            , ("style:auto-text-indent" , "false"  )]
                       else []
-      attributes = indent ++ tight
+      attributes = indent <> tight
       paraProps = if null attributes
                      then mempty
                      else selfClosingTag
                              "style:paragraph-properties" attributes
-  addParaStyle $ inTags True "style:style" (styleAttr ++ attrs) paraProps
+  addParaStyle $ inTags True "style:style" (styleAttr <> attrs) paraProps
   return pn
 
 paraListStyle :: PandocMonad m => Int -> OD m Int
 paraListStyle l = paraStyle
   [("style:parent-style-name","Text_20_body")
-  ,("style:list-style-name", "L" ++ show l )]
+  ,("style:list-style-name", "L" <> tshow l)]
 
-paraTableStyles :: String -> Int -> [Alignment] -> [(String, Doc)]
+paraTableStyles :: Text -> Int -> [Alignment] -> [(Text, Doc Text)]
 paraTableStyles _ _ [] = []
 paraTableStyles t s (a:xs)
     | AlignRight  <- a = (         pName s, res s "end"   ) : paraTableStyles t (s + 1) xs
     | AlignCenter <- a = (         pName s, res s "center") : paraTableStyles t (s + 1) xs
-    | otherwise        = ("Table_20_" ++ t, empty         ) : paraTableStyles t  s      xs
-    where pName sn = "P" ++ show (sn + 1)
+    | otherwise        = ("Table_20_" <> t, empty         ) : paraTableStyles t  s      xs
+    where pName sn = "P" <> tshow (sn + 1)
           res sn x = inTags True "style:style"
                      [ ("style:name"             , pName sn        )
                      , ("style:family"           , "paragraph"     )
-                     , ("style:parent-style-name", "Table_20_" ++ t)] $
+                     , ("style:parent-style-name", "Table_20_" <> t)] $
                      selfClosingTag "style:paragraph-properties"
                      [ ("fo:text-align", x)
                      , ("style:justify-single-word", "false")]
@@ -702,25 +698,27 @@ data TextStyle = Italic
                | Language Lang
                deriving ( Eq,Ord )
 
-textStyleAttr :: TextStyle -> [(String,String)]
-textStyleAttr s
-    | Italic <- s = [("fo:font-style"                ,"italic"    )
-                    ,("style:font-style-asian"       ,"italic"    )
-                    ,("style:font-style-complex"     ,"italic"    )]
-    | Bold   <- s = [("fo:font-weight"               ,"bold"      )
-                    ,("style:font-weight-asian"      ,"bold"      )
-                    ,("style:font-weight-complex"    ,"bold"      )]
-    | Strike <- s = [("style:text-line-through-style", "solid"    )]
-    | Sub    <- s = [("style:text-position"          ,"sub 58%"   )]
-    | Sup    <- s = [("style:text-position"          ,"super 58%" )]
-    | SmallC <- s = [("fo:font-variant"              ,"small-caps")]
-    | Pre    <- s = [("style:font-name"              ,"Courier New")
-                    ,("style:font-name-asian"        ,"Courier New")
-                    ,("style:font-name-complex"      ,"Courier New")]
+textStyleAttr :: Map.Map Text Text
+              -> TextStyle
+              -> Map.Map Text Text
+textStyleAttr m s
+    | Italic <- s = Map.insert "fo:font-style" "italic" .
+                    Map.insert "style:font-style-asian" "italic" .
+                    Map.insert "style:font-style-complex" "italic" $ m
+    | Bold   <- s = Map.insert "fo:font-weight" "bold" .
+                    Map.insert "style:font-weight-asian" "bold" .
+                    Map.insert "style:font-weight-complex" "bold" $ m
+    | Strike <- s = Map.insert "style:text-line-through-style" "solid" m
+    | Sub    <- s = Map.insert "style:text-position" "sub 58%" m
+    | Sup    <- s = Map.insert "style:text-position" "super 58%" m
+    | SmallC <- s = Map.insert "fo:font-variant" "small-caps" m
+    | Pre    <- s = Map.insert "style:font-name" "Courier New" .
+                    Map.insert "style:font-name-asian" "Courier New" .
+                    Map.insert "style:font-name-complex" "Courier New" $ m
     | Language lang <- s
-                  = [("fo:language"                  ,langLanguage lang)
-                    ,("fo:country"                   ,langRegion lang)]
-    | otherwise   = []
+                  = Map.insert "fo:language" (langLanguage lang) .
+                    Map.insert "fo:country" (langRegion lang) $ m
+    | otherwise   = m
 
 withLangFromAttr :: PandocMonad m => Attr -> OD m a -> OD m a
 withLangFromAttr (_,_,kvs) action =
