@@ -1,28 +1,10 @@
-{-# LANGUAGE NoImplicitPrelude          #-}
+{-# LANGUAGE OverloadedStrings          #-}
 {-# LANGUAGE PatternGuards              #-}
+{-# LANGUAGE ViewPatterns               #-}
 {-# LANGUAGE GeneralizedNewtypeDeriving #-}
-
-{-
-Copyright (C) 2017-2018 Jesse Rosenthal <jrosenthal@jhu.edu>
-
-This program is free software; you can redistribute it and/or modify
-it under the terms of the GNU General Public License as published by
-the Free Software Foundation; either version 2 of the License, or
-(at your option) any later version.
-
-This program is distributed in the hope that it will be useful,
-but WITHOUT ANY WARRANTY; without even the implied warranty of
-MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-GNU General Public License for more details.
-
-You should have received a copy of the GNU General Public License
-along with this program; if not, write to the Free Software
-Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--}
-
 {- |
    Module      : Text.Pandoc.Writers.Powerpoint.Presentation
-   Copyright   : Copyright (C) 2017-2018 Jesse Rosenthal
+   Copyright   : Copyright (C) 2017-2020 Jesse Rosenthal
    License     : GNU GPL, version 2 or above
 
    Maintainer  : Jesse Rosenthal <jrosenthal@jhu.edu>
@@ -52,6 +34,7 @@ module Text.Pandoc.Writers.Powerpoint.Presentation ( documentToPresentation
                                                    , TableProps(..)
                                                    , Strikethrough(..)
                                                    , Capitals(..)
+                                                   , Pixels
                                                    , PicProps(..)
                                                    , URL
                                                    , TeXString(..)
@@ -59,7 +42,6 @@ module Text.Pandoc.Writers.Powerpoint.Presentation ( documentToPresentation
                                                    ) where
 
 
-import Prelude
 import Control.Monad.Reader
 import Control.Monad.State
 import Data.List (intercalate)
@@ -72,7 +54,9 @@ import Text.Pandoc.Logging
 import Text.Pandoc.Walk
 import Data.Time (UTCTime)
 import qualified Text.Pandoc.Shared as Shared -- so we don't overlap "Element"
-import Text.Pandoc.Writers.Shared (lookupMetaInlines)
+import Text.Pandoc.Shared (tshow)
+import Text.Pandoc.Writers.Shared (lookupMetaInlines, lookupMetaBlocks
+                                 , lookupMetaString, toTableOfContents)
 import qualified Data.Map as M
 import qualified Data.Set as S
 import Data.Maybe (maybeToList, fromMaybe)
@@ -110,7 +94,7 @@ instance Default WriterEnv where
 
 data WriterState = WriterState { stNoteIds :: M.Map Int [Block]
                                -- associate anchors with slide id
-                               , stAnchorMap :: M.Map String SlideId
+                               , stAnchorMap :: M.Map T.Text SlideId
                                , stSlideIdSet :: S.Set SlideId
                                , stLog :: [LogMessage]
                                , stSpeakerNotes :: SpeakerNotes
@@ -140,17 +124,17 @@ reservedSlideIds = S.fromList [ metadataSlideId
                               , endNotesSlideId
                               ]
 
-uniqueSlideId' :: Integer -> S.Set SlideId -> String -> SlideId
+uniqueSlideId' :: Integer -> S.Set SlideId -> T.Text -> SlideId
 uniqueSlideId' n idSet s =
-  let s' = if n == 0 then s else s ++ "-" ++ show n
+  let s' = if n == 0 then s else s <> "-" <> tshow n
   in if SlideId s' `S.member` idSet
      then uniqueSlideId' (n+1) idSet s
      else SlideId s'
 
-uniqueSlideId :: S.Set SlideId -> String -> SlideId
+uniqueSlideId :: S.Set SlideId -> T.Text -> SlideId
 uniqueSlideId = uniqueSlideId' 0
 
-runUniqueSlideId :: String -> Pres SlideId
+runUniqueSlideId :: T.Text -> Pres SlideId
 runUniqueSlideId s = do
   idSet <- gets stSlideIdSet
   let sldId = uniqueSlideId idSet s
@@ -176,11 +160,14 @@ type Pixels = Integer
 data Presentation = Presentation DocProps [Slide]
   deriving (Show)
 
-data DocProps = DocProps { dcTitle :: Maybe String
-                         , dcSubject :: Maybe String
-                         , dcCreator :: Maybe String
-                         , dcKeywords :: Maybe [String]
+data DocProps = DocProps { dcTitle :: Maybe T.Text
+                         , dcSubject :: Maybe T.Text
+                         , dcCreator :: Maybe T.Text
+                         , dcKeywords :: Maybe [T.Text]
+                         , dcDescription :: Maybe T.Text
+                         , cpCategory :: Maybe T.Text
                          , dcCreated :: Maybe UTCTime
+                         , customProperties :: Maybe [(T.Text, T.Text)]
                          } deriving (Show, Eq)
 
 
@@ -189,7 +176,7 @@ data Slide = Slide { slideId :: SlideId
                    , slideSpeakerNotes :: SpeakerNotes
                    } deriving (Show, Eq)
 
-newtype SlideId = SlideId String
+newtype SlideId = SlideId T.Text
   deriving (Show, Eq, Ord)
 
 -- In theory you could have anything on a notes slide but it seems
@@ -198,25 +185,20 @@ newtype SlideId = SlideId String
 newtype SpeakerNotes = SpeakerNotes {fromSpeakerNotes :: [Paragraph]}
   deriving (Show, Eq, Monoid, Semigroup)
 
-data Layout = MetadataSlide { metadataSlideTitle :: [ParaElem]
-                            , metadataSlideSubtitle :: [ParaElem]
-                            , metadataSlideAuthors :: [[ParaElem]]
-                            , metadataSlideDate :: [ParaElem]
-                            }
-           | TitleSlide { titleSlideHeader :: [ParaElem]}
-           | ContentSlide { contentSlideHeader :: [ParaElem]
-                          , contentSlideContent :: [Shape]
-                          }
-           | TwoColumnSlide { twoColumnSlideHeader :: [ParaElem]
-                            , twoColumnSlideLeft   :: [Shape]
-                            , twoColumnSlideRight  :: [Shape]
-                            }
-           deriving (Show, Eq)
+data Layout = MetadataSlide [ParaElem] [ParaElem] [[ParaElem]] [ParaElem]
+            --              title      subtitle   authors      date
+            | TitleSlide [ParaElem]
+            --           heading
+            | ContentSlide [ParaElem] [Shape]
+            --             heading    content
+            | TwoColumnSlide [ParaElem] [Shape] [Shape]
+            --               heading    left    right
+            deriving (Show, Eq)
 
 data Shape = Pic PicProps FilePath [ParaElem]
            | GraphicFrame [Graphic] [ParaElem]
            | TextBox [Paragraph]
-           | RawOOXMLShape String
+           | RawOOXMLShape T.Text
   deriving (Show, Eq)
 
 type Cell = [Paragraph]
@@ -246,6 +228,7 @@ data ParaProps = ParaProps { pPropMarginLeft :: Maybe Pixels
                            , pPropBullet :: Maybe BulletType
                            , pPropAlign :: Maybe Algnment
                            , pPropSpaceBefore :: Maybe Pixels
+                           , pPropIndent :: Maybe Pixels
                            } deriving (Show, Eq)
 
 instance Default ParaProps where
@@ -255,19 +238,20 @@ instance Default ParaProps where
                   , pPropBullet = Nothing
                   , pPropAlign = Nothing
                   , pPropSpaceBefore = Nothing
+                  , pPropIndent = Just 0
                   }
 
-newtype TeXString = TeXString {unTeXString :: String}
+newtype TeXString = TeXString {unTeXString :: T.Text}
   deriving (Eq, Show)
 
 data ParaElem = Break
-              | Run RunProps String
+              | Run RunProps T.Text
               -- It would be more elegant to have native TeXMath
               -- Expressions here, but this allows us to use
               -- `convertmath` from T.P.Writers.Math. Will perhaps
               -- revisit in the future.
               | MathElem MathType TeXString
-              | RawOOXMLParaElem String
+              | RawOOXMLParaElem T.Text
               deriving (Show, Eq)
 
 data Strikethrough = NoStrike | SingleStrike | DoubleStrike
@@ -276,9 +260,9 @@ data Strikethrough = NoStrike | SingleStrike | DoubleStrike
 data Capitals = NoCapitals | SmallCapitals | AllCapitals
   deriving (Show, Eq)
 
-type URL = String
+type URL = T.Text
 
-data LinkTarget = ExternalTarget (URL, String)
+data LinkTarget = ExternalTarget (URL, T.Text)
                 | InternalTarget SlideId
                 deriving (Show, Eq)
 
@@ -377,7 +361,10 @@ inlineToParElems (Note blks) = do
         curNoteId = maxNoteId + 1
     modify $ \st -> st { stNoteIds = M.insert curNoteId blks notes }
     local (\env -> env{envRunProps = (envRunProps env){rLink = Just $ InternalTarget endNotesSlideId}}) $
-      inlineToParElems $ Superscript [Str $ show curNoteId]
+      inlineToParElems $ Superscript [Str $ tshow curNoteId]
+inlineToParElems (Span (_, ["underline"], _) ils) =
+  local (\r -> r{envRunProps = (envRunProps r){rPropUnderline=True}}) $
+  inlinesToParElems ils
 inlineToParElems (Span _ ils) = inlinesToParElems ils
 inlineToParElems (Quoted quoteType ils) =
   inlinesToParElems $ [Str open] ++ ils ++ [Str close]
@@ -403,11 +390,11 @@ isListType (BulletList _) = True
 isListType (DefinitionList _) = True
 isListType _ = False
 
-registerAnchorId :: String -> Pres ()
+registerAnchorId :: T.Text -> Pres ()
 registerAnchorId anchor = do
   anchorMap <- gets stAnchorMap
   sldId <- asks envCurSlideId
-  unless (null anchor) $
+  unless (T.null anchor) $
     modify $ \st -> st {stAnchorMap = M.insert anchor sldId anchorMap}
 
 -- Currently hardcoded, until I figure out how to make it dynamic.
@@ -428,18 +415,23 @@ blockToParagraphs (LineBlock ilsList) = do
   pProps <- asks envParaProps
   return [Paragraph pProps parElems]
 -- TODO: work out the attributes
-blockToParagraphs (CodeBlock attr str) =
-  local (\r -> r{ envParaProps = def{pPropMarginLeft = Just 100}
+blockToParagraphs (CodeBlock attr str) = do
+  pProps <- asks envParaProps
+  local (\r -> r{ envParaProps = def{ pPropMarginLeft = Nothing
+                                    , pPropBullet = Nothing
+                                    , pPropLevel = pPropLevel pProps
+                                    , pPropIndent = Just 0
+                                    }
                 , envRunProps = (envRunProps r){rPropCode = True}}) $ do
-  mbSty <- writerHighlightStyle <$> asks envOpts
-  synMap <- writerSyntaxMap <$> asks envOpts
-  case mbSty of
-    Just sty ->
-      case highlight synMap (formatSourceLines sty) attr str of
-        Right pElems -> do pProps <- asks envParaProps
-                           return [Paragraph pProps pElems]
-        Left _ -> blockToParagraphs $ Para [Str str]
-    Nothing -> blockToParagraphs $ Para [Str str]
+    mbSty <- writerHighlightStyle <$> asks envOpts
+    synMap <- writerSyntaxMap <$> asks envOpts
+    case mbSty of
+      Just sty ->
+        case highlight synMap (formatSourceLines sty) attr str of
+          Right pElems -> do pPropsNew <- asks envParaProps
+                             return [Paragraph pPropsNew pElems]
+          Left _ -> blockToParagraphs $ Para [Str str]
+      Nothing -> blockToParagraphs $ Para [Str str]
 -- We can't yet do incremental lists, but we should render a
 -- (BlockQuote List) as a list to maintain compatibility with other
 -- formats.
@@ -448,7 +440,9 @@ blockToParagraphs (BlockQuote (blk : blks)) | isListType blk = do
   ps' <- blockToParagraphs $ BlockQuote blks
   return $ ps ++ ps'
 blockToParagraphs (BlockQuote blks) =
-  local (\r -> r{ envParaProps = (envParaProps r){pPropMarginLeft = Just 100}
+  local (\r -> r{ envParaProps = (envParaProps r){ pPropMarginLeft = Just 100
+                                                 , pPropIndent = Just 0
+                                                 }
                 , envRunProps = (envRunProps r){rPropForceSize = Just blockQuoteSize}})$
   concatMapM blockToParagraphs blks
 -- TODO: work out the format
@@ -472,6 +466,7 @@ blockToParagraphs (BulletList blksLst) = do
                     , envParaProps = pProps{ pPropLevel = lvl + 1
                                            , pPropBullet = Just Bullet
                                            , pPropMarginLeft = Nothing
+                                           , pPropIndent = Nothing
                                            }}) $
     concatMapM multiParBullet blksLst
 blockToParagraphs (OrderedList listAttr blksLst) = do
@@ -481,6 +476,7 @@ blockToParagraphs (OrderedList listAttr blksLst) = do
                     , envParaProps = pProps{ pPropLevel = lvl + 1
                                            , pPropBullet = Just (AutoNumbering listAttr)
                                            , pPropMarginLeft = Nothing
+                                           , pPropIndent = Nothing
                                            }}) $
     concatMapM multiParBullet blksLst
 blockToParagraphs (DefinitionList entries) = do
@@ -536,11 +532,11 @@ withAttr _ sp = sp
 blockToShape :: Block -> Pres Shape
 blockToShape (Plain ils) = blockToShape (Para ils)
 blockToShape (Para (il:_))  | Image attr ils (url, _) <- il =
-      (withAttr attr . Pic def url) <$> inlinesToParElems ils
+      (withAttr attr . Pic def (T.unpack url)) <$> inlinesToParElems ils
 blockToShape (Para (il:_))  | Link _ (il':_) target <- il
                             , Image attr ils (url, _) <- il' =
-      (withAttr attr . Pic def{picPropLink = Just $ ExternalTarget target} url) <$>
-      inlinesToParElems ils
+      (withAttr attr . Pic def{picPropLink = Just $ ExternalTarget target} (T.unpack url))
+      <$> inlinesToParElems ils
 blockToShape (Table caption algn _ hdrCells rows) = do
   caption' <- inlinesToParElems caption
   hdrCells' <- rowToParagraphs algn hdrCells
@@ -582,15 +578,16 @@ isImage (Link _ (Image{} : _) _) = True
 isImage _ = False
 
 splitBlocks' :: [Block] -> [[Block]] -> [Block] -> Pres [[Block]]
-splitBlocks' cur acc [] = return $ acc ++ (if null cur then [] else [cur])
+splitBlocks' cur acc [] = return $ acc ++ ([cur | not (null cur)])
 splitBlocks' cur acc (HorizontalRule : blks) =
-  splitBlocks' [] (acc ++ (if null cur then [] else [cur])) blks
+  splitBlocks' [] (acc ++ ([cur | not (null cur)])) blks
 splitBlocks' cur acc (h@(Header n _ _) : blks) = do
   slideLevel <- asks envSlideLevel
+  let (nts, blks') = span isNotesDiv blks
   case compare n slideLevel of
-    LT -> splitBlocks' [] (acc ++ (if null cur then [] else [cur]) ++ [[h]]) blks
-    EQ -> splitBlocks' [h] (acc ++ (if null cur then [] else [cur])) blks
-    GT -> splitBlocks' (cur ++ [h]) acc blks
+    LT -> splitBlocks' [] (acc ++ ([cur | not (null cur)]) ++ [h : nts]) blks'
+    EQ -> splitBlocks' (h:nts) (acc ++ ([cur | not (null cur)])) blks'
+    GT -> splitBlocks' (cur ++ (h:nts)) acc blks'
 -- `blockToParagraphs` treats Plain and Para the same, so we can save
 -- some code duplication by treating them the same here.
 splitBlocks' cur acc (Plain ils : blks) = splitBlocks' cur acc (Para ils : blks)
@@ -605,7 +602,7 @@ splitBlocks' cur acc (Para (il:ils) : blks) | isImage il = do
                             (acc ++ [cur ++ [Para [il]] ++ nts])
                             (if null ils then blks' else Para ils : blks')
     _ -> splitBlocks' []
-         (acc ++ (if null cur then [] else [cur]) ++ [[Para [il]] ++ nts])
+         (acc ++ ([cur | not (null cur)]) ++ [Para [il] : nts])
          (if null ils then blks' else Para ils : blks')
 splitBlocks' cur acc (tbl@Table{} : blks) = do
   slideLevel <- asks envSlideLevel
@@ -613,14 +610,14 @@ splitBlocks' cur acc (tbl@Table{} : blks) = do
   case cur of
     [Header n _ _] | n == slideLevel ->
                             splitBlocks' [] (acc ++ [cur ++ [tbl] ++ nts]) blks'
-    _ ->  splitBlocks' [] (acc ++ (if null cur then [] else [cur]) ++ [[tbl] ++ nts]) blks'
+    _ ->  splitBlocks' [] (acc ++ ([cur | not (null cur)]) ++ [tbl : nts]) blks'
 splitBlocks' cur acc (d@(Div (_, classes, _) _): blks) | "columns" `elem` classes =  do
   slideLevel <- asks envSlideLevel
   let (nts, blks') = span isNotesDiv blks
   case cur of
     [Header n _ _] | n == slideLevel ->
                             splitBlocks' [] (acc ++ [cur ++ [d] ++ nts]) blks'
-    _ ->  splitBlocks' [] (acc ++ (if null cur then [] else [cur]) ++ [[d] ++ nts]) blks'
+    _ ->  splitBlocks' [] (acc ++ ([cur | not (null cur)]) ++ [d : nts]) blks'
 splitBlocks' cur acc (blk : blks) = splitBlocks' (cur ++ [blk]) acc blks
 
 splitBlocks :: [Block] -> Pres [[Block]]
@@ -632,7 +629,7 @@ blocksToSlide' lvl (Header n (ident, _, _) ils : blks) spkNotes
       registerAnchorId ident
       sldId <- asks envCurSlideId
       hdr <- inlinesToParElems ils
-      return $ Slide sldId TitleSlide {titleSlideHeader = hdr} spkNotes
+      return $ Slide sldId (TitleSlide hdr) spkNotes
   | n == lvl = do
       registerAnchorId ident
       hdr <- inlinesToParElems ils
@@ -663,10 +660,7 @@ blocksToSlide' _ (blk : blks) spkNotes
       sldId <- asks envCurSlideId
       return $ Slide
         sldId
-        TwoColumnSlide { twoColumnSlideHeader = []
-                       , twoColumnSlideLeft = shapesL
-                       , twoColumnSlideRight = shapesR
-                       }
+        (TwoColumnSlide [] shapesL shapesR)
         spkNotes
 blocksToSlide' _ (blk : blks) spkNotes = do
       inNoteSlide <- asks envInNoteSlide
@@ -677,48 +671,48 @@ blocksToSlide' _ (blk : blks) spkNotes = do
       return $
         Slide
         sldId
-        ContentSlide { contentSlideHeader = []
-                     , contentSlideContent = shapes
-                     }
+        (ContentSlide [] shapes)
         spkNotes
 blocksToSlide' _ [] spkNotes = do
   sldId <- asks envCurSlideId
   return $
     Slide
     sldId
-    ContentSlide { contentSlideHeader = []
-                 , contentSlideContent = []
-                 }
+    (ContentSlide [] [])
     spkNotes
 
-handleNotes :: Block -> Pres ()
-handleNotes (Div (_, ["notes"], _) blks) =
-  local (\env -> env{envInSpeakerNotes=True}) $ do
-  spNotes <- SpeakerNotes <$> concatMapM blockToParagraphs blks
-  modify $ \st -> st{stSpeakerNotes = (stSpeakerNotes st) <> spNotes}
-handleNotes _ = return ()
+blockToSpeakerNotes :: Block -> Pres SpeakerNotes
+blockToSpeakerNotes (Div (_, ["notes"], _) blks) =
+  local (\env -> env{envInSpeakerNotes=True}) $
+  SpeakerNotes <$> concatMapM blockToParagraphs blks
+blockToSpeakerNotes _ = return mempty
 
-handleAndFilterNotes' :: [Block] -> Pres [Block]
-handleAndFilterNotes' blks = do
-  mapM_ handleNotes blks
+handleSpeakerNotes :: Block -> Pres ()
+handleSpeakerNotes blk = do
+  spNotes <- blockToSpeakerNotes blk
+  modify $ \st -> st{stSpeakerNotes = stSpeakerNotes st <> spNotes}
+
+handleAndFilterSpeakerNotes' :: [Block] -> Pres [Block]
+handleAndFilterSpeakerNotes' blks = do
+  mapM_ handleSpeakerNotes blks
   return $ filter (not . isNotesDiv) blks
 
-handleAndFilterNotes :: [Block] -> Pres ([Block], SpeakerNotes)
-handleAndFilterNotes blks = do
+handleAndFilterSpeakerNotes :: [Block] -> Pres ([Block], SpeakerNotes)
+handleAndFilterSpeakerNotes blks = do
   modify $ \st -> st{stSpeakerNotes = mempty}
-  blks' <- walkM handleAndFilterNotes' blks
+  blks' <- walkM handleAndFilterSpeakerNotes' blks
   spkNotes <- gets stSpeakerNotes
   return (blks', spkNotes)
 
 blocksToSlide :: [Block] -> Pres Slide
 blocksToSlide blks = do
-  (blks', spkNotes) <- handleAndFilterNotes blks
+  (blks', spkNotes) <- handleAndFilterSpeakerNotes blks
   slideLevel <- asks envSlideLevel
   blocksToSlide' slideLevel blks' spkNotes
 
-makeNoteEntry :: Int -> [Block] -> [Block]
-makeNoteEntry n blks =
-  let enum = Str (show n ++ ".")
+makeNoteEntry :: (Int, [Block]) -> [Block]
+makeNoteEntry (n, blks) =
+  let enum = Str (tshow n <> ".")
   in
     case blks of
       (Para ils : blks') -> (Para $ enum : Space : ils) : blks'
@@ -746,7 +740,7 @@ makeEndNotesSlideBlocks = do
                        ls -> ls
              ident = Shared.uniqueIdent exts title anchorSet
              hdr = Header slideLevel (ident, [], []) title
-             blks = concatMap (\(n, bs) -> makeNoteEntry n bs) $
+             blks = concatMap makeNoteEntry $
                     M.toList noteIds
          in return $ hdr : blks
 
@@ -763,30 +757,20 @@ getMetaSlide  = do
          Just $
          Slide
          metadataSlideId
-         MetadataSlide { metadataSlideTitle = title
-                       , metadataSlideSubtitle = subtitle
-                       , metadataSlideAuthors = authors
-                       , metadataSlideDate = date
-                       }
+         (MetadataSlide title subtitle authors date)
          mempty
 
--- adapted from the markdown writer
-elementToListItem :: Shared.Element -> Pres [Block]
-elementToListItem (Shared.Sec lev _nums (ident,_,_) headerText subsecs) = do
-  opts <- asks envOpts
-  let headerLink = if null ident
-                   then walk Shared.deNote headerText
-                   else [Link nullAttr (walk Shared.deNote headerText)
-                          ('#':ident, "")]
-  listContents <- if null subsecs || lev >= writerTOCDepth opts
-                  then return []
-                  else mapM elementToListItem subsecs
-  return [Plain headerLink, BulletList listContents]
-elementToListItem (Shared.Blk _) = return []
+addSpeakerNotesToMetaSlide :: Slide -> [Block] -> Pres (Slide, [Block])
+addSpeakerNotesToMetaSlide (Slide sldId layout@(MetadataSlide{}) spkNotes) blks =
+  do let (ntsBlks, blks') = span isNotesDiv blks
+     spkNotes' <- mconcat <$> mapM blockToSpeakerNotes ntsBlks
+     return (Slide sldId layout (spkNotes <> spkNotes'), blks')
+addSpeakerNotesToMetaSlide sld blks = return (sld, blks)
 
 makeTOCSlide :: [Block] -> Pres Slide
 makeTOCSlide blks = local (\env -> env{envCurSlideId = tocSlideId}) $ do
-  contents <- BulletList <$> mapM elementToListItem (Shared.hierarchicalize blks)
+  opts <- asks envOpts
+  let contents = toTableOfContents opts blks
   meta <- asks envMetadata
   slideLevel <- asks envSlideLevel
   let tocTitle = case lookupMetaInlines "toc-title" meta of
@@ -803,7 +787,7 @@ combineParaElems' (Just pElem') (pElem : pElems)
   | Run rPr' s' <- pElem'
   , Run rPr s <- pElem
   , rPr == rPr' =
-    combineParaElems' (Just $ Run rPr' $ s' ++ s) pElems
+    combineParaElems' (Just $ Run rPr' $ s' <> s) pElems
   | otherwise =
     pElem' : combineParaElems' (Just pElem) pElems
 
@@ -848,7 +832,8 @@ applyToSlide f slide = do
 
 replaceAnchor :: ParaElem -> Pres ParaElem
 replaceAnchor (Run rProps s)
-  | Just (ExternalTarget ('#':anchor, _)) <- rLink rProps = do
+  | Just (ExternalTarget (T.uncons -> Just ('#', anchor), _)) <- rLink rProps
+  = do
       anchorMap <- gets stAnchorMap
       -- If the anchor is not in the anchormap, we just remove the
       -- link.
@@ -860,9 +845,9 @@ replaceAnchor pe = return pe
 
 emptyParaElem :: ParaElem -> Bool
 emptyParaElem (Run _ s) =
-  null $ Shared.trim s
+  T.null $ Shared.trim s
 emptyParaElem (MathElem _ ts) =
-  null $ Shared.trim $ unTeXString ts
+  T.null $ Shared.trim $ unTeXString ts
 emptyParaElem _ = False
 
 emptyParagraph :: Paragraph -> Bool
@@ -890,12 +875,21 @@ emptyLayout layout = case layout of
     all emptyShape shapes2
 
 emptySlide :: Slide -> Bool
-emptySlide (Slide _ layout notes) = (notes == mempty) && (emptyLayout layout)
+emptySlide (Slide _ layout notes) = (notes == mempty) && emptyLayout layout
 
 blocksToPresentationSlides :: [Block] -> Pres [Slide]
 blocksToPresentationSlides blks = do
   opts <- asks envOpts
-  metadataslides <- maybeToList <$> getMetaSlide
+  mbMetadataSlide <- getMetaSlide
+  -- if the metadata slide exists, we try to add any speakerNotes
+  -- which immediately follow it. We also convert from maybe to a
+  -- list, so that it will be able to add together more easily with
+  -- the other lists of slides.
+  (metadataslides, blks') <- case mbMetadataSlide of
+                                 Just sld ->
+                                   do (s, bs) <- addSpeakerNotesToMetaSlide sld blks
+                                      return ([s], bs)
+                                 Nothing -> return ([], blks)
   -- As far as I can tell, if we want to have a variable-length toc in
   -- the future, we'll have to make it twice. Once to get the length,
   -- and a second time to include the notes slide. We can't make the
@@ -906,9 +900,9 @@ blocksToPresentationSlides blks = do
   -- For now, though, since the TOC slide is only length 1, if it
   -- exists, we'll just get the length, and then come back to make the
   -- slide later
-  blksLst <- splitBlocks blks
+  blksLst <- splitBlocks blks'
   bodySlideIds <- mapM
-                  (\n -> runUniqueSlideId $ "BodySlide" ++ show n)
+                  (\n -> runUniqueSlideId $ "BodySlide" <> tshow n)
                   (take (length blksLst) [1..] :: [Integer])
   bodyslides <- mapM
                 (\(bs, ident) ->
@@ -917,7 +911,7 @@ blocksToPresentationSlides blks = do
   endNotesSlideBlocks <- makeEndNotesSlideBlocks
   -- now we come back and make the real toc...
   tocSlides <- if writerTableOfContents opts
-               then do toc <- makeTOCSlide $ blks ++ endNotesSlideBlocks
+               then do toc <- makeTOCSlide $ blks' ++ endNotesSlideBlocks
                        return [toc]
                else return []
   -- ... and the notes slide. We test to see if the blocks are empty,
@@ -943,13 +937,26 @@ metaToDocProps meta =
 
       authors = case map Shared.stringify $ docAuthors meta of
                   [] -> Nothing
-                  ss -> Just $ intercalate ";" ss
+                  ss -> Just $ T.intercalate "; " ss
+
+      description = case map Shared.stringify $ lookupMetaBlocks "description" meta of
+                  [] -> Nothing
+                  ss -> Just $ T.intercalate "_x000d_\n" ss
+
+      customProperties' = case [(k, lookupMetaString k meta) | k <- M.keys (unMeta meta)
+                               , k `notElem` ["title", "author", "keywords", "description"
+                                             , "subject","lang","category"]] of
+                  [] -> Nothing
+                  ss -> Just ss
   in
     DocProps{ dcTitle = Shared.stringify <$> lookupMeta "title" meta
             , dcSubject = Shared.stringify <$> lookupMeta "subject" meta
             , dcCreator = authors
             , dcKeywords = keywords
+            , dcDescription = description
+            , cpCategory = Shared.stringify <$> lookupMeta "category" meta
             , dcCreated = Nothing
+            , customProperties = customProperties'
             }
 
 documentToPresentation :: WriterOptions
@@ -982,7 +989,7 @@ formatToken sty (tokType, txt) =
         Just tokSty -> applyTokStyToRunProps tokSty rProps
         Nothing     -> rProps
   in
-    Run rProps' $ T.unpack txt
+    Run rProps' txt
 
 formatSourceLine :: Style -> FormatOptions -> SourceLine -> [ParaElem]
 formatSourceLine sty _ srcLn = map (formatToken sty) srcLn
