@@ -1,27 +1,7 @@
-{-# LANGUAGE NoImplicitPrelude #-}
 {-# LANGUAGE OverloadedStrings #-}
-{-
-Copyright (C) 2008-2018 John MacFarlane
-              2012 Peter Wang
-
-This program is free software; you can redistribute it and/or modify
-it under the terms of the GNU General Public License as published by
-the Free Software Foundation; either version 2 of the License, or
-(at your option) any later version.
-
-This program is distributed in the hope that it will be useful,
-but WITHOUT ANY WARRANTY; without even the implied warranty of
-MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-GNU General Public License for more details.
-
-You should have received a copy of the GNU General Public License
-along with this program; if not, write to the Free Software
-Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--}
-
 {- |
    Module      : Text.Pandoc.Writers.Texinfo
-   Copyright   : Copyright (C) 2008-2018 John MacFarlane
+   Copyright   : Copyright (C) 2008-2020 John MacFarlane
                                2012 Peter Wang
    License     : GNU GPL, version 2 or above
 
@@ -32,7 +12,6 @@ Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 Conversion of 'Pandoc' format into Texinfo.
 -}
 module Text.Pandoc.Writers.Texinfo ( writeTexinfo ) where
-import Prelude
 import Control.Monad.Except (throwError)
 import Control.Monad.State.Strict
 import Data.Char (chr, ord)
@@ -40,24 +19,25 @@ import Data.List (maximumBy, transpose)
 import Data.Ord (comparing)
 import qualified Data.Set as Set
 import Data.Text (Text)
+import qualified Data.Text as T
 import Network.URI (unEscapeString)
 import System.FilePath
-import Text.Pandoc.Class (PandocMonad, report)
+import Text.Pandoc.Class.PandocMonad (PandocMonad, report)
 import Text.Pandoc.Definition
 import Text.Pandoc.Error
 import Text.Pandoc.ImageSize
 import Text.Pandoc.Logging
 import Text.Pandoc.Options
-import Text.Pandoc.Pretty
+import Text.DocLayout
 import Text.Pandoc.Shared
-import Text.Pandoc.Templates (renderTemplate')
+import Text.Pandoc.Templates (renderTemplate)
 import Text.Pandoc.Writers.Shared
 import Text.Printf (printf)
 
 data WriterState =
   WriterState { stStrikeout   :: Bool  -- document contains strikeout
               , stEscapeComma :: Bool -- in a context where we need @comma
-              , stIdentifiers :: Set.Set String -- header ids used already
+              , stIdentifiers :: Set.Set Text -- header ids used already
               , stOptions     :: WriterOptions -- writer options
               }
 
@@ -87,26 +67,23 @@ pandocToTexinfo options (Pandoc meta blocks) = do
   let colwidth = if writerWrapText options == WrapAuto
                     then Just $ writerColumns options
                     else Nothing
-  let render' :: Doc -> Text
-      render' = render colwidth
-  metadata <- metaToJSON options
-              (fmap render' . blockListToTexinfo)
-              (fmap render' . inlineListToTexinfo)
+  metadata <- metaToContext options
+              blockListToTexinfo
+              (fmap chomp .inlineListToTexinfo)
               meta
-  main <- blockListToTexinfo blocks
+  body <- blockListToTexinfo blocks
   st <- get
-  let body = render colwidth main
   let context = defField "body" body
               $ defField "toc" (writerTableOfContents options)
               $ defField "titlepage" titlePage
-              $
-        defField "strikeout" (stStrikeout st) metadata
-  case writerTemplate options of
-       Nothing  -> return body
-       Just tpl -> renderTemplate' tpl context
+              $ defField "strikeout" (stStrikeout st) metadata
+  return $ render colwidth $
+    case writerTemplate options of
+       Nothing  -> body
+       Just tpl -> renderTemplate tpl context
 
 -- | Escape things as needed for Texinfo.
-stringToTexinfo :: String -> String
+stringToTexinfo :: Text -> Text
 stringToTexinfo = escapeStringUsing texinfoEscapes
   where texinfoEscapes = [ ('{', "@{")
                          , ('}', "@}")
@@ -118,7 +95,7 @@ stringToTexinfo = escapeStringUsing texinfoEscapes
                          , ('\x2019', "'")
                          ]
 
-escapeCommas :: PandocMonad m => TI m Doc -> TI m Doc
+escapeCommas :: PandocMonad m => TI m (Doc Text) -> TI m (Doc Text)
 escapeCommas parser = do
   oldEscapeComma <- gets stEscapeComma
   modify $ \st -> st{ stEscapeComma = True }
@@ -127,13 +104,13 @@ escapeCommas parser = do
   return res
 
 -- | Puts contents into Texinfo command.
-inCmd :: String -> Doc -> Doc
-inCmd cmd contents = char '@' <> text cmd <> braces contents
+inCmd :: Text -> Doc Text -> Doc Text
+inCmd cmd contents = char '@' <> literal cmd <> braces contents
 
 -- | Convert Pandoc block element to Texinfo.
 blockToTexinfo :: PandocMonad m
                => Block     -- ^ Block to convert
-               -> TI m Doc
+               -> TI m (Doc Text)
 
 blockToTexinfo Null = return empty
 
@@ -143,13 +120,14 @@ blockToTexinfo (Plain lst) =
   inlineListToTexinfo lst
 
 -- title beginning with fig: indicates that the image is a figure
-blockToTexinfo (Para [Image attr txt (src,'f':'i':'g':':':tit)]) = do
-  capt <- if null txt
-             then return empty
-             else (\c -> text "@caption" <> braces c) `fmap`
-                    inlineListToTexinfo txt
-  img  <- inlineToTexinfo (Image attr txt (src,tit))
-  return $ text "@float" $$ img $$ capt $$ text "@end float"
+blockToTexinfo (Para [Image attr txt (src,tgt)])
+  | Just tit <- T.stripPrefix "fig:" tgt = do
+      capt <- if null txt
+              then return empty
+              else (\c -> text "@caption" <> braces c) `fmap`
+                     inlineListToTexinfo txt
+      img  <- inlineToTexinfo (Image attr txt (src,tit))
+      return $ text "@float" $$ img $$ capt $$ text "@end float"
 
 blockToTexinfo (Para lst) =
   inlineListToTexinfo lst    -- this is handled differently from Plain in blockListToTexinfo
@@ -166,13 +144,13 @@ blockToTexinfo (BlockQuote lst) = do
 blockToTexinfo (CodeBlock _ str) =
   return $ blankline $$
          text "@verbatim" $$
-         flush (text str) $$
+         flush (literal str) $$
          text "@end verbatim" <> blankline
 
 blockToTexinfo b@(RawBlock f str)
-  | f == "texinfo" = return $ text str
+  | f == "texinfo" = return $ literal str
   | f == "latex" || f == "tex" =
-                      return $ text "@tex" $$ text str $$ text "@end tex"
+                      return $ text "@tex" $$ literal str $$ text "@end tex"
   | otherwise      = do
       report $ BlockNotRendered b
       return empty
@@ -225,23 +203,25 @@ blockToTexinfo (Header 0 _ lst) = do
   return $ text "@node Top" $$
            text "@top " <> txt <> blankline
 
-blockToTexinfo (Header level _ lst)
+blockToTexinfo (Header level (ident,_,_) lst)
   | level < 1 || level > 4 = blockToTexinfo (Para lst)
   | otherwise = do
     node <- inlineListForNode lst
     txt <- inlineListToTexinfo lst
     idsUsed <- gets stIdentifiers
     opts <- gets stOptions
-    let id' = uniqueIdent (writerExtensions opts) lst idsUsed
+    let id' = if T.null ident
+                 then uniqueIdent (writerExtensions opts) lst idsUsed
+                 else ident
     modify $ \st -> st{ stIdentifiers = Set.insert id' idsUsed }
     sec <- seccmd level
     return $ if (level > 0) && (level <= 4)
                 then blankline <> text "@node " <> node $$
-                     text sec <> txt $$
-                     text "@anchor" <> braces (text $ '#':id')
+                     literal sec <> txt $$
+                     text "@anchor" <> braces (literal $ "#" <> id')
                 else txt
     where
-      seccmd :: PandocMonad m => Int -> TI m String
+      seccmd :: PandocMonad m => Int -> TI m Text
       seccmd 1 = return "@chapter "
       seccmd 2 = return "@section "
       seccmd 3 = return "@subsection "
@@ -257,7 +237,7 @@ blockToTexinfo (Table caption aligns widths heads rows) = do
   colDescriptors <-
     if all (== 0) widths
        then do -- use longest entry instead of column widths
-            cols <- mapM (mapM (liftM (render Nothing . hcat) . mapM blockToTexinfo)) $
+            cols <- mapM (mapM (liftM (T.unpack . render Nothing . hcat) . mapM blockToTexinfo)) $
                         transpose $ heads : rows
             return $ concatMap ((\x -> "{"++x++"} ") .  maximumBy (comparing length)) cols
        else return $ "@columnfractions " ++ concatMap (printf "%.2f ") widths
@@ -275,29 +255,28 @@ blockToTexinfo (Table caption aligns widths heads rows) = do
 tableHeadToTexinfo :: PandocMonad m
                    => [Alignment]
                    -> [[Block]]
-                   -> TI m Doc
+                   -> TI m (Doc Text)
 tableHeadToTexinfo = tableAnyRowToTexinfo "@headitem "
 
 tableRowToTexinfo :: PandocMonad m
                   => [Alignment]
                   -> [[Block]]
-                  -> TI m Doc
+                  -> TI m (Doc Text)
 tableRowToTexinfo = tableAnyRowToTexinfo "@item "
 
 tableAnyRowToTexinfo :: PandocMonad m
-                     => String
+                     => Text
                      -> [Alignment]
                      -> [[Block]]
-                     -> TI m Doc
+                     -> TI m (Doc Text)
 tableAnyRowToTexinfo itemtype aligns cols =
-  zipWithM alignedBlock aligns cols >>=
-  return . (text itemtype $$) . foldl (\row item -> row $$
-  (if isEmpty row then empty else text " @tab ") <> item) empty
+  (literal itemtype $$) . foldl (\row item -> row $$
+  (if isEmpty row then empty else text " @tab ") <> item) empty <$> zipWithM alignedBlock aligns cols
 
 alignedBlock :: PandocMonad m
              => Alignment
              -> [Block]
-             -> TI m Doc
+             -> TI m (Doc Text)
 -- XXX @flushleft and @flushright text won't get word wrapped.  Since word
 -- wrapping is more important than alignment, we ignore the alignment.
 alignedBlock _ = blockListToTexinfo
@@ -314,7 +293,7 @@ alignedBlock _ col = blockListToTexinfo col
 -- | Convert Pandoc block elements to Texinfo.
 blockListToTexinfo :: PandocMonad m
                    => [Block]
-                   -> TI m Doc
+                   -> TI m (Doc Text)
 blockListToTexinfo [] = return empty
 blockListToTexinfo (x:xs) = do
   x' <- blockToTexinfo x
@@ -356,7 +335,7 @@ collectNodes level (x:xs) =
 
 makeMenuLine :: PandocMonad m
              => Block
-             -> TI m Doc
+             -> TI m (Doc Text)
 makeMenuLine (Header _ _ lst) = do
   txt <- inlineListForNode lst
   return $ text "* " <> txt <> text "::"
@@ -364,7 +343,7 @@ makeMenuLine _ = throwError $ PandocSomeError "makeMenuLine called with non-Head
 
 listItemToTexinfo :: PandocMonad m
                   => [Block]
-                  -> TI m Doc
+                  -> TI m (Doc Text)
 listItemToTexinfo lst = do
   contents <- blockListToTexinfo lst
   let spacer = case reverse lst of
@@ -374,7 +353,7 @@ listItemToTexinfo lst = do
 
 defListItemToTexinfo :: PandocMonad m
                      => ([Inline], [[Block]])
-                     -> TI m Doc
+                     -> TI m (Doc Text)
 defListItemToTexinfo (term, defs) = do
     term' <- inlineListToTexinfo term
     let defToTexinfo bs = do d <- blockListToTexinfo bs
@@ -387,15 +366,15 @@ defListItemToTexinfo (term, defs) = do
 -- | Convert list of inline elements to Texinfo.
 inlineListToTexinfo :: PandocMonad m
                     => [Inline]  -- ^ Inlines to convert
-                    -> TI m Doc
+                    -> TI m (Doc Text)
 inlineListToTexinfo lst = hcat <$> mapM inlineToTexinfo lst
 
 -- | Convert list of inline elements to Texinfo acceptable for a node name.
 inlineListForNode :: PandocMonad m
                   => [Inline]  -- ^ Inlines to convert
-                  -> TI m Doc
-inlineListForNode = return . text . stringToTexinfo .
-                    filter (not . disallowedInNode) . stringify
+                  -> TI m (Doc Text)
+inlineListForNode = return . literal . stringToTexinfo .
+                    T.filter (not . disallowedInNode) . stringify
 
 -- periods, commas, colons, and parentheses are disallowed in node names
 disallowedInNode :: Char -> Bool
@@ -404,7 +383,7 @@ disallowedInNode c = c `elem` (".,:()" :: String)
 -- | Convert inline element to Texinfo
 inlineToTexinfo :: PandocMonad m
                 => Inline    -- ^ Inline to convert
-                -> TI m Doc
+                -> TI m (Doc Text)
 
 inlineToTexinfo (Span _ lst) =
   inlineListToTexinfo lst
@@ -432,7 +411,7 @@ inlineToTexinfo (SmallCaps lst) =
   inCmd "sc" <$> inlineListToTexinfo lst
 
 inlineToTexinfo (Code _ str) =
-  return $ text $ "@code{" ++ stringToTexinfo str ++ "}"
+  return $ literal $ "@code{" <> stringToTexinfo str <> "}"
 
 inlineToTexinfo (Quoted SingleQuote lst) = do
   contents <- inlineListToTexinfo lst
@@ -444,12 +423,12 @@ inlineToTexinfo (Quoted DoubleQuote lst) = do
 
 inlineToTexinfo (Cite _ lst) =
   inlineListToTexinfo lst
-inlineToTexinfo (Str str) = return $ text (stringToTexinfo str)
-inlineToTexinfo (Math _ str) = return $ inCmd "math" $ text str
+inlineToTexinfo (Str str) = return $ literal (stringToTexinfo str)
+inlineToTexinfo (Math _ str) = return $ inCmd "math" $ literal str
 inlineToTexinfo il@(RawInline f str)
   | f == "latex" || f == "tex" =
-                      return $ text "@tex" $$ text str $$ text "@end tex"
-  | f == "texinfo" =  return $ text str
+                      return $ text "@tex" $$ literal str $$ text "@end tex"
+  | f == "texinfo" =  return $ literal str
   | otherwise      =  do
       report $ InlineNotRendered il
       return empty
@@ -462,35 +441,36 @@ inlineToTexinfo SoftBreak = do
       WrapPreserve -> return cr
 inlineToTexinfo Space = return space
 
-inlineToTexinfo (Link _ txt (src@('#':_), _)) = do
-  contents <- escapeCommas $ inlineListToTexinfo txt
-  return $ text "@ref" <>
-           braces (text (stringToTexinfo src) <> text "," <> contents)
-inlineToTexinfo (Link _ txt (src, _)) =
-  case txt of
-        [Str x] | escapeURI x == src ->  -- autolink
-             return $ text $ "@url{" ++ x ++ "}"
-        _ -> do contents <- escapeCommas $ inlineListToTexinfo txt
-                let src1 = stringToTexinfo src
-                return $ text ("@uref{" ++ src1 ++ ",") <> contents <>
-                         char '}'
+inlineToTexinfo (Link _ txt (src, _))
+  | Just ('#', _) <- T.uncons src = do
+      contents <- escapeCommas $ inlineListToTexinfo txt
+      return $ text "@ref" <>
+        braces (literal (stringToTexinfo src) <> text "," <> contents)
+  | otherwise = case txt of
+      [Str x] | escapeURI x == src ->  -- autolink
+                  return $ literal $ "@url{" <> x <> "}"
+      _ -> do
+        contents <- escapeCommas $ inlineListToTexinfo txt
+        let src1 = stringToTexinfo src
+        return $ literal ("@uref{" <> src1 <> ",") <> contents <>
+                 char '}'
 
 inlineToTexinfo (Image attr alternate (source, _)) = do
   content <- escapeCommas $ inlineListToTexinfo alternate
   opts <- gets stOptions
   let showDim dim = case dimension dim attr of
-                      (Just (Pixel a))   -> showInInch opts (Pixel a) ++ "in"
+                      (Just (Pixel a))   -> showInInch opts (Pixel a) <> "in"
                       (Just (Percent _)) -> ""
-                      (Just d)           -> show d
+                      (Just d)           -> tshow d
                       Nothing            -> ""
-  return $ text ("@image{" ++ base ++ ',':(showDim Width) ++ ',':(showDim Height) ++ ",")
-           <> content <> text "," <> text (ext ++ "}")
+  return $ literal ("@image{" <> base <> "," <> showDim Width <> "," <> showDim Height <> ",")
+           <> content <> text "," <> literal (ext <> "}")
   where
-    ext     = drop 1 $ takeExtension source'
-    base    = dropExtension source'
+    ext     = T.drop 1 $ T.pack $ takeExtension source'
+    base    = T.pack $ dropExtension source'
     source' = if isURI source
-                 then source
-                 else unEscapeString source
+              then T.unpack source
+              else unEscapeString $ T.unpack source
 
 inlineToTexinfo (Note contents) = do
   contents' <- blockListToTexinfo contents
