@@ -1,4 +1,16 @@
 {-# LANGUAGE NoImplicitPrelude #-}
+{-# LANGUAGE OverloadedStrings #-}
+{- |
+   Module      : Tests.Readers.Docx
+   Copyright   : © 2017-2020 Jesse Rosenthal, John MacFarlane
+   License     : GNU GPL, version 2 or above
+
+   Maintainer  : Jesse Rosenthal <jrosenthal@jhu.edu>
+   Stability   : alpha
+   Portability : portable
+
+Tests for the word docx reader.
+-}
 module Tests.Readers.Docx (tests) where
 
 import Prelude
@@ -35,7 +47,7 @@ instance ToString NoNormPandoc where
    where s = case d of
                   NoNormPandoc (Pandoc (Meta m) _)
                     | M.null m  -> Nothing
-                    | otherwise -> Just "" -- need this to get meta output
+                    | otherwise -> Just mempty -- need this to get meta output
 
 instance ToPandoc NoNormPandoc where
   toPandoc = unNoNorm
@@ -68,7 +80,7 @@ testForWarningsWithOptsIO opts name docxFile expected = do
   df <- B.readFile docxFile
   logs <-  runIOorExplode $ setVerbosity ERROR >> readDocx opts df >> P.getLog
   let warns = [m | DocxParserWarning m <- logs]
-  return $ test id name (unlines warns, unlines expected)
+  return $ test id name (T.unlines warns, unlines expected)
 
 testForWarningsWithOpts :: ReaderOptions -> String -> FilePath -> [String] -> TestTree
 testForWarningsWithOpts opts name docxFile expected =
@@ -79,8 +91,8 @@ testForWarningsWithOpts opts name docxFile expected =
 
 getMedia :: FilePath -> FilePath -> IO (Maybe B.ByteString)
 getMedia archivePath mediaPath = do
-  zf <- B.readFile archivePath >>= return . toArchive
-  return $ findEntryByPath ("word/" ++ mediaPath) zf >>= (Just . fromEntry)
+  zf <- toArchive <$> B.readFile archivePath
+  return $ fromEntry <$> findEntryByPath ("word/" ++ mediaPath) zf
 
 compareMediaPathIO :: FilePath -> MediaBag -> FilePath -> IO Bool
 compareMediaPathIO mediaPath mediaBag docxPath = do
@@ -115,7 +127,13 @@ testMediaBag :: String -> FilePath -> TestTree
 testMediaBag name docxFile = unsafePerformIO $ testMediaBagIO name docxFile
 
 tests :: [TestTree]
-tests = [ testGroup "inlines"
+tests = [ testGroup "document"
+          [ testCompare
+            "allow different document.xml file as defined in _rels/.rels"
+            "docx/alternate_document_path.docx"
+            "docx/alternate_document_path.native"
+          ]
+        , testGroup "inlines"
           [ testCompare
             "font formatting"
             "docx/inline_formatting.docx"
@@ -168,6 +186,10 @@ tests = [ testGroup "inlines"
             "move trailing spaces outside of formatting"
             "docx/trailing_spaces_in_formatting.docx"
             "docx/trailing_spaces_in_formatting.native"
+          , testCompare
+            "remove trailing spaces from last inline"
+            "docx/trim_last_inline.docx"
+            "docx/trim_last_inline.native"
           , testCompare
             "inline code (with VerbatimChar style)"
             "docx/inline_code.docx"
@@ -180,6 +202,10 @@ tests = [ testGroup "inlines"
             "inlines inside of Structured Document Tags"
             "docx/sdt_elements.docx"
             "docx/sdt_elements.native"
+          , testCompare
+            "Structured Document Tags in footnotes"
+            "docx/sdt_in_footnote.docx"
+            "docx/sdt_in_footnote.native"
           , testCompare
             "nested Structured Document Tags"
             "docx/nested_sdt.docx"
@@ -230,6 +256,14 @@ tests = [ testGroup "inlines"
             "lists"
             "docx/lists.docx"
             "docx/lists.native"
+          , testCompare
+            "compact lists"
+            "docx/lists-compact.docx"
+            "docx/lists-compact.native"
+          , testCompare
+            "lists with level overrides"
+            "docx/lists_level_override.docx"
+            "docx/lists_level_override.native"
           , testCompare
             "lists continuing after interruption"
             "docx/lists_continuing.docx"
@@ -238,6 +272,10 @@ tests = [ testGroup "inlines"
             "lists restarting after interruption"
             "docx/lists_restarting.docx"
             "docx/lists_restarting.native"
+          , testCompare
+            "sublists reset numbering to 1"
+            "docx/lists_sublist_reset.docx"
+            "docx/lists_sublist_reset.native"
           , testCompare
             "definition lists"
             "docx/definition_list.docx"
@@ -396,6 +434,11 @@ tests = [ testGroup "inlines"
             "custom styles (`+styles`) enabled"
             "docx/custom-style-reference.docx"
             "docx/custom-style-with-styles.native"
+          , testCompareWithOpts
+            def{readerExtensions=extensionsFromList [Ext_styles]}
+            "custom styles (`+styles`): Compact style is removed from output"
+            "docx/compact-style-removal.docx"
+            "docx/compact-style-removal.native"
           ]
         , testGroup "metadata"
           [ testCompareWithOpts def{readerStandalone=True}
@@ -407,5 +450,4 @@ tests = [ testGroup "inlines"
             "docx/metadata_after_normal.docx"
             "docx/metadata_after_normal.native"
           ]
-
         ]
