@@ -1,5 +1,16 @@
 {-# LANGUAGE NoImplicitPrelude #-}
 {-# LANGUAGE OverloadedStrings #-}
+{- |
+   Module      : Tests.Readers.HTML
+   Copyright   : © 2006-2020 John MacFarlane
+   License     : GNU GPL, version 2 or above
+
+   Maintainer  : John MacFarlane <jgm@berkeley.edu>
+   Stability   : alpha
+   Portability : portable
+
+Tests for the HTML reader.
+-}
 module Tests.Readers.HTML (tests) where
 
 import Prelude
@@ -10,6 +21,7 @@ import Test.Tasty.QuickCheck
 import Test.Tasty.Options (IsOption(defaultValue))
 import Tests.Helpers
 import Text.Pandoc
+import Text.Pandoc.Shared (isHeaderBlock)
 import Text.Pandoc.Arbitrary ()
 import Text.Pandoc.Builder
 import Text.Pandoc.Walk (walk)
@@ -24,6 +36,9 @@ makeRoundTrip :: Block -> Block
 makeRoundTrip CodeBlock{} = Para [Str "code block was here"]
 makeRoundTrip LineBlock{} = Para [Str "line block was here"]
 makeRoundTrip RawBlock{} = Para [Str "raw block was here"]
+makeRoundTrip (Div attr bs) = Div attr $ filter (not . isHeaderBlock) bs
+-- avoids round-trip failures related to makeSections
+-- e.g. with [Div ("loc",[],[("a","11"),("b_2","a b c")]) [Header 3 ("",[],[]) []]]
 makeRoundTrip x           = x
 
 removeRawInlines :: Inline -> Inline
@@ -72,12 +87,24 @@ tests = [ testGroup "base tag"
           , test htmlNativeDivs "<main role=X> becomes <div role=X>" $ "<main role=foobar>hello</main>" =?>
             doc (divWith ("", [], [("role", "foobar")]) (plain (text "hello")))
           , test htmlNativeDivs "<main> has attributes preserved" $ "<main id=foo class=bar data-baz=qux>hello</main>" =?>
-            doc (divWith ("foo", ["bar"], [("role", "main"), ("data-baz", "qux")]) (plain (text "hello")))
+            doc (divWith ("foo", ["bar"], [("role", "main"), ("baz", "qux")]) (plain (text "hello")))
           , test htmlNativeDivs "<main> closes <p>" $ "<p>hello<main>main content</main>" =?>
             doc (para (text "hello") <> divWith ("", [], [("role", "main")]) (plain (text "main content")))
           , test htmlNativeDivs "<main> followed by text" $ "<main>main content</main>non-main content" =?>
             doc (divWith ("", [], [("role", "main")]) (plain (text "main content")) <> plain (text "non-main content"))
           ]
+        , testGroup "samp"
+          [
+            test html "inline samp block" $
+            "<samp>Answer is 42</samp>" =?>
+            plain (codeWith ("",["sample"],[]) "Answer is 42")
+          ]
+        , testGroup "var"
+        [
+          test html "inline var block" $
+          "<var>result</var>" =?>
+          plain (codeWith ("",["variable"],[]) "result")
+        ]
         , askOption $ \(QuickCheckTests numtests) ->
             testProperty "Round trip" $
               withMaxSuccess (if QuickCheckTests numtests == defaultValue
