@@ -1,5 +1,16 @@
 {-# LANGUAGE NoImplicitPrelude #-}
 {-# LANGUAGE OverloadedStrings #-}
+{- |
+   Module      : Tests.Readers.LaTeX
+   Copyright   : © 2006-2020 John MacFarlane
+   License     : GNU GPL, version 2 or above
+
+   Maintainer  : John MacFarlane <jgm@berkeley.edu>
+   Stability   : alpha
+   Portability : portable
+
+Tests for the LaTeX reader.
+-}
 module Tests.Readers.LaTeX (tests) where
 
 import Prelude
@@ -158,10 +169,10 @@ tests = [ testGroup "tokenization"
           testGroup "Character Escapes"
           [ "Two-character escapes" =:
             mconcat ["^^" <> T.pack [i,j] | i <- hex, j <- hex] =?>
-            para (str ['\0'..'\255'])
+            para (str $ T.pack ['\0'..'\255'])
           , "One-character escapes" =:
             mconcat ["^^" <> T.pack [i] | i <- hex] =?>
-            para (str $ ['p'..'y']++['!'..'&'])
+            para (str $ T.pack $ ['p'..'y']++['!'..'&'])
           ]
         , testGroup "memoir scene breaks"
           [ "plainbreak" =:
@@ -244,7 +255,7 @@ baseCitation = Citation{ citationId      = "item1"
                        }
 
 rt :: String -> Inlines
-rt = rawInline "latex"
+rt = rawInline "latex" . T.pack
 
 natbibCitations :: TestTree
 natbibCitations = testGroup "natbib"
