@@ -1,5 +1,16 @@
 {-# LANGUAGE NoImplicitPrelude #-}
 {-# LANGUAGE OverloadedStrings #-}
+{- |
+   Module      : Tests.Readers.Markdown
+   Copyright   : © 2006-2020 John MacFarlane
+   License     : GNU GPL, version 2 or above
+
+   Maintainer  : John MacFarlane <jgm@berkeley.edu>
+   Stability   : alpha
+   Portability : portable
+
+Tests for the Markdown reader.
+-}
 module Tests.Readers.Markdown (tests) where
 
 import Prelude
@@ -42,7 +53,8 @@ autolink :: String -> Inlines
 autolink = autolinkWith ("",["uri"],[])
 
 autolinkWith :: Attr -> String -> Inlines
-autolinkWith attr s = linkWith attr s "" (str s)
+autolinkWith attr s = linkWith attr s' "" (str s')
+  where s' = T.pack s
 
 bareLinkTests :: [(Text, Inlines)]
 bareLinkTests =
@@ -319,9 +331,9 @@ tests = [ testGroup "inline code"
                        Ext_literate_haskell pandocExtensions })
               "inverse bird tracks and html" $
               "> a\n\n< b\n\n<div>\n"
-              =?> codeBlockWith ("",["sourceCode","literate","haskell"],[]) "a"
+              =?> codeBlockWith ("",["haskell","literate"],[]) "a"
                   <>
-                  codeBlockWith ("",["sourceCode","haskell"],[]) "b"
+                  codeBlockWith ("",["haskell"],[]) "b"
                   <>
                   rawBlock "html" "<div>\n\n"
           ]
