@@ -1,5 +1,16 @@
 {-# LANGUAGE NoImplicitPrelude #-}
 {-# LANGUAGE OverloadedStrings #-}
+{- |
+   Module      : Tests.Readers.Muse
+   Copyright   : © 2017-2020 Alexander Krotov
+   License     : GNU GPL, version 2 or above
+
+   Maintainer  : Alexander Krotov <ilabdsf@gmail.com>
+   Stability   : alpha
+   Portability : portable
+
+Tests for the Muse reader.
+-}
 module Tests.Readers.Muse (tests) where
 
 import Prelude
@@ -122,6 +133,38 @@ tests =
         "**foo *bar* baz**" =?>
         para (strong (text "foo " <> emph (text "bar") <> text " baz"))
 
+      , "Opening asterisk can't be preceded by another one" =:
+        "**foo*" =?>
+        para "**foo*"
+
+      , "Asterisk between words does not terminate emphasis" =:
+        "*foo*bar*" =?>
+        para (emph "foo*bar")
+
+      , "Two asterisks between words do not terminate emphasis" =:
+        "*foo**bar*" =?>
+        para (emph "foo**bar")
+
+      , "Three asterisks between words do not terminate emphasis" =:
+        "*foo***bar*" =?>
+        para (emph "foo***bar")
+
+      , "Two asterisks between words do not terminate strong" =:
+        "**foo**bar**" =?>
+        para (strong "foo**bar")
+
+      , "Three asterisks between words do not terminate strong" =:
+        "**foo***bar**" =?>
+        para (strong "foo***bar")
+
+      , "Three asterisks between words do not terminate strong emphasis" =:
+        "***foo***bar***" =?>
+        para (strong . emph $ "foo***bar")
+
+      , "Six asterisks between words do not terminate strong emphasis" =:
+        "***foo******bar***" =?>
+        para (strong . emph $ "foo******bar")
+
       , test emacsMuse "Underline"
         ("_Underline_" =?> para (underlineSpan "Underline"))
 
@@ -215,9 +258,19 @@ tests =
 
       , "Verbatim tag after text" =: "Foo <verbatim>bar</verbatim>" =?> para "Foo bar"
 
+      , "Verbatim tag escapes block level markup" =:
+        T.unlines [ "Foo <verbatim>bar"
+                  , "* Not a heading"
+                  , "</verbatim>baz"
+                  ] =?>
+        para "Foo bar\n* Not a heading\nbaz"
+
       , "Class tag" =: "<class name=\"foo\">bar</class>" =?> para (spanWith ("", ["foo"], []) "bar")
       , "Class tag without name" =: "<class>foobar</class>" =?> para (spanWith ("", [], []) "foobar")
 
+      , "RTL" =: "<<<foo bar>>>" =?> para (spanWith ("", [], [("dir", "rtl")]) "foo bar")
+      , "LTR" =: ">>>foo bar<<<" =?> para (spanWith ("", [], [("dir", "ltr")]) "foo bar")
+
       -- <em> tag should match with the last </em> tag, not verbatim one
       , "Nested \"</em>\" inside em tag" =: "<em>foo<verbatim></em></verbatim>bar</em>" =?> para (emph "foo</em>bar")
 
@@ -271,6 +324,12 @@ tests =
         , "No footnotes inside links" =:
           "[[https://amusewiki.org/][foo[1]]" =?>
           para (link "https://amusewiki.org/" "" (text "foo[1"))
+        , "Image inside link" =:
+          "[[https://amusewiki.org/][Image [[image.png][with it's own description]] inside link description]]" =?>
+          para (link "https://amusewiki.org/" "" (text "Image " <> image "image.png" "" (text "with it's own description") <> text " inside link description"))
+        , "Link inside image description" =:
+          "[[image.jpg][Image from [[https://amusewiki.org/]]]]" =?>
+          para (image "image.jpg" "" (text "Image from " <> link "https://amusewiki.org/" "" (str "https://amusewiki.org/")))
         ]
 
       , testGroup "Literal"
@@ -726,6 +785,11 @@ tests =
           Pandoc (setMeta "title" (MetaInlines $ toList "Document title") $
                   setMeta "notes" (MetaInlines $ toList "First line\nand second line") $
                   setMeta "author" (MetaInlines $ toList "Name") nullMeta) mempty
+        , "Amusewiki's #cover is translated to pandoc's #cover-image" =:
+          "#cover cover.png" =?>
+          let titleInline = toList "cover.png"
+              meta = setMeta "cover-image" (MetaInlines titleInline) nullMeta
+          in Pandoc meta mempty
         ]
       , testGroup "Anchors"
         [ "Anchor" =:
@@ -761,6 +825,23 @@ tests =
           para (text "Here is a footnote" <>
                 note (para "Footnote contents") <>
                 str ".")
+        , "Simple secondary footnote" =:
+          T.unlines [ "Here is a secondary note{1}."
+                    , ""
+                    , "{1} Secondary note contents"
+                    ] =?>
+          para (text "Here is a secondary note" <>
+                note (para "Secondary note contents") <>
+                str ".")
+        , "Missing footnote" =: "Foo[1]" =?> para "Foo[1]"
+        , "Missing secondary note" =: "Foo{1}" =?> para "Foo{1}"
+        , "Wrong note type" =:
+          T.unlines [ "Here is a secondary note{1}"
+                    , ""
+                    , "Footnote contents[1]"
+                    ] =?>
+          para "Here is a secondary note{1}" <>
+          para "Footnote contents[1]"
         , "Recursive footnote" =:
           T.unlines [ "Start recursion here[1]"
                     , ""
