@@ -1,5 +1,16 @@
 {-# LANGUAGE NoImplicitPrelude #-}
 {-# LANGUAGE OverloadedStrings #-}
+{- |
+   Module      : Tests.Readers.Org.Directive
+   Copyright   : © 2014-2020 Albert Krewinkel
+   License     : GNU GPL, version 2 or above
+
+   Maintainer  : Albert Krewinkel <albert@zeitkraut.de>
+   Stability   : alpha
+   Portability : portable
+
+Tests parsing of org directives (like @#+OPTIONS@).
+-}
 module Tests.Readers.Org.Directive (tests) where
 
 import Prelude
@@ -173,6 +184,24 @@ tests =
                   ] =?>
         headerWith ("wichtig", mempty, mempty) 1 "Wichtig"
       ]
+
+    , testGroup "unknown options"
+      [ "unknown options are ignored" =:
+          T.unlines [ "#+OPTIONS: does-not-exist:t "] =?>
+          (mempty :: Pandoc)
+
+      , "highlighting after unknown option" =:
+          T.unlines [ "#+OPTIONS: nope"
+                    , "/yup/"
+                    ] =?>
+          para (emph "yup")
+
+      , "unknown option interleaved with known" =:
+          T.unlines [ "#+OPTIONS: tags:nil foo:bar todo:nil"
+                    , "* DONE ignore things  :easy:"
+                    ] =?>
+          headerWith ("ignore-things", [], mempty) 1 "ignore things"
+      ]
     ]
 
   , testGroup "Include"
@@ -192,10 +221,15 @@ tests =
        headerWith ("level3", [], []) 3 "Level3")
 
     , testWithFiles [("./level3.org", "*** Level3\n\n")]
-      "Minlevel shifts level"
+      "Minlevel shifts level leftward"
       (T.unlines [ "#+include: \"level3.org\" :minlevel 1" ] =?>
        headerWith ("level3", [], []) 1 "Level3")
 
+    , testWithFiles [("./level1.org", "* Level1\n\n")]
+      "Minlevel shifts level rightward"
+      (T.unlines [ "#+include: \"level1.org\" :minlevel 3" ] =?>
+       headerWith ("level1", [], []) 3 "Level1")
+
     , testWithFiles [("./src.hs", "putStrLn outString\n")]
       "Include file as source code snippet"
       (T.unlines [ "#+include: \"src.hs\" src haskell" ] =?>
