@@ -1,5 +1,16 @@
 {-# LANGUAGE NoImplicitPrelude #-}
 {-# LANGUAGE OverloadedStrings #-}
+{- |
+   Module      : Tests.Readers.Org.Inline
+   Copyright   : © 2014-2020 Albert Krewinkel
+   License     : GNU GPL, version 2 or above
+
+   Maintainer  : Albert Krewinkel <albert@zeitkraut.de>
+   Stability   : alpha
+   Portability : portable
+
+Tests parsing of org inlines.
+-}
 module Tests.Readers.Org.Inline (tests) where
 
 import Prelude
@@ -184,26 +195,26 @@ tests =
 
   , testGroup "Images"
     [ "Image" =:
-        "[[./sunset.jpg]]" =?>
-    para (image "./sunset.jpg" "" "")
+      "[[./sunset.jpg]]" =?>
+      para (image "./sunset.jpg" "" "")
 
     , "Image with explicit file: prefix" =:
-        "[[file:sunrise.jpg]]" =?>
-    para (image "sunrise.jpg" "" "")
+      "[[file:sunrise.jpg]]" =?>
+      para (image "sunrise.jpg" "" "")
 
     , "Multiple images within a paragraph" =:
-        T.unlines [ "[[file:sunrise.jpg]]"
-                  , "[[file:sunset.jpg]]"
-                  ] =?>
-    para ((image "sunrise.jpg" "" "")
+      T.unlines [ "[[file:sunrise.jpg]]"
+                , "[[file:sunset.jpg]]"
+                ] =?>
+      para (image "sunrise.jpg" "" ""
              <> softbreak
-             <> (image "sunset.jpg" "" ""))
+             <> image "sunset.jpg" "" "")
 
     , "Image with html attributes" =:
-        T.unlines [ "#+ATTR_HTML: :width 50%"
-                  , "[[file:guinea-pig.gif]]"
-                  ] =?>
-    para (imageWith ("", [], [("width", "50%")]) "guinea-pig.gif" "" "")
+      T.unlines [ "#+ATTR_HTML: :width 50%"
+                , "[[file:guinea-pig.gif]]"
+                ] =?>
+      para (imageWith ("", [], [("width", "50%")]) "guinea-pig.gif" "" "")
     ]
 
   , "Explicit link" =:
@@ -215,6 +226,10 @@ tests =
       "[[http://zeitlens.com/]]" =?>
     para (link "http://zeitlens.com/" "" "http://zeitlens.com/")
 
+  , "Internal self-link (reference)" =:
+    "[[#rabbit]]" =?>
+    para (link "#rabbit" "" "#rabbit")
+
   , "Absolute file link" =:
       "[[/url][hi]]" =?>
     para (link "file:///url" "" "hi")
@@ -235,6 +250,10 @@ tests =
       "[[http://example.com][./logo.png]]" =?>
     para (link "http://example.com" "" (image "./logo.png" "" ""))
 
+  , "Link to image" =:
+    "[[https://example.com/image.jpg][Look!]]" =?>
+    para (link "https://example.com/image.jpg" "" (str "Look!"))
+
   , "Plain link" =:
       "Posts on http://zeitlens.com/ can be funny at times." =?>
     para (spcSep [ "Posts", "on"
@@ -309,9 +328,13 @@ tests =
       "\\emph{Emphasis mine}" =?>
       para (emph "Emphasis mine")
 
-  , "Inline LaTeX math symbol" =:
-      "\\tau" =?>
-      para (emph "τ")
+  , "Inline math symbols" =:
+      "\\tau \\oplus \\alpha" =?>
+      para "τ ⊕ α"
+
+  , "Inline LaTeX math command" =:
+      "\\crarr" =?>
+      para "↵"
 
   , "Unknown inline LaTeX command" =:
       "\\notacommand{foo}" =?>
