@@ -1,4 +1,16 @@
 {-# LANGUAGE NoImplicitPrelude #-}
+{-# LANGUAGE OverloadedStrings #-}
+{- |
+   Module      : Tests.Readers.Org.Shared
+   Copyright   : © 2014-2020 Albert Krewinkel
+   License     : GNU GPL, version 2 or above
+
+   Maintainer  : Albert Krewinkel <albert@zeitkraut.de>
+   Stability   : alpha
+   Portability : portable
+
+Helper functions used by other org tests.
+-}
 module Tests.Readers.Org.Shared
   ( (=:)
   , org
@@ -27,5 +39,5 @@ spcSep :: [Inlines] -> Inlines
 spcSep = mconcat . intersperse space
 
 -- | Create a span for the given tag.
-tagSpan :: String -> Inlines
+tagSpan :: Text -> Inlines
 tagSpan t = spanWith ("", ["tag"], [("tag-name", t)]) . smallcaps $ str t
