@@ -6,6 +6,9 @@ import Tests.Writers.OOXML (ooxmlTest)
 import Text.Pandoc
 import Test.Tasty
 import System.FilePath
+import Text.DocTemplates (ToContext(toVal), Context(..))
+import qualified Data.Map as M
+import Data.Text (pack)
 
 -- templating is important enough, and can break enough things, that
 -- we want to run all our tests with both default formatting and a
@@ -59,6 +62,10 @@ tests = groupPptxTests [ pptxTests "Inline formatting"
                          def
                          "pptx/lists.native"
                          "pptx/lists.pptx"
+                       , pptxTests "start ordered list at specified num"
+                         def
+                         "pptx/start_numbering_at.native"
+                         "pptx/start_numbering_at.pptx"
                        , pptxTests "tables"
                          def
                          "pptx/tables.native"
@@ -91,6 +98,14 @@ tests = groupPptxTests [ pptxTests "Inline formatting"
                          def
                          "pptx/speaker_notes_afterseps.native"
                          "pptx/speaker_notes_afterseps.pptx"
+                       , pptxTests "speaker notes after a separating header"
+                         def
+                         "pptx/speaker_notes_afterheader.native"
+                         "pptx/speaker_notes_afterheader.pptx"
+                       , pptxTests "speaker notes after metadata"
+                         def
+                         "pptx/speaker_notes_after_metadata.native"
+                         "pptx/speaker_notes_after_metadata.pptx"
                        , pptxTests "remove empty slides"
                          def
                          "pptx/remove_empty_slides.native"
@@ -99,4 +114,21 @@ tests = groupPptxTests [ pptxTests "Inline formatting"
                          def
                          "pptx/raw_ooxml.native"
                          "pptx/raw_ooxml.pptx"
+                       , pptxTests "metadata, custom properties"
+                         def
+                         "pptx/document-properties.native"
+                         "pptx/document-properties.pptx"
+                       , pptxTests "metadata, short description"
+                         def
+                         "pptx/document-properties-short-desc.native"
+                         "pptx/document-properties-short-desc.pptx"
+                       , pptxTests "inline code and code blocks"
+                         def
+                         "pptx/code.native"
+                         "pptx/code.pptx"
+                       , pptxTests "inline code and code blocks, custom formatting"
+                         def { writerVariables = Context $ M.fromList
+                                 [(pack "monofont", toVal $ pack "Consolas")] }
+                         "pptx/code.native"
+                         "pptx/code-custom.pptx"
                        ]
