@@ -3,6 +3,7 @@
 module Tests.Writers.RST (tests) where
 
 import Prelude
+import Control.Monad.Identity
 import Test.Tasty
 import Test.Tasty.HUnit
 import Tests.Helpers
@@ -10,6 +11,7 @@ import Text.Pandoc
 import Text.Pandoc.Arbitrary ()
 import Text.Pandoc.Builder
 import Text.Pandoc.Writers.RST
+import qualified Data.Text as T
 
 infix 4 =:
 (=:) :: (ToString a, ToPandoc a)
@@ -18,8 +20,15 @@ infix 4 =:
 
 testTemplate :: (ToString a, ToString c, ToPandoc a) =>
                 String -> String -> (a, c) -> TestTree
-testTemplate t =
-  test (purely (writeRST def{ writerTemplate = Just t }) . toPandoc)
+testTemplate t = case runIdentity (compileTemplate [] (T.pack t)) of
+    Left e -> error $ "Could not compile RST template: " ++ e
+    Right templ -> test (purely (writeRST def{ writerTemplate = Just templ }) . toPandoc)
+
+bodyTemplate :: Template T.Text
+bodyTemplate = case runIdentity (compileTemplate [] "$body$\n") of
+                    Left e      -> error $
+                      "Could not compile RST bodyTemplate" ++ e
+                    Right templ -> templ
 
 tests :: [TestTree]
 tests = [ testGroup "rubrics"
@@ -96,6 +105,9 @@ tests = [ testGroup "rubrics"
           , "keeps quotes" =:
             strong (str "f" <> doubleQuoted (str "d") <> str "l") =?>
             "**f“d”l**"
+          , "backslash inserted between str and code" =:
+            str "/api?query=" <> code "foo" =?>
+            "/api?query=\\ ``foo``"
           ]
         , testGroup "headings"
           [ "normal heading" =:
@@ -104,7 +116,8 @@ tests = [ testGroup "rubrics"
               [ "foo"
               , "==="]
           -- note: heading normalization is only done in standalone mode
-          , test (purely (writeRST def{ writerTemplate = Just "$body$\n" }) . toPandoc)
+          , test (purely (writeRST def{ writerTemplate = Just bodyTemplate })
+                       . toPandoc)
             "heading levels" $
               header 1 (text "Header 1") <>
               header 3 (text "Header 2") <>
@@ -134,7 +147,7 @@ tests = [ testGroup "rubrics"
               , ""
               , "Header 2"
               , "--------"]
-          , test (purely (writeRST def{ writerTemplate = Just "$body$\n" }) . toPandoc)
+          , test (purely (writeRST def{ writerTemplate = Just bodyTemplate }) . toPandoc)
             "minimal heading levels" $
               header 2 (text "Header 1") <>
               header 3 (text "Header 2") <>
@@ -162,6 +175,6 @@ tests = [ testGroup "rubrics"
               , "--------"]
           ]
         , testTemplate "$subtitle$\n" "subtitle" $
-          (setMeta "subtitle" ("subtitle" :: Inlines) $ doc $ plain "") =?>
+          setMeta "subtitle" ("subtitle" :: Inlines) (doc $ plain "") =?>
           ("subtitle" :: String)
         ]
