@@ -8,13 +8,16 @@ import GHC.IO.Encoding
 import Test.Tasty
 import qualified Tests.Command
 import qualified Tests.Lua
+import qualified Tests.Lua.Module
 import qualified Tests.Old
 import qualified Tests.Readers.Creole
 import qualified Tests.Readers.Docx
+import qualified Tests.Readers.DokuWiki
 import qualified Tests.Readers.EPUB
 import qualified Tests.Readers.FB2
 import qualified Tests.Readers.HTML
 import qualified Tests.Readers.JATS
+import qualified Tests.Readers.Jira
 import qualified Tests.Readers.LaTeX
 import qualified Tests.Readers.Markdown
 import qualified Tests.Readers.Muse
@@ -40,11 +43,13 @@ import qualified Tests.Writers.Plain
 import qualified Tests.Writers.Powerpoint
 import qualified Tests.Writers.RST
 import qualified Tests.Writers.TEI
+import Tests.Helpers (findPandoc)
 import Text.Pandoc.Shared (inDirectory)
 
-tests :: TestTree
-tests = testGroup "pandoc tests" [ Tests.Command.tests
-        , testGroup "Old" Tests.Old.tests
+tests :: FilePath -> TestTree
+tests pandocPath = testGroup "pandoc tests"
+        [ Tests.Command.tests pandocPath
+        , testGroup "Old" (Tests.Old.tests pandocPath)
         , testGroup "Shared" Tests.Shared.tests
         , testGroup "Writers"
           [ testGroup "Native" Tests.Writers.Native.tests
@@ -69,6 +74,7 @@ tests = testGroup "pandoc tests" [ Tests.Command.tests
           , testGroup "Markdown" Tests.Readers.Markdown.tests
           , testGroup "HTML" Tests.Readers.HTML.tests
           , testGroup "JATS" Tests.Readers.JATS.tests
+          , testGroup "Jira" Tests.Readers.Jira.tests
           , testGroup "Org" Tests.Readers.Org.tests
           , testGroup "RST" Tests.Readers.RST.tests
           , testGroup "Docx" Tests.Readers.Docx.tests
@@ -79,11 +85,18 @@ tests = testGroup "pandoc tests" [ Tests.Command.tests
           , testGroup "Creole" Tests.Readers.Creole.tests
           , testGroup "Man" Tests.Readers.Man.tests
           , testGroup "FB2" Tests.Readers.FB2.tests
+          , testGroup "DokuWiki" Tests.Readers.DokuWiki.tests
+          ]
+        , testGroup "Lua"
+          [ testGroup "Lua filters" Tests.Lua.tests
+          , testGroup "Lua modules" Tests.Lua.Module.tests
           ]
-        , testGroup "Lua filters" Tests.Lua.tests
         ]
 
 main :: IO ()
 main = do
   setLocaleEncoding utf8
-  inDirectory "test" $ defaultMain tests
+  inDirectory "test" $ do
+    fp <- findPandoc
+    putStrLn $ "Using pandoc executable at " ++ fp
+    defaultMain $ tests fp
